#include <MFAtlas.h>
#include <IMFExpansion.h>
#include <IMFFlow.h>
#include <math.h>
#ifdef DMALLOC
#include <dmalloc.h>
#endif

struct IMFExpansionSt {int n;
                       int k;
                       double *u;
                       double *du;
                       double *ddu;
                       double *dddu;
                       int dataLn;
                       double *data;
                       int nRefs;
                      };

double *MFNV_CStar(MFNVector);
double *MFKV_CStar(MFKVector);

IMFExpansion IMFCreateExpansion(int n, int k)
 {
  IMFExpansion result;

  result=(IMFExpansion)malloc(sizeof(struct IMFExpansionSt));
  if(result==(IMFExpansion)NULL)abort();

  result->n=n;
  result->k=k;

  result->u=(double*)NULL;
  result->du=(double*)NULL;
  result->ddu=(double*)NULL;
  result->dddu=(double*)NULL;
  result->dataLn=0;
  result->data=(double*)NULL;

  result->nRefs=1;

  return result;
 }

void IMFFreeExpansion(IMFExpansion E)
 {
  E->nRefs--;

  if(E->nRefs<=0)
   {
    if(E->data!=(double*)NULL){free(E->data);E->data=(double*)NULL;}
    E->dataLn=0;
    E->u=(double*)NULL;
    E->du=(double*)NULL;
    E->ddu=(double*)NULL;
    E->dddu=(double*)NULL;
    E->n=0;
    E->k=0;

    free(E);
   }

  return;
 }

void IMFRefExpansion(IMFExpansion E)
 {
  E->nRefs++;
  return;
 }

void IMFEvaluateExpansion(IMFExpansion E,MFKVector vs,MFNVector vu)
 {
  double *s;
  double *u;
  int i,p,q,r;

  s=MFKV_CStar(vs);
  u=MFNV_CStar(vu);

  for(i=0;i<E->n;i++)
   {
    u[i]=E->u[i];
    if(E->du!=(double*)NULL)
     {
      for(p=0;p<E->k;p++)
       {
        u[i]+=E->du[i+E->n*p]*s[p];
        if(E->ddu!=(double*)NULL)
         {
          for(q=0;q<E->k;q++)
           {
            u[i]+=.5*E->ddu[i+E->n*(p+E->k*q)]*s[p]*s[q];
            if(E->dddu!=(double*)NULL)
             {
              for(r=0;r<E->k;r++)
               {
                u[i]+=E->dddu[i+E->n*(p+E->k*(q+E->k*r))]*s[p]*s[q]*s[r]/6.;
               }
             }
           }
         }
       }
     }
   }
  return;

 }

void IMFEvaluateExpansionDirectionalDerivative(IMFExpansion E,MFKVector vs,MFKVector vds,MFNVector vdu)
 {
  double *s;
  double *ds;
  double *du;
  int i,p,q,r;

  s=MFKV_CStar(vs);
  ds=MFKV_CStar(vds);
  du=MFNV_CStar(vdu);

  for(i=0;i<E->n;i++)
   {
    du[i]=0.;
    if(E->du!=(double*)NULL)
     {
      for(p=0;p<E->k;p++)
       {
        du[i]+=E->du[i+E->n*p]*ds[p];
        if(E->ddu!=(double*)NULL)
         {
          for(q=0;q<E->k;q++)
           {
            du[i]+=E->ddu[i+E->n*(p+E->k*q)]*ds[p]*s[q];
            if(E->dddu!=(double*)NULL)
             {
              for(r=0;r<E->k;r++)
                du[i]+=.5*E->dddu[i+E->n*(p+E->k*(q+E->k*r))]*ds[p]*s[q]*s[r];
             }
           }
         }
       }
     }
   }
  return;
 }

void IMFEvaluateExpansionSecondDirectionalDerivative(IMFExpansion E,MFKVector vs,MFKVector vds0,MFKVector vds1,MFNVector vddu)
 {
  double *s;
  double *ds0,*ds1;
  double *ddu;
  int i,p,q,r;

  s=MFKV_CStar(vs);
  ds0=MFKV_CStar(vds0);
  ds1=MFKV_CStar(vds1);
  ddu=MFNV_CStar(vddu);

  for(i=0;i<E->n;i++)
   {
    ddu[i]=0.;
    if(E->du!=(double*)NULL)
     {
      for(p=0;p<E->k;p++)
       {
        if(E->ddu!=(double*)NULL)
         {
          for(q=0;q<E->k;q++)
           {
            ddu[i]+=E->ddu[i+E->n*(p+E->k*q)]*ds0[p]*ds1[q];

            if(E->dddu!=(double*)NULL)
             {
              for(r=0;r<E->k;r++)
                ddu[i]+=E->dddu[i+E->n*(p+E->k*(q+E->k*r))]*ds0[p]*ds1[q]*s[r];
             }
           }
         }
       }
     }
   }
  return;
 }

void IMFEvaluateExpansionDerivative(IMFExpansion E,MFKVector vs,int d,MFNVector vdu)
 {
  double *s;
  double *du;
  int i,p,q;

  s=MFKV_CStar(vs);
  du=MFNV_CStar(vdu);

  if(d<0||d>E->k){fprintf(stderr,"d out of range in EvaluateExpansion Derivative, 0<=%d<%d\n",d,E->k);fflush(stderr);abort();}

  for(i=0;i<E->n;i++)
   {
    du[i]=0.;
    if(E->du!=(double*)NULL)  
     {
      du[i]+=E->du[i+E->n*d];
      for(p=0;p<E->k;p++)
       {
        if(E->ddu!=(double*)NULL)
         {
          du[i]+=E->ddu[i+E->n*(p+E->k*d)]*s[p];
          if(E->dddu!=(double*)NULL)
           {
            for(q=0;q<E->k;q++)
             {
              du[i]+=E->dddu[i+E->n*(p+E->k*(q+E->k*d))]*s[p]*s[q]/2.;
             }
           }
         }
       }
     }
   }
  return;
 }

void IMFEvaluateExpansionSecondDerivative(IMFExpansion E,MFKVector vs,int d0,int d1,MFNVector vddu)
 {
  double *s;
  double *ddu;
  int i,p;

  s=MFKV_CStar(vs);
  ddu=MFNV_CStar(vddu);

  for(i=0;i<E->n;i++)
   {
    ddu[i]=0.;
    if(E->du!=(double*)NULL && E->ddu!=(double*)NULL)
     {
      ddu[i]=E->ddu[i+E->n*(d0+E->k*d1)];
      if(E->dddu!=(double*)NULL)
       {
        for(p=0;p<E->k;p++)
          ddu[i]+=E->dddu[i+E->n*(p+E->k*(d0+E->k*d1))]*s[p];
       }
     }
   }
  return;
 }

void IMFEvaluateExpansionThirdDerivative(IMFExpansion E,MFKVector vs,int d0,int d1,int d2,MFNVector vdddu)
 {
  double *s;
  double *dddu;
  int i;

  s=MFKV_CStar(vs);
  dddu=MFNV_CStar(vdddu);

  for(i=0;i<E->n;i++)
   {
    if(E->du!=(double*)NULL && E->ddu!=(double*)NULL&&dddu!=(double*)NULL)
      dddu[i]=E->dddu[i+E->n*(d0+E->k*(d1+E->k*d2))];
     else
      dddu[i]=0.;
   }
  return;
 }

int IMFExpansionN(IMFExpansion E)
 {
  return E->n;
 }

int IMFExpansionK(IMFExpansion E)
 {
  return E->k;
 }

int IMFExpansionOrder(IMFExpansion E)
 {
  int O;

  O=0;
  if(E->du!=(double*)NULL)
   {
    O=1;
    if(E->ddu!=(double*)NULL)
     {O=2;if(E->dddu!=(double*)NULL)O=3;}
   }

  return O;
 }

double *IMFExpansionU(IMFExpansion E)
 {
  return E->u;
 }

double *IMFExpansionDu(IMFExpansion E)
 {
  return E->du;
 }

double *IMFExpansionDDu(IMFExpansion E)
 {
  return E->ddu;
 }

double *IMFExpansionDDDu(IMFExpansion E)
 {
  return E->dddu;
 }

double *IMFExpansionData(IMFExpansion E)
 {
  return E->data;
 }

int IMFExpansionDataLn(IMFExpansion E)
 {
  return E->dataLn;
 }

void IMFExpansionSetDerivatives(IMFExpansion E,double *u,
                                               double *du,
                                               double *ddu,
                                               double *dddu)
 {
  int i;

  if(E->data!=(double*)NULL)free(E->data);

  E->dataLn=E->n;
  if(du!=(double*)NULL)
   {
    E->dataLn+=E->n*E->k;
    if(ddu!=(double*)NULL)
     {
      E->dataLn+=E->n*E->k*E->k;
      if(dddu!=(double*)NULL)
        E->dataLn+=E->n*E->k*E->k*E->k;
     }
   }

  E->data=(double*)malloc(E->dataLn*sizeof(double));
  if(E->data==(double*)NULL && E->dataLn!=0)
   {
    abort();
   }

  E->u=E->data;
  if(du!=(double*)NULL)
   {
    E->du=E->u+E->n;
    if(ddu!=(double*)NULL)
     {
      E->ddu=E->du+E->n*E->k;
      if(dddu!=(double*)NULL)
       {
        E->dddu=E->ddu+E->n*E->k*E->k;
       }else{
        E->dddu=(double*)NULL;
       }
     }else{
      E->ddu=(double*)NULL;
      E->dddu=(double*)NULL;
     }
   }else{
    E->du=(double*)NULL;
    E->ddu=(double*)NULL;
    E->dddu=(double*)NULL;
   }

  if(u!=(double*)NULL)
   {
    for(i=0;i<E->n;i++)(E->u)[i]=u[i];
    if(du!=(double*)NULL)
     {
      for(i=0;i<E->n*E->k;i++)(E->du)[i]=du[i];
      if(ddu!=(double*)NULL)
       {
        for(i=0;i<E->n*E->k*E->k;i++)(E->ddu)[i]=ddu[i];
        if(dddu!=(double*)NULL)
         {
          for(i=0;i<E->n*E->k*E->k*E->k;i++)(E->dddu)[i]=dddu[i];
         }
       }
     }
   }

  return;
 }

char *IMFC(double a, char *suffix,int *leading)
 {
  static char result[1024];
  int ia;

  result[0]=0x0;
  ia=a;
  if(suffix[0]!=0x0 && !(*leading))
   {
    if(fabs(a-1)<1.e-7)
      sprintf(result,"+%s",suffix);
     else if(fabs(a+1)<1.e-7)
      sprintf(result,"-%s",suffix);
     else if(fabs(a-ia)<1.e-7 && ia > 0)
      sprintf(result,"+ %d*%s",ia,suffix);
     else if(fabs(a-ia)<1.e-7 && ia < 0)
      sprintf(result,"- %d*%s",-ia,suffix);
     else if(fabs(a-ia-1)<1.e-7 && ia+1 > 0)
      sprintf(result,"+ %d*%s",ia+1,suffix);
     else if(fabs(a-ia-1)<1.e-7 && ia+1 < 0)
      sprintf(result,"- %d*%s",-ia-1,suffix);
     else if(fabs(a-ia+1)<1.e-7 && ia-1 > 0)
      sprintf(result,"+ %d*%s",ia-1,suffix);
     else if(fabs(a-ia+1)<1.e-7 && ia-1 < 0)
      sprintf(result,"- %d*%s",-ia+1,suffix);
     else if(a>1.e-7)
      sprintf(result,"+ %le*%s",a,suffix);
     else if(a<-1.e-7)
      sprintf(result,"- %le*%s",-a,suffix);
     else
      sprintf(result,"");
   }else if(suffix[0]==0x0 && !(*leading))
   {
    if(fabs(a-1)<1.e-7)
      sprintf(result,"+1");
     else if(fabs(a+1)<1.e-7)
      sprintf(result,"-1");
     else if(fabs(a-ia)<1.e-7 && ia > 0)
      sprintf(result,"+ %d",ia);
     else if(fabs(a-ia)<1.e-7 && ia < 0)
      sprintf(result,"- %d",-ia);
     else if(fabs(a-ia-1)<1.e-7 && ia+1 > 0)
      sprintf(result,"+ %d",ia+1);
     else if(fabs(a-ia-1)<1.e-7 && ia+1 < 0)
      sprintf(result,"- %d",-ia-1);
     else if(fabs(a-ia+1)<1.e-7 && ia-1 > 0)
      sprintf(result,"+ %d",ia-1);
     else if(fabs(a-ia+1)<1.e-7 && ia-1 < 0)
      sprintf(result,"- %d",-ia+1);
     else if(a>1.e-7)
      sprintf(result,"+ %le",a);
     else if(a<-1.e-7)
      sprintf(result,"- %le",-a);
     else
      sprintf(result,"");
   }else if(suffix[0]!=0x0 && *leading)
   {
    if(fabs(a-1)<1.e-7)
      {sprintf(result,"%s",suffix);*leading=0;}
     else if(fabs(a+1)<1.e-7)
      {sprintf(result,"-%s",suffix);*leading=0;}
     else if(fabs(a-ia)<1.e-7 && ia > 0)
      {sprintf(result," %d*%s",ia,suffix);*leading=0;}
     else if(fabs(a-ia)<1.e-7 && ia < 0)
      {sprintf(result,"- %d*%s",-ia,suffix);*leading=0;}
     else if(fabs(a-ia-1)<1.e-7 && ia+1 > 0)
      {sprintf(result," %d*%s",ia+1,suffix);*leading=0;}
     else if(fabs(a-ia-1)<1.e-7 && ia+1 < 0)
      {sprintf(result,"- %d*%s",-ia-1,suffix);*leading=0;}
     else if(fabs(a-ia+1)<1.e-7 && ia-1 > 0)
      {sprintf(result," %d*%s",ia-1,suffix);*leading=0;}
     else if(fabs(a-ia+1)<1.e-7 && ia-1 < 0)
      {sprintf(result,"- %d*%s",-ia+1,suffix);*leading=0;}
     else if(a>1.e-7)
      {sprintf(result," %le*%s",a,suffix);*leading=0;}
     else if(a<-1.e-7)
      {sprintf(result,"- %le*%s",-a,suffix);*leading=0;}
     else
      sprintf(result,"");
   }else if(suffix[0]==0x0 && *leading)
   {
    if(fabs(a-1)<1.e-7)
      {sprintf(result,"1");*leading=0;}
     else if(fabs(a+1)<1.e-7)
      {sprintf(result,"-1");*leading=0;}
     else if(fabs(a-ia)<1.e-7 && ia > 0)
      {sprintf(result," %d",ia);*leading=0;}
     else if(fabs(a-ia)<1.e-7 && ia < 0)
      {sprintf(result,"- %d",-ia);*leading=0;}
     else if(fabs(a-ia-1)<1.e-7 && ia+1 > 0)
      {sprintf(result," %d",ia+1);*leading=0;}
     else if(fabs(a-ia-1)<1.e-7 && ia+1 < 0)
      {sprintf(result,"- %d",-ia-1);*leading=0;}
     else if(fabs(a-ia+1)<1.e-7 && ia-1 > 0)
      {sprintf(result," %d",ia-1);*leading=0;}
     else if(fabs(a-ia+1)<1.e-7 && ia-1 < 0)
      {sprintf(result,"- %d",-ia+1);*leading=0;}
     else if(a>1.e-7)
      {sprintf(result," %le",a);*leading=0;}
     else if(a<-1.e-7)
      {sprintf(result,"- %le",-a);*leading=0;}
     else
      sprintf(result,"");
   }
  return result;
 }

void IMFPrintExpansion(FILE *fid,IMFExpansion E)
 {
  int i;
  int leading;

  fprintf(fid,"IMFPrintExpansion, data=0x%8.8x, length %d\n",E->data,E->dataLn);
  fprintf(fid,"                   u=0x%8.8x\n",E->u);
  fprintf(fid,"                  du=0x%8.8x\n",E->du);
  fprintf(fid,"                 ddu=0x%8.8x\n",E->ddu);
  fprintf(fid,"                dddu=0x%8.8x\n",E->dddu);

  switch(E->k)
   {
    case 1:
     printf("u(s) = (");
     for(i=0;i<E->n;i++)
      {
       leading=1;
       if(i>0)fprintf(fid,"       ");
       fprintf(fid,"%s",IMFC(E->u[i],"",&leading));
       fprintf(fid,"%s",IMFC(E->du[i],"s",&leading));
       fprintf(fid,"%s",IMFC(E->ddu[i]/2.,"s*s",&leading));
       fprintf(fid,"%s",IMFC(E->dddu[i]/6.,"s*s*s",&leading));
       if(i<E->n-1)fprintf(fid,",\n");
        else fprintf(fid,") + O(s^4)\n");
      }

     break;
    case 2:
     fprintf(fid,"u(s,t) = (");
     for(i=0;i<E->n;i++)
      {
       leading=1;
       if(i>0)fprintf(fid,"         ");
       fprintf(fid,"%s",IMFC(E->u[i],"",&leading));
       if(E->du!=(double*)NULL)
        {
         fprintf(fid,"%s",IMFC(E->du[i],"s",&leading));
         fprintf(fid,"%s",IMFC(E->du[i+E->n],"t",&leading));
         if(E->ddu!=(double*)NULL)
          {
           fprintf(fid,"%s",IMFC(E->ddu[i]/2,"s*s",&leading));
           fprintf(fid,"%s",IMFC(E->ddu[i+E->n*(1+E->k*0)],"s*t",&leading));
           fprintf(fid,"%s",IMFC(E->ddu[i+E->n*(1+E->k*1)],"t*t",&leading));
           if(E->dddu!=(double*)NULL)
            {
             fprintf(fid,"%s",IMFC(E->dddu[i+E->n*(0+E->k*(0+E->k*0))]/6,"s*s*s",&leading));
             fprintf(fid,"%s",IMFC(E->dddu[i+E->n*(0+E->k*(0+E->k*1))]/2,"s*s*t",&leading));
             fprintf(fid,"%s",IMFC(E->dddu[i+E->n*(0+E->k*(1+E->k*1))]/2,"s*t*t",&leading));
             fprintf(fid,"%s",IMFC(E->dddu[i+E->n*(1+E->k*(1+E->k*1))]/6,"t*t*t",&leading));
            }
          }
        }
       if(i<E->n-1)fprintf(fid,",\n");
        else fprintf(fid,") + O(s^4,s^3t,s^2t^2,s t^3,t^4)\n");
      }
     break;
   }
  return;
 }

IMFExpansion IMFCloneExpansion(IMFExpansion E)
 {
  IMFExpansion result;
  int i;

  result=IMFCreateExpansion(E->n,E->k);
  result->data=(double*)malloc(E->dataLn*sizeof(double));if(result->data==(double*)NULL&&E->dataLn>0)abort();
  result->dataLn=E->dataLn;
  result->u=(double*)NULL;
  result->du=(double*)NULL;
  result->ddu=(double*)NULL;
  result->dddu=(double*)NULL;

  if(E->dataLn>0&&E->u!=(double*)NULL)
   {
    result->u=result->data;
    for(i=0;i<E->n;i++)(result->u)[i]=(E->u)[i];
   }
  if(E->dataLn>0&&E->du!=(double*)NULL)
   {
    result->du=result->u+result->n;
    for(i=0;i<E->n*E->k;i++)(result->du)[i]=(E->du)[i];
   }
  if(E->dataLn>0&&E->ddu!=(double*)NULL)
   {
    result->ddu=result->du+result->n*result->k;
    for(i=0;i<E->n*E->k*E->k;i++)(result->ddu)[i]=(E->ddu)[i];
   }
  if(E->dataLn>0&&E->dddu!=(double*)NULL)
   {
    result->dddu=result->ddu+result->n*result->k*result->k;
    for(i=0;i<E->n*E->k*E->k*E->k;i++)(result->dddu)[i]=(E->dddu)[i];
   }

  return result;
 }

IMFExpansion IMFRectify(IMFExpansion E0)
 {
  double t;

  double *g0;
  double *G0;

  double *G;

  double *ng;
  double *nG;

  double *dX;
  double *ddX;

  double *A;
  double *b;
  int i,j,k,p,q,r,w;
  int verbose=0;
  IMFExpansion E;
  int n,m;

  double *u;
  double *du;
  double *ddu;
  double *dddu;

  n=E0->n;
  m=E0->k;

  g0=(double*)malloc(m*m*sizeof(double));if(g0==(double*)NULL)abort();
  G0=(double*)malloc(m*m*m*sizeof(double));if(G0==(double*)NULL)abort();
  G =(double*)malloc(m*m*m*sizeof(double));if(G==(double*)NULL)abort();
  ng=(double*)malloc(m*m*sizeof(double));if(ng==(double*)NULL)abort();
  nG=(double*)malloc(m*m*m*sizeof(double));if(nG==(double*)NULL)abort();
  dX=(double*)malloc(m*m*sizeof(double));if(dX==(double*)NULL)abort();
  ddX=(double*)malloc(m*m*m*sizeof(double));if(ddX==(double*)NULL)abort();
  A=(double*)malloc(m*m*sizeof(double));if(A==(double*)NULL)abort();
  b=(double*)malloc(m*sizeof(double));if(b==(double*)NULL)abort();
  u=(double*)malloc(n*sizeof(double));if(u==(double*)NULL)abort();
  du=(double*)malloc(n*m*sizeof(double));if(du==(double*)NULL)abort();
  ddu=(double*)malloc(n*m*m*sizeof(double));if(ddu==(double*)NULL)abort();
  dddu=(double*)malloc(n*m*m*m*sizeof(double));if(dddu==(double*)NULL)abort();

  if(verbose){printf("IMFRectify n=%d, k=%d\n",n,m);fflush(stdout);}

  for(i=0;i<m;i++)
   {
    for(j=0;j<m;j++)
     {
      g0[i+m*j]=0.;for(p=0;p<n;p++)g0[i+m*j]+=E0->du[p+n*i]*E0->du[p+n*j];
     }
   }

  if(verbose)
   {
    printf("\n");
    printf("Metric in old coordinates\n\n");
    for(i=0;i<m;i++)
     {
      printf("   [%lf",g0[i+m*0]);
      for(j=1;j<m;j++)printf(" %lf",g0[i+m*j]);
      printf("]\n");
     }
    if(m==2)printf(" determinant %le\n",g0[0+m*0]*g0[1+m*1]-g0[1+m*0]*g0[0+m*1]);
    fflush(stdout);
   }

/* Gram-Schmidt */

  for(j=0;j<m;j++)
   {
    for(i=0;i<n;i++)du[i+n*j]=E0->du[i+n*j];

    for(k=0;k<j;k++)
     {
      t=0.;for(i=0;i<n;i++)t+=du[i+n*j]*du[i+n*k];
      for(i=0;i<n;i++)du[i+n*j]=du[i+n*j]-t*du[i+n*k];
     }
  
    t=0.;for(i=0;i<n;i++)t+=du[i+n*j]*du[i+n*j];t=1./sqrt(t);
    for(i=0;i<n;i++)du[i+n*j]=t*du[i+n*j];
   }

  for(j=0;j<m;j++)
   {
    for(k=0;k<m;k++)
     {
      t=0.;for(i=0;i<n;i++)t+=du[i+n*j]*du[i+n*k];
      ng[j+m*k]=t;
     }
   }

  if(verbose)
   {
    printf("\n");
    printf("Metric in new coordinates\n\n");
    for(i=0;i<m;i++)
     {
      printf("   [%lf",ng[i+m*0]);
      for(j=1;j<m;j++)printf(" %lf",ng[i+m*j]);
      printf("]\n");
     }
    if(m==2)printf(" new determinant %le\n",ng[0+m*0]*ng[1+m*1]-ng[1+m*0]*ng[0+m*1]);
    fflush(stdout);
   }


  for(k=0;k<m;k++)
   {
    for(i=0;i<m;i++)for(j=0;j<m;j++)A[i+m*j]=g0[j+m*i];
    for(j=0;j<m;j++){b[j]=0.;for(p=0;p<n;p++)b[j]+=E0->du[p+n*j]*du[p+n*k];}
    if(!MFSolveFull(m,A,b))exit(12);

    for(j=0;j<m;j++)dX[j+m*k]=b[j];
   }

  if(verbose)
   {
    printf("\n");
    printf("Check of first derivative of coordinate transformation X^i_{,j}\n\n");
    for(j=0;j<m;j++)
     {
      printf("   u_{,%d}             = [%9.6lf",j,du[0+n*j]);
      for(i=1;i<n;i++)printf(",%9.6lf",du[i+n*j]);
      printf("]\n");
  
      printf("   X^p_{,%d} u_{,\\xi_p}= [",j);
      for(i=0;i<n;i++)
       {
        t=0.;for(p=0;p<m;p++)t+=dX[p+m*j]*E0->du[i+n*p];
        if(i>0)printf(",");
        printf("%9.6lf",t);
       }
      printf("]\n");
     }
    fflush(stdout);
   }

  for(i=0;i<m;i++)
   {
    for(j=0;j<m;j++)
     {
      for(k=0;k<m;k++)
       {
        G0[i+m*(j+m*k)]=0.;for(p=0;p<n;p++)G0[i+m*(j+m*k)]+=E0->du[p+n*i]*E0->ddu[p+n*(j+m*k)];
       }
     }
   }

  if(verbose)
   {
    printf("\n");
    printf("Connection in old coordinates\n\n");
    for(i=0;i<m;i++)
     {
      for(j=0;j<m;j++)
       {
        if(j==0)printf("  %d [",i);
         else printf("    [");
        for(k=0;k<m;k++)
         {
          if(k>0)printf(" ");
          printf("%9.6lf",G0[i+m*(j+m*k)]);
         }
        printf("]\n");
       }
      printf("\n");
     }
    fflush(stdout);
   }

  for(i=0;i<m*m*m;i++)G[i]=0.;

  for(i=0;i<m;i++)
   {
    for(j=0;j<m;j++)
     {
      for(k=0;k<m;k++)
       {
        ddX[i+m*(j+m*k)]=0.;
        for(p=0;p<m;p++)
         {
          ddX[i+m*(j+m*k)]+=dX[i+m*p]*G[p+m*(j+m*k)];
          for(q=0;q<m;q++)
           {
            for(r=0;r<m;r++)
             {
              for(w=0;w<m;w++)
               {
                ddX[i+m*(j+m*k)]-=dX[i+m*w]*dX[p+m*w]*dX[q+m*j]*dX[r+m*k]*G0[p+m*(q+m*r)];
               }
             }
           }
         }
       }
     }
   }


  for(i=0;i<m;i++)
   {
    for(j=0;j<m;j++)
     {
      for(k=0;k<m;k++)
       {
        nG[i+m*(j+m*k)]=0.;
        for(p=0;p<m;p++)
         {
          for(q=0;q<m;q++)
           {
            nG[i+m*(j+m*k)]+=dX[p+m*i]*ddX[q+m*(j+m*k)]*g0[p+m*q];
            for(r=0;r<m;r++)
             {
              nG[i+m*(j+m*k)]+=dX[p+m*i]*dX[q+m*j]*dX[r+m*k]*G0[p+m*(q+m*r)];
             }
           }
         }
       }
     }
   }

  if(verbose)
   {
    printf("\n");
    printf("Connection in new coordinates (requested) \n\n");
    for(i=0;i<m;i++)
     {
      for(j=0;j<m;j++)
       {
        if(j==0)printf("  %d [",i);
         else printf("    [");
        for(k=0;k<m;k++)
         {
          if(k>0)printf(" ");
          printf("%9.6lf",G[i+m*(j+m*k)]);
         }
        printf("]\n");
       }
      printf("\n");
     }
  
    printf("\n");
    printf("Connection in new coordinates (transformed)\n\n");
    for(i=0;i<m;i++)
     {
      for(j=0;j<m;j++)
       {
        if(j==0)printf("  %d [",i);
         else printf("    [");
        for(k=0;k<m;k++)
         {
          if(k>0)printf(" ");
          printf("%9.6lf",nG[i+m*(j+m*k)]);
         }
        printf("]\n");
       }
      printf("\n");
     }
    fflush(stdout);
   }
  
  for(i=0;i<n;i++)
   {
    for(j=0;j<m;j++)
     {
      for(k=0;k<m;k++)
       {
        ddu[i+n*(j+m*k)]=0.;
        for(p=0;p<m;p++)
         {
          ddu[i+n*(j+m*k)]+=ddX[p+m*(j+m*k)]*E0->du[i+n*p];
          for(q=0;q<m;q++)
           ddu[i+n*(j+m*k)]+=dX[p+m*j]*dX[q+m*k]*E0->ddu[i+n*(p+m*q)];
         }
       }
     }
   }

  if(verbose)
   {
    printf("\n");
    printf("Connection in new coordinates (direct)\n\n");
    for(i=0;i<m;i++)
     {
      for(j=0;j<m;j++)
       {
        if(j==0)printf("  %d [",i);
         else printf("    [");
        for(k=0;k<m;k++)
         {
          if(k>0)printf(" ");
          t=0.;for(p=0;p<n;p++)t+=du[p+n*i]*ddu[p+n*(j+m*k)];
          printf("%9.6lf",t);
         }
        printf("]\n");
       }
      printf("\n");
     }
    fflush(stdout);
   }

/* Transform third derivatives too ? */

  free(g0);
  free(G0);
  free(ng);
  free(G);
  free(nG);
  free(dX);
  free(ddX);
  free(A);
  free(b);

  if(verbose){printf("done IMFRectify\n");fflush(stdout);fflush(stdout);}
  E=IMFCreateExpansion(n,m);
  for(i=0;i<n;i++)u[i]=(E0->u)[i];
  IMFExpansionSetDerivatives(E,u,du,ddu,(double*)NULL);
  free(u);
  free(du);
  free(ddu);
  free(dddu);

  return E;
 }

IMFExpansion IMFInflateExpansionWithFlow(IMFExpansion E,IMFFlow f)
 {
  double *u;
  double *du;
  double *ddu;
  double *F;
  double *dF;
  double d;
  IMFExpansion tmp;
  IMFExpansion result;
  int n,m,k;
  int i,j,l;
  MFNVector v;

  n=IMFExpansionN(E);
  k=IMFExpansionK(E);
  m=k+1;

  tmp=IMFCreateExpansion(n,m);
  u=E->u;
  du=(double*)malloc(n*m*sizeof(double));
  ddu=(double*)malloc(n*m*m*sizeof(double));
  F=(double*)malloc(n*sizeof(double));
  dF=(double*)malloc(n*n*sizeof(double));

  v=MFCreateWrappedNVector(n,E->u);
  IMFEvaluateFlow(f,v,F);
  IMFEvaluateDerivativeOfFlow(f,v,dF);
  MFFreeNVector(v);

  d=0.;for(i=0;i<n;i++)d+=F[i]*F[i];d=sqrt(d);
  for(i=0;i<n;i++)F[i]=F[i]/d;

  for(i=0;i<n;i++)
   {
    for(j=0;j<k;j++)
     {
      du[i+n*j]=E->du[i+n*j];
     }
    du[i+n*k]=F[i];
   }

  for(i=0;i<n;i++)
   {
    for(j=0;j<k;j++)
     {
      for(l=0;l<k;l++)
       {
        ddu[i+n*(j+m*l)]=E->ddu[i+n*(j+k*l)];
       }
      ddu[i+n*(j+m*k)]=0.;
      ddu[i+n*(k+m*j)]=0.;
     }
    ddu[i+n*(k+m*k)]=0.;
    for(j=0;j<n;j++)ddu[i+n*(k+m*k)]+=dF[i+n*j]*F[j]/d; /* Hadn't divided by d before REMOVEME?? */
   }

  IMFExpansionSetDerivatives(tmp,u,du,ddu,(double*)NULL);
  free(du);
  free(ddu);
  free(F);
  free(dF);
  result=IMFRectify(tmp);
  IMFFreeExpansion(tmp);
  return result;
 }

double IMFExpansionR(IMFExpansion E, double epsilon)
 {
  int i,j;
  double A;
  double t;

  A=0.;
  for(j=0;j<E->k*E->k;j++)
   {
    t=0.;
    for(i=0;i<E->n;i++)t+=pow(E->ddu[i+E->n*j],2.);
    t=sqrt(t);
    if(t<A||j==0)A=t;
   }
  return sqrt(2*epsilon/A);
 }

MFNKMatrix IMFExpansionTS(IMFExpansion E)
 {
  return MFCreateNKMatrixWithData(E->n,E->k,E->du);
 }

void IMFOLDExpansionSetDerivatives(IMFExpansion E,double *u,
                                               double *du,
                                               double *ddu,
                                               double *dddu)
 {
  int i;

  if(E->u!=(double*)NULL)free(E->u);
  if(u!=(double*)NULL)
   {
    E->u=(double*)malloc(E->n*sizeof(double));if(E->u==(double*)NULL)abort();
    for(i=0;i<E->n;i++)(E->u)[i]=u[i];
   }else
    E->u=(double*)NULL;

  if(E->du!=(double*)NULL)free(E->du);
  if(du!=(double*)NULL)
   {
    E->du=(double*)malloc(E->n*E->k*sizeof(double));if(E->du==(double*)NULL)abort();
    for(i=0;i<E->n*E->k;i++)(E->du)[i]=du[i];
   }else
    E->du=(double*)NULL;

  if(E->ddu!=(double*)NULL)free(E->ddu);
  if(ddu!=(double*)NULL)
   {
    E->ddu=(double*)malloc(E->n*E->k*E->k*sizeof(double));if(E->ddu==(double*)NULL)abort();
    for(i=0;i<E->n*E->k*E->k;i++)(E->ddu)[i]=ddu[i];
   }else
    E->ddu=(double*)NULL;

  if(E->dddu!=(double*)NULL)free(E->dddu);
  if(dddu!=(double*)NULL)
   {
    E->dddu=(double*)malloc(E->n*E->k*E->k*E->k*sizeof(double));if(E->dddu==(double*)NULL)abort();
    for(i=0;i<E->n*E->k*E->k*E->k;i++)(E->dddu)[i]=dddu[i];
   }else
    E->dddu=(double*)NULL;

  return;
 }

void IMFEvaluateExpansionDirectionalDerivativeE(IMFExpansion E,double *s,double *ds,double *du)
 {
  int i,p,q,r;
  int j,k,l,m;

  for(i=0;i<E->n;i++)
   {
    du[i]=0.;
    if(E->du!=(double*)NULL)
     {
      for(p=0;p<E->k;p++)
       {
        du[i]+=E->du[i+E->n*p]*ds[p];
        if(E->ddu!=(double*)NULL)
         {
          for(q=0;q<E->k;q++)
           {
            du[i]+=E->ddu[i+E->n*(p+E->k*q)]*ds[p]*s[q];
            if(E->dddu!=(double*)NULL)
             {
              for(r=0;r<E->k;r++)
                du[i]+=.5*E->dddu[i+E->n*(p+E->k*(q+E->k*r))]*ds[p]*s[q]*s[r];
             }
           }
         }
       }
     }

    for(j=0;j<E->k;j++)
     {
      du[i+E->n*j]=0.;
      if(E->du!=(double*)NULL)
       {
        for(p=0;p<E->k;p++)
         {
          if(E->ddu!=(double*)NULL)
           {
            du[i+E->n*j]+=E->ddu[i+E->n*(p+E->k*j)]*ds[p];
            for(l=0;l<E->k;l++)
             {
              if(E->dddu!=(double*)NULL)
                du[i+E->n*(j+E->k*l)]+=E->dddu[i+E->n*(p+E->k*(l+E->k*j))]*ds[p];
             }
           }
         }
       }
     }

   }
  return;
 }

