/* 
    %W%
    %D% %T%
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   September 15, 2003                    */

#include <MFError.h>
#include <MFNVector.h>
#include <IMFExpansion.h>
#include <IMFExpansionSpace.h>
#include <float.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#ifdef DMALLOC
#include <dmalloc.h>
#endif
#ifndef DBL_QNAN
#define DBL_QNAN 1.e200
#endif

struct IMFExpansionNVectorData
 {
  int nC;
  int k;
  IMFExpansion E;
  double t;
  MFNVector sigma;
  MFKVector s0;
  int chart0;
  int prevChart;
  int type;
 };

void IMFFreeExpansionNVectorData(void*);
int IMFExpansionNVGetNC(void*);
double IMFExpansionNVGetC(int,void*);
void IMFExpansionNVSetC(int,double,void*);
void IMFExpansionNVDiff(void*,void*,void*);
void IMFExpansionNVAdd(void*,void*,void*);
void IMFPrintExpansionNVector(FILE*,void*);
MFNVector MFCloneExpansionNVector(void*);

static char MFNVectorErrorMsg[256]="";

MFNVector IMFCreateExpansionNVector(IMFExpansion E,double t, MFNVector sigma, int prevChart, int type)
 {
  static char RoutineName[]={"IMFCreateExpansionNVector"};
  struct IMFExpansionNVectorData *data;
  int i;
  MFNVector this;

  data=(struct IMFExpansionNVectorData *)malloc(sizeof(struct IMFExpansionNVectorData));
  if(data==(struct IMFExpansionNVectorData*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct IMFExpansionNVectorData));fflush(stdout);
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return(this);
   }

  data->nC=IMFExpansionN(E);
  data->k=IMFExpansionK(E);
  data->E=E;
  IMFRefExpansion(E);

  data->t=t;
  data->sigma=MFCloneNVector(sigma);
  data->s0=(MFKVector)NULL;
  data->chart0=-1;
  data->prevChart=prevChart;
  data->type=type;

  this=MFCreateNVectorBaseClass("IMFExpansionVector");

  MFNVectorSetData(this,data);
  MFNVectorSetFreeData(this,IMFFreeExpansionNVectorData);
  MFNVectorSetGetNC(this,IMFExpansionNVGetNC);
  MFNVectorSetGetC(this,IMFExpansionNVGetC);
  MFNVectorSetSetC(this,IMFExpansionNVSetC);
  MFNVectorSetDiff(this,IMFExpansionNVDiff);
  MFNVectorSetAdd(this,IMFExpansionNVAdd);
  MFNVectorSetClone(this,MFCloneExpansionNVector);
  MFNVectorSetPrint(this,IMFPrintExpansionNVector);

  return this;
 }

void IMFFreeExpansionNVectorData(void *data)
 {
  static char RoutineName[]={"IMFFreeExpansionNVectorData"};
  struct IMFExpansionNVectorData *this;

  this=(struct IMFExpansionNVectorData*)data;

  if(this==(struct IMFExpansionNVectorData*)NULL)return;
  IMFFreeExpansion(this->E);
  if(this->sigma!=(MFNVector)NULL)MFFreeNVector(this->sigma);
  if(this->s0!=(MFKVector)NULL)MFFreeKVector(this->s0);
  free(this);

  return;
 }

int IMFExpansionNVGetNC(void *data)
 {
  static char RoutineName[]={"IMFNV_NC"};
  struct IMFExpansionNVectorData *this;

  this=(struct IMFExpansionNVectorData*)data;

  if(this==(struct IMFExpansionNVectorData*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector Data (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return -1;
   }

  return(this->nC);
 }

double IMFExpansionNVGetC(int i,void *data)
 {
  static char RoutineName[]={"IMFExpansionNVGetC"};
  struct IMFExpansionNVectorData *this;
  double result;

  this=(struct IMFExpansionNVectorData*)data;

  if(this==(struct IMFExpansionNVectorData*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector Data (argument 2) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return DBL_QNAN;
   }

  if(i<0|| !(i<this->nC))
   {
    sprintf(MFNVectorErrorMsg,"Coordinate %d (argument 1) is illegal. Must be in 0 to %d",i,this->nC-1);
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(8,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return DBL_QNAN;
   }

  result=IMFExpansionU(this->E)[i];

  return result;
 }

void IMFExpansionNVSetC(int i,double vl,void *data)
 {
  static char RoutineName[]={"IMFExpansionNVSetC"};
  struct IMFExpansionNVectorData *this;

  this=(struct IMFExpansionNVectorData*)data;

  if(this==(struct IMFExpansionNVectorData*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector Data (argument 3) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(i<0|| !(i<this->nC))
   {
    sprintf(MFNVectorErrorMsg,"Coordinate %d (argument 1) is illegal. Must be in 0 to %d",i,this->nC-1);
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(8,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  IMFExpansionU(this->E)[i]=vl;

  return;
 }

void IMFExpansionNVDiff(void *adata,void *bdata, void *cdata)
 {
  static char RoutineName[]={"IMFExpansionNVDiff"};
  int i,n;
  struct IMFExpansionNVectorData *a;
  struct IMFExpansionNVectorData *b;
  struct IMFExpansionNVectorData *c;

  a=(struct IMFExpansionNVectorData*)adata;
  if(a==(struct IMFExpansionNVectorData*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector Data for a (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  b=(struct IMFExpansionNVectorData*)bdata;
  if(b==(struct IMFExpansionNVectorData*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector Data for b (argument 2) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  c=(struct IMFExpansionNVectorData*)cdata;
  if(c==(struct IMFExpansionNVectorData*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector Data for c (argument 3) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(a->nC!=b->nC || a->nC!=c->nC || b->nC!=c->nC)
   {
    sprintf(MFNVectorErrorMsg,"Vectors must all be the same length a=%d, b=%d, c=%d",a->nC,b->nC,c->nC);
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }
  n=a->nC;

  for(i=0;i<n;i++)
    IMFExpansionU(c->E)[i]=IMFExpansionU(a->E)[i]-IMFExpansionU(b->E)[i];

  return;
 }

void IMFExpansionNVAdd(void *adata,void *bdata,void *cdata)
 {
  static char RoutineName[]={"IMFExpansionNVAdd"};
  int i,n;
  struct IMFExpansionNVectorData *a;
  struct IMFExpansionNVectorData *b;
  struct IMFExpansionNVectorData *c;

  a=(struct IMFExpansionNVectorData*)adata;
  if(a==(struct IMFExpansionNVectorData*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector Data for a (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  b=(struct IMFExpansionNVectorData*)bdata;
  if(b==(struct IMFExpansionNVectorData*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector Data for b (argument 2) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  c=(struct IMFExpansionNVectorData*)cdata;
  if(c==(struct IMFExpansionNVectorData*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"Pointer to Vector Data for c (argument 3) is NULL");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(a->nC!=b->nC || a->nC!=c->nC || b->nC!=c->nC)
   {
    sprintf(MFNVectorErrorMsg,"Vectors must all be the same length a=%d, b=%d, c=%d",a->nC,b->nC,c->nC);
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);
    MFSetError(4,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }
  n=a->nC;

  for(i=0;i<n;i++)
    IMFExpansionU(c->E)[i]=IMFExpansionU(a->E)[i]-IMFExpansionU(b->E)[i];

  return;
 }

void IMFPrintExpansionNVector(FILE *fid,void *data)
 {
  static char RoutineName[]={"IMFPrintExpansionNVector"};
  int i;
  struct IMFExpansionNVectorData *u;

  u=(struct IMFExpansionNVectorData*)data;

  if(fid==(FILE*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"fid (argument 1) is NULL.");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(data==(void*)NULL)
   {
    sprintf(MFNVectorErrorMsg,"data (argument 2) is NULL.");
    printf("%s -- %s\n",RoutineName,MFNVectorErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNVectorErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(u->nC<1)return;

  if(u->nC<10)
   {
    fprintf(fid,"(%lf",IMFExpansionU(u->E)[0]);
    for(i=1;i<u->nC;i++)
      fprintf(fid,",%lf",IMFExpansionU(u->E)[i]);
    fprintf(fid,")");
   }else{
    fprintf(fid,"(%lf",IMFExpansionU(u->E)[0]);
    for(i=1;i<5;i++)
      fprintf(fid,",%lf",IMFExpansionU(u->E)[i]);
    fprintf(fid,",...");
    for(i=u->nC-5;i<u->nC;i++)
      fprintf(fid,",%lf",IMFExpansionU(u->E)[i]);
    fprintf(fid,")");
   }
  fprintf(fid,"\n");

  return;
 }

IMFExpansion IMFExpansionNVGetE(MFNVector this)
 {
  static char RoutineName[]={"IMFExpansionNVGetE"};
  struct IMFExpansionNVectorData *data;

  data=(struct IMFExpansionNVectorData*)MFNVectorGetData(this);

  return data->E;
 }

double IMFExpansionNVGetT(MFNVector this)
 {
  static char RoutineName[]={"IMFExpansionNVGetE"};
  struct IMFExpansionNVectorData *data;

  data=(struct IMFExpansionNVectorData*)MFNVectorGetData(this);

  return data->t;
 }

MFNVector IMFExpansionNVGetSigma(MFNVector this)
 {
  static char RoutineName[]={"IMFExpansionNVGetSigma"};
  struct IMFExpansionNVectorData *data;

  data=(struct IMFExpansionNVectorData*)MFNVectorGetData(this);

  return data->sigma;
 }

MFNVector MFCloneExpansionNVector(void *d)
 {
  static char RoutineName[]={"MFCloneExpansionNVector"};
  struct IMFExpansionNVectorData *data;
  IMFExpansion Eclone;
  MFNVector result;
  MFKVector tmp;
  int i;

  data=(struct IMFExpansionNVectorData*)d;

  Eclone=IMFCloneExpansion(data->E);
  result=IMFCreateExpansionNVector(Eclone,data->t,data->sigma,data->prevChart,data->type);
  IMFFreeExpansion(Eclone);
  IMFExpansionNVSetChart0(result,data->chart0);
  if(data->s0!=(MFKVector)NULL)
   {
    tmp=MFCreateKVector(MFKV_NC(data->s0));
    for(i=0;i<MFKV_NC(data->s0);i++)MFKVSetC(tmp,i,MFKV_C(data->s0,i));
    IMFExpansionNVSetS0(result,tmp);
    MFFreeKVector(tmp);
   }
  return result;
 }

int IMFExpansionNVGetType(MFNVector this)
 {
  static char RoutineName[]={"IMFExpansionNVGetE"};
  struct IMFExpansionNVectorData *data;

  data=(struct IMFExpansionNVectorData*)MFNVectorGetData(this);

  return data->type;
 }

void IMFExpansionNVSetType(MFNVector this, int type)
 {
  static char RoutineName[]={"IMFExpansionNVGetE"};
  struct IMFExpansionNVectorData *data;

  data=(struct IMFExpansionNVectorData*)MFNVectorGetData(this);

  data->type=type;

  return;
 }

MFKVector IMFExpansionNVGetS0(MFNVector this)
 {
  static char RoutineName[]={"IMFExpansionNVGetS0"};
  struct IMFExpansionNVectorData *data;

  data=(struct IMFExpansionNVectorData*)MFNVectorGetData(this);

  return data->s0;
 }

void IMFExpansionNVSetS0(MFNVector this, MFKVector S)
 {
  static char RoutineName[]={"IMFExpansionNVSetS0"};
  struct IMFExpansionNVectorData *data;

  data=(struct IMFExpansionNVectorData*)MFNVectorGetData(this);

  MFRefKVector(S);
  if(data->s0!=(MFKVector)NULL)MFFreeKVector(data->s0);
  data->s0=S;
 }

int IMFExpansionNVGetChart0(MFNVector this)
 {
  static char RoutineName[]={"IMFExpansionNVGetChart0"};
  struct IMFExpansionNVectorData *data;

  data=(struct IMFExpansionNVectorData*)MFNVectorGetData(this);

  return data->chart0;
 }

void IMFExpansionNVSetChart0(MFNVector this, int chart0)
 {
  static char RoutineName[]={"IMFExpansionNVSetChart0"};
  struct IMFExpansionNVectorData *data;

  data=(struct IMFExpansionNVectorData*)MFNVectorGetData(this);

  data->chart0=chart0;
 }

int IMFExpansionNVGetPrev(MFNVector this)
 {
  static char RoutineName[]={"IMFExpansionNVGetPrev"};
  struct IMFExpansionNVectorData *data;

  data=(struct IMFExpansionNVectorData*)MFNVectorGetData(this);

  return data->prevChart;
 }

void IMFExpansionNVSetPrev(MFNVector this, int prevChart)
 {
  static char RoutineName[]={"IMFExpansionNVSetPrev"};
  struct IMFExpansionNVectorData *data;

  data=(struct IMFExpansionNVectorData*)MFNVectorGetData(this);

  data->prevChart=prevChart;
 }
