/*
    %W%
    %D% %T%
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   August 26, 1999 (MFExpression)      */
/*              February 25, 2003 changed name and added subroutine support  */

#include <MFImplicitMF.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <ExpCmp.h>
#include <MFFortran.h>

#ifndef MAX
#define MAX(X, Y) ((X) < (Y) ? (Y) : (X))
#endif

double *MFKV_CStar(MFKVector);
double *MFNV_CStar(MFNVector);
double *MFNKM_CStar(MFNKMatrix);

int MFFullNRealEV(int,double*);
int MFFullNPosSV(int,int,double*);
int MFFullNZeroSV(int,int,double*);
double MFFullProdEV(int,int,double*);
int MFFullNPosEV(int,int,double*);
int MFFullNUnitEV(int,int,double*);

static char MFAlgebraicMFErrorMsg[256]="";
void MFSetError(int,char*,char*,int,char*);

void MFFreeAlgebraicData(void*);
int MFProjectAlgebraic(int,int,MFNVector,MFNKMatrix,MFNVector,void*,int*);
int MFStopAlgebraic(MFImplicitMF,MFNVector,MFNKMatrix,MFNVector,MFNKMatrix,void*);
int MFTangentAlgebraic(int,int,MFNVector,MFNKMatrix,void*);
int MFTangentAlgebraicWithGuess(int,int,MFNVector,MFNKMatrix,MFNKMatrix,void*);
double MFScaleAlgebraic(int,int,MFNVector,MFNKMatrix,void*);
void MFWriteAlgebraicData(FILE*,void*);
MFImplicitMF MFReadAlgebraic(FILE*);
int MFSingularAlgebraic(int,int,MFNVector,MFNKMatrix,MFNVector,void*);

int MFAlgebraicProjectToSave(MFNVector,double*,void*);
int MFAlgebraicProjectToDraw(MFNVector,double*,void*);
int MFAlgebraicProjectForBB(MFNVector,double*,void*);
void MFAlgebraicSetStability(MFImplicitMF,MFNVector,MFNKMatrix,void*);

double MFAtlasDet(int,double*);

struct MFAlgebraicData
 {
  int expr;

  char *variables;
  char *expression;
  int n;
  int k;
  ECFn F;
  ECFn *dF;
  ECFn *ddF;
  void (*FSub)(int*,double*,int*,double*);
  void (*dFSub)(int*,double*,int*,double*);
 };

MFImplicitMF MFIMFCreateAlgebraicExpressionWithRadius(char *variables,char *expression,double R)
 {
  static char RoutineName[]={"MFIMFCreateAlgebraicExpressionWithRadius"};
  MFImplicitMF this;
  MFNSpace space;
  ECFn f;
  ECFn df;
  int i,j,n,k;
  struct MFAlgebraicData *data;

  f=ECCreateFunction(variables,expression);
  n=ECFunctionM(f);
  k=ECFunctionM(f)-ECFunctionN(f);

  this=MFIMFCreateBaseClass(n,k,"Algebraic");

  space=MFCreateNSpace(n);
  MFIMFSetSpace(this,space);
  MFFreeNSpace(space);

  data=(struct MFAlgebraicData*)malloc(sizeof(struct MFAlgebraicData)); /*done*/
  if(data==(void*)NULL)
   {
    sprintf(MFAlgebraicMFErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFAlgebraicData));
    printf("%s -- %s\n",RoutineName,MFAlgebraicMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFAlgebraicMFErrorMsg,__LINE__,__FILE__);
    free(this);
    return (MFImplicitMF)NULL;
   }

  data->variables=(char*)malloc((strlen(variables)+1)*sizeof(char));
  if(data->variables==(char*)NULL)
   {
    sprintf(MFAlgebraicMFErrorMsg,"Out of memory, trying to allocate %d bytes",(strlen(variables)+1)*sizeof(char));
    printf("%s -- %s\n",RoutineName,MFAlgebraicMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFAlgebraicMFErrorMsg,__LINE__,__FILE__);
    free(this);
    return (MFImplicitMF)NULL;
   }
  strcpy(data->variables,variables);

  data->expression=(char*)malloc((strlen(expression)+1)*sizeof(char));
  if(data->variables==(char*)NULL)
   {
    sprintf(MFAlgebraicMFErrorMsg,"Out of memory, trying to allocate %d bytes",(strlen(expression)+1)*sizeof(char));
    printf("%s -- %s\n",RoutineName,MFAlgebraicMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFAlgebraicMFErrorMsg,__LINE__,__FILE__);
    free(this);
    free(data->variables);
    return (MFImplicitMF)NULL;
   }
  strcpy(data->expression,expression);

  data->n=n;
  data->k=k;
  data->expr=1;
  data->F=f;
  data->dF=(ECFn*)malloc(n*sizeof(ECFn));
  if(data->dF==(ECFn*)NULL)
   {
    sprintf(MFAlgebraicMFErrorMsg,"Out of memory, trying to allocate %d bytes",n*sizeof(ECFn));
    printf("%s -- %s\n",RoutineName,MFAlgebraicMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFAlgebraicMFErrorMsg,__LINE__,__FILE__);
    free(this);
    free(data->variables);
    free(data->expression);
    return (MFImplicitMF)NULL;
   }

  data->ddF=(ECFn*)malloc(n*n*sizeof(ECFn));
  if(data->ddF==(ECFn*)NULL)
   {
    sprintf(MFAlgebraicMFErrorMsg,"Out of memory, trying to allocate %d bytes",n*n*sizeof(ECFn));
    printf("%s -- %s\n",RoutineName,MFAlgebraicMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFAlgebraicMFErrorMsg,__LINE__,__FILE__);
    free(this);
    free(data->variables);
    free(data->expression);
    free(data->dF);
    return (MFImplicitMF)NULL;
   }

  for(i=0;i<n;i++)
   {
    df=ECCreateDerivativeOfFunction(f,i);
    data->dF[i]=df;
    for(j=0;j<n;j++)
     {
      data->ddF[i+n*j]=ECCreateDerivativeOfFunction(df,j);
     }
   }

  data->FSub=NULL;
  data->dFSub=NULL;

  MFIMFSetData(this,(void*)data);
  MFIMFSetFreeData(this,MFFreeAlgebraicData);
  MFIMFSetStop(this,MFStopAlgebraic);
  MFIMFSetSingular(this,MFSingularAlgebraic);
  MFIMFSetProject(this,MFProjectAlgebraic);
  MFIMFSetTangent(this,MFTangentAlgebraic);
  MFIMFSetTangentWithGuess(this,MFTangentAlgebraicWithGuess);
  MFIMFSetScale(this,MFScaleAlgebraic);
  MFIMFSetR(this,R);
  MFIMFSetWriteData(this,MFWriteAlgebraicData);

  MFIMFSetProjectForSave(this,MFAlgebraicProjectToSave);
  MFIMFSetProjectForDraw(this,MFAlgebraicProjectToDraw);
  MFIMFSetProjectForBB(this,MFAlgebraicProjectForBB);

  MFIMFSetSetStability(this,MFAlgebraicSetStability);

  return this;
 }

MFImplicitMF MFIMFCreateAlgebraicExpression(char *variables,char *expression)
 {
  static char RoutineName[]={"MFIMFCreateAlgebraicExpression"};
  MFImplicitMF this;

  this=MFIMFCreateAlgebraicExpressionWithRadius(variables,expression,.1);

  return this;
 }

MFImplicitMF MFIMFCreateAlgebraicSubroutine(int n, int k, void (*FSub)(int*,double*,int*,double*), void (*dFSub)(int*,double*,int*,double*))
 {
  static char RoutineName[]={"MFIMFCreateAlgebraicSubroutine"};
  MFImplicitMF this;

  this=MFIMFCreateAlgebraicSubroutineWithRadius(n,k,FSub,dFSub,.1);

  return this;
 }

MFImplicitMF MFIMFCreateAlgebraicSubroutineWithRadius(int n, int k,void (*FSub)(int*,double*,int*,double*),
               void (*dFSub)(int*,double*,int*,double*), double R)
 {
  static char RoutineName[]={"MFIMFCreateAlgebraicSubroutineWithRadius"};
  MFImplicitMF this;
  MFNSpace space;
  int i,j;
  struct MFAlgebraicData *data;

  this=MFIMFCreateBaseClass(n,k,"Algebraic");

  space=MFCreateNSpace(n);
  MFIMFSetSpace(this,space);
  MFFreeNSpace(space);

  data=(struct MFAlgebraicData*)malloc(sizeof(struct MFAlgebraicData)); /*done*/
  if(data==(void*)NULL)
   {
    sprintf(MFAlgebraicMFErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFAlgebraicData));
    printf("%s -- %s\n",RoutineName,MFAlgebraicMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFAlgebraicMFErrorMsg,__LINE__,__FILE__);
    free(this);
    return (MFImplicitMF)NULL;
   }

  data->FSub=FSub;
  if(data->FSub==NULL)
   {
    sprintf(MFAlgebraicMFErrorMsg,"Function (arg 3), must be provided");
    printf("%s -- %s\n",RoutineName,MFAlgebraicMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFAlgebraicMFErrorMsg,__LINE__,__FILE__);
    free(this);
    return (MFImplicitMF)NULL;
   }

  data->dFSub=dFSub;
  if(data->dFSub==NULL)
   {
    sprintf(MFAlgebraicMFErrorMsg,"Jacobian (arg 4), must be provided");
    printf("%s -- %s\n",RoutineName,MFAlgebraicMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFAlgebraicMFErrorMsg,__LINE__,__FILE__);
    free(this);
    return (MFImplicitMF)NULL;
   }

  data->n=n;
  data->k=k;
  data->expr=0;

  MFIMFSetData(this,(void*)data);
  MFIMFSetFreeData(this,MFFreeAlgebraicData);
  MFIMFSetStop(this,MFStopAlgebraic);
  MFIMFSetSingular(this,MFSingularAlgebraic);
  MFIMFSetProject(this,MFProjectAlgebraic);
  MFIMFSetTangent(this,MFTangentAlgebraic);
  MFIMFSetScale(this,MFScaleAlgebraic);
  MFIMFSetR(this,R);
  MFIMFSetWriteData(this,MFWriteAlgebraicData);

  MFIMFSetProjectForSave(this,MFAlgebraicProjectToSave);
  MFIMFSetProjectForDraw(this,MFAlgebraicProjectToDraw);
  MFIMFSetProjectForBB(this,MFAlgebraicProjectForBB);

  MFIMFSetSetStability(this,MFAlgebraicSetStability);

  return this;
 }

void MFFreeAlgebraicData(void *d)
 {
  static char RoutineName[]={"MFFreeAlgebraicData"};
  struct MFAlgebraicData *data;
  int i,n;

  data=(struct MFAlgebraicData*)d;

  if(data->expr)
   {
    if(data->F!=(ECFn)NULL)ECFreeFunction(data->F);

    n=data->n;
    if(data->dF!=(ECFn*)NULL)
     {
      for(i=0;i<n;i++)
        if((data->dF)[i]!=(ECFn)NULL)ECFreeFunction(data->dF[i]);
      free(data->dF);
     }

    if(data->ddF!=(ECFn*)NULL)
     {
      for(i=0;i<n*n;i++)
        if((data->ddF)[i]!=(ECFn)NULL)ECFreeFunction(data->ddF[i]);
      free(data->ddF);
     }
   }

  if(data->variables!=(char*)NULL)free(data->variables);
  if(data->expression!=(char*)NULL)free(data->expression);
  free(data);
  return;
 }

void MFWriteAlgebraicData(FILE *fid,void *d)
 {
  static char RoutineName[]={"MFWriteAlgebraicData"};
  struct MFAlgebraicData *data;

  data=(struct MFAlgebraicData*)d;

  fprintf(fid,"%d %d %d\n",data->n,data->k,data->expression);
  if(data->expr)
   {
    fprintf(fid,"%d %d\n",strlen(data->variables),strlen(data->expression));
    fprintf(fid,"%s\n",data->variables);
    fprintf(fid,"%s\n",data->expression);
   }
  fflush(fid);
  return;
 }

MFImplicitMF MFReadAlgebraic(FILE *fid)
 {
  static char RoutineName[]={"MFReadAlgebraic"};
  MFImplicitMF expressionMF=(MFImplicitMF)NULL;
  int lnvariables=0;
  int lnexpression=0;
  char *variables;
  char *expression;
  int i;
  int n,k,flag;
  char c=' ';

  fscanf(fid,"%d %d\n",&n,&k,&flag);
  if(flag)
   {
    fscanf(fid,"%d %d\n",&lnvariables,&lnexpression);
    variables=(char*)malloc((lnvariables+1)*sizeof(char));
    expression=(char*)malloc((lnexpression+1)*sizeof(char));
    for(i=0;i<lnvariables;i++)fscanf(fid,"%c",variables+i);
    variables[lnvariables]=0x0;
    fscanf(fid,"%c",&c);
    for(i=0;i<lnexpression;i++)fscanf(fid,"%c",expression+i);
    fscanf(fid,"%c",&c);
    expression[lnexpression]=0x0;
   }else{
    sprintf(MFAlgebraicMFErrorMsg,"You can't read an algebraic manifold defined by subroutines");
    printf("%s -- %s\n",RoutineName,MFAlgebraicMFErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFAlgebraicMFErrorMsg,__LINE__,__FILE__);
    return (MFImplicitMF)NULL;
   }

  expressionMF=MFIMFCreateAlgebraicExpression(variables,expression);
  free(variables);
  free(expression);

  return expressionMF;
 }

int MFProjectAlgebraic(int n,int k,MFNVector vu0,MFNKMatrix mPhi,MFNVector vu,void *d,int *index)
 {
  static char RoutineName[]={"MFProjectAlgebraic"};
  double err;
  double *A=(double*)NULL; /* [n*n];*/
  double *a=(double*)NULL; /* [n*n];*/
  double *b=(double*)NULL; /* [n];  */
  int *ipvt=(int*)NULL;    /* [n];  */
  int one=1;
  int info=0;
  int itimes;
  int i,j,l,m;
  char trans='N';
  struct MFAlgebraicData *data;
  double *f=(double*)NULL;
  double *df=(double*)NULL;
  double t,dnorm,unorm;
  double *u0,*Phi,*u;
  int verbose;

  verbose=0;

  u0=MFNV_CStar(vu0);
  Phi=MFNKM_CStar(mPhi);
  u=MFNV_CStar(vu);

  data=(struct MFAlgebraicData*)d;

  A=malloc(n*n*sizeof(double));
  a=malloc(n*n*sizeof(double));
  b=malloc(n*sizeof(double));
  ipvt=malloc(n*sizeof(int));
  f=malloc((n-k)*sizeof(double));
  df=malloc(n*(n-k)*sizeof(double));

  for(i=0;i<n;i++)u[i]=u0[i];

  err=1.;
  itimes=0;
  if(verbose)
   {
    printf("%s\n",RoutineName);fflush(stdout);
    printf("u0:(%lf",u[0]);for(l=1;l<n;l++)printf(",%lf",u[l]);printf(")\n");fflush(stdout);
   }
  while(err>1.e-10)
   {
    if(data->expr)
     {
      for(j=0;j<n;j++)
       {
        ECEvaluateFunction(data->dF[j],u,f);
        for(i=0;i<n-k;i++)A[i+n*j]=f[i];
       }
      ECEvaluateFunction(data->F,u,f);
     }else{
      m=n-k;
      (data->FSub)(&n,u,&m,f);
      (data->dFSub)(&n,u,&m,df);
      for(i=0;i<m;i++)
       for(j=0;j<n;j++)A[i+n*j]=df[i+m*j];
     }

    if(0&&verbose){printf("f:(%lf",f[0]);for(l=1;l<n-k;l++)printf(",%lf",f[l]);printf(")\n");fflush(stdout);}

    for(i=0;i<n-k;i++)b[i]=-f[i];

    for(i=n-k;i<n;i++)b[i]=0.;
    for(j=0;j<n;j++)
     {
      for(i=0;i<n-k;i++)a[i+n*j]=A[i+n*j];
      for(i=n-k;i<n;i++)
       {
        A[i+n*j]=Phi[j+n*(i-n+k)];
        b[i]-=Phi[j+n*(i-n+k)]*(u[j]-u0[j]);
       }
     }

    for(j=0;j<n;j++)
      for(i=0;i<n;i++)
        a[i+n*j]=A[i+n*j];

    if(0&&verbose)
     {
      printf("A:\n");
      for(i=0;i<n;i++)
       {
        printf(" [%lf",A[i+n*0]);for(l=1;l<n;l++)printf(",%lf",A[i+n*l]);printf("]\n");fflush(stdout);
       }
      printf("u: (%lf",u[0]);for(l=1;l<n;l++)printf(",%lf",u[l]);printf(")\n");fflush(stdout);
      printf("b: (%lf",b[0]);for(l=1;l<n;l++)printf(",%lf",b[l]);printf(")\n");fflush(stdout);
     }

    err=0.;
    for(i=0;i<n;i++)err+=b[i]*b[i];
    err=sqrt(err);

    F77_FUNC(dgetrf,DGETRF)(&n,&n,A,&n,ipvt,&info);
    F77_FUNC(dgetrs,DGETRS)(&trans,&n,&one,A,&n,ipvt,b,&n,&info);

    t=1.;
    dnorm=fabs(b[0]);
    unorm=fabs(u[0]);
    for(i=1;i<n;i++)
     {
      if(fabs(b[i])>dnorm)dnorm=fabs(b[i]);
      if(fabs(u[i])>unorm)unorm=fabs(u[i]);
     }
    if(dnorm>.1*(unorm+1))t=.1*(unorm+1)/dnorm; /* Damping */
    if(verbose){printf("  %d %le %lf Du=%le, u=%le\n",itimes,err,t,dnorm,unorm);fflush(stdout);}

    for(i=0;i<n;i++)u[i]=u[i]+t*b[i];
    itimes++;
    if(itimes>10)
     {
      free(A);
      free(a);
      free(b);
      free(ipvt);
      free(f);
      free(df);

      if(verbose){printf("done %s. Too many iterations.\n",RoutineName);fflush(stdout);}
      return 0;
     }
   }

  if(0&&verbose)
   {
    printf("u:(%lf",u[0]);for(l=1;l<n;l++)printf(",%lf",u[l]);printf(")\n");fflush(stdout);
    printf("A:\n");
    for(i=0;i<n;i++)
     {
      printf(" [%lf",A[i+n*0]);for(l=1;l<n;l++)printf(",%lf",A[i+n*l]);printf("]\n");fflush(stdout);
     }
    printf("\n");
   }

  *index=MFFullNPosEV(n,k,a);
  if(verbose){printf("%s, niter=%d, index is %d\n",RoutineName,itimes,*index);fflush(stdout);}

  free(A);
  free(a);
  free(b);
  free(ipvt);
  free(f);
  free(df);

  if(verbose){printf("done %s\n",RoutineName);fflush(stdout);}
  return 1;
 }

int MFTangentAlgebraic(int n,int k,MFNVector vu,MFNKMatrix mPhi,void *d)
 {
  static char RoutineName[]={"MFTangentAlgebraic"};
  double *A;             /* n*n */
  char jobvl='N';                   /* No left  eigenvectors */
  char jobvr='A';                   /*    right eigenvectors */
  int lda;                        /* Leading dimension of A */
  double *s;    /* n */
  double *U;    /* n*n */
  int ldU;
  double *vr;    /* n*k */
  double *V;    /* n*n */
  int ldV;
  double *work;                     /* Work array */
  int lwork;                     /* Length of work array */
  int info=0;
  int i,j,l,m,mm;
  double t;
  struct MFAlgebraicData *data;
  double *f=(double*)NULL;
  double *df=(double*)NULL;
  double *u,*Phi;
  int verbose;
  double *a;

  verbose=0;

  if(verbose){printf("%s, k=%d, n=%d\n",RoutineName,k,n);fflush(stdout);}

  Phi=MFNKM_CStar(mPhi);
  u=MFNV_CStar(vu);

  data=(struct MFAlgebraicData*)d;
  if(verbose){printf("%s, %s\n",data->variables,data->expression);fflush(stdout);}

  A=malloc(n*n*sizeof(double));
  a=malloc(n*n*sizeof(double));
  lda=n;
  ldU=n;

  U=malloc(n*n*sizeof(double));
  if(U==(double*)NULL)abort();

  ldV=n;
  V=malloc(n*n*sizeof(double));
  if(V==(double*)NULL)abort();

  vr=malloc(n*k*sizeof(double));
  if(vr==(double*)NULL)abort();

  f=malloc((n-k)*sizeof(double));
  if(f==(double*)NULL)abort();

  df=malloc(n*(n-k)*sizeof(double));
  if(df==(double*)NULL)abort();

  s=(double*)malloc(n*sizeof(double));
  if(s==(double*)NULL)abort();

  if(data->expr)
   {
    for(j=0;j<n;j++)
     {
      ECEvaluateFunction(data->dF[j],u,f);
      for(i=0;i<n-k;i++)A[i+n*j]=f[i];
     }
   }else{
    m=n-k;
    (data->dFSub)(&n,u,&m,df);
    for(i=0;i<m;i++)
      for(j=0;j<n;j++)A[i+n*j]=df[i+m*j];
   }
  for(j=0;j<n;j++)
    for(i=n-k;i<n;i++)A[i+n*j]=0.;
  for(j=0;j<n;j++)
    for(i=0;i<n;i++)a[i+n*j]=A[i+n*j];

  if(verbose)
   {
    printf("u:(%lf",u[0]);for(l=1;l<n;l++)printf(",%lf",u[l]);printf(")\n");fflush(stdout);
    printf("A:\n");
    for(i=0;i<n;i++)
     {
      printf(" [%lf",A[i+n*0]);for(l=1;l<n;l++)printf(",%lf",A[i+n*l]);printf("]\n");fflush(stdout);
     }
    printf("\n");
   }

#ifdef MFDEBUG
  _heap_check();
#endif

  i=-1;
  info=0;
  jobvl='N';
  jobvr='A';
  F77_FUNC(dgesvd,DGESVD)(&jobvl,&jobvr,&n,&n,A,&lda,s,U,&ldU,V,&ldV,&t,&i,&info);
  lwork=MAX(4*n,5*n-4);
  lwork=t;

  work=malloc(lwork*sizeof(double));
  if(work==(double*)NULL)abort();

  info=0;
  jobvl='N';
  jobvr='A';
  F77_FUNC(dgesvd,DGESVD)(&jobvl,&jobvr,&n,&n,A,&lda,s,U,&ldU,V,&ldV,work,&lwork,&info);

  if(verbose){printf("\n      Null Vectors\n");fflush(stdout);}
  m=0;
  for(i=0;i<n;i++)
   {
    if(fabs(s[i])<1.e-7)
     {
      if(verbose){printf("     Phi[%d]=v[%d]=(%lf",m,i,V[0+n*i]);for(j=1;j<n;j++)printf(",%lf",V[i+n*j]);printf(")\n");fflush(stdout);}
      for(j=0;j<n;j++)vr[j+n*m]=V[i+n*j];
      m++;
     }
   }

  if(m>k)
   {
    sprintf(MFAlgebraicMFErrorMsg,"Too many tangent vectors (%d when k is only %d) - Jacobian is not full rank!",m,k);
    printf("%s -- %s\n",RoutineName,MFAlgebraicMFErrorMsg);fflush(stdout);
    printf("Singular Values [%le",s[0]);for(i=1;i<n;i++)printf(",%le",s[i]);printf("]\n");fflush(stdout);
    printf("U: ");MFPrintNVector(stdout,vu);printf("\n");fflush(stdout);
    MFSetError(4,RoutineName,MFAlgebraicMFErrorMsg,__LINE__,__FILE__);
    return 0;
   }

  if(verbose){printf("\n      Gram-Schmidt\n");fflush(stdout);}
#ifdef MFDEBUG
  _heap_check();
#endif
  t=0.;
  for(j=0;j<n;j++)t+=vr[j]*vr[j];
  if(fabs(t)>1.e-15)
   {
    t=1./sqrt(t);
    for(j=0;j<n;j++){vr[j]=vr[j]*t;Phi[j+n*0]=vr[j];}
    if(verbose){printf("\n     Phi[%d]=(%lf",0,Phi[0+n*0]);for(j=1;j<n;j++)printf(",%lf",Phi[j+n*0]);printf(")\n");fflush(stdout);}
   }

#ifdef MFDEBUG
  _heap_check();
#endif
  for(i=1;i<m;i++)
   {
    for(l=0;l<i;l++)
     {
      t=0.;
      for(j=0;j<n;j++)t+=vr[j+n*l]*vr[j+n*i];
      for(j=0;j<n;j++)vr[j+n*i]=vr[j+n*i]-t*vr[j+n*l];
     }

    t=0.;
    for(j=0;j<n;j++)t+=vr[j+n*i]*vr[j+n*i];
    if(fabs(t)>1.e-15)
     {
      t=1./sqrt(t);
      for(j=0;j<n;j++){vr[j+n*i]=vr[j+n*i]*t;Phi[j+n*i]=vr[j+n*i];}
      if(verbose){printf("     Phi[%d]=(%lf",i,Phi[0+n*i]);for(j=1;j<n;j++)printf(",%lf",Phi[j+n*i]);printf(")\n");fflush(stdout);}
     }
   }

#ifdef MFDEBUG
  _heap_check();
#endif
  if(verbose)
   {
    printf("\n      test Gram-Schmidt\n");
    printf("\n");
    for(i=0;i<k;i++)
     {
      printf("[");
      for(j=0;j<k;j++)
       {
        t=0.;
        for(m=0;m<n;m++)t+=Phi[m+n*i]*Phi[m+n*j];
        if(j>0)printf(" ");
        printf("%lf",t);
       }
      printf("]\n");
     }
    fflush(stdout);
   }

  free(a);
  free(A);
  free(U);
  free(V);
  free(s);
  free(vr);
  free(work);
  free(f);
  free(df);

  if(verbose){printf("done %s\n",RoutineName);fflush(stdout);}
  return 1;
 }

extern double MFEpsilon;

double MFScaleAlgebraic(int n,int k,MFNVector vu,MFNKMatrix mPhi,void *d)
 {
  static char RoutineName[]={"MFScaleAlgebraic"};
  double r=1.;
  double *A;
  double *b;
  double *c;
  double *u,*Phi;

  char jobvl='N';                   /* No left  eigenvectors */
  char jobvr='V';                   /*    right eigenvectors */
  double *wr;    /* k */            /* Array to hold real part of eigenvalues */
  double *wi;    /* k */            /* Array to hold imag part of eigenvalues */
  int ldvl;                       /* Leading dimension of vl */
  double *vl;   /* k*k */            /* Array to hold left eigenvectors */
  int ldvr;                       /* Leading dimension of vr */
  double *vr;   /* k*k */           /* Array to hold right eigenvectors */
  double *work;    /* k*k */        /* Work array */
  int lwork;                     /* Length of work array */
  int info=0;

  double cn;
  double tu=0.;
  double t1=0.;
  double t2=0.;
  double t3=0.;
  double su=0.;
  double s1=0.;
  double s2=0.;
  double s3=0.;
  int i,j,l;
  int ii,jj;
  int *ipvt;
  char trans='N';
  int one=1;
  struct MFAlgebraicData *data;
  double *f=(double*)NULL;
  double *df=(double*)NULL;
  int m;
  int verbose;

  verbose=0;

  if(verbose){printf("%s, k=%d, n=%d\n",RoutineName,k,n);fflush(stdout);}

  Phi=MFNKM_CStar(mPhi);
  u=MFNV_CStar(vu);

  data=(struct MFAlgebraicData*)d;

  A=malloc(n*n*sizeof(double));
  b=malloc(n*sizeof(double));
  ipvt=malloc(n*sizeof(int));
  f=malloc((n-k)*sizeof(double));
  df=malloc(n*(n-k)*sizeof(double));

  c=malloc(k*k*sizeof(double));
  wr=malloc(k*sizeof(double));
  wi=malloc(k*sizeof(double));
  ldvl=k;
  vl=malloc(k*k*sizeof(double));
  ldvr=k;
  vr=malloc(k*k*sizeof(double));
  lwork=4*k;
  work=malloc(lwork*sizeof(double));

  if(verbose)
   {printf("u:(%lf",u[0]);for(l=1;l<n;l++)printf(",%lf",u[l]);printf(")\n");fflush(stdout);}
  if(data->expr)
   {
    for(j=0;j<n;j++)
     {
      ECEvaluateFunction(data->dF[j],u,f);
      for(i=0;i<n-k;i++)A[i+n*j]=f[i];
     }
   }else{
    m=n-k;
    (data->dFSub)(&n,u,&m,df);
    for(i=0;i<m;i++)
     for(j=0;j<n;j++)A[i+n*j]=df[i+m*j];
   }
  for(j=0;j<n;j++)
    for(i=n-k;i<n;i++)
      A[i+n*j]=Phi[j+n*(i-n+k)];

  if(verbose)
   {
    printf("A:\n");
    for(i=0;i<n;i++)
     {
      printf(" [%lf",A[i+n*0]);for(l=1;l<n;l++)printf(",%lf",A[i+n*l]);printf("]\n");fflush(stdout);
     }
    printf("\n");
   }

  if(verbose){printf("    factor\n");fflush(stdout);}
  F77_FUNC(dgetrf,DGETRF)(&n,&n,A,&n,ipvt,&info);
  if(verbose){printf("    done factor\n");fflush(stdout);}

  for(i=0;i<k;i++)
   {
    for(j=i;j<k;j++)
     {
      if(verbose){printf("    (%d,%d) ",i,j);fflush(stdout);}
      for(l=0;l<n;l++)b[l]=0.;
      if(verbose)
       {
        printf("\n Phi[%d]=(%lf",i,Phi[0+n*i]);for(l=1;l<n;l++)printf(",%lf",Phi[l+n*i]);printf(")\n");fflush(stdout);
        printf(" Phi[%d]=(%lf",j,Phi[0+n*j]);for(l=1;l<n;l++)printf(",%lf",Phi[l+n*j]);printf(")\n");fflush(stdout);
       }
      for(ii=0;ii<n;ii++)
       {
        for(jj=0;jj<n;jj++)
         {
          if(verbose)
           {
            printf(" dF/dx_%d/dx_%d(%lf",ii,jj,u[0]);
            for(l=1;l<n;l++)printf(",%lf",u[l]);
            printf(") = (%lf",f[0]);
            for(l=1;l<n-k;l++)printf(",%lf",f[l]);
            printf(")\n");
            fflush(stdout);
           }
          for(l=0;l<n-k;l++)b[l]-=f[l]*Phi[ii+n*i]*Phi[jj+n*j];
         }
       }

      if(verbose)
       {
        printf("      backsolve ");fflush(stdout);
        printf(" rhs=(%lf",b[0]);for(l=1;l<n;l++)printf(",%lf",b[l]);printf("), info=%d\n",info);fflush(stdout);
       }
      F77_FUNC(dgetrs,DGETRS)(&trans,&n,&one,A,&n,ipvt,b,&n,&info);
      if(verbose)
       {
        printf(" b=(%lf",b[0]);for(l=1;l<n;l++)printf(",%lf",b[l]);printf("), info=%d\n",info);fflush(stdout);
       }

      c[i+k*j]=0.;
      for(l=0;l<n;l++)c[i+k*j]+=b[l]*b[l];
      c[i+k*j]=sqrt(c[i+k*j]);
      c[j+k*i]=c[i+k*j];
      if(verbose){printf("c[%d,%d]=%lf\n",i,j,c[i+k*j]);fflush(stdout);}
     }
   }

  if(verbose)
   {
    printf("c:\n");
    for(i=0;i<k;i++)
     {
      printf(" [%lf",c[i+k*0]);for(l=1;l<k;l++)printf(",%lf",c[i+k*l]);printf("]\n");fflush(stdout);
     }
    printf("\n");
   }

  info=0;
  jobvl='N';
  jobvr='N';
  if(verbose)
   {
    printf("    find largest eigenvalue\n");fflush(stdout);
    printf("      call dgeev\n");fflush(stdout);
   }
  F77_FUNC(dgeev,DGEEV)(&jobvl,&jobvr,&k,c,&k,wr,wi,vl,&ldvl,vr,&ldvr,work,&lwork,&info);
  if(verbose){printf("      info from dgeev is %d\n",info);fflush(stdout);}
  if(verbose){printf("      find max\n");fflush(stdout);}

  cn=sqrt(wr[0]*wr[0]+wi[0]*wi[0]);
  if(verbose){printf("ScaleAlgebraic, ev= %lf",cn);}
  for(l=1;l<k;l++)
   {
    tu=sqrt(wr[l]*wr[l]+wi[l]*wi[l]);
    if(tu>cn)cn=tu;
    if(verbose){printf(",%lf",tu);}
   }

  r=sqrt(2*MFEpsilon/cn);
  r=10.*r;
  if(verbose){printf(", max ev=%lf, r=%lf\n",cn,r);fflush(stdout);}

  free(A);
  free(b);
  free(c);
  free(ipvt);
  free(f);
  free(df);
  free(wr);
  free(wi);
  free(vl);
  free(vr);
  free(work);

  if(verbose){printf("done %s\n",RoutineName);fflush(stdout);}
  return r;
 }

int MFStopAlgebraic(MFImplicitMF this,MFNVector u0,MFNKMatrix Phi0,MFNVector u1,MFNKMatrix Phi1,void *d)
 {
  static char RoutineName[]={"MFStopAlgebraic"};
  int verbose=0;

/*MFAlgebraicSetStability(this,u1,Phi0,d);*/

  if(verbose)printf("%s index 0=%d, index 1=%d\n",RoutineName,MFNVGetIndex(u0),MFNVGetIndex(u1));fflush(stdout);

  if(MFNVGetIndex(u0)!=MFNVGetIndex(u1))return 1;
   else return 0;
 }

int MFFullNRealEV(int n,double *A)
 {
  static char RoutineName[]={"MFFullNRealEV"};
  int i,j;
  int lda;
  char jobvl='N';                   /* No left  eigenvectors */
  char jobvr='V';                   /*    right eigenvectors */
  static double *wr=(double*)NULL;    /* k */            /* Array to hold real part of eigenvalues */
  static double *wi=(double*)NULL;    /* k */            /* Array to hold imag part of eigenvalues */
  int ldvl;                       /* Leading dimension of vl */
  static double *vl=(double*)NULL;   /* k*k */            /* Array to hold left eigenvectors */
  int ldvr;                       /* Leading dimension of vr */
  static double *vr=(double*)NULL;   /* k*k */           /* Array to hold right eigenvectors */
  static double *work=(double*)NULL;    /* k*k */        /* Work array */
  int lwork;                     /* Length of work array */
  int info=0;
  int verbose=0;
  int result;

  if(verbose)
   {
    for(i=0;i<n;i++)
     {
      printf("[");
      for(j=0;j<n;j++)
       {
        if(j>0)printf(" ");
        printf("%lf",A[i+n*j]);
       }
      printf("]\n");
     }
   }

  lda=n;
  wr=(double*)realloc((void*)wr,n*sizeof(double));
  wi=(double*)realloc((void*)wi,n*sizeof(double));
  ldvl=n;
  vl=(double*)realloc((void*)vl,n*n*sizeof(double));
  ldvr=n;
  vr=(double*)realloc((void*)vr,n*n*sizeof(double));
  lwork=4*n;
  work=(double*)realloc((void*)work,lwork*sizeof(double));

  info=0;
  jobvl='N';
  jobvr='V';
  if(verbose){printf("      call dgeev\n");fflush(stdout);}
  F77_FUNC(dgeev,DGEEV)(&jobvl,&jobvr,&n,A,&lda,wr,wi,vl,&ldvl,vr,&ldvr,work,&lwork,&info);
  if(verbose){printf("      done dgeev\n");fflush(stdout);}

  if(verbose){printf("%s\n",RoutineName);}
  result=0;
  for(i=0;i<n;i++)
   {
    if(verbose){printf("   %d (%le,%le)\n",i,wr[i],wi[i]);fflush(stdout);}
    if(wr[i]>0.)result++;
   }
  if(verbose){printf("  number with positive real part =%d\n",result);fflush(stdout);}

  return result;
 }

double *MFNV_CStar(MFNVector);
void MFAlgebraicSetStability(MFImplicitMF M,MFNVector u, MFNKMatrix mPhi,void *d)
 {
  static char RoutineName[]={"MFAlgebraicSetStability"};
  static double *A=(double*)NULL; /* [n*n];*/
  struct MFAlgebraicData *data;
  static double *f=(double*)NULL; /* [k];  */
  static double *df=(double*)NULL; /* [k];  */
  int n,k;
  int i,j,l,m;
  int result;
  int verbose=0;
  double *Phi;

  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

  n=MFIMF_N(M);
  k=MFIMF_K(M);
  data=(struct MFAlgebraicData*)MFIMFGetData(M);

  A=(double*)realloc((void*)A,n*n*sizeof(double));
  f=(double*)realloc((void*)f,(n-k)*sizeof(double));
  df=(double*)realloc((void*)df,(n-k)*n*sizeof(double));
  Phi=MFNKM_CStar(mPhi);

  if(data->expr)
   {
    for(j=0;j<n;j++)
     {
      ECEvaluateFunction(data->dF[j],MFNV_CStar(u),f);
      for(i=0;i<n-k;i++)A[i+n*j]=f[i];
     }
   }else{
    m=n-k;
    (data->dFSub)(&n,MFNV_CStar(u),&m,df);
    for(i=0;i<n-k;i++)
     for(j=0;j<n;j++)A[i+n*j]=df[i+(n-k)*j];
   }
  for(j=0;j<n;j++)
    for(i=0;i<k;i++)A[n-k+i+n*j]=Phi[j+n*i];

  result=MFFullNPosEV(n,k,A);
  MFNVSetIndex(u,result);
  if(verbose){printf("index is %d\n",result);fflush(stdout);}
  return;
 }

int MFAlgebraicProjectToSave(MFNVector u, double *x, void *d)
 {
  int i;
  struct MFAlgebraicData *data;

  data=(struct MFAlgebraicData*)d;

  if(x==(double*)NULL)return data->n;
  for(i=0;i<data->n;i++)x[i]=MFNV_C(u,i);

  return 0;
 }

int MFAlgebraicProjectToDraw(MFNVector u, double *x, void *d)
 {
  int i;
  struct MFAlgebraicData *data;

  data=(struct MFAlgebraicData*)d;

  if(x==(double*)NULL)return data->n;
  for(i=0;i<data->n;i++)x[i]=MFNV_C(u,i);

  return 0;
 }

int MFAlgebraicProjectForBB(MFNVector u, double *x, void *d)
 {
  int i;
  struct MFAlgebraicData *data;

  data=(struct MFAlgebraicData*)d;

  if(x==(double*)NULL)return data->n;
  for(i=0;i<data->n;i++)x[i]=MFNV_C(u,i);

  return 0;
 }

int MFSingularAlgebraic(int n,int k,MFNVector vu,MFNKMatrix mPhi,MFNVector perp,void *d)
 {
  static char RoutineName[]={"MFSingularAlgebraic"};
  double *A;             /* n*n */
  char jobvl='N';                   /* No left  eigenvectors */
  char jobvr='V';                   /*    right eigenvectors */
  int lda;                        /* Leading dimension of A */
  double *wr;    /* n */            /* Array to hold real part of eigenvalues */
  double *wi;    /* n */            /* Array to hold imag part of eigenvalues */
  int ldvl;                       /* Leading dimension of vl */
  double *vl;   /* n*n */            /* Array to hold left eigenvectors */
  int ldvr=4;                       /* Leading dimension of vr */
  double *vr;   /* n*n */           /* Array to hold right eigenvectors */
  double *work;    /* n*n */        /* Work array */
  int lwork;                     /* Length of work array */
  int info=0;
  int i,j,l,m,mm;
  double t;
  struct MFAlgebraicData *data;
  double *f=(double*)NULL; /* [k];  */
  double *u,*Phi,*phi;
  int verbose;
  double *s;    /* n */
  double *U;    /* n*n */
  int ldU;
  double *V;    /* n*n */
  int ldV;
  double *df=(double*)NULL;

/* Finds a NULL vector at u, perpendicular to the tangent plane Phi (assuming the Jacobian is singular) */

  verbose=0;

  if(verbose){printf("%s, k=%d, n=%d\n",RoutineName,k,n);fflush(stdout);}

  Phi=MFNKM_CStar(mPhi);
  u=MFNV_CStar(vu);
  phi=MFNV_CStar(perp);

  data=(struct MFAlgebraicData*)d;

  A=malloc(n*n*sizeof(double));
  lda=n;
  ldU=n;

  U=malloc(n*n*sizeof(double));
  if(U==(double*)NULL)abort();

  ldV=n;
  V=malloc(n*n*sizeof(double));
  if(V==(double*)NULL)abort();
  ldvr=n;
  vr=malloc(n*n*sizeof(double));
  f=malloc((n-k)*sizeof(double));
  if(f==(double*)NULL)abort();

  df=malloc(n*(n-k)*sizeof(double));
  if(df==(double*)NULL)abort();
  s=(double*)malloc(n*sizeof(double));
  if(s==(double*)NULL)abort();

  if(data->expr)
   {
    if(verbose){printf("%s, %s\n",data->variables,data->expression);fflush(stdout);}
    for(j=0;j<n;j++)
     {
      ECEvaluateFunction(data->dF[j],u,f);
      for(i=0;i<n-k;i++)A[i+n*j]=f[i];
     }
   }else{
    if(verbose){printf("Functions\n");fflush(stdout);}
    m=n-k;
    (data->dFSub)(&n,u,&m,df);
    for(i=0;i<n-k;i++)
     for(j=0;j<n;j++)A[i+n*j]=df[i+(n-k)*j];
   }
  for(j=0;j<n;j++)
    for(i=0;i<k;i++)A[n-k+i+n*j]=Phi[j+n*i];

  if(verbose)
   {
    printf("u:(%lf",u[0]);for(l=1;l<n;l++)printf(",%lf",u[l]);printf(")\n");fflush(stdout);
    printf("Phi:\n");MFPrintNKMatrix(stdout,mPhi);
    printf("A:\n");
    for(i=0;i<n;i++)
     {
      printf(" [%lf",A[i+n*0]);for(l=1;l<n;l++)printf(",%lf",A[i+n*l]);printf("]\n");fflush(stdout);
     }
    printf("\n");
   }

#ifdef MFDEBUG
  _heap_check();
#endif

  i=-1;
  info=0;
  jobvl='N';
  jobvr='A';
  F77_FUNC(dgesvd,DGESVD)(&jobvl,&jobvr,&n,&n,A,&lda,s,U,&ldU,V,&ldV,&t,&i,&info);
  lwork=MAX(4*n,5*n-4);
  lwork=t;

  work=malloc(lwork*sizeof(double));
  if(work==(double*)NULL)abort();
  for(i=0;i<lwork;i++)work[i]=0.;

  info=0;
  jobvl='N';
  jobvr='A';
  F77_FUNC(dgesvd,DGESVD)(&jobvl,&jobvr,&n,&n,A,&lda,s,U,&ldU,V,&ldV,work,&lwork,&info);

  if(verbose){printf("\n      Null Vectors\n");fflush(stdout);}
  m=0;
  for(i=0;i<n;i++)
   {
    if(verbose){printf("  s[%d]=%lf,   v[%d]=(%lf",i,s[i],i,V[i+n*0]);for(j=1;j<n;j++)printf(",%lf",V[i+n*j]);printf(")\n");fflush(stdout);}
    if(fabs(s[i])<1.e-3)
     {
      for(j=0;j<n;j++)vr[j+n*m]=V[i+n*j];
      m++;
     }
   }

  if(m<1)
   {
/*  sprintf(MFAlgebraicMFErrorMsg,"Jacobian is full rank!");
    for(i=0;i<n;i++)
      printf("  s[%d]=%lf\n",i,s[i]);
    printf("%s -- %s\n",RoutineName,MFAlgebraicMFErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFAlgebraicMFErrorMsg,__LINE__,__FILE__);*/
    free(A);
    free(U);
    free(V);
    free(vr);
    free(s);
    free(work);
    free(f);
    free(df);
    return 0;
   }

  if(m>1)
   {
    sprintf(MFAlgebraicMFErrorMsg,"Jacobian is rank deficient by more than one! m=%d",m);
    printf("%s -- %s\n",RoutineName,MFAlgebraicMFErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFAlgebraicMFErrorMsg,__LINE__,__FILE__);
    free(A);
    free(U);
    free(V);
    free(vr);
    free(s);
    free(work);
    free(f);
    free(df);
    return 0;
   }

  t=0.;
  for(j=0;j<n;j++)t+=vr[j]*vr[j];
  if(fabs(t)>1.e-15)
   {
    t=1./sqrt(t);
    for(j=0;j<n;j++){vr[j]=vr[j]*t;phi[j]=vr[j];}
    if(verbose){printf("     phi=(%lf",phi[0]);for(j=1;j<n;j++)printf(",%lf",phi[j]);printf(")\n");fflush(stdout);}
   }

  free(A);
  free(U);
  free(V);
  free(vr);
  free(s);
  free(work);
  free(f);
  free(df);

  if(verbose){printf("done %s\n",RoutineName);fflush(stdout);}
  return 1;
 }

int MFFullNPosSV(int n,int k, double *A)
 {
  static char RoutineName[]={"MFFullNPosSV"};
  int i,j;
  int lda;
  char jobvl='N';
  char jobvr='N';
  double *s=(double*)NULL;
  double *U=(double*)NULL;
  double *V=(double*)NULL;
  double *work=(double*)NULL;
  double t;
  int lwork;
  int info=0;
  int verbose=0;
  int result;

  if(verbose){printf("%s\n",RoutineName);}

  if(0&&verbose)
   {
    for(i=0;i<n;i++)
     {
      printf("[");
      for(j=0;j<n;j++)
       {
        if(j>0)printf(" ");
        printf("%lf",A[i+n*j]);
       }
      printf("]\n");
     }
   }

  s=(double*)malloc(n*sizeof(double));
  U=(double*)malloc(n*n*sizeof(double));
  V=(double*)malloc(n*n*sizeof(double));
  i=-1;
  info=0;
  jobvl='N';
  jobvr='A';
  F77_FUNC(dgesvd,DGESVD)(&jobvl,&jobvr,&n,&n,A,&n,s,U,&n,V,&n,&t,&i,&info);
  lwork=MAX(4*n,5*n-4);
  lwork=t;
  work=(double*)malloc(lwork*sizeof(double));

  info=0;
  jobvl='N';
  jobvr='N';
  F77_FUNC(dgesvd,DGESVD)(&jobvl,&jobvr,&n,&n,A,&n,s,U,&n,V,&n,work,&lwork,&info);

  result=0;
  for(i=0;i<n;i++)
   {
    if(verbose&&i<n-k){printf("   %d %le\n",i,s[i]);fflush(stdout);}
    if(verbose&&i>=n-k){printf("   %d %le *\n",i,s[i]);fflush(stdout);}
    if(i<n-k&&s[i]>0.)result++;
   }
  if(verbose){printf("  number positive = %d\n",result);fflush(stdout);}
  free(s);
  free(U);
  free(V);
  free(work);

  return result;
 }

double MFFullProdEV(int n,int k,double *A)
 {
  static char RoutineName[]={"MFFullProdEV"};
  int i,j,l;
  int lda;
  char jobvl='N';                   /* No left  eigenvectors */
  char jobvr='V';                   /*    right eigenvectors */
  static double *wr=(double*)NULL;    /* k */            /* Array to hold real part of eigenvalues */
  static double *wi=(double*)NULL;    /* k */            /* Array to hold imag part of eigenvalues */
  int ldvl;                       /* Leading dimension of vl */
  static double *vl=(double*)NULL;   /* k*k */            /* Array to hold left eigenvectors */
  int ldvr;                       /* Leading dimension of vr */
  static double *vr=(double*)NULL;   /* k*k */           /* Array to hold right eigenvectors */
  static double *work=(double*)NULL;    /* k*k */        /* Work array */
  int lwork;                     /* Length of work array */
  int info=0;
  int verbose=0;
  int result;
  double resultI,resultR;
  double I,R;
  double prod;

  if(verbose)
   {
    printf("%s\n",RoutineName);
    for(i=0;i<n;i++)
     {
      printf("[");
      for(j=0;j<n;j++)
       {
        if(j>0)printf(" ");
        printf("%lf",A[i+n*j]);
       }
      printf("]\n");
     }

/*  i=0;
    printf("So x     =(%lf",A[i+n*0]/2);
    for(j=1;j<n;j++)printf(",%lf",A[i+n*j]/2);
    printf(")\n");
    i++;
    printf("   phi[1]=(%lf",A[i+n*0]);
    for(j=1;j<n;j++)printf(",%lf",A[i+n*j]);
    printf(")\n");
    i++;
    printf("   phi[0]=(%lf",A[i+n*0]);
    for(j=1;j<n;j++)printf(",%lf",A[i+n*j]);
    printf(")\n");
    printf("   2x.phi[0]=%lf\n",A[0+n*0]*A[1+n*0]+A[0+n*1]*A[1+n*1]+A[0+n*2]*A[1+n*2]);
    printf("   2x.phi[1]=%lf\n",A[0+n*0]*A[2+n*0]+A[0+n*1]*A[2+n*1]+A[0+n*2]*A[2+n*2]);
    printf("   4x.x     =%lf\n",A[0+n*0]*A[0+n*0]+A[0+n*1]*A[0+n*1]+A[0+n*2]*A[0+n*2]);
    printf("   phi[0]^2 =%lf\n",A[1+n*0]*A[1+n*0]+A[1+n*1]*A[1+n*1]+A[1+n*2]*A[1+n*2]);
    printf("   phi[1]^2 =%lf\n",A[2+n*0]*A[2+n*0]+A[2+n*1]*A[2+n*1]+A[2+n*2]*A[2+n*2]);
    printf("   phi[0].phi[1]=%lf\n",A[1+n*0]*A[2+n*0]+A[1+n*1]*A[2+n*1]+A[1+n*2]*A[2+n*2]);*/
   }

  lda=n;
  wr=(double*)realloc((void*)wr,n*sizeof(double));
  wi=(double*)realloc((void*)wi,n*sizeof(double));
  ldvl=n;
  vl=(double*)realloc((void*)vl,n*n*sizeof(double));
  ldvr=n;
  vr=(double*)realloc((void*)vr,n*n*sizeof(double));
  lwork=4*n;
  work=(double*)realloc((void*)work,lwork*sizeof(double));

  info=0;
  jobvl='N';
  jobvr='V';
  if(verbose){printf("      call dgeev\n");fflush(stdout);}
  F77_FUNC(dgeev,DGEEV)(&jobvl,&jobvr,&n,A,&lda,wr,wi,vl,&ldvl,vr,&ldvr,work,&lwork,&info);
  if(verbose){printf("      done dgeev info=%d\n",info);fflush(stdout);}

  if(verbose){printf("%s\n",RoutineName);}
  resultR=1.;
  resultI=0.;
  for(i=0;i<n;i++)
   {
    if(verbose){printf("   %d (%le,%le)\n",i,wr[i],wi[i]);fflush(stdout);}
    R=resultR*wr[i]-resultI*wi[i];
    I=resultR*wi[i]+resultI*wr[i];
    resultR=R;
    resultI=I;
   }
  if(verbose){printf("  product of eigenvalues =(%lf,%lf)\n",resultR,resultI);fflush(stdout);}

  return resultR;
 }

int MFTangentAlgebraicWithGuess(int n,int k,MFNVector u,MFNKMatrix Phi0,MFNKMatrix Phi,void *d)
 {
  static char RoutineName[]={"MFTangentAlgebraicWithGuess"};
  double *A;
  int i,j,l;
  double *p0;
  double *p;
  int verbose=0;
 
  MFTangentAlgebraic(n,k,u,Phi,d);

/* Now, multiply the first basis vector by -1 if det(Phi0^*Phi)<0 */

  A=(double*)malloc(k*k*sizeof(double));
  if(A==(double*)NULL)abort();

  p0=MFNKM_CStar(Phi0);
  p=MFNKM_CStar(Phi);

  for(i=0;i<k;i++)
   {
    for(j=0;j<k;j++)
     {
      A[i+k*j]=0.;
      for(l=0;l<n;l++)A[i+k*j]+=p0[l+n*i]*p[l+n*j];
     }
   }

  if(MFAtlasDet(k,A)<0)
    for(i=0;i<n;i++)p[i+n*0]=-p[i+n*0];

  if(verbose)
   {
    printf("Tangent with guess, det of Phi0^*Phi was %lf.\n",MFAtlasDet(k,A));
    for(i=0;i<k;i++)
     {
      for(j=0;j<k;j++)
       {
        A[i+k*j]=0.;
          for(l=0;l<n;l++)A[i+k*j]+=p0[l+n*i]*p[l+n*j];
       }
     }
    printf("      is now                             %lf.\n",MFAtlasDet(k,A));
   }

  free(A);

  return 1;
 }

int MFFullNPosEV(int n,int k,double *A)
 {
  static char RoutineName[]={"MFFullNPosEV"};
  int i,j,l;
  int lda;
  char jobvl='N';                   /* No left  eigenvectors */
  char jobvr='V';                   /*    right eigenvectors */
  static double *wr=(double*)NULL;    /* k */            /* Array to hold real part of eigenvalues */
  static double *wi=(double*)NULL;    /* k */            /* Array to hold imag part of eigenvalues */
  int ldvl;                       /* Leading dimension of vl */
  static double *vl=(double*)NULL;   /* k*k */            /* Array to hold left eigenvectors */
  int ldvr;                       /* Leading dimension of vr */
  static double *vr=(double*)NULL;   /* k*k */           /* Array to hold right eigenvectors */
  static double *work=(double*)NULL;    /* k*k */        /* Work array */
  int lwork;                     /* Length of work array */
  int info=0;
  int verbose=0;
  int result;

  if(verbose)
   {
    printf("%s\n",RoutineName);
    for(i=0;i<n;i++)
     {
      printf("[");
      for(j=0;j<n;j++)
       {
        if(j>0)printf(" ");
        printf("%lf",A[i+n*j]);
       }
      printf("]\n");
     }
   }

  lda=n;
  wr=(double*)realloc((void*)wr,n*sizeof(double));
  wi=(double*)realloc((void*)wi,n*sizeof(double));
  ldvl=n;
  vl=(double*)realloc((void*)vl,n*n*sizeof(double));
  ldvr=n;
  vr=(double*)realloc((void*)vr,n*n*sizeof(double));
  lwork=4*n;
  work=(double*)realloc((void*)work,lwork*sizeof(double));

  info=0;
  jobvl='N';
  jobvr='V';
  F77_FUNC(dgeev,DGEEV)(&jobvl,&jobvr,&n,A,&lda,wr,wi,vl,&ldvl,vr,&ldvr,work,&lwork,&info);

  if(verbose){printf("%s, info back from dgeev is %d\n",RoutineName,info);}
  result=0;
  for(i=0;i<n;i++)
   {
    int j;
    if(verbose){printf("   %d (%le,%le)",i,wr[i],wi[i]);
                printf("          (%le",vr[0+n*i]);for(j=1;j<n;j++)printf(",%le",vr[j+n*i]);printf(")\n");fflush(stdout);}
    if(wr[i]>0.)result++;
   }
  if(verbose){printf("  number of eigenvalues with positive real part = %d\n",result);fflush(stdout);}

  return result;
 }

int MFFullNUnitEV(int n,int k,double *A)
 {
  static char RoutineName[]={"MFFullNUnitEV"};
  int i,j,l;
  int lda;
  char jobvl='N';                   /* No left  eigenvectors */
  char jobvr='V';                   /*    right eigenvectors */
  static double *wr=(double*)NULL;    /* k */            /* Array to hold real part of eigenvalues */
  static double *wi=(double*)NULL;    /* k */            /* Array to hold imag part of eigenvalues */
  int ldvl;                       /* Leading dimension of vl */
  static double *vl=(double*)NULL;   /* k*k */            /* Array to hold left eigenvectors */
  int ldvr;                       /* Leading dimension of vr */
  static double *vr=(double*)NULL;   /* k*k */           /* Array to hold right eigenvectors */
  static double *work=(double*)NULL;    /* k*k */        /* Work array */
  int lwork;                     /* Length of work array */
  int info=0;
  int verbose=0;
  int result;

  lda=n;
  wr=(double*)realloc((void*)wr,n*sizeof(double));
  wi=(double*)realloc((void*)wi,n*sizeof(double));
  ldvl=n;
  vl=(double*)realloc((void*)vl,n*n*sizeof(double));
  ldvr=n;
  vr=(double*)realloc((void*)vr,n*n*sizeof(double));
  lwork=4*n;
  work=(double*)realloc((void*)work,lwork*sizeof(double));

  info=0;
  jobvl='N';
  jobvr='V';
  F77_FUNC(dgeev,DGEEV)(&jobvl,&jobvr,&n,A,&lda,wr,wi,vl,&ldvl,vr,&ldvr,work,&lwork,&info);

  if(verbose){printf("%s\n",RoutineName);}
  result=0;
  for(i=0;i<n;i++)
   {
    if(verbose){printf("   %d (%le,%le)\n",i,wr[i],wi[i]);fflush(stdout);}
    if(wr[i]*wr[i]+wi[i]*wi[i]>1.)result++;
   }
  if(verbose){printf("  number of eigenvalues with modulus greater than one = %d\n",result);fflush(stdout);}

  return result;
 }

int MFFullNZeroSV(int n,int k, double *A)
 {
  static char RoutineName[]={"MFFullNZeroSV"};
  int i,j;
  int lda;
  char jobvl='N';
  char jobvr='N';
  double *s=(double*)NULL;
  double *U=(double*)NULL;
  double *V=(double*)NULL;
  double *work=(double*)NULL;
  double t;
  int lwork;
  int info=0;
  int verbose=1;
  int result;

  if(verbose){printf("%s\n",RoutineName);}

  if(0&&verbose)
   {
    for(i=0;i<n;i++)
     {
      printf("[");
      for(j=0;j<n;j++)
       {
        if(j>0)printf(" ");
        printf("%lf",A[i+n*j]);
       }
      printf("]\n");
     }
   }

  s=(double*)malloc(n*sizeof(double));
  U=(double*)malloc(n*n*sizeof(double));
  V=(double*)malloc(n*n*sizeof(double));
  i=-1;
  info=0;
  jobvl='N';
  jobvr='A';
  F77_FUNC(dgesvd,DGESVD)(&jobvl,&jobvr,&n,&n,A,&n,s,U,&n,V,&n,&t,&i,&info);
  lwork=MAX(4*n,5*n-4);
  lwork=t;
  work=(double*)malloc(lwork*sizeof(double));

  info=0;
  jobvl='N';
  jobvr='N';
  F77_FUNC(dgesvd,DGESVD)(&jobvl,&jobvr,&n,&n,A,&n,s,U,&n,V,&n,work,&lwork,&info);

  result=0;
  for(i=0;i<n;i++)
   {
    if(verbose&&i<n-k){printf("   %d %le\n",i,s[i]);fflush(stdout);}
    if(verbose&&i>=n-k){printf("   %d %le *\n",i,s[i]);fflush(stdout);}
    if(fabs(s[i])<1.e-10)result++;
   }
  if(verbose){printf("  number zero = %d\n",result);fflush(stdout);}
  free(s);
  free(U);
  free(V);
  free(work);

  return result;
 }
