/* 
    %W%
    %D% %T%
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

    Author: Michael E. Henderson, mhender@watson.ibm.com
*/


#include <MFAtlas.h>
#include <MFError.h>
#include <MFNVector.h>
#include <MFKVector.h>
#include <MFNSpace.h>
#include <MFNRegion.h>
#include <MFImplicitMF.h>
#include <MFListOfCharts.h>
#include <MFPolytope.h>
#include <MFPrint.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include <MFTime.h>
#include <MFAtlasFriends.h>
double MFNKMatrixDot(MFNSpace,MFNKMatrix,MFNKMatrix);
double *MFNV_CStar(MFNVector);
double *MFKV_CStar(MFKVector);
void MFReportMemory();
double MFFullProdEV(int,int,double*);

#ifdef MFDEBUG
#include <umalloc.h>
#endif

#ifndef DBL_QNAN
#define DBL_QNAN 1.e200
#endif

#ifndef DBL_INFINITY
#define DBL_INFINITY 1.e200
#endif

void MFSetError(int,char*,char*,int,char*);
static char MFAtlasErrorMsg[256]="";
void MFSetChartPolytope(MFChart,MFPolytope);
void MFSetChartRadius(MFChart,double);
void MFAtlasSetK(MFAtlas,int);
void MFAtlasSetN(MFAtlas,int);
void MFAtlasPageOutChartsNotNearBoundary(MFAtlas,int,int,char*);
void MFAtlasWriteOutChartsNotPaged(MFAtlas,int,int,char*);
void MFWriteChartToPlotFile(FILE*,MFAtlas,MFChart,int);
int *MFPolytopeVertexIndexSet(MFPolytope,int);

void MFAtlasHalfSpaceSetIndex(MFAtlas,int,int);
void MFAtlasHalfSpaceSetOrigin(MFAtlas,int,double);
void MFAtlasHalfSpaceSetNormal(MFAtlas,int,MFKVector);
void MFAtlasHalfSpaceSetLeftChart(MFAtlas,int,int);
void MFAtlasHalfSpaceSetRightChart(MFAtlas,int,int);
void MFAtlasHalfSpaceIncreaseSize(MFAtlas);
void MFAtlasHalfSpaceFree(MFAtlas);
void MFAtlasHalfSpaceInit(MFAtlas);

#define REALLYPAGE
/* #define TEST 1*/

#ifndef NOBINARYTREE
#include <MFBinaryTree.h>
#endif

/* Internal Routines */

void MFAddInitialHyperplanes(MFAtlas);
int MFAtlasAddChartToList(MFAtlas,MFNVector,MFNKMatrix,double);
MFChart MFAtlasChart(MFAtlas,int);
int MFPolytopeSmallestVertexIndex(MFPolytope);
int MFPolytopeLargestVertexIndex(MFPolytope);
int MFPolytopeVertexLargestIndex(MFPolytope,int);
int MFTestPolytope(MFPolytope);

/* Parameters */

double MFEpsilon=0.2;
double MFRMin=0.001;
double MFDotMin=0.2;
int MFVerbose=1;

/* For Testing */

int MFTestBoundaryPoint(MFAtlas,MFNVector,int);
int MFTestAtlas(MFAtlas);      /* No vertex may be in another PT */
int MFChartTotallyInterior(MFChart,MFKVector,double);

/* For the BoundaryList */

#define MFAtlasIncrementToAllocateBoundaryList 1000
void MFAtlasAddChartToBoundaryList(MFAtlas,int);
void MFAtlasRemoveChartFromBoundaryList(MFAtlas,int);
void MFAtlasRemoveChartFromBoundaryListByChart(MFAtlas,int);

/* For the ChartList */

#define MFAtlasIncrementToAllocate 1000
int MFAddHalfSpaceToAtlas(MFAtlas,int,int,int,MFKVector,double);

struct MFAtlasSt
 {
  int    n;
  int    k;

  MFImplicitMF M;
  MFNSpace NSpace;

  int    nCharts;
  int    mCharts;
  MFChart *chart;

/* List of Half Spaces */

  int    offset;
  int    nHS;
  int    mHS;
  int    nGlobalIndex;

#ifdef OLDHS
  int    *globalIndex;
  double *onrm;
  MFKVector *nrm;
  int    *leftChart;
  int    *rightChart;
#else
  int nnHS;
  int mmHS;
  int    **globalIndex;
  double **onrm;
  MFKVector **nrm;
  int    **leftChart;
  int    **rightChart;
#endif

#ifndef NOBINARYTREE
/* For binary tree */
  MFBinaryTree BTree;
#endif

  FILE *plotFileFid;
  double expFactor;
  FILE *pageFileFid;
  FILE *centerFileFid;
  int   pageFileIndex;
  int   nPagedOut;

#ifndef NOBOUNDARYLIST
/* For list of boundarys */

  int nBnd;
  int mBnd;
  int *bnd;

/* For post processing */

  int nclipf;
  int mclipf;
  double (**clipf)(MFNVector);
  int nclipIndx;
  int *clipIndx;
#endif
/* For additional culling of nearby points */
  int (*isNear)(MFAtlas,MFChart,MFChart);
 };

void MFPrintChartSummary(MFAtlas A)
 {
  int i;
  printf("Atlas charts\n");
  for(i=0;i<A->nCharts;i++)
   {
    printf(" Chart %d is 0x%8.8x, ",i,A->chart[i]);
    if(!MFChartPaged(A->chart[i]))
     {
      printf(" Center is 0x%8.8x, TS is 0x%8.8x",MFChartCenter(A->chart[i]),MFChartTangentSpace(A->chart[i]));
      printf(" index %d, bif? %d\n",MFNVGetIndex(MFChartCenter(A->chart[i])),MFNVGetIndex2(MFChartCenter(A->chart[i])));
     }else
      printf(" Paged out\n");
   }
  fflush(stdout);
  return;
 }

MFAtlas MFCreateAtlas(MFImplicitMF M)
 {
  static char RoutineName[]={"MFCreateAtlas"};
  MFAtlas A;
  int i,j,ij;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFCreateAtlas++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(M==(MFImplicitMF)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Manifold (argument 1) is NULL.");
    printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
    MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateAtlas+=clock()-starttime;
#endif
    return (MFAtlas)NULL;
   }

  A=(MFAtlas)malloc(sizeof(struct MFAtlasSt)); /*done*/
  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFAtlasSt));
    printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
    MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateAtlas+=clock()-starttime;
#endif
    return (MFAtlas)NULL;
   }

  A->M=M;
  MFRefImplicitMF(M);

  A->NSpace=MFIMFNSpace(M);
  MFRefNSpace(A->NSpace);

  A->n=MFIMF_N(M);
  A->k=MFIMF_K(M);

/* List of Charts */

  A->nCharts=0;
  A->mCharts=0;
  A->chart=(MFChart*)NULL;

/* List of Half spaces */

  A->nHS=0;
  A->mHS=0;
  MFAtlasHalfSpaceInit(A);
  A->nGlobalIndex=0;
  A->offset=0;

/* Paging */

  A->pageFileFid=(FILE*)NULL;
  A->pageFileIndex=0;
  A->nPagedOut=0;
  A->plotFileFid=(FILE*)NULL;
  A->expFactor=1.;
  A->centerFileFid=(FILE*)NULL;

/* Half spaces defining a hypercube */

  if(A->k>0)MFAddInitialHyperplanes(A);

/* BinaryTree (created when first chart is added) */

#ifndef NOBINARYTREE
  A->BTree=(MFBinaryTree)NULL;
#endif

/* List of boundary Charts (created when first chart is added) */

  A->nBnd=0;
  A->mBnd=0;
  A->bnd=(int*)NULL;

  A->nclipf=0;
  A->mclipf=0;
  A->clipf=NULL;
  A->nclipIndx=-10;
  A->clipIndx=NULL;

  A->isNear=NULL;

#ifdef MFTIMINGS
  MFTimeMFCreateAtlas+=clock()-starttime;
#endif
  return(A);
 }

void MFFreeAtlas(MFAtlas A)
 {
  static char RoutineName[]={"MFFreeAtlas"};
  int i;
  int verbose;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFFreeAtlas++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  verbose=0;
  if(verbose){printf("MFFreeAtlas\n");fflush(stdout);}

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFFreeAtlas+=clock()-starttime;
#endif
    return;
   }


  if(verbose){printf("  Delete M\n");fflush(stdout);}
  if(A->M!=(MFImplicitMF)NULL)MFFreeImplicitMF(A->M);
  if(A->NSpace!=(MFNSpace)NULL)MFFreeNSpace(A->NSpace);

  if(verbose){printf("  Delete List of charts\n");fflush(stdout);}
  if(A->chart!=(MFChart*)NULL)
   {
    for(i=0;i<A->nCharts;i++)
      MFFreeChart(A->chart[i]);
    free(A->chart);
   }

/* List of half spaces */

  MFAtlasHalfSpaceFree(A);

#ifndef NOBINARYTREE
  if(verbose){printf("  Delete BTree\n");fflush(stdout);}
  if(A->BTree!=(MFBinaryTree)NULL)MFFreeBinaryTree(A->BTree);
#endif

  if(A->bnd!=(int*)NULL)free(A->bnd);

  if(A->clipIndx!=(int*)NULL)free(A->clipIndx);
  if(A->clipf!=(double (**)(MFNVector))NULL)free(A->clipf);

  if(A->pageFileFid!=(FILE*)NULL)fclose(A->pageFileFid);

  free(A);

  if(verbose)printf("done MFFreeAtlas\n");
#ifdef MFTIMINGS
  MFTimeMFFreeAtlas+=clock()-starttime;
#endif
  return;
 }

int MFAtlasK(MFAtlas A)
 {
  static char RoutineName[]={"MFAtlasK"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasK++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasK+=clock()-starttime;
#endif
    return -1;
   }

#ifdef MFTIMINGS
  MFTimeMFAtlasK+=clock()-starttime;
#endif
  return(A->k);
 }

int MFAtlasN(MFAtlas A)
 {
  static char RoutineName[]={"MFAtlasN"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasN++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasN+=clock()-starttime;
#endif
    return -1;
   }

#ifdef MFTIMINGS
  MFTimeMFAtlasN+=clock()-starttime;
#endif
  return(A->n);
 }

int MFAtlasNumberOfCharts(MFAtlas A)
 {
  static char RoutineName[]={"MFNumberOfCharts"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNumberOfCharts++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasNumberOfCharts+=clock()-starttime;
#endif
    return -1;
   }

#ifdef MFTIMINGS
  MFTimeMFAtlasNumberOfCharts+=clock()-starttime;
#endif
  return(A->nCharts);
 }

double MFAtlasChartRadius(MFAtlas A,int chart)
 {
  static char RoutineName[]={"MFAtlasChartRadius"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasChartRadius++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasChartRadius+=clock()-starttime;
#endif
    return -1.;
   }

  if(chart<0||chart>=A->nCharts)
   {
    sprintf(MFAtlasErrorMsg,"Chart %d (argument 2) is invalid. Must be positive and less than %d (total number of charts).",chart,A->nCharts);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasChartRadius+=clock()-starttime;
#endif
    return -1.;
   }

#ifdef MFTIMINGS
  MFTimeMFAtlasChartRadius+=clock()-starttime;
#endif
  return MFChartRadius(A->chart[chart]);
 }

MFNVector MFAtlasCenterOfChart(MFAtlas A,int chart)
 {
  static char RoutineName[]={"MFAtlasCenterOfChart"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasCenterOfChart++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFAtlasCenterOfChart+=clock()-starttime;
#endif
    return (MFNVector)NULL;
   }

  if(chart<0||chart>=A->nCharts)
   {
    sprintf(MFAtlasErrorMsg,"Chart %d (argument 2) is invalid. Must be positive and less than %d (total number of charts).",chart,A->nCharts);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFAtlasCenterOfChart+=clock()-starttime;
#endif
    return (MFNVector)NULL;
   }

#ifdef MFTIMINGS
  MFTimeMFAtlasCenterOfChart+=clock()-starttime;
#endif
  return MFChartCenter(A->chart[chart]);
 }

int MFAtlasIsPointInChart(MFAtlas A,int chart, MFKVector pt)
 {
  static char RoutineName[]={"MFAtlasIsPointInChart"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasIsPointInChart++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFAtlasIsPointInChart+=clock()-starttime;
#endif
    return -1;
   }

  if(chart<0||chart>=A->nCharts)
   {
    sprintf(MFAtlasErrorMsg,"Chart %d (argument 2) is invalid. Must be positive and less than %d (total number of charts).",chart,A->nCharts);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFAtlasIsPointInChart+=clock()-starttime;
#endif
    return -1;
   }

  if(pt==(MFKVector)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Point (argument 3) is NULL.");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFAtlasIsPointInChart+=clock()-starttime;
#endif
    return -1;
   }

#ifdef MFTIMINGS
  MFTimeMFAtlasIsPointInChart+=clock()-starttime;
#endif
  return MFChartInterior(A->chart[chart],pt);
 }

void MFAtlasEvaluateChart(MFAtlas A,int chart,MFKVector s,MFNVector u)
 {
  static char RoutineName[]={"MFAtlasEvaluateChart"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasEvaluateChart++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFAtlasEvaluateChart+=clock()-starttime;
#endif
    return;
   }

  if(chart<0||chart>=A->nCharts)
   {
    sprintf(MFAtlasErrorMsg,"Chart %d (argument 2) is invalid. Must be positive and less than %d (total number of charts).",chart,A->nCharts);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFAtlasEvaluateChart+=clock()-starttime;
#endif
    return;
   }

  if(s==(MFKVector)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Point in domain (argument 3) is NULL.");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFAtlasEvaluateChart+=clock()-starttime;
#endif
    return;
   }

  if(u==(MFNVector)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Point in range (argument 4) is NULL.");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFAtlasEvaluateChart+=clock()-starttime;
#endif
    return;
   }

  MFChartEvaluate(A->chart[chart],s,u);

#ifdef MFTIMINGS
  MFTimeMFAtlasEvaluateChart+=clock()-starttime;
#endif
  return;
 }

int MFAtlasNumberOfChartsWithBoundary(MFAtlas A)
 {
  static char RoutineName[]={"MFAtlasNumberOfChartsWithBoundary"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasNumberOfChartsWithBoundary++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasNumberOfCharts+=clock()-starttime;
#endif
    return -1;
   }

#ifdef MFTIMINGS
  MFTimeMFAtlasNumberOfCharts+=clock()-starttime;
#endif
  return(A->nBnd);
 }

int MFAtlasChartWithBoundary(MFAtlas A,int i)
 {
  static char RoutineName[]={"MFAtlasChartWithBoundary"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasChartWithBoundary++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    printf("%s : %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasChartWithBoundary+=clock()-starttime;
#endif
    return -1;
   }

  if(i<0||i>=A->nBnd)
   {
    sprintf(MFAtlasErrorMsg,"Index of boundary chart %d (argument 2) is invalid. Must be positive and less than %d (number of charts on boundary)",i,A->nBnd);
    printf("%s : %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasChartWithBoundary+=clock()-starttime;
#endif
    return -1;
   }

#ifdef MFTIMINGS
  MFTimeMFAtlasChartWithBoundary+=clock()-starttime;
#endif
  return(A->bnd[i]);
 }

int MFAtlasAddChart(MFAtlas A,MFNVector u)
 {
  static char RoutineName[]={"MFAtlasAddChart"};
  int result;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasAddChart++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  result=MFAtlasAddChartWithApproxTS(A,u,(MFNKMatrix)NULL,-1.,-1.);

#ifdef MFTIMINGS
  MFTimeMFAtlasAddChart+=clock()-starttime;
#endif

  return result;
 }

/* Add radius of chart approx came from and distance 
   between u and the chart */
 
int MFAtlasAddChartWithApproxTS(MFAtlas A,MFNVector u,MFNKMatrix TS,double delta, double R0)
 {
  static char RoutineName[]={"MFAtlasAddChartWithApproxTS"};
  double R;
  MFNKMatrix Phi;
  int chart;

  int verbose;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasAddChartWithApproxTS++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  verbose=0;
  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasAddChartWithApproxTS+=clock()-starttime;
#endif
    return -1;
   }

  if(u==(MFNVector)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to u (argument 2) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasAddChartWithApproxTS+=clock()-starttime;
#endif
    return -1;
   }

  if(verbose){printf("       calculate TS \n");fflush(stdout);}
  if(TS==(MFNKMatrix)NULL)
   {
    if(verbose){printf("         without guess\n");fflush(stdout);}
    Phi=MFIMFTangentSpace(A->M,u);
   }else{
    if(verbose){printf("         with guess\n");fflush(stdout);}
    Phi=MFIMFTangentSpaceWithGuess(A->M,u,TS);
   }
  if(Phi==(MFNKMatrix)NULL)return -1;

  R=MFIMFScale(MFAtlasMF(A),u,Phi);
  chart=MFAtlasAddChartWithAll(A,u,Phi,R);
  MFFreeNKMatrix(Phi);

  if(verbose){printf(" done %s\n",RoutineName);fflush(stdout);}
  return(chart);
 }

/* --------------------------------------------------------- */
#define FROMTOP

int MFAtlasPointOnBoundaryInsideRegion(MFAtlas A,MFNRegion Omega,MFNVector u, MFNKMatrix *TS, double *dist)
 {
  static char RoutineName[]={"MFAtlasPointOnBoundaryInsideRegion"};
  int v,l,n;
  int pass;
  int nChartsTried;
  int chart;
  int bchart;
  MFNVector up;
  MFKVector s;
  MFKVector s2;
  MFNVector w;
  MFNKMatrix Phi;
  double dot;
  double rv;
  double R;
  double delta;
  int rc;
  int abandon;
  int verbose;
  double t,tl,tr;
  int bisection;
  double tol=1.e-10;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasPointOnBoundaryInsideRegion++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  verbose=0;
  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

  up=MFCloneNVector(u);
  s=MFCreateKVector(A->k);
  nChartsTried=0;
  bisection=0;

  pass=1;
/*  pass=0; */
#ifndef NOBOUNDARYLIST
  bchart=0;
  if(verbose){printf("%s, there are %d charts on the boundary list\n",RoutineName,A->nBnd);fflush(stdout);}
  while(bchart<A->nBnd && pass<2)
   {
#ifdef FROMTOP
    chart=A->bnd[bchart];
#else
    chart=A->bnd[A->nBnd-bchart-1];
    printf("bchart=%d, which is %d on the list out of %d. This is chart %d\n",bchart,A->nBnd-bchart-1,A->nBnd,chart);
#endif
#else
  chart=0;
  while(chart<A->nCharts && pass<2)
   {
#endif
    nChartsTried++;
    if(MFChartIsSingular(A->chart[chart]))goto nextchart;
tryagain:
    n=MFPolytopeNumberOfVertices(MFChartPolytope(A->chart[chart]));
    if(verbose){printf(" chart %d has %d vertices\n",chart,n);fflush(stdout);}
    if(verbose)
     {
      for(v=0;v<n;v++)
      printf("   Vertex %d has mark %d, index %d, radius %le/%le\n",v,
           MFPolytopeGetVertexMark(MFChartPolytope(A->chart[chart]),v),
           MFNVGetIndex2(MFChartCenter(A->chart[chart])),
           MFChartRadius(A->chart[chart]),MFRMin);
     }
    v=0;
    while(v<n 
       && MFChartRadius(A->chart[chart])>MFRMin 
       && MFPolytopeGetVertexMark(MFChartPolytope(A->chart[chart]),v)==0
       && MFNVGetIndex2(MFChartCenter(A->chart[chart]))!=-98
       && MFNVGetIndex2(MFChartCenter(A->chart[chart]))!=-99)
     {
      if(verbose){printf(" vertex %d \n",v);fflush(stdout);}
      if(pass==1 || ( pass==0 && MFPolytopeVertexLargestIndex(MFChartPolytope(A->chart[chart]),v)>= A->offset))
       {
        rv=MFPolytopeRadiusOfVertex(MFChartPolytope(A->chart[chart]),v);
        R=MFChartRadius(A->chart[chart]);
        if(verbose){printf("  Trying vertex %d of chart %d (Radius %le)\n",v,chart,R);fflush(stdout);}

        if(rv>=R)
         {
          if(verbose){printf("  Exterior vertex. ");fflush(stdout);}
          t=1.;
bsect:
          if(verbose){printf("t=%le, bisection=%d\n",t,bisection);fflush(stdout);}
          if(t<.001&&!bisection)goto nextvertex;
          MFPolytopeVertex(MFChartPolytope(A->chart[chart]),v,s);
          MFKVScale(t*R/rv,s);
          MFChartPointInTangentSpace((A->chart[chart]),s,up);
          if(verbose){printf("up(");MFPrintKVector(stdout,s);printf(")=");MFPrintNVector(stdout,up);printf("\n");fflush(stdout);}

          if(!MFNRegionInterior(Omega,up))
           {
            if(verbose){printf("     Not in Omega.\n");fflush(stdout);}
            goto nextvertex;
           }
          if(verbose){printf("     In Omega.\n");fflush(stdout);}

          rc=MFChartEvaluate(A->chart[chart],s,u);
          if(!rc)
           {
            if(verbose){printf("%s - ChartEvaluate failed at vertex %d of chart %d, chart R=%lf vertex R %lf\n",RoutineName,chart,v,R,rv);fflush(stdout);}
            if(!bisection)
             {
              t=.8*t;
              if(t<.001)goto nextvertex;
              MFSetSuggestedChartRadius(A->chart[chart],t*R);
              goto bsect;
             }else{
              tr=t+.1*tol;
              tl=t-.1*tol;
              goto donebsect;
             }
           }

/* Was correction orthogonal? */

          if(verbose)
           {
            s2=MFCreateKVector(A->k);
            MFChartProjectIntoTangentSpace(A->chart[chart],up,s2);
            printf("     s(up) is ");MFPrintKVector(stdout,s2);printf(" should be the same as s ");MFPrintKVector(stdout,s);printf("\n");fflush(stdout);
            MFChartProjectIntoTangentSpace(A->chart[chart],u,s2);
            printf("     s(u) is ");MFPrintKVector(stdout,s2);printf(" should be the same as s ");MFPrintKVector(stdout,s);printf("\n");fflush(stdout);
            MFFreeKVector(s2);
           }
 
          delta=MFNSpaceDistance(A->NSpace,u,up);
          if(verbose){printf("     The correction was %lf, epsilon=%lf\n",delta,MFEpsilon);fflush(stdout);}
          if(0&&verbose)
           {
            printf("     u0 is ");MFPrintNVector(stdout,MFChartCenter(A->chart[chart]));printf("\n");fflush(stdout);
            printf("     s is ");MFPrintKVector(stdout,s);printf("\n");fflush(stdout);
            printf("     up is ");MFPrintNVector(stdout,up);printf("\n");fflush(stdout);
            printf("     u  is ");MFPrintNVector(stdout,u);printf("\n");fflush(stdout);
            printf("     Phi[%d] is 0x%8.8x\n",chart,MFChartTangentSpace(A->chart[chart]));fflush(stdout);
            printf("     Phi is \n");MFPrintNKMatrix(stdout,MFChartTangentSpace(A->chart[chart]));printf("\n");fflush(stdout);
           }
          if(delta>MFEpsilon&&!bisection)
           {
            if(verbose){printf("%s - Correction too large at R=%lf, (%lf>%lf) trying again at %lf. Radius of vertex is %lf\n",RoutineName,R,delta,MFEpsilon,.8*t*R,rv);fflush(stdout);}
            t=.8*t;
            if(t<.001)goto nextvertex;
            MFSetSuggestedChartRadius(A->chart[chart],t*R);
            goto bsect;
           }else if(bisection)
           {
            if(verbose){printf("     Bisection: interval [%lf, %lf, %lf ]\n",tl,t,tr);fflush(stdout);}
            if(delta<MFEpsilon&&!MFIMFStop(A->M,MFChartCenter(A->chart[chart]),MFChartTangentSpace(A->chart[chart]),u,MFChartTangentSpace(A->chart[chart])))
             {
              if(verbose)printf("                 Keep Right\n");
              tl=t;
             }else{
              if(verbose)printf("                 Keep Left\n");
              tr=t;
             }
            t=(tr+tl)/2;
            if(verbose){printf("            new interval [%lf, %lf, %lf ] width %le \n",tl,t,tr,tr-tl);fflush(stdout);}
donebsect:
            if(tr-tl<tol)
             {
              if(fabs(t)>1.e-6)
               {
#ifdef MFTIMINGS
                MFTimeMFAtlasPointOnBoundaryInsideRegion+=clock()-starttime;
#endif
                if(verbose){printf("     Interval less than 1.e-10 (%le), declaring success, t in [%lf,%lf]\n",tr-tl,tl,tr);}
                MFPolytopeSetVertexMark(MFChartPolytope(A->chart[chart]),v,1);

                MFNVSetIndex(u,MFNVGetIndex(MFChartCenter(A->chart[chart])));
/*              MFNVSetIndex (u,-99);*/
                MFNVSetIndex2(u,-99);
                *TS=MFCloneNKMatrix(MFChartTangentSpace(A->chart[chart]));if(*TS==(MFNKMatrix)NULL)abort();

                w=MFCloneNVector(up);
                rc=MFIMFSingular(A->M,u,*TS,w); /* w is the "other" null vector */
                if(verbose){printf("     return code from MFIMFSingular %d\n",rc);fflush(stdout);}
                if(rc==1)
                 {
                  MFNVector v;
                  MFKVector s;
                  MFNKMatrix Psi;
                  double vnorm;

                  MFNSpaceScale(A->NSpace,.5*MFChartRadius(A->chart[chart]),w,up);
                  MFNSpaceAdd(A->NSpace,u,up,up);
                  Psi=MFIMFTangentSpace(A->M,up);if(Psi==(MFNKMatrix)NULL)abort();
                  Phi=MFCloneNKMatrix(MFChartTangentSpace(A->chart[chart]));if(Phi==(MFNKMatrix)NULL)abort();

                  v=MFCloneNVector(w);
                  s=MFCreateKVector(MFAtlasK(A));
                  MFMVMulT(A->NSpace,Psi,w,s);
                  MFMVMul (A->NSpace,Psi,s,v);
                  MFMVMulT(A->NSpace,Phi,v,s);
                  MFMVMul (A->NSpace,Phi,s,v);

                  MFFreeNVector(w);
                  w=MFCloneNVector(u);
                  MFNVSetIndex2(w,-98);
                  vnorm=sqrt(MFNSpaceInner(A->NSpace,v,v));
                  if(fabs(vnorm)>1.e-10)
                   {
                    double R;
                    int bchart;
                    R=1.*MFChartRadius(A->chart[chart]);
                    MFIMFSetStability(A->M,w,Psi);
                    bchart=MFAtlasAddChartWithAll(A,w,Psi,R);
                    printf("%d) Bifurcating branch          ",bchart);MFPrintNVector(stdout,w);printf(" index %d, bif? %d\n",MFNVGetIndex(w),MFNVGetIndex2(w));fflush(stdout);
                    if(verbose){printf("Using Psi (w lies in the tangent space of u+ew.)\n");}
                   }else{
                    double R;
                    int bchart;
                    MFNSpaceScale(A->NSpace,1./vnorm,v,v);
                    MFGramSchmidtReplace(A->NSpace,Phi,v,w);
                    R=1.*MFChartRadius(A->chart[chart]);
                    MFIMFSetStability(A->M,w,Psi);
                    bchart=MFAtlasAddChartWithAll(A,w,Phi,R);
                    printf("%d) Bifurcating branch          ",bchart);MFPrintNVector(stdout,w);printf(" index %d, bif? %d\n",MFNVGetIndex(w),MFNVGetIndex2(w));fflush(stdout);
                    if(verbose){printf("Using modified TS (Phi)\n");}
                   }

                  MFFreeNVector(v);
                  MFFreeNKMatrix(Phi);
                  MFFreeNKMatrix(Psi);
                 }
                MFFreeNVector(w);
                MFFreeNVector(up);
                MFFreeKVector(s);

                return(chart);
               }else{
                if(verbose){printf("     Same as center (index %d -- type %d), going on to next vertex.\n",MFNVGetIndex(MFChartCenter(A->chart[chart])),MFNVGetIndex2(MFChartCenter(A->chart[chart])));}
/*              MFChartSetSingular(A->chart[chart]);*/
                bisection=0;
                goto nextvertex;
               }
             }
            goto bsect;
           }else{
            int in0,in1;
            int out0,out1;
            if(verbose){printf("     Check for stability change.\n");fflush(stdout);}
            in0=MFNVGetIndex(MFChartCenter(A->chart[chart]));
            in1=MFNVGetIndex(u);
            MFIMFSetStability(A->M,MFChartCenter(A->chart[chart]),MFChartTangentSpace(A->chart[chart]));
            MFIMFSetStability(A->M,u,MFChartTangentSpace(A->chart[chart]));
            out0=MFNVGetIndex(MFChartCenter(A->chart[chart]));
            out1=MFNVGetIndex(u);
            if(0&&in0!=out0){printf("   Index of center of chart %d changed from %d to %d\n",chart,in0,out0);fflush(stdout);}
            if(0&&in1!=out1){printf("   Index of new point           changed from %d to %d\n",in1,out1);fflush(stdout);}
            if(MFIMFStop(A->M,MFChartCenter(A->chart[chart]),MFChartTangentSpace(A->chart[chart]),u,MFChartTangentSpace(A->chart[chart])))
             {
              if(verbose){printf("     Stepped over stability change, invoke bisection\n");fflush(stdout);}
              t=.5;
              tl=0.;
              tr=1.;
              if(verbose){printf("            interval [%lf, %lf, %lf ] width %le \n",tl,t,tr,tr-tl);fflush(stdout);}
              bisection=1;
              goto bsect;
             }else{
              if(verbose){printf("     No stability change, invoke bisection\n");fflush(stdout);}

/*          Phi=MFIMFTangentSpaceWithGuess(A->M,u,MFChartTangentSpace(A->chart[chart]));*/
/*          dot=MFNKMatrixDot(A->NSpace,Phi,MFChartTangentSpace(A->chart[chart]));*/

              if(0&&dot>.2)
               {
                printf("%s - TangentSpaces too different (dot=%lf), at t=%lf, trying again at %lf.\n",RoutineName,dot,t,t/2);fflush(stdout);
                if(verbose)
                 {
                  printf("%s - TangentSpaces too different (dot=%lf), at R=%lf, trying again at %lf. Radius of vertex is %lf\n",RoutineName,dot,R,.8*R,rv);fflush(stdout);
                  printf("     Old is \n");MFPrintNKMatrix(stdout,MFChartTangentSpace(A->chart[chart]));printf("\n");fflush(stdout);
                  printf("     New is \n");MFPrintNKMatrix(stdout,Phi);printf("\n");fflush(stdout);
                 }
                t=t/2;
/*          MFFreeNKMatrix(Phi);*/
                goto bsect;
               }else{
                if(verbose){printf("%s - TangentSpaces OK (dot=%lf), at t=%lf\n",RoutineName,dot,t);fflush(stdout);}
  /*          MFFreeNKMatrix(Phi);*/
               }
             }
           }

          if(verbose){printf("      %s - Succeeded at R=%lf, delta=%lf, chart radius=%lf. Radius of vertex is %lf and is on chart %d\n",RoutineName,R,delta,MFChartRadius(A->chart[chart]),rv,chart);fflush(stdout);}
          if(dist!=(double*)NULL)
            *dist=MFNSpaceDistance(A->NSpace,u,up);
          if(verbose){printf("Distance from point to tangent space is %le\n",*dist);fflush(stdout);}

#ifdef MFTEST
          if(!MFTestBoundaryPoint(A,u,chart))
           {
            sprintf(MFAtlasErrorMsg,"Test of Manifold BPt failed.");
            printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
    printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
            MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
            abort();
           }
#endif

          if(verbose){printf("    found after examining %d charts\n",nChartsTried);fflush(stdout);}
          if(0&&verbose)
           {
            printf("Chart %d is now \n",chart);
            MFPrintChart(stdout,A->chart[chart]);
            fflush(stdout);
           }

          *TS=MFIMFTangentSpaceWithGuess(A->M,u,MFChartTangentSpace(A->chart[chart]));
          if(*TS==(MFNKMatrix)NULL)
           {
            t=.8*t;
            goto bsect;
           }

#ifdef MFTIMINGS
  MFTimeMFAtlasPointOnBoundaryInsideRegion+=clock()-starttime;
#endif
          MFPolytopeSetVertexMark(MFChartPolytope(A->chart[chart]),v,1);
          MFFreeNVector(up);
          MFFreeKVector(s);
          return(chart);
         }else{
          if(verbose){printf("  Interior vertex. ");fflush(stdout);}
         }
       }
      if(verbose)printf("   Fall through to nextvertex (n=%d)\n",n);
nextvertex:
      t=1.;
      v++;
      if(verbose&&v<n)
             printf("     Vertex %d has mark %d, index %d, radius %le/%le\n",v,
             MFPolytopeGetVertexMark(MFChartPolytope(A->chart[chart]),v),
             MFNVGetIndex2(MFChartCenter(A->chart[chart])),
             MFChartRadius(A->chart[chart]),MFRMin);
     }
      if(verbose&&v<n)
             printf("     Vertex %d has mark %d, index %d, radius %le/%le\n",v,
             MFPolytopeGetVertexMark(MFChartPolytope(A->chart[chart]),v),
             MFNVGetIndex2(MFChartCenter(A->chart[chart])),
             MFChartRadius(A->chart[chart]),MFRMin);
       else if(verbose)printf(" !(%d<%d)\n",v,n);
    if(verbose)printf("   Fall through to nextchart\n");
nextchart:
    if(verbose){printf("     Remove this chart (%d) from the boundary list\n",bchart);}
#ifdef FROMTOP
    MFAtlasRemoveChartFromBoundaryList(A,bchart); /*!!!*/
#else
    MFAtlasRemoveChartFromBoundaryList(A,A->nBnd-bchart-1); /*!!!*/
#endif
#ifndef NOBOUNDARYLIST
/*  bchart++;*/
    if(bchart>=A->nBnd)
     {
      bchart=0;
#else
/*  chart++;*/
    if(chart>=A->nCharts)
     {
      chart=0;
#endif
      chart=-1;
      pass++;
     }
   }
  MFFreeNVector(up);
  MFFreeKVector(s);

#ifdef MFTIMINGS
  MFTimeMFAtlasPointOnBoundaryInsideRegion+=clock()-starttime;
#endif
  return(-1);
 }

int MFAddHalfSpaceToAtlas(MFAtlas A,int gI,int left,int right,MFKVector nrm, double onrm)
 {
  static char RoutineName[]={"MFAddHalfSpaceToAtlas"};
  int result;
  int i,n;
  int verbose;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAddHalfSpaceToAtlas++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  verbose=0;
  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

  result=A->nHS;
  MFAtlasHalfSpaceSetIndex(A,result,gI);
  MFAtlasHalfSpaceSetOrigin(A,result,onrm);
  MFAtlasHalfSpaceSetNormal(A,result,nrm);
  MFAtlasHalfSpaceSetLeftChart(A,result,left);
  MFAtlasHalfSpaceSetRightChart(A,result,right);
  A->nHS++;

  if(verbose){printf("done %s\n",RoutineName);fflush(stdout);}
#ifdef MFTIMINGS
  MFTimeMFAddHalfSpaceToAtlas+=clock()-starttime;
#endif
  return(result);
 }

MFNVector MFAtlasChartCenter(MFAtlas A,int chart)
 {
  static char RoutineName[]={"MFAtlasChartCenter"};
  MFNVector result;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasChartCenter++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  result=MFChartCenter(A->chart[chart]);

#ifdef MFTIMINGS
  MFTimeMFAtlasChartCenter+=clock()-starttime;
#endif
  return result;
 }

/*
int MFAtlasIsPointInPolytope(MFAtlas A, int p,MFKVector pt)
 {
  static char RoutineName[]={"MFAtlasIsPointInPolytope"};
  int result;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasIsPointInPolytope++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  result=MFPolytopeInterior(MFChartPolytope(A->chart[p]),pt);

#ifdef MFTIMINGS
  MFTimeMFAtlasIsPointInPolytope+=clock()-starttime;
#endif
  return result;
 }
*/

int MFAtlasOffset(MFAtlas A)
 {
  static char RoutineName[]={"MFAtlasOffset"};
  int result;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasOffset++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  result=A->offset;

#ifdef MFTIMINGS
  MFTimeMFAtlasOffset+=clock()-starttime;
#endif
  return result;
 }

int MFAtlasIsHalfSpaceHyperCube(MFAtlas A,int i)
 {
  static char RoutineName[]={"MFAtlasIsHalfSpaceHyperCube"};
  int result;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasIsHalfSpaceHyperCube++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  result=i<A->offset;

#ifdef MFTIMINGS
  MFTimeMFAtlasIsHalfSpaceHyperCube+=clock()-starttime;
#endif
  return result;
 }

int MFAtlasLeftPolytope(MFAtlas A,int i)
 {
  static char RoutineName[]={"MFAtlasLeftPolytope"};
  int result;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasLeftPolytope++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  result=MFAtlasHalfSpaceLeftChart(A,i);

#ifdef MFTIMINGS
  MFTimeMFAtlasLeftPolytope+=clock()-starttime;
#endif
  return result;
 }

int MFAtlasRightPolytope(MFAtlas A,int i)
 {
  static char RoutineName[]={"MFAtlasRightPolytope"};
  int result;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasRightPolytope++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  result=MFAtlasHalfSpaceRightChart(A,i);

#ifdef MFTIMINGS
  MFTimeMFAtlasRightPolytope+=clock()-starttime;
#endif

  return result;
 }

int MFAtlasIsHalfSpaceInBothPolytopes(MFAtlas A,int hs)
 {
  static char RoutineName[]={"MFAtlasIsHalfSpaceInBothPolytopes"};
  int inLeft,inRight;
  int i,n;
  int result;
  int ghs;
  int g;
  MFPolytope P;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasIsHalfSpaceInBothPolytopes++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  ghs=MFAtlasHalfSpaceIndex(A,hs);

  inLeft=0;
  P=MFChartPolytope(A->chart[MFAtlasHalfSpaceLeftChart(A,hs)]);
  n=MFPolytopeNumberOfFaces(P);
  for(i=0;i<n;i++)
   {
    g=MFAtlasHalfSpaceIndex(A,MFPolytopeFaceIndex(P,i));
    if(g==ghs)inLeft=1;
   }

  inRight=0;
  P=MFChartPolytope(A->chart[MFAtlasHalfSpaceRightChart(A,hs)]);
  n=MFPolytopeNumberOfFaces(P);
  for(i=0;i<n;i++)
   {
    g=MFAtlasHalfSpaceIndex(A,MFPolytopeFaceIndex(P,i));
    if(g==ghs)inRight=1;
   }

  result=inLeft&&inRight;

#ifdef MFTIMINGS
  MFTimeMFAtlasIsHalfSpaceInBothPolytopes+=clock()-starttime;
#endif
  return(result);
 }

void MFAtlasAddChartToBoundaryList(MFAtlas A,int ball)
 {
  static char RoutineName[]={"MFAtlasAddChartToBoundaryList"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasAddChartToBoundaryList++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(A->nBnd>=A->mBnd)
   {
    A->mBnd+=MFAtlasIncrementToAllocateBoundaryList;
    A->bnd=(int*)realloc((void*)(A->bnd),A->mBnd*sizeof(int)); /*done*/
    if(A->bnd==(int*)NULL)
     {
      sprintf(MFAtlasErrorMsg,"Out of memory, trying to allocate %d bytes",A->mBnd*sizeof(int));
      printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
      MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasAddChartToBoundaryList+=clock()-starttime;
#endif
      return;
     }
   }

  A->bnd[A->nBnd]=ball;
  MFChartSetPositionInBoundaryList(A->chart[ball],A->nBnd);
  (A->nBnd)++;

#ifdef MFTIMINGS
  MFTimeMFAtlasAddChartToBoundaryList+=clock()-starttime;
#endif
  return;
 }

void MFAtlasRemoveChartFromBoundaryList(MFAtlas A,int ball)
 {
  static char RoutineName[]={"MFAtlasRemoveChartFromBoundaryList"};
  int i;
  int verbose=0;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasRemoveChartFromBoundaryList++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(verbose){printf("%s, boundary chart # %d\n",RoutineName,ball);fflush(stdout);}
  MFChartSetPositionInBoundaryList(A->chart[A->bnd[ball]],-1);
  if(ball<A->nBnd-1)
   {
/*  A->bnd[ball]=A->bnd[A->nBnd-1];*/
    for(i=ball;i<A->nBnd-1;i++)
     {
      A->bnd[i]=A->bnd[i+1];
      MFChartSetPositionInBoundaryList(A->chart[A->bnd[i]],i);
     }
    A->bnd[A->nBnd-1]=-1;
/*  MFChartSetPositionInBoundaryList(A->chart[A->bnd[ball]],ball);*/
    (A->nBnd)--;
   }else{
    A->bnd[ball]=-1;
    A->nBnd--;
   }

#ifdef MFTIMINGS
  MFTimeMFAtlasRemoveChartFromBoundaryList+=clock()-starttime;
#endif
  return;
 }

void MFAddInitialHyperplanes(MFAtlas A)
 {
  static char RoutineName[]={"MFAddInitialHyperplanes"};
  int i,j,ij;
  MFKVector nrm;
  double onrm;
#ifndef MFSIMPLEX
  double *c;
  int carry;
#else
  MFPolytope simplex;
#endif

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAddInitialHyperplanes++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

#ifndef MFSIMPLEX
  A->offset=1;for(i=0;i<A->k;i++)A->offset*=2;

  c=(double*)malloc(A->k*sizeof(double)); /*done*/
  if(c==(double*)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Out of memory, trying to allocate %d bytes",A->k*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
    MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAddInitialHyperplanes+=clock()-starttime;
#endif
    return;
   }
  for(j=0;j<A->k;j++)c[j]=-1.;

  for(i=0;i<A->offset;i++)
   {
    nrm=MFCreateKVector(A->k);
    for(j=0;j<A->k;j++)MFKVSetC(nrm,j,c[j]);
    onrm=1;
    if(i%2==0)onrm=-1;

    carry=1;
    j=0;
    while(carry&&j<A->k)
     {
      if(c[j]<0)
       {
        c[j]=1;
        carry=0;
       }else{
        c[j]=-1;
        carry=1;
        j++;
       }
     }
    ij=MFAddHalfSpaceToAtlas(A,-1,0,0,nrm,onrm);
    MFFreeKVector(nrm);
   }
  free(c);
#else
  A->offset=A->k+1;
  simplex=MFCreateSimplexAtOrigin(A->k,1.);
  for(i=0;i<A->offset;i++)
   {
    nrm=MFPolytopeFaceNormal(simplex,i);
    MFRefKVector(nrm);
    onrm=MFPolytopeFaceOrigin(simplex,i);
    ij=MFAddHalfSpaceToAtlas(A,-1,0,0,nrm,onrm);
    MFFreeKVector(nrm);
   }
  MFFreePolytope(simplex);
#endif

#ifdef MFTIMINGS
  MFTimeMFAddInitialHyperplanes+=clock()-starttime;
#endif
  return;
 }

int MFAtlasAddChartToList(MFAtlas A,MFNVector u, MFNKMatrix Phi,double R)
 {
  static char RoutineName[]={"MFAtlasAddChartToList"};
  int i;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasAddChartToList++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(!(A->nCharts<A->mCharts))
   {
    A->mCharts+=MFAtlasIncrementToAllocate;

    A->chart=(MFChart*)realloc((void*)A->chart,A->mCharts*sizeof(MFChart)); /*done*/
    if(A->chart==(MFChart*)NULL)
     {
      sprintf(MFAtlasErrorMsg,"Out of memory, trying to allocate %d bytes",A->mCharts*sizeof(MFChart));
      printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
      MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasAddChartToList+=clock()-starttime;
#endif
      return -1;
     }
    for(i=A->nCharts;i<A->mCharts;i++)A->chart[i]=(MFChart)NULL;
   }

  A->chart[A->nCharts]=MFCreateChart(A->M,u,Phi,R);
  A->nCharts++;

#ifdef MFTIMINGS
  MFTimeMFAtlasAddChartToList+=clock()-starttime;
#endif
  return(A->nCharts-1);
 }

int MFTestBoundaryPoint(MFAtlas A,MFNVector u, int chart)
 {
  static char RoutineName[]={"MFTestBoundaryPoint"};
  int i,n;
  double ds,du,dn;
  MFKVector s;
  MFNVector up;
  MFNVector diff;
  int good;
  int verbose;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFTestBoundaryPoint++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  verbose=0;

  n=MFAtlasNumberOfCharts(A);
  s=MFCreateKVector(A->k);
  up=MFCloneNVector(u);
  diff=MFCloneNVector(u);

  if(verbose){printf("Testing point ");MFPrintNVector(stdout,u);printf(" on chart %d\n",chart);fflush(stdout);}
  good=1;
  for(i=0;i<n;i++)
   {
    du=MFNSpaceDistance(A->NSpace,MFChartCenter(A->chart[i]),u);
/*  if(du<.98*MFChartRadius(A->chart[i]))*/
     {
      MFChartProjectIntoTangentSpace(A->chart[i],u,s);
      ds=MFKVNorm(s);
      if(verbose){printf("   %d, center=",i);MFPrintNVector(stdout,MFChartCenter(A->chart[i]));printf(" R=%lf, distance=%lf",MFChartRadius(A->chart[i]),ds);}
/*    if(ds<.98*MFChartRadius(A->chart[i]))*/
       {
        MFChartPointInTangentSpace(A->chart[i],s,up);
        dn=MFNSpaceDistance(A->NSpace,up,u);
        if(i!=chart &&
           ds<.98*MFChartRadius(A->chart[i]) && 
           du<.98*MFChartRadius(A->chart[i]) &&
           dn<MFEpsilon
          )
         {
          printf("   Test point is on chart %d, ",chart);
          MFPrintNVector(stdout,u);
          printf("\n");
          printf("   Failed against chart %d, center=",i,chart);
          MFPrintNVector(stdout,MFChartCenter(A->chart[i]));
          printf(" R=%lf, distance=%lf",MFChartRadius(A->chart[i]),du);
          printf(" tang=%lf",ds);fflush(stdout);
          printf(" perp=%lf\n",dn);fflush(stdout);
          printf("NO GOOD!\n");fflush(stdout);
          if(ds<.5*MFChartRadius(A->chart[i]))good=0;
          if(du<MFChartRadius(A->chart[i]))
             printf(" distance in N-space %lf is < radius %lf!\n",du,MFChartRadius(A->chart[i]));fflush(stdout);
          if(ds<MFChartRadius(A->chart[i]))
             printf(" distance in tangent space %lf is < radius %lf!\n",ds,MFChartRadius(A->chart[i]));fflush(stdout);
          if(dn<MFEpsilon)
             printf(" distance in normal space %lf is < epsilon %lf!\n",dn,MFEpsilon);fflush(stdout);
          MFChartProjectIntoTangentSpace(A->chart[chart],u,s);
          printf(" In chart %d point is ",chart);MFPrintKVector(stdout,s);printf("\n");fflush(stdout);
          MFChartProjectIntoTangentSpace(A->chart[i],u,s);
          printf(" In chart %d point is ",i);MFPrintKVector(stdout,s);printf("\n");fflush(stdout);
         }
       }
     }
   }
  MFFreeKVector(s);
  MFFreeNVector(up);
  MFFreeNVector(diff);
#ifdef MFTIMINGS
  MFTimeMFTestBoundaryPoint+=clock()-starttime;
#endif
  return(good);
 }

MFChart MFAtlasChart(MFAtlas A, int chart)
 {
  static char RoutineName[]={"MFAtlasChart"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasChart++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasChart+=clock()-starttime;
#endif
    return (MFChart)NULL;
   }

  if(chart<0||chart>=A->nCharts)
   {
    sprintf(MFAtlasErrorMsg,"Chart %d (argument 2) is invalid. Must be positive and less than %d (total number of charts).",chart,A->nCharts);
    MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasChart+=clock()-starttime;
#endif
    return (MFChart)NULL;
   }

#ifdef MFTIMINGS
  MFTimeMFAtlasChart+=clock()-starttime;
#endif
  return A->chart[chart];
 }

int MFAtlasNumberOfHalfSpaces(MFAtlas A)
 {
  static char RoutineName[]={"MFAtlasNumberOfHalfSpaces"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasNumberOfHalfSpaces++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasNumberOfHalfSpaces+=clock()-starttime;
#endif
    return -1;
   }

#ifdef MFTIMINGS
  MFTimeMFAtlasNumberOfHalfSpaces+=clock()-starttime;
#endif
  return A->nHS;
 }

MFNKMatrix MFAtlasChartTangentSpace(MFAtlas A,int chart)
 {
  static char RoutineName[]={"MFAtlasChartTangentSpace"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasChartTangentSpace++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasChartTangentSpace+=clock()-starttime;
#endif
    return (MFNKMatrix)NULL;
   }

  if(chart<0||chart>=A->nCharts)
   {
    sprintf(MFAtlasErrorMsg,"Chart %d (argument 2) is invalid. Must be positive and less than %d (total number of charts).",chart,A->nCharts);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasChartTangentSpace+=clock()-starttime;
#endif
    return (MFNKMatrix)NULL;
   }

#ifdef MFTIMINGS
  MFTimeMFAtlasChartTangentSpace+=clock()-starttime;
#endif
  return MFChartTangentSpace(A->chart[chart]);
 }

void MFAtlasSetVerbose(MFAtlas A, int verbose)
 {
  static char RoutineName[]={"MFAtlasSetVerbose"};

#ifdef MFDEBUG
  _heap_check();
#endif

  MFVerbose=verbose;
  return;
 }

void MFAtlasSetDotMin(MFAtlas A, double d)
 {
  static char RoutineName[]={"MFAtlasSetDotMin"};

#ifdef MFDEBUG
  _heap_check();
#endif

  MFDotMin=d;
  return;
 }

void MFAtlasSetEpsilon(MFAtlas A, double e)
 {
  static char RoutineName[]={"MFAtlasSetEpsilon"};

#ifdef MFDEBUG
  _heap_check();
#endif

  MFEpsilon=e;
  return;
 }

void MFAtlasSetRMin(MFAtlas A, double r)
 {
  static char RoutineName[]={"MFAtlasSetRMin"};

#ifdef MFDEBUG
  _heap_check();
#endif

  MFRMin=r;
  return;
 }

void MFIMFSetSpace(MFImplicitMF,MFNSpace);

MFAtlas MFReadAtlas(FILE *fid)
 {
  static char RoutineName[]={"MFReadAtlas"};
  MFAtlas A;
  int i;
  int verbose;
  double tmp;
  int itmp;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFReadAtlas++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  verbose=0;

  if(verbose){printf("%s\n");fflush(stdout);}

  A=(MFAtlas)malloc(sizeof(struct MFAtlasSt)); /*done*/
  if(verbose){printf("   A=0x%8.8x\n",A);fflush(stdout);}
  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFAtlasSt));
    printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
    MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFReadAtlas+=clock()-starttime;
#endif
    return (MFAtlas)NULL;
   }

  fscanf(fid,"%d\n",&(A->n));
  if(verbose){printf("   n=%d\n",A->n);fflush(stdout);}
  fscanf(fid,"%d\n",&(A->k));
  if(verbose){printf("   k=%d\n",A->k);fflush(stdout);}
  A->M=MFReadImplicitMF(fid);
  if(verbose){printf("   M=%8.8x\n",A->M);fflush(stdout);}
  A->NSpace=MFReadNSpace(fid);
  if(verbose){printf("   NSpace=%8.8x\n",A->NSpace);fflush(stdout);}
  MFIMFSetSpace(A->M,A->NSpace);

/* List of Charts */

  fscanf(fid,"%d\n",&(A->nCharts));
  if(verbose){printf("   nCharts=%d\n",A->nCharts);fflush(stdout);}
  fscanf(fid,"%d\n",&(A->mCharts));
  if(verbose){printf("   mCharts=%d\n",A->mCharts);fflush(stdout);}
  A->chart=(MFChart*)malloc(A->mCharts*sizeof(MFChart)); /*done*/
  if(A->chart==(MFChart*)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Out of memory, trying to allocate %d bytes",A->mCharts*sizeof(MFChart));
    printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
    MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFReadAtlas+=clock()-starttime;
#endif
    return (MFAtlas)NULL;
   }
  for(i=0;i<A->nCharts;i++)
   {
    A->chart[i]=MFReadChart(fid,A);
    if(A->chart[i]==(MFChart)NULL){printf("   chart %d is NULL!\n",i);fflush(stdout);}
    if(verbose){printf("   chart[%d]=0x%8.8x\n",i,A->chart[i]);fflush(stdout);}
   }

/* List of HalfSpaces */

  fscanf(fid,"%d\n",&(A->offset));
  if(verbose){printf("   offset=%d\n",A->offset);fflush(stdout);}
  fscanf(fid,"%d\n",&(A->nHS));
  if(verbose){printf("   nHS=%d\n",A->nHS);fflush(stdout);}
  fscanf(fid,"%d\n",&(A->mHS));
  if(verbose){printf("   mHS=%d\n",A->mHS);fflush(stdout);}

#ifdef OLDHS
  MFAtlasHalfSpaceInit(A);
  A->onrm=(double*)malloc(A->mHS*sizeof(double)); /*done*/
  if(A->onrm==(double*)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Out of memory, trying to allocate %d bytes",A->k*A->mHS*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
    MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFReadAtlas+=clock()-starttime;
#endif
    return (MFAtlas)NULL;
   }
  if(verbose){printf("   read onrm\n");fflush(stdout);}
  for(i=0;i<A->nHS;i++)fscanf(fid,"%lf\n",&(A->onrm[i]));
#ifdef DONTREAD
  for(i=0;i<A->nHS;i++)fscanf(fid,"%lf\n",&tmp);
#endif
  fscanf(fid,"%d\n",&(A->nGlobalIndex));

  A->globalIndex=(int*)malloc(A->mHS*sizeof(int)); /*done*/
  if(A->globalIndex==(int*)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Out of memory, trying to allocate %d bytes",A->mHS*sizeof(int));
    printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
    MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFReadAtlas+=clock()-starttime;
#endif
    return (MFAtlas)NULL;
   }
  if(verbose){printf("   read globalIndex\n");fflush(stdout);}
  for(i=0;i<A->nHS;i++)fscanf(fid,"%ld\n",&(A->globalIndex[i]));

  A->nrm=(MFKVector*)malloc(A->mHS*sizeof(MFKVector)); /*done*/
  if(A->nrm==(MFKVector*)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Out of memory, trying to allocate %d bytes",A->mHS*sizeof(MFKVector));
    printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
    MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFReadAtlas+=clock()-starttime;
#endif
    return (MFAtlas)NULL;
   }
  if(verbose){printf("   read normals\n");fflush(stdout);}
  for(i=0;i<A->nHS;i++)A->nrm[i]=MFReadKVector(fid);
#ifdef DONTREAD
  A->nrm=(MFKVector*)malloc(sizeof(MFKVector)); /*done*/
  for(i=0;i<A->nHS;i++)A->nrm[0]=MFReadKVector(fid);
#endif

  A->leftChart=(int*)malloc(A->mHS*sizeof(int)); /*done*/
  if(A->leftChart==(int*)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Out of memory, trying to allocate %d bytes",A->mHS*sizeof(int));
    printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
    MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFReadAtlas+=clock()-starttime;
#endif
    return (MFAtlas)NULL;
   }
  if(verbose){printf("   read chart on left\n");fflush(stdout);}
  for(i=0;i<A->nHS;i++)fscanf(fid,"%ld\n",&(A->leftChart[i]));
#ifdef DONTREAD
  for(i=0;i<A->nHS;i++)fscanf(fid,"%ld\n",&itmp);
#endif

  A->rightChart=(int*)malloc(A->mHS*sizeof(int)); /*done*/
  if(A->rightChart==(int*)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Out of memory, trying to allocate %d bytes",A->mHS*sizeof(int));
    printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
    MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFReadAtlas+=clock()-starttime;
#endif
    return (MFAtlas)NULL;
   }
  if(verbose){printf("   read chart on right\n");fflush(stdout);}
  for(i=0;i<A->nHS;i++)fscanf(fid,"%ld\n",&(A->rightChart[i]));
#ifdef DONTREAD
  for(i=0;i<A->nHS;i++)fscanf(fid,"%ld\n",&itmp);
#endif

#else
  MFAtlasHalfSpaceInit(A);
  if(verbose){printf("   read onrm\n");fflush(stdout);}
  for(i=0;i<A->nHS;i++)
   {
    fscanf(fid,"%lf\n",&tmp);
    MFAtlasHalfSpaceSetOrigin(A,i,tmp);
   }
  fscanf(fid,"%d\n",&(A->nGlobalIndex));

  if(verbose){printf("   read globalIndex\n");fflush(stdout);}
  for(i=0;i<A->nHS;i++)
   {
    fscanf(fid,"%ld\n",&itmp);
    MFAtlasHalfSpaceSetIndex(A,i,itmp);
   }

  if(verbose){printf("   read normals\n");fflush(stdout);}
  for(i=0;i<A->nHS;i++)
   {
    MFAtlasHalfSpaceSetNormal(A,i,MFReadKVector(fid));
   }

  if(verbose){printf("   read chart on left\n");fflush(stdout);}
  for(i=0;i<A->nHS;i++)
   {
    fscanf(fid,"%ld\n",&itmp);
    MFAtlasHalfSpaceSetLeftChart(A,i,itmp);
   }

  if(verbose){printf("   read chart on right\n");fflush(stdout);}
  for(i=0;i<A->nHS;i++)
   {
    fscanf(fid,"%ld\n",&itmp);
    MFAtlasHalfSpaceSetRightChart(A,i,itmp);
   }
#endif

/* Binary tree */

#ifndef NOBINARYTREE
  if(verbose){printf("   read binary tree\n");fflush(stdout);}
  A->BTree=MFReadBinaryTree(fid);
#endif

#ifndef NOBOUNDARYLIST
/* List of Boundarys */

  fscanf(fid,"%d\n",&(A->nBnd));
  fscanf(fid,"%d\n",&(A->mBnd));
  if(verbose){printf("   nBnd=%d\n",A->nBnd);fflush(stdout);}

  A->bnd=(int*)malloc(A->mBnd*sizeof(int)); /*done*/
  if(A->bnd==(int*)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Out of memory, trying to allocate %d bytes",A->mBnd*sizeof(int));
    printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
    MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFReadAtlas+=clock()-starttime;
#endif
    return (MFAtlas)NULL;
   }
  for(i=0;i<A->nBnd;i++)
   {
    fscanf(fid,"%d\n",&(A->bnd[i]));
    MFChartSetPositionInBoundaryList(A->chart[A->bnd[i]],i);
   }
#endif

  if(verbose){printf("   done %s\n",RoutineName);fflush(stdout);}
#ifdef MFTIMINGS
  MFTimeMFReadAtlas+=clock()-starttime;
#endif
  return A;
 }

void MFWriteAtlas(FILE *fid,MFAtlas A)
 {
  static char RoutineName[]={"MFWriteAtlas"};
  int i;
  int verbose;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFWriteAtlas++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  verbose=0;

  if(verbose){printf("%s\n");fflush(stdout);}

  fprintf(fid,"%d\n",A->n);
  fprintf(fid,"%d\n",A->k);
  MFWriteImplicitMF(fid,A->M);
  MFWriteNSpace(fid,A->NSpace);

/* List of Charts */

  fprintf(fid,"%d\n",A->nCharts);
  fprintf(fid,"%d\n",A->mCharts);
  for(i=0;i<A->nCharts;i++)
   {
    MFWriteChart(fid,A->chart[i]);
    MFChartClean(A->chart[i]);
   }

/* List of HalfSpaces */

  fprintf(fid,"%d\n",A->offset);
  fprintf(fid,"%d\n",A->nHS);
  fprintf(fid,"%d\n",A->mHS);
  for(i=0;i<A->nHS;i++)
    fprintf(fid,"%lf\n",MFAtlasHalfSpaceOrigin(A,i));
  fprintf(fid,"%d\n",A->nGlobalIndex);
  for(i=0;i<A->nHS;i++)fprintf(fid,"%ld\n",A->globalIndex[i]);
  for(i=0;i<A->nHS;i++)MFWriteKVector(fid,MFAtlasHalfSpaceNormal(A,i));
  for(i=0;i<A->nHS;i++)fprintf(fid,"%ld\n",MFAtlasHalfSpaceLeftChart(A,i));
  for(i=0;i<A->nHS;i++)fprintf(fid,"%ld\n",MFAtlasHalfSpaceRightChart(A,i));

/* Binary tree */

#ifndef NOBINARYTREE
  MFWriteBinaryTree(fid,A->BTree);
#endif

#ifndef NOBOUNDARYLIST
/* List of Boundarys */

  fprintf(fid,"%d\n",A->nBnd);
  fprintf(fid,"%d\n",A->mBnd);
  for(i=0;i<A->nBnd;i++)fprintf(fid,"%d\n",A->bnd[i]);
#endif

#ifdef MFTIMINGS
  MFTimeMFWriteBinaryTree+=clock()-starttime;
#endif
  return;
 }

MFImplicitMF MFAtlasMF(MFAtlas A)
 {
  static char RoutineName[]={"MFAtlasMF"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasMF++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasMF+=clock()-starttime;
#endif
    return (MFImplicitMF)NULL;
   }

#ifdef MFTIMINGS
  MFTimeMFAtlasMF+=clock()-starttime;
#endif
  return(A->M);
 }

int MFTestAtlas(MFAtlas A)
 {
  static char RoutineName[]={"MFTestAtlas"};
  int i,n,j,m,l;
  MFPolytope P;
  MFChart chart;
  MFKVector s;
  MFNVector u;
  int good;
  int verbose;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFTestAtlas++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  verbose=0;

  n=MFAtlasNumberOfCharts(A);
  s=MFCreateKVector(A->k);
  if(n>0)u=MFCloneNVector(MFChartCenter(A->chart[0]));
   else u=(MFNVector)NULL;

  if(verbose){printf("Testing Atlas ");fflush(stdout);}

/* 1. Is any polytope vertex inside another any other polytope? */

  good=1;
  for(i=0;i<n;i++)
   {
    chart=MFAtlasChart(A,i);
    P=MFChartPolytope(chart);
    good=good&&MFTestPolytope(P);
#ifdef MFJ
    m=MFPolytopeNumberOfVertices(P);
    for(l=0;l<m;l++)
     {
      MFPolytopeVertex(P,l,s);
      MFChartPointInTangentSpace(chart,s,u);
      for(j=0;j<n;j++)
       {
        if(i!=j)
         {
          MFChartProjectIntoTangentSpace(A->chart[j],u,s);
          if(MFChartTotallyInterior(MFAtlasChart(A,j),s,.001))
           {
            good=0;
            printf("Test Atlas, vertex %d of chart %d is in the polytope of chart %d\n",l,i,j);fflush(stdout);
           }
         }
       }
     }
#endif
   }

  MFFreeKVector(s);
  if(u!=(MFNVector)NULL)MFFreeNVector(u);

#ifdef MFTIMINGS
  MFTimeMFTestAtlas+=clock()-starttime;
#endif
  return(good);
 }

void MFAtlasReduceChartRadius(MFAtlas A,int chart,double R)
 {
  static char RoutineName[]={"MFAtlasReduceChartRadius"};
  int k;
  MFChart c;
  MFPolytope P,Q;
  double R0;
  double Ri,Rj;
  MFNVector ui,uj;
  MFNKMatrix Phij;

  int nn,f;
  int *ncharts;
  double *no;
  MFKVector *nnrm;
  int *nindx;
  int i,j,m;
  int nHS;
  double *o;
  MFKVector *nrm;
  int qindx;
  double qo;
  MFKVector qnrm;
  MFNVector dir;
  MFKVector projdir;
  int *indx;
  double dist;
  int BBDimension=0;
  double *BBCenter=(double*)NULL;
  int verbose;

#ifdef MFDEBUG
  _heap_check();
#endif

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasReduceChartRadius++;
  starttime=clock();
#endif

  verbose=0;
  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

  c=MFAtlasChart(A,chart);
  k=MFAtlasK(A);
  R0=MFChartRadius(c);
  P=MFChartPolytope(c);
  dir=MFCloneNVector(MFAtlasChartCenter(A,chart));
  projdir=MFCreateKVector(A->k);

  /* Check that R<R_chart */

  if(R>=R0)
   {
    sprintf(MFAtlasErrorMsg,"New radius (%lf) must be less than old (%lf), chart will not be changed.",R,R0);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasReduceChartRadius+=clock()-starttime;
#endif
    return;
   }

  /* Make a list of all charts that border chart */

  if(verbose){printf("Reducing chart %d from R=%lf to %lf\n",chart,R0,R);fflush(stdout);}
  if(0&&verbose){printf("Atlas is\n");MFPrintAtlas(stdout,A);fflush(stdout);}

  nn=MFPolytopeNumberOfFaces(P);
  ncharts=(int*)malloc(nn*sizeof(int));
  nindx=(int*)malloc(nn*sizeof(int));
  no=(double*)malloc(nn*sizeof(double));
  nnrm=(MFKVector*)malloc(nn*sizeof(MFKVector));
  if(verbose){printf(" chart %d adjoins %d other charts\n",chart,nn);fflush(stdout);}
  nn=0;
  for(i=0;i<MFPolytopeNumberOfFaces(P);i++)
   {
    f=MFPolytopeFaceIndex(P,i);
    if(!MFAtlasIsHalfSpaceHyperCube(A,f))
     {
      nindx[nn]=f;
      if((ncharts[nn]=MFAtlasHalfSpaceRightChart(A,f))==chart)
          ncharts[nn]=MFAtlasHalfSpaceLeftChart(A,f);
      no[nn]=MFAtlasHalfSpaceOrigin(A,f);
      nnrm[nn]=MFAtlasHalfSpaceNormal(A,f);
      if(verbose){printf("    %d\n",ncharts[nn]);fflush(stdout);}
      nn++;
     }
   }

  /* Remove all halfspaces from chart */

  if(verbose){printf(" Remove the polytope of this chart %d",chart);fflush(stdout);}

  if(P!=(MFPolytope)NULL)MFFreePolytope(P);

  /* Set P_chart=C_R */
  if(verbose){printf(" Create a new hypercube for the polytope of this chart %d\n",chart);fflush(stdout);}

#ifndef MFSIMPLEX
  P=MFCreateHyperCubeAtOrigin(k,1.05*R);
#else
  P=MFCreateSimplexAtOrigin(k,1.05*R);
#endif
  MFSetChartPolytope(c,P);
  MFSetChartRadius(c,R);

  if(verbose){printf(" For each adjoining chart:\n");fflush(stdout);}
  for(i=0;i<nn;i++)
   {
    if(verbose){printf("chart %d:\n",ncharts[i]);fflush(stdout);}
    Q=MFChartPolytope(A->chart[ncharts[i]]);

  /* Make list of half spaces to be removed from Q */

    nHS=MFPolytopeNumberOfFaces(Q);
    o=(double*)malloc(nHS*sizeof(double));
    nrm=(MFKVector*)malloc(nHS*sizeof(MFKVector));
    indx=(int*)malloc(nHS*sizeof(int));
    if(verbose){printf("     there are %d faces:\n",nHS);fflush(stdout);}
    m=0;
    qindx=-1;
    for(j=0;j<nHS;j++)
     {
      f=MFPolytopeFaceIndex(Q,j);
      if(!MFAtlasIsHalfSpaceHyperCube(A,f))
       {
        if(MFAtlasHalfSpaceRightChart(A,f)!=chart
                        && MFAtlasHalfSpaceLeftChart(A,f)!= chart)
         {
          indx[m]=f;
          o[m]=MFAtlasHalfSpaceOrigin(A,f);
          nrm[m]=MFAtlasHalfSpaceNormal(A,f);
          m++;
         }else{
          qindx=f;
          qo=MFAtlasHalfSpaceOrigin(A,f);
          qnrm=MFAtlasHalfSpaceNormal(A,f);
         }
       }
     }
    if(0&&qindx<0){printf("Problem, qindx[%d] is <0\n",i);fflush(stdout);abort();}
    if(Q!=(MFPolytope)NULL)MFFreePolytope(Q);

  /* Set Q=C_R */
   if(verbose){printf("  Create a new Hypercube for chart %d\n",ncharts[i]);fflush(stdout);}

    R0=MFAtlasChartRadius(A,ncharts[i]);
#ifndef MFSIMPLEX
    Q=MFCreateHyperCubeAtOrigin(k,1.05*R0);
#else
    Q=MFCreateSimplexAtOrigin(k,1.05*R0);
#endif
    MFSetChartPolytope(A->chart[ncharts[i]],Q);

  /* Subtract each HS from Q, omitting HS_P */
    if(verbose){printf("  Subtract %d halfspaces from chart %d\n",m,ncharts[i]);fflush(stdout);}

    for(j=0;j<m;j++)
     {
      if(verbose){printf("     %d, %d\n",j,indx[j]);fflush(stdout);}
      MFSubtractHalfSpaceFromPolytope(Q,indx[j],nrm[j],o[j]);
     }

    if(verbose){MFPrintPolytope(stdout,Q);fflush(stdout);}

    free(o);
    free(nrm);
    free(indx);

    if(verbose){printf("Done, now subtract reduced/old intersections\n");fflush(stdout);}

/* Now deal with reduced polytope */

    Ri=MFChartRadius(c);
    ui=MFChartCenter(c);
    Rj=MFChartRadius(A->chart[ncharts[i]]);
    uj=MFChartCenter(A->chart[ncharts[i]]);
    Phij=MFChartTangentSpace(A->chart[ncharts[i]]);

  /* Subtract new HS_P from Q */

    if(qindx>-1)
     {
      if(verbose){printf("  Subtract updated halfspaces from chart %d\n",ncharts[i]);fflush(stdout);}

      MFNSpaceDirection(A->NSpace,uj,ui,dir);
      MFChartProjectVectorIntoTangentSpace(A->chart[ncharts[i]],dir,qnrm);
      dist=MFKVDot(qnrm,qnrm);
      qo=.5*(dist+Rj*Rj-Ri*Ri);
      MFAtlasHalfSpaceSetOrigin(A,qindx,qo);

      if(verbose){printf("  qindx=%d\n",qindx);fflush(stdout);}
  
      if(verbose){printf("   direction from %d to %d is ",ncharts[i],chart);MFPrintNVector(stdout,dir);printf("\n");fflush(stdout);}
      if(verbose){printf("   in TS %d is ",ncharts[i]);MFPrintKVector(stdout,qnrm);printf("\n");fflush(stdout);}
      if(verbose){printf("   distance in TS is %lf\n",sqrt(dist));fflush(stdout);}
  
      if(1)MFSubtractHalfSpaceFromPolytope(Q,qindx,qnrm,qo);

  /* Subtract new HS_Q from P */
      if(verbose){printf("  Subtract updated halfspaces from chart %d\n",chart);fflush(stdout);}

      MFNSpaceDirection(A->NSpace,ui,uj,dir);
      MFChartProjectVectorIntoTangentSpace(c,dir,nnrm[i]);
      dist=MFKVDot(nnrm[i],nnrm[i]);
      no[i]=.5*(dist+Ri*Ri-Rj*Rj);
      MFAtlasHalfSpaceSetOrigin(A,nindx[i],no[i]);

      if(verbose){printf("  nindx[i]=%d\n",nindx[i]);fflush(stdout);}
      if(verbose){printf("   direction from %d to %d is ",chart,ncharts[i]);MFPrintNVector(stdout,dir);printf("\n");fflush(stdout);}
      if(verbose){printf("   in TS %d is ",chart);MFPrintKVector(stdout,nnrm[i]);printf("\n");fflush(stdout);}
      if(verbose){printf("   distance in TS is %lf\n",sqrt(dist));fflush(stdout);}

      if(1)MFSubtractHalfSpaceFromPolytope(P,nindx[i],nnrm[i],no[i]);
     }
   }
  if(verbose){printf("  free temporary\n");fflush(stdout);}

  free(ncharts);
  free(nindx);
  free(no);
  free(nnrm);
  if(verbose){printf("done %s\n",RoutineName);fflush(stdout);}

  BBDimension=MFIMFProjectToBB(A->M,MFChartCenter(A->chart[chart]),(double*)NULL);
  BBCenter=(double*)malloc(BBDimension*sizeof(double));
  if(verbose){printf("       %s, BBDimension=%d\n",RoutineName,BBDimension);fflush(stdout);}

  MFIMFProjectToBB(A->M,MFChartCenter(A->chart[chart]),BBCenter);
  if(verbose)
   {
    printf("      BBCenter=(%lf",BBCenter[0]);
    for(i=1;i<BBDimension;i++)printf(",%lf",BBCenter[i]);
    printf(" )\n");fflush(stdout);
   }
  MFFreeNVector(dir);
  MFFreeKVector(projdir);

  MFRecomputeBoundingBoxes(A->BTree,chart,BBCenter,MFChartRadius(A->chart[chart]));

#ifdef MFTIMINGS
  MFTimeMFAtlasReduceChartRadius+=clock()-starttime;
#endif
  return;
 }

MFNSpace MFAtlasNSpace(MFAtlas A)
 {
  static char RoutineName[]={"MFAtlasNSpace"};

#ifdef MFDEBUG
  _heap_check();
#endif

  return(A->NSpace);
 }

double MFAtlasDet(int,double*);

double MFNKMatrixDot(MFNSpace space,MFNKMatrix A,MFNKMatrix B)
 {
  static char RoutineName[]={"MFNKMatrixDot"};
  double *Q;
  double dot;
  MFNVector col;
  MFKVector C;
  int k,n;
  int i,j;
  int verbose=0;

#ifdef MFDEBUG
  _heap_check();
#endif

  if(verbose)
   {
    printf("%s\n",RoutineName);fflush(stdout);
    printf("Phi_i:\n");MFPrintNKMatrix(stdout,A);
    printf("Phi_j:\n");MFPrintNKMatrix(stdout,B);
   }

  n=MFNKMatrixN(A);
  k=MFNKMatrixK(A);
  if(k>3)return 0.;
  Q=(double*)malloc(k*k*sizeof(double));
  if(Q==(double*)NULL)abort();
  C=MFCreateKVector(k);
  for(i=0;i<k;i++)
   {
    col=MFMColumn(B,i);
    MFMVMulT(space,A,col,C);
    for(j=0;j<k;j++)Q[i+j*k]=MFKV_C(C,j);
    MFFreeNVector(col);
   }

  if(verbose)
   {
    printf("Phi_i^TPhi_j=\n");
    for(i=0;i<k;i++)
     {
      printf("[");
      for(j=0;j<k;j++)
       {
        if(j>0)printf(",");
        printf("%lf",Q[i+k*j]);
       }
      printf("]\n");
     }
    fflush(stdout);
   }
/*for(i=0;i<k;i++)Q[i+k*i]-=1.;*/
  dot=pow(fabs(MFAtlasDet(k,Q)),1./k);
  free(Q);
  MFFreeKVector(C);
  return dot;
 }

double MFAtlasDet(int n, double *A)
 {
  static char RoutineName[]={"MFAtlasDet"};

#ifdef MFDEBUG
  _heap_check();
#endif

  if(n==1)return A[0];
   else if(n==2)return A[0]*A[3]-A[1]*A[2];
   else if(n==3)return A[0]*(A[4]*A[8]-A[5]*A[7])
                      -A[1]*(A[3]*A[8]-A[5]*A[6])
                      +A[2]*(A[3]*A[7]-A[4]*A[6]);
  
  return MFFullProdEV(n,0,A);
 }

int MFAtlasAddChartWithAll(MFAtlas A,MFNVector u,MFNKMatrix Phi,double R)
 {
  static char RoutineName[]={"MFAtlasAddChartWithAll"};
#ifdef NOBINARYTREE
  int i;
#endif
  int j,jj;
  int ii,n;
  MFKVector nrm;
  double onrm;
  double dist,sdist;
  double ddist;
  double tdist;
  double dot;
  MFNVector dir;
  MFPolytope P;
  int chart;
  int bchart;
  double Rstar;
  MFNVector uj;
  MFNKMatrix Phij;
  double Rj;
  int nInt,ibnd;
  int ni;
  double RMin;
  double R0;
  double RMax;
  double sqrtTwo;
  int BBDimension=0;
  double *BBCenter=(double*)NULL;
  int mark;

  int ij,ji;
  MFListOfCharts L;

  int verbose;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasAddChartWithAll++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  verbose=0;
  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

  sqrtTwo=sqrt(2.);

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasAddChartWithApproxTS+=clock()-starttime;
#endif
    return -1;
   }

  if(u==(MFNVector)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to u (argument 2) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasAddChartWithApproxTS+=clock()-starttime;
#endif
    return -1;
   }

  if(Phi==(MFNKMatrix)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Phi (argument 3) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasAddChartWithApproxTS+=clock()-starttime;
#endif
    return -1;
   }

  RMin=-1.;
  RMax=-1.;
  Rstar=-1.;

  if(verbose){printf("     R=%lf\n",R);}

/* Add it to the list */

/* Defer to end? */
  if(verbose){printf("       add chart to list\n");fflush(stdout);}
  chart=MFAtlasAddChartToList(A,u,Phi,R);
  MFChartSetNearBoundary(A->chart[chart],1); /* Only if not paging! @@@ */
  if(MFNVGetIndex2(u)==-99||MFNVGetIndex2(u)==-98)
   {
    MFChartSetSingular(A->chart[chart]);
    MFChartSetNearBoundary(A->chart[chart],1);
   }

  dir=MFCloneNVector(u);

/* Intersections with other charts */

  nInt=0;
  mark=-1;
#ifndef NOBINARYTREE

/* Add new chart to binary tree */

  BBDimension=MFIMFProjectToBB(A->M,u,(double*)NULL);
  BBCenter=(double*)malloc(BBDimension*sizeof(double));
  if(BBCenter==(double*)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Out of memory trying to allocate BBCenter (%d)*sizeof(double)",BBDimension);
    MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasAddChartWithApproxTS+=clock()-starttime;
#endif
    return -1;
   }

  if(A->BTree==(MFBinaryTree)NULL)A->BTree=MFCreateBinaryTree(BBDimension);

  if(verbose){printf("       Project chart center onto space of binary tree\n");fflush(stdout);}
  if(verbose){printf("       %s, BBDimension=%d\n",RoutineName,BBDimension);fflush(stdout);}
  MFIMFProjectToBB(A->M,u,BBCenter);
  if(verbose)
   {
    printf("      BBCenter=(%lf",BBCenter[0]);
    for(jj=1;jj<BBDimension;jj++)printf(",%lf",BBCenter[jj]);
    printf(" )\n");fflush(stdout);
   }

/* Find minimum radius of intersecting charts */
/* Also find minimum index of intersecting charts (add one to get distance from last labeled pt)*/
/*                                                (reset index to zero after saving)*/

#define CHECKNEIGHBORS
#ifdef CHECKNEIGHBORS
  R0=R;
  if(A->nCharts>0)
   {
    if(verbose){printf("checking neighboring radii\n");fflush(stdout);}
    L=MFCreateListOfNearbyCharts(A->BTree,BBCenter,R);
    ni=MFNumberOfIntersectingCharts(L);
    if(verbose){printf("  There are %d neighbors\n",ni);fflush(stdout);}
    RMin=0.;
    RMax=0.;
    for(jj=0;jj<ni;jj++)
     {
      j=MFIntersectingChart(L,jj);
      if(jj==0||MFChartSuggestedRadius(A->chart[j])<RMin)
        RMin=MFChartSuggestedRadius(A->chart[j]);
      if(jj==0||MFChartSuggestedRadius(A->chart[j])>RMax)
        RMax=MFChartSuggestedRadius(A->chart[j]);
      if(verbose){printf("    neighbor %d, radius=%lf\n",j,MFChartSuggestedRadius(A->chart[j]));fflush(stdout);}
     }
    MFFreeListOfIntersectingCharts(L);
/*  if(ni>0&&R>sqrt(2.)*RMin)R=.8*sqrt(2.)*RMin;*/
   }
  if(verbose){printf("Done checking neighboring radii, minimum nearby radius=%lf, max. %lf. reduced from %lf to %lf\n",RMin,RMax,R0,R);}
  if(verbose){printf("Done checking neighboring radii\n");fflush(stdout);}

#endif

  if(verbose){printf("       do intersections\n");fflush(stdout);}

  if(verbose){printf("       Add chart to binary tree\n");fflush(stdout);}
/* Defer to end? */
  MFBinaryTreeAddChart(A->BTree,chart,BBCenter,MFChartRadius(A->chart[chart]));

  if(A->nCharts>1)
   {
    if(verbose){printf("       Create list of intersecting charts\n");fflush(stdout);}
    L=MFCreateListOfIntersectingCharts(A->BTree,chart,BBCenter,MFChartRadius(A->chart[chart]));
    if(verbose){printf("       Loop through list\n");fflush(stdout);}
    ni=MFNumberOfIntersectingCharts(L);
    if(MFVerbose){printf("  There are %d nearby charts out of %d\n",ni,A->nCharts);fflush(stdout);}
    for(jj=0;jj<ni;jj++)
     {
      j=MFIntersectingChart(L,jj);
      if(MFChartPaged(A->chart[j]))continue;
#else
  if(A->n>1)
   {
    for(j=0;j<A->nCharts;j++)
     {
      if(chart==j)continue;
#endif

      if(A->isNear!=NULL && !(A->isNear)(A,A->chart[j],A->chart[chart]))continue;

      if(!MFChartNearBoundary(A->chart[j]))
       {
        sprintf(MFAtlasErrorMsg,"Intersecting chart is not near the boundary.");
        printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
        printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
        MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
       }
      if(!MFChartNearBoundary(A->chart[j]))continue;
      if(MFNVGetIndex2(MFChartCenter(A->chart[j]))==-99&&MFNVGetIndex2(MFChartCenter(A->chart[chart]))==-98)continue;
      if(MFNVGetIndex2(MFChartCenter(A->chart[j]))==-98&&MFNVGetIndex2(MFChartCenter(A->chart[chart]))==-99)continue;

/* Is the norm of Phi_j^T Phi_i within epsilon of 1? */

      Rj=MFChartRadius(A->chart[j]);
      uj=MFChartCenter(A->chart[j]);
      Phij=MFChartTangentSpace(A->chart[j]);
      if(verbose){printf("\n   chart %d (R=%lf) and %d(R=%lf)\n",chart,R,j,Rj);fflush(stdout);}
      MFNSpaceDirection(A->NSpace,u,uj,dir);
      if(verbose){printf("   direction from %d to %d is ",chart,j);MFPrintNVector(stdout,dir);printf("\n");fflush(stdout);}
      nrm=MFCreateKVector(A->k);
      MFChartProjectVectorIntoTangentSpace(A->chart[chart],dir,nrm);
      if(verbose){printf("   in TS %d is ",chart);MFPrintKVector(stdout,nrm);printf("\n");fflush(stdout);}
      dist=MFKVDot(nrm,nrm);
      sdist=sqrt(dist);
      ddist=MFNSpaceDistance(A->NSpace,u,uj);
      if(verbose){printf("   distance in n-space is %lf, in TS is %lf\n",ddist,sdist);fflush(stdout);}
      if(ddist<1.e-10)
       {
        sprintf(MFAtlasErrorMsg,"Two points the same! %d and %d.",chart,j);
        printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
        printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
/*      MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);*/
        MFFreeKVector(nrm);
        return -12;
       }
  
      if(verbose){printf("   check dist > (R+Rj)*(R+Rj) || ddist > R+Rj\n",dist,(R+Rj)*(R+Rj),ddist,R+Rj);
                  printf("   check %lf > %lf || %lf > %lf\n",dist,(R+Rj)*(R+Rj),ddist,R+Rj);fflush(stdout);}
      if(dist>(R+Rj)*(R+Rj) || ddist>R+Rj)
       {
        if(verbose){printf("  These do not intersect - sphere test\n");fflush(stdout);}
        MFFreeKVector(nrm);
        continue;
       }
  
#ifdef MFUSETANGENTTEST
      if(MFIMFStop(A->M,u,Phi,uj,Phij))
       {
        if(verbose){printf("  These do not intersect - index changes\n");fflush(stdout);}
        MFFreeKVector(nrm);
        continue;
       }
      if((tdist=MFNSpaceTangentDistance(A->NSpace,Phi, MFChartTangentSpace(A->chart[j])))>.1)
       {
        if(verbose){printf("  These do not intersect - tangent test %le\n",tdist);fflush(stdout);}
        MFFreeKVector(nrm);
        continue;
       }
#endif

#ifdef MFUSECYLINDERTEST
      if(ddist*ddist-dist>16*MFEpsilon*MFEpsilon)
       {
        if(verbose){printf("  These do not intersect - cylinder test\n");fflush(stdout);}
        MFFreeKVector(nrm);
        continue;
       }
#endif

      dot=MFNKMatrixDot(A->NSpace,MFChartTangentSpace(A->chart[j]),Phi);
      if(verbose)printf("Matrix Dot with %d is %lf\n",j,dot);fflush(stdout);
      if(1&&dot<MFDotMin) /* Was .9 for Rod */
       {
        MFFreeKVector(nrm);
        if(verbose){printf("WARNING: This pair (%d,%d) has too different a TS %lf\n",chart,j,dot);fflush(stdout);}
        continue;   /*  Tangent Space  */
       }
      if(verbose){printf("This pair (%d,%d) has an OK difference in TS %lf\n",chart,j,dot);fflush(stdout);}

      if(!MFChartNearBoundary(A->chart[j]))
       {
        printf("   This pair intersect %d and %d, and %d is not near the boundary!\n",chart,j,j);fflush(stdout);
       }
      if(verbose){printf("   These Intersect\n");fflush(stdout);}
      if(RMin<0.||Rj/sqrtTwo<RMin)RMin=Rj/sqrtTwo;
      if(RMax<0.||Rj*sqrtTwo>RMax)RMax=Rj*sqrtTwo;
      nInt++;
      if(R>sdist+Rj && -R<sdist-Rj)
       {
        printf("%s -- Warning, New chart (%d) covers chart %d\n",RoutineName,chart,j);fflush(stdout);
        printf("      new chart %d is interval [%lf,%lf], chart %d is interval [%lf,%lf]\n",chart,-R,R,j,sdist-Rj,sdist+Rj);fflush(stdout);
        if(R>sdist+Rj)Rstar=sdist+Rj;
        if(-R<sdist-Rj)Rstar=Rj-sdist;
       }
      if(sdist-Rj<-R && R<sdist+Rj)
       {
        printf("%s -- Warning, chart %d covers new chart %d\n",RoutineName,j,chart);fflush(stdout);
        printf("      new chart %d is interval [%lf,%lf], chart %d is interval [%lf,%lf]\n",chart,-R,R,j,sdist-Rj,sdist+Rj);fflush(stdout);
        MFFreeKVector(nrm);
        continue;
       }

      if(MFChartIsSingular(A->chart[chart]))
       {
        if(MFChartIsSingular(A->chart[j])&&mark==-1||MFChartReferenceNumber(A->chart[j])<mark)mark=MFChartReferenceNumber(A->chart[j]);
       }else{
        if(mark==-1||MFChartReferenceNumber(A->chart[j])<mark)mark=MFChartReferenceNumber(A->chart[j]);
       }
  
      onrm=.5*(dist+R*R-Rj*Rj);

      if(verbose){printf("   onrm_ij=%lf nrm_ij=",onrm);MFPrintKVector(stdout,nrm);printf("\n");fflush(stdout);}
      ij=MFAddHalfSpaceToAtlas(A,A->nGlobalIndex,chart,j,nrm,onrm);
      MFSubtractHalfSpaceFromChart(A->chart[chart],ij,nrm,onrm);
  
      MFFreeKVector(nrm);
      MFNSpaceDirection(A->NSpace,uj,u,dir);
      if(verbose){printf("   direction from %d to %d is ",j,chart);MFPrintNVector(stdout,dir);printf("\n");fflush(stdout);}
      nrm=MFCreateKVector(A->k);
      MFChartProjectVectorIntoTangentSpace(A->chart[j],dir,nrm);
      if(verbose){printf("   in TS %d is ",j);MFPrintKVector(stdout,nrm);printf("\n");fflush(stdout);}
      dist=MFKVDot(nrm,nrm);
      if(verbose){printf("   distance in TS is %lf\n",sqrt(dist));fflush(stdout);}
  
      onrm=.5*(dist+Rj*Rj-R*R);

      if(verbose){printf("   onrm_ji=%lf nrm_ji=",onrm);MFPrintKVector(stdout,nrm);printf("\n");fflush(stdout);}
      ji=MFAddHalfSpaceToAtlas(A,A->nGlobalIndex,j,chart,nrm,onrm);
      MFSubtractHalfSpaceFromChart(A->chart[j],ji,nrm,onrm);
      MFFreeKVector(nrm);

      if( (ibnd=MFChartGetPositionInBoundaryList(A->chart[j]))>-1 )
       {
        P=MFChartPolytope(A->chart[j]);
        if(MFPolytopeLargestRadiusOfVertex(P)<MFChartRadius(A->chart[j]))
            MFAtlasRemoveChartFromBoundaryList(A,ibnd);
       }
  
      A->nGlobalIndex++;
     }

#ifndef NOBINARYTREE
    MFFreeListOfIntersectingCharts(L);
#endif
   }
  if(MFVerbose){printf("   %d actually intersected\n",nInt);fflush(stdout);}

  if(Rstar>=0.)
   {
    printf("       Rstar=%lf\n",Rstar);
   }

/*if(RMax>=0.)
   {
    printf("         RMin=%lf, R=%lf, RMax=%lf\n",RMin,R,RMax);
    fflush(stdout);
    if(R>RMax)
     {
      printf("         @@@@@@@RESIZE@@@@@@\n");
      MFAtlasReduceChartRadius(A,chart,RMax);
     }
   }*/

  if(verbose){printf("       update boundary list\n");fflush(stdout);}

  P=MFChartPolytope(A->chart[chart]);
  if(  MFPolytopeNumberOfVertices(P)>0
    && MFPolytopeSmallestVertexIndex(P)<A->offset 
    && MFPolytopeLargestRadiusOfVertex(P)>MFChartRadius(A->chart[chart]))
      MFAtlasAddChartToBoundaryList(A,chart);

  MFFreeNVector(dir);
  free(BBCenter);

  MFChartSetReferenceNumber(A->chart[chart],mark+1);

  if(verbose){printf(" done %s\n",RoutineName);fflush(stdout);}
  return(chart);
 }

/* --------------------------------------------------------- */

int MFAtlasPointOnBoundaryWOProject(MFAtlas A,MFNRegion Omega,MFKVector s)
 {
  static char RoutineName[]={"MFAtlasPointOnBoundaryWOProject"};
  int v,l,n;
  int nChartsTried;
  int chart;
  int bchart;
  MFNVector up;
  MFNKMatrix Phi;
  double rv;
  double R;
  int rc;
  int abandon;
  int verbose;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasPointOnBoundaryWOProject++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  verbose=0;
  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

  if(MFAtlasNumberOfCharts(A)>0)up=MFCloneNVector(MFAtlasChartCenter(A,0));
   else               up=(MFNVector)NULL;
  nChartsTried=0;

#ifndef NOBOUNDARYLIST
  bchart=0;
  {printf("%s, there are %d charts on the boundary list\n",RoutineName,A->nBnd);fflush(stdout);}
  while(bchart<A->nBnd)
   {
    chart=A->bnd[bchart];
#else
  chart=0;
  while(chart<A->nCharts)
   {
#endif
tryagain:
    nChartsTried++;
    n=MFPolytopeNumberOfVertices(MFChartPolytope(A->chart[chart]));
    v=0;
    while(v<n && MFChartRadius(A->chart[chart])>MFRMin)
     {
      if(verbose){printf("  Trying vertex %d of chart %d\n",v,chart);fflush(stdout);}
      rv=MFPolytopeRadiusOfVertex(MFChartPolytope(A->chart[chart]),v);
      R=MFChartRadius(A->chart[chart]);

      if(rv>=R)
       {
        if(verbose){printf("  Exterior!\n");fflush(stdout);}
        MFPolytopeVertex(MFChartPolytope(A->chart[chart]),v,s);
        MFKVScale(R/rv,s);
        MFChartPointInTangentSpace((A->chart[chart]),s,up);

        if(MFNRegionInterior(Omega,up))
         {
          if(verbose){printf("    In Omega!\n");fflush(stdout);}

          MFFreeNVector(up);

          {printf("    found after examining %d chart\n",nChartsTried);fflush(stdout);}
          if(verbose)
           {
            printf("Chart %d is now \n",chart);
            MFPrintChart(stdout,A->chart[chart]);
            fflush(stdout);
           }

#ifdef MFTIMINGS
  MFTimeMFAtlasPointOnBoundaryWOProject+=clock()-starttime;
#endif
          return(chart);
         }else
          if(verbose){printf("    Not in Omega!\n");fflush(stdout);}
       }
      v++;
     }
#ifndef NOBOUNDARYLIST
    MFAtlasRemoveChartFromBoundaryList(A,0);
    if(bchart>=A->nBnd)
     {
      bchart=0;
#else
    if(chart>=A->nCharts)
     {
      chart=0;
#endif
      chart=-1;
     }
   }
  if(up!=(MFNVector)NULL)MFFreeNVector(up);

#ifdef MFTIMINGS
  MFTimeMFAtlasPointOnBoundaryWOProject+=clock()-starttime;
#endif
  return(-1);
 }

void MFAtlasSetK(MFAtlas A,int k)
 {
  static char RoutineName[]={"MFAtlasSetK"};

#ifdef MFDEBUG
  _heap_check();
#endif

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(A->nCharts>0&&k!=A->k)
   {
    sprintf(MFAtlasErrorMsg,"Changing k after Atlas has charts may lead to errors later.");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
   }

  A->k=k;
  MFAddInitialHyperplanes(A);

  return;
 }

void MFAtlasSetN(MFAtlas A,int n)
 {
  static char RoutineName[]={"MFAtlasSetN"};

#ifdef MFDEBUG
  _heap_check();
#endif

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(A->nCharts>0&&n!=A->n)
   {
    sprintf(MFAtlasErrorMsg,"Changing n after Atlas has charts may lead to errors later.");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
   }

  A->n=n;
  MFAddInitialHyperplanes(A);

  return;
 }

void MFWriteChartCenters(FILE *fid,MFAtlas A)
 {
  static char RoutineName[]={"MFWriteChartCenters"};
  int i;
  int verbose;

#ifdef MFDEBUG
  _heap_check();
#endif

  verbose=0;

  if(verbose){printf("%s\n");fflush(stdout);}

  fprintf(fid,"%d\n",A->n);
  fprintf(fid,"%d\n",A->k);

/* List of Charts */

  fprintf(fid,"%d\n",A->nCharts);
  fprintf(fid,"%d\n",A->mCharts);
  for(i=0;i<A->nCharts;i++)
   {
    MFWriteNVector(fid,MFChartCenter(A->chart[i]));
    MFWriteNKMatrix(fid,MFChartTangentSpace(A->chart[i]));
    fprintf(fid,"%lf\n",MFChartRadius(A->chart[i]));
   }

  return;
 }


int MFReadChartCenters(FILE *fid,MFNVector **c,MFNKMatrix **t,double**R)
 {
  static char RoutineName[]={"MFReadChartCenters"};
  int i,n,k;
  int nCharts;
  int mCharts;
  int verbose;

#ifdef MFDEBUG
  _heap_check();
#endif

  verbose=0;

  if(verbose){scanf("%s\n");fflush(stdout);}

  fscanf(fid,"%d\n",&n);
  fscanf(fid,"%d\n",&k);

/* List of Charts */

  fscanf(fid,"%d\n",&nCharts);
  fscanf(fid,"%d\n",&mCharts);
  *c=(MFNVector*)realloc((void*)(*c),nCharts*sizeof(MFNVector));
  *t=(MFNKMatrix*)realloc((void*)(*t),nCharts*sizeof(MFNKMatrix));
  *R=(double*)realloc((void*)(*R),nCharts*sizeof(double));
  for(i=0;i<nCharts;i++)
   {
    (*c)[i]=MFReadNVector(fid);
    (*t)[i]=MFReadNKMatrix(fid);
    fscanf(fid,"%lf\n",&((*R)[i]));
   }

  return nCharts;
 }

void MFAtlasPageOutChartsNotNearBoundary(MFAtlas A, int dumpToPlotFile, int dumpToCenterFile, char *FileName)
 {
  static char RoutineName[]={"MFAtlasPageOutChartsNotNearBoundary"};
  int i,ii,j,n,l,m;
  int nmoved;
  MFPolytope P,Q;
  int verbose=0;
  char name[10000];
  int right;

#ifdef MFDEBUG
  _heap_check();
#endif

  if(verbose){fprintf(stdout,"PAGING!!! (%s) -- %d charts in Atlas\n",RoutineName,A->nCharts);MFReportMemory();fflush(stdout);}
  /*MFPrintChartSummary(A);*/

  for(i=0;i<A->nCharts;i++)
    MFChartSetNearBoundary(A->chart[i],0);

  for(i=0;i<A->nCharts;i++)
   {
    if(!MFChartPaged(A->chart[i]))
     {
      if(MFNVGetIndex2(MFChartCenter(A->chart[i]))==-99||MFNVGetIndex2(MFChartCenter(A->chart[i]))==-98)
       {
        MFChartSetNearBoundary(A->chart[i],1);

        P=MFChartPolytope(A->chart[i]);
        n=MFPolytopeNumberOfFaces(P);
        for(j=0;j<n;j++)
         {
          ii=MFPolytopeFaceIndex(P,j);
          if(ii>-1 && ii< A->nHS)right=MFAtlasHalfSpaceRightChart(A,ii);
           else right=-1;
          if(ii>-1 && ii< A->nHS && right>-1 && right<A->nCharts)
           {
            MFChartSetNearBoundary(A->chart[right],1);
            Q=MFChartPolytope(A->chart[right]);
            if(Q!=(MFPolytope)NULL)
             {
              m=MFPolytopeNumberOfFaces(Q);
              for(l=0;l<m;l++)
               {
                ii=MFPolytopeFaceIndex(Q,l);
                if(ii>-1 && ii< A->nHS && ii >= A->offset && right>-1 && right<A->nCharts)
                 {
                  MFChartSetNearBoundary(A->chart[right],1);
                 }
               }
             }
           }
         }

       }
     }
   }

  if(verbose){fprintf(stdout,"PAGING mark boundary charts(%d on boundary)!!!\n",A->nBnd);fflush(stdout);}
  for(i=0;i<A->nBnd;i++)
   {
    MFChartSetNearBoundary(A->chart[A->bnd[i]],1);
    P=MFChartPolytope(A->chart[A->bnd[i]]);
    n=MFPolytopeNumberOfFaces(P);
    for(j=0;j<n;j++)
     {
      ii=MFPolytopeFaceIndex(P,j);
      if(ii>-1 && ii< A->nHS)right=MFAtlasHalfSpaceRightChart(A,ii);
       else right=-1;
      if(ii>-1 && ii< A->nHS && right>-1 && right<A->nCharts)
       {
        MFChartSetNearBoundary(A->chart[right],1);
        Q=MFChartPolytope(A->chart[right]);
        if(Q!=(MFPolytope)NULL)
         {
          m=MFPolytopeNumberOfFaces(Q);
          for(l=0;l<m;l++)
           {
            ii=MFPolytopeFaceIndex(Q,l);
            if(ii>-1 && ii< A->nHS && ii >= A->offset && right>-1 && right<A->nCharts)
             {
              MFChartSetNearBoundary(A->chart[right],1);
             }
           }
         }
       }
     }
   }
  if(verbose)
   {
    m=0;
    for(i=0;i<A->nCharts;i++)if(MFChartNearBoundary(A->chart[i]))m++;
    fprintf(stdout,"       %d charts marked as near boundary\n",m);fflush(stdout);
   }

  if(verbose){fprintf(stdout,"PAGING those not marked\n");fflush(stdout);}
#ifdef REALLYPAGE
  if(verbose){printf("page those not marked\n");fflush(stdout);}
  nmoved=0;
  for(i=0;i<A->nCharts;i++)
   {
    if(!MFChartNearBoundary(A->chart[i]))
     {
      if(dumpToPlotFile && A->plotFileFid==(FILE*)NULL)
       {
        strcpy(name,FileName);
        strcat(name,".plotfile");
        A->plotFileFid=fopen(name,"w+b");
        if(A->plotFileFid==(FILE*)NULL)abort();
        if(verbose){fprintf(stdout,"Opened plotfile, 0x%8.8x\n",A->plotFileFid);fflush(stdout);}
        fprintf(A->plotFileFid,"Dimension of vertices, %d\n",MFIMFProjectToDraw(A->M,MFAtlasChartCenter(A,0),(double*)NULL));
        fprintf(A->plotFileFid,"Dimension of manifold, %d\n",A->k);
       }
      if(dumpToCenterFile && A->centerFileFid==(FILE*)NULL)
       {
        strcpy(name,FileName);
        strcat(name,".centers");
        A->centerFileFid=fopen(name,"w+b");
        if(A->centerFileFid==(FILE*)NULL)abort();
        if(verbose){fprintf(stdout,"Opened centerfile, 0x%8.8x\n",A->centerFileFid);fflush(stdout);}
        fprintf(A->centerFileFid,"%d\n",A->n);
        fprintf(A->centerFileFid,"%d\n",A->k);
        fprintf(A->centerFileFid,"nCharts goes here\n");
        fprintf(A->centerFileFid,"mCharts goes here\n");
       }
      if(verbose){printf(" chart %d is not marked\n",i);fflush(stdout);}
      if(!MFChartPaged(A->chart[i]))
       {
        if(dumpToPlotFile)
         {
          MFNVector col;
          if(verbose){fprintf(stdout,"Paging Chart %d\n",i);fflush(stdout);}

          MFWriteChartToPlotFile(A->plotFileFid,A,A->chart[i],i);
          fflush(A->plotFileFid);
         }
        if(dumpToCenterFile)
         {
          MFWriteNVector(A->centerFileFid,MFChartCenter(A->chart[i]));
          MFWriteNKMatrix(A->centerFileFid,MFChartTangentSpace(A->chart[i]));
          fprintf(A->centerFileFid,"%lf\n",MFChartRadius(A->chart[i]));
          fflush(A->centerFileFid);
         }
        if(MFChartPageOut(A->chart[i],A->pageFileFid,A->pageFileIndex))A->pageFileIndex++;
        nmoved++;
       }
     }
   }
  A->nPagedOut+=nmoved;
  if(verbose){printf("%s, paged %d. A total of %d are out, out of %d\n",RoutineName,nmoved,A->nPagedOut,A->nCharts);MFReportMemory();fflush(stdout);}
#endif
/*MFPrintChartSummary(A);*/

/* @@@ Now clean half space list. If left and right are paged, add to garbage list. */

  return;
 }

int MFAtlasIsChartSingular(MFAtlas this,int chart)
 {
  static char RoutineName[]={"MFAtlasIsChartSingular"};

#ifdef MFDEBUG
  _heap_check();
#endif

  if(this==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
    return;
   }

  return MFChartIsSingular(this->chart[chart]);
 }

int MFAtlasIsChartNearBoundary(MFAtlas this,int chart)
 {
  static char RoutineName[]={"MFAtlasIsChartNearBoundary"};

#ifdef MFDEBUG
  _heap_check();
#endif

  if(this==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
    return;
   }

  return MFChartNearBoundary(this->chart[chart]);
 }

int MFAtlasNumberOfNeighborsOfChart(MFAtlas this,int chart)
 {
  static char RoutineName[]={"MFAtlasNumberOfNeighborsOfChart"};

#ifdef MFDEBUG
  _heap_check();
#endif

  if(this==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(chart<0 || chart>=this->nCharts)
   {
    sprintf(MFAtlasErrorMsg,"Chart number %d (argument 2) is invalid, must be in [0,%d)",chart,this->nCharts);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
    return;
   }

  return MFPolytopeNumberOfFaces(MFChartPolytope(this->chart[chart]));
 }

int MFAtlasChartNeighbor(MFAtlas this,int chart,int neighbor)
 {
  static char RoutineName[]={"MFAtlasChartNeighbor"};
  int n;

#ifdef MFDEBUG
  _heap_check();
#endif

  if(this==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(chart<0 || chart>= this->nCharts)
   {
    sprintf(MFAtlasErrorMsg,"Chart number %d (argument 2) is invalid, must be in [0,%d)",chart,this->nCharts);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
    return;
   }

  n=MFChartNumberOfFaces(this->chart[chart]);
  if(neighbor<0 || neighbor>=n)
   {
    sprintf(MFAtlasErrorMsg,"Neighbor number %d (argument 3) is invalid, must be in [0,%d)",neighbor,n);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
    return;
   }

  n=MFPolytopeFaceIndex(MFChartPolytope(this->chart[chart]),neighbor);

  if(MFAtlasHalfSpaceLeftChart(this,n)==chart)return MFAtlasHalfSpaceRightChart(this,n);
   else return MFAtlasHalfSpaceLeftChart(this,n);
 }

int MFPolytopeIntersectIndexSets(MFPolytope,int,int,int*);

void MFWriteChartToPlotFile(FILE *fid, MFAtlas A, MFChart this, int chart)
 {
  static char RoutineName[]={"MFWriteChartToPlotFile"};
  int d;
  double *x;
  MFKVector s;
  double *ps;
  MFNVector u;
  int i,j,k,l;
  int n,m,ne,ie;
  int nv;
  int c;
  int *index;
  int *indx;
  int verbose=0;
  MFNKMatrix Phi;
  MFNVector col;
  int bnd,sing;
  double *dv;
  int trace=0;

#ifdef MFDEBUG
  _heap_check();
#endif
  if(MFChartPaged(this))
   {
    sprintf(MFAtlasErrorMsg,"Chart (argument 3) has been paged out");
    printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFChartPointInTangentSpace+=clock()-starttime;
#endif
    return;
   }

  k=MFChartK(this);
  if(MFPolytopeNumberOfVertices(MFChartPolytope(this))<=k)return;

  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}
  if(fid==(FILE*)NULL){printf("%s - fid is NULL %s(%d)\n",RoutineName,__LINE__,__FILE__);fflush(stdout);abort();}
  if(A==(MFAtlas)NULL){printf("%s - A is NULL %s(%d)\n",RoutineName,__LINE__,__FILE__);fflush(stdout);abort();}
  if(this==(MFChart)NULL){printf("%s - Chart is NULL %s(%d)\n",RoutineName,__LINE__,__FILE__);fflush(stdout);abort();}

  d=MFIMFProjectToDraw(A->M,MFChartCenter(this),(double*)NULL);
  if(A->M==(MFImplicitMF)NULL){printf("%s - A->M is NULL %s(%d)\n",RoutineName,__LINE__,__FILE__);fflush(stdout);abort();}

  s=MFCreateKVector(MFChartK(this));
  ps=MFKV_CStar(s);
  u=MFCloneNVector(MFChartCenter(this));
  x=(double*)malloc(d*sizeof(double));
  if(x==(double*)NULL)abort();

  for(j=0;j<A->nclipf;j++)
   {
    n=MFPolytopeNumberOfVertices(MFChartPolytope(this));
    if(n>0)
     {
      dv=(double*)malloc(n*sizeof(double));
      for(i=0;i<n;i++)
       {
        MFPolytopeVertex(MFChartPolytope(this),i,s);
        MFChartPointInTangentSpace(this,s,u);
        dv[i]=A->clipf[j](u);
       }
      MFClipPolytope(MFChartPolytope(this),A->clipIndx[j],dv,0);
      if(dv!=(double*)NULL)free(dv);
     }
   }

  if(MFPolytopeNumberOfVertices(MFChartPolytope(this))<=k)
   {
    MFFreeKVector(s);
    MFFreeNVector(u);
    free(x);
    return;
   }

  Phi=MFChartTangentSpace(this);
  if(Phi==(MFNKMatrix)NULL)abort();
  col=MFMColumn(Phi,0);
  if(col==(MFNVector)NULL)abort();
  MFFreeNVector(col);
  if(k>1)
   {
    col=MFMColumn(Phi,1);
    if(col==(MFNVector)NULL)abort();
    MFFreeNVector(col);
   }

  n=MFPolytopeNumberOfVertices(MFChartPolytope(this));

  indx=(int*)malloc(k*sizeof(int));
  ne=0;
  for(i=0;i<n-1;i++)
    for(j=i+1;j<n;j++)
      if(MFPolytopeIntersectIndexSets(MFChartPolytope(this),i,j,indx)>=k-1)ne++;

  m=MFPolytopeNumberOfFaces(MFChartPolytope(this));
  bnd=MFPolytopeLargestRadiusOfVertex(MFChartPolytope(this))>MFChartRadius(this);
  sing=MFChartIsSingular(this);

  nv=n+1;
  fprintf(fid,"Polyhedron %d, %d vertices, %d edges, %d faces, boundary %d, singular %d\n",chart,nv,ne,m,bnd,sing);
  if(trace){printf("Polyhedron %d, %d vertices, %d edges, %d faces, boundary %d, singular %d\n",chart,n,ne,m,bnd,sing);fflush(stdout);}
  for(i=0;i<n;i++)
   {
       if(trace){printf("Vertex %d, dimension %d\n",i,d);fflush(stdout);}
       if(trace){printf(" call MFPolytopeVertex\n");fflush(stdout);}
    MFPolytopeVertex(MFChartPolytope(this),i,s);
    fflush(fid);
       if(trace){printf(" call MFChartPointInTangentSpace\n");fflush(stdout);}
    for(j=0;j<k;j++)ps[j]=(A->expFactor)*ps[j]; /* This allows for eliminating gaps between plates */
    MFChartPointInTangentSpace(this,s,u);
    MFNVSetIndex(u,MFNVGetIndex(MFChartCenter(this)));
    MFNVSetIndex2(u,MFNVGetIndex2(MFChartCenter(this)));
       if(trace){printf(" call MFIMFProjectToDraw\n");fflush(stdout);}
    MFIMFProjectToDraw(A->M,u,x);
    for(j=0;j<d;j++)if(x[j]!=x[j]){printf("In %s, x[%d] is NaN!\n",RoutineName,j);fflush(stdout);abort();}
       if(trace){printf(" write vertex\n");fflush(stdout);}
    fprintf(fid,"Vertex %d (%lf",i,x[0]);
    for(j=1;j<d;j++)fprintf(fid,",%lf",x[j]);
    fprintf(fid,")");
       if(trace){printf(" call MFPolytopeVertexNumberOfIndices\n");fflush(stdout);}
    m=MFPolytopeVertexNumberOfIndices(MFChartPolytope(this),i);
       if(trace){printf("           %d indices\n",m);fflush(stdout);}
    index=MFPolytopeVertexIndexSet(MFChartPolytope(this),i);
    fprintf(fid,", %d [%d",m,index[0]);
    for(j=1;j<m;j++)fprintf(fid,",%d",index[j]);
    fprintf(fid,"]\n");
       if(trace){printf("  done vertex\n",m);fflush(stdout);}
   }

/* Center */

  MFIMFProjectToDraw(A->M,MFChartCenter(this),x);
     if(trace){printf(" write vertex\n");fflush(stdout);}
  fprintf(fid,"Vertex %d (%lf",i,x[0]);
  for(j=1;j<d;j++)fprintf(fid,",%lf",x[j]);
  fprintf(fid,")");
  fprintf(fid,", %d [ ]\n",0);

  if(trace){printf("done writing vertices\n");fflush(stdout);}
  MFFreeNVector(u);
  MFFreeKVector(s);
  free(x);
  fflush(fid);

  ie=0;
  for(i=0;i<n-1;i++)
   {
    for(j=i+1;j<n;j++)
     {
      if((m=MFPolytopeIntersectIndexSets(MFChartPolytope(this),i,j,indx))>=k-1)
       {
        fprintf(fid,"Edge %d (%d,%d)",ie,i,j);
        fprintf(fid,", %d [",m);
        if(m>0)fprintf(fid,"%d",indx[0]);
        for(l=1;l<m;l++)fprintf(fid,",%d",indx[l]);
        fprintf(fid,"]\n");
        ie++;
       }
     }
   }
  fflush(fid);
  free(indx);

  m=MFPolytopeNumberOfFaces(MFChartPolytope(this));
  for(i=0;i<m;i++)
   {
    j=MFPolytopeFaceIndex(MFChartPolytope(this),i);
    if(j<A->offset)c=-1;
    else if(MFAtlasHalfSpaceLeftChart(A,j)==chart)c=MFAtlasHalfSpaceRightChart(A,j);
     else c=MFAtlasHalfSpaceLeftChart(A,j);
    fprintf(fid,"Face %d neighbor %d\n",j,c);
   }

  fflush(fid);
  if(verbose){printf("done %s\n",RoutineName);fflush(stdout);}
  return;
 }

void MFAtlasPageOutAllCharts(MFAtlas A)
 {
  static char RoutineName[]={"MFAtlasPageOutAllCharts"};
  int i,ii,j,n,l,m;
  int nmoved;
  MFPolytope P,Q;
  int verbose=0;

#ifdef MFDEBUG
  _heap_check();
#endif

  if(verbose){fprintf(stdout,"PAGING!!! (%s) -- %d charts in Atlas\n",RoutineName,A->nCharts);fflush(stdout);}

#ifdef REALLYPAGE
  if(verbose){printf("       Page all not paged out\n");fflush(stdout);}
  nmoved=0;
  for(i=0;i<A->nCharts;i++)
   {
    if(A->plotFileFid==(FILE*)NULL)
     {
      A->plotFileFid=fopen("Atlas.plotfile","w+b");
      if(A->plotFileFid==(FILE*)NULL)abort();
      if(verbose){fprintf(stdout,"Opened plotfile, 0x%8.8x\n",A->plotFileFid);fflush(stdout);}
      fprintf(A->plotFileFid,"Dimension of vertices, %d\n",MFIMFProjectToDraw(A->M,MFAtlasChartCenter(A,0),(double*)NULL));
      fprintf(A->plotFileFid,"Dimension of manifold, %d\n",A->k);
     }
    if(A->centerFileFid==(FILE*)NULL)
     {
      A->centerFileFid=fopen("Atlas.centers","w+b");
      if(A->centerFileFid==(FILE*)NULL)abort();
      if(verbose){fprintf(stdout,"Opened centerfile, 0x%8.8x\n",A->centerFileFid);fflush(stdout);}
      fprintf(A->centerFileFid,"%d\n",A->n);
      fprintf(A->centerFileFid,"%d\n",A->k);
      fprintf(A->centerFileFid,"nCharts goes here\n");
      fprintf(A->centerFileFid,"mCharts goes here\n");
     }
    if(!MFChartPaged(A->chart[i]))
     {
      MFWriteChartToPlotFile(A->plotFileFid,A,A->chart[i],i);
      MFWriteNVector(A->centerFileFid,MFChartCenter(A->chart[i]));
      MFWriteNKMatrix(A->centerFileFid,MFChartTangentSpace(A->chart[i]));
      fprintf(A->centerFileFid,"%lf\n",MFChartRadius(A->chart[i]));
      printf("Page out chart %d\n",i);fflush(stdout);
      if(MFChartPageOut(A->chart[i],A->pageFileFid,A->pageFileIndex))A->pageFileIndex++;
      MFReportMemory();
      nmoved++;
      fflush(A->plotFileFid);
      fflush(A->centerFileFid);
     }
   }
  A->nPagedOut+=nmoved;
  if(verbose){printf("done PAGING!!! (%s) moved %d out. A total of %d are out, out of %d\n",RoutineName,nmoved,A->nPagedOut,A->nCharts);fflush(stdout);}
#endif

  return;
 }

void MFAtlasWriteOutChartsNotPaged(MFAtlas A, int dumpToPlotFile, int dumpToCenterFile, char *FileName)
 {
  static char RoutineName[]={"MFAtlasWriteOutChartsNotPaged"};
  int i,ii,j,n,l,m;
  int nmoved;
  MFPolytope P,Q;
  int verbose=0;
  char name[10000];

#ifdef MFDEBUG
  _heap_check();
#endif

  if(verbose){fprintf(stdout,"Dumping charts not paged!!! (%s) -- %d charts in Atlas\n",RoutineName,A->nCharts);fflush(stdout);}

  nmoved=0;
  for(i=0;i<A->nCharts;i++)
   {
    if(dumpToPlotFile && A->plotFileFid==(FILE*)NULL)
     {
      strcpy(name,FileName);
      strcat(name,".plotfile");
      A->plotFileFid=fopen(name,"w+b");
      if(A->plotFileFid==(FILE*)NULL)abort();
      if(verbose){fprintf(stdout,"Opened plotfile, 0x%8.8x\n",A->plotFileFid);fflush(stdout);}
      fprintf(A->plotFileFid,"Dimension of vertices, %d\n",MFIMFProjectToDraw(A->M,MFAtlasChartCenter(A,0),(double*)NULL));
      fprintf(A->plotFileFid,"Dimension of manifold, %d\n",A->k);
     }
    if(dumpToCenterFile && A->centerFileFid==(FILE*)NULL)
     {
      strcpy(name,FileName);
      strcat(name,".centers");
      A->centerFileFid=fopen(name,"w+b");
      if(A->centerFileFid==(FILE*)NULL)abort();
      if(verbose){fprintf(stdout,"Opened centerfile, 0x%8.8x\n",A->centerFileFid);fflush(stdout);}
      fprintf(A->centerFileFid,"%d\n",A->n);
      fprintf(A->centerFileFid,"%d\n",A->k);
      fprintf(A->centerFileFid,"nCharts goes here\n");
      fprintf(A->centerFileFid,"mCharts goes here\n");
     }
    if(!MFChartPaged(A->chart[i]))
     {
      if(dumpToPlotFile)
       {
        MFWriteChartToPlotFile(A->plotFileFid,A,A->chart[i],i);
        fflush(A->plotFileFid);
       }
      if(dumpToCenterFile)
       {
        MFWriteNVector(A->centerFileFid,MFChartCenter(A->chart[i]));
        MFWriteNKMatrix(A->centerFileFid,MFChartTangentSpace(A->chart[i]));
        fprintf(A->centerFileFid,"%lf\n",MFChartRadius(A->chart[i]));
        fflush(A->centerFileFid);
       }
      if(verbose){printf("Chart %d/%d written to plotfile and center file\n",i,A->nCharts);}
      nmoved++;
     }else{
      if(verbose){printf("Chart %d/%d has already been written to plotfile and center file\n",i,A->nCharts);}
     }
   }
  if(verbose){printf("done Dumping!!! (%s) wrote %d out. A total of %d are out, out of %d\n",RoutineName,nmoved,A->nPagedOut,A->nCharts);fflush(stdout);}

  return;
 }

#ifndef __FreeBSD__
#include <malloc.h>
#endif

void MFReportMemory()
 {
#ifndef __FreeBSD__
  struct mallinfo j;

  j=mallinfo();
  printf(
"       %d arena;            /* total space in arena */\n"
"       %d usmblks;          /* space in small blocks in use */\n"
"       %d fsmblks;          /* space in free small blocks */\n"
"       %d uordblks;         /* space in ordinary blocks in use */\n"
"       %d fordblks;         /* space in free ordinary blocks */\n",
        j.arena,
        j.usmblks,
        j.fsmblks,
        j.uordblks,
        j.fordblks);fflush(stdout);
#endif
  return;
 }

int MFAtlasGetSingularChartWithBoundary(MFAtlas A, MFNRegion Omega)
 {
  static char RoutineName[]={"MFAtlasGetSingularChartWithBoundary"};
  int chart,n,v;
  double R,rv;
  MFKVector s;
  MFNVector up;
  int pass;
  int verbose;

  verbose=0;
  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFAtlasSt));
    printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
    MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateAtlas+=clock()-starttime;
#endif
    return -1;
   }

  s=MFCreateKVector(A->k);
  up=(MFNVector)NULL;
  for(chart=0;chart<A->nCharts;chart++)
   if(up==(MFNVector)NULL&&!MFChartPaged(A->chart[chart]))up=MFCloneNVector(MFChartCenter(A->chart[chart]));
  if(up==(MFNVector)NULL)return -1;

  for(pass=0;pass<2;pass++)
   {
    for(chart=0;chart<A->nCharts;chart++)
     {
      if(verbose){printf("chart %d/%d\n",chart,A->nCharts);fflush(stdout);}
      if(!MFChartPaged(A->chart[chart])
        &&MFChartIsSingular(A->chart[chart])&&(pass==1||pass==0&&MFNVGetIndex2(MFChartCenter(A->chart[chart]))==-99)
        &&MFChartRadius(A->chart[chart])>MFRMin)
       {
        if(verbose){printf("  Is Singular\n");fflush(stdout);}
        n=MFPolytopeNumberOfVertices(MFChartPolytope(A->chart[chart]));
        for(v=0;v<n;v++)
         {
          rv=MFPolytopeRadiusOfVertex(MFChartPolytope(A->chart[chart]),v);
          R=MFChartRadius(A->chart[chart]);
          if(rv>=R && MFPolytopeGetVertexMark(MFChartPolytope(A->chart[chart]),v)==0)
           {
            if(verbose){printf("  Vertex %d is exterior\n",v);fflush(stdout);}
            MFPolytopeVertex(MFChartPolytope(A->chart[chart]),v,s);
            MFKVScale(R/rv,s);
            MFChartPointInTangentSpace((A->chart[chart]),s,up);
            if(MFNRegionInterior(Omega,up))
             {
              MFFreeKVector(s);
              MFFreeNVector(up);
              if(verbose){printf("done %s\n",RoutineName);fflush(stdout);}
              MFPolytopeSetVertexMark(MFChartPolytope(A->chart[chart]),v,1);
              return chart;
             }
           }
         }
        if(verbose){printf("  No exterior vertices\n");fflush(stdout);}
       }else
        if(verbose){printf("  Not singular\n");fflush(stdout);}
     }
   }

  MFFreeKVector(s);
  MFFreeNVector(up);
  if(verbose){printf("done %s\n",RoutineName);fflush(stdout);}
  return -1;
 }

MFNVector MFAtlasGetPointOnBoundaryChart(MFAtlas A, MFNRegion Omega, int chart, double t0)
 {
  static char RoutineName[]={"MFAtlasGetPointOnBoundaryChart"};
  int n,v;
  int rc;
  double R,rv,t;
  MFKVector s;
  MFNVector up;
  MFNVector u;
  int verbose;

  verbose=0;
  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFAtlasSt));
    printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
    MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
    return (MFNVector)NULL;
   }

  if(chart<0||chart>=A->nCharts)
   {
    sprintf(MFAtlasErrorMsg,"Chart %d is bad, must be in range [0,%d)",chart,A->nCharts);
    printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
    MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
    return (MFNVector)NULL;
   }

  if(!MFChartIsSingular(A->chart[chart]))
   {
    sprintf(MFAtlasErrorMsg,"Chart %d is bad, must be a singular chart",chart);
    printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
    MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
    return (MFNVector)NULL;
   }

  if(MFChartRadius(A->chart[chart])<MFRMin)
   {
    sprintf(MFAtlasErrorMsg,"Chart %d is bad, radius is too small (%lf<%lf)",chart,
                                              MFChartRadius(A->chart[chart]),MFRMin);
    printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
    MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
    return (MFNVector)NULL;
   }

  s=MFCreateKVector(A->k);
  up=MFCloneNVector(MFChartCenter(A->chart[0]));
  u=MFCloneNVector(up);

  n=MFPolytopeNumberOfVertices(MFChartPolytope(A->chart[chart]));
  for(v=0;v<n;v++)
   {
    rv=MFPolytopeRadiusOfVertex(MFChartPolytope(A->chart[chart]),v);
    R=MFChartRadius(A->chart[chart]);
    if(verbose){printf(" vertex %d/%d  radius=%lf/%lf\n",v,n,rv,R);fflush(stdout);}
    if(rv>=R)
     {
      if(verbose){printf(" exterior\n");fflush(stdout);}
      MFPolytopeVertex(MFChartPolytope(A->chart[chart]),v,s);
      MFKVScale(R/rv,s);
      MFChartPointInTangentSpace((A->chart[chart]),s,up);
      if(MFNRegionInterior(Omega,up))
       {
        if(verbose){printf(" in Omega\n");fflush(stdout);}
        t=t0;
        while(t>1.e-4)
         {
          rc=MFChartEvaluate(A->chart[chart],s,u);
          if(verbose){printf(" rc from MFChartEvaluate=%d, t=%lf, s=",rc,t);MFPrintKVector(stdout,s);printf("\n");fflush(stdout);}
          if(rc)
           {
            MFFreeKVector(s);
            MFFreeNVector(up);
            if(verbose){printf("done %s\n",RoutineName);fflush(stdout);}
            return u;
           }else{
            t=t*.8;
            MFKVScale(.8,s);
           }
         }
       }else
        if(verbose){printf(" not in Omega\n");fflush(stdout);}
     }
   }

  sprintf(MFAtlasErrorMsg,"Couldn't get a point on boundary of chart %d",chart);
  printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
    printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
  MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
  return (MFNVector)NULL;
 }

void MFAtlasRemoveChartFromBoundaryListByChart(MFAtlas A,int chart)
 {
  static char RoutineName[]={"MFAtlasRemoveChartFromBoundaryListByChart"};
  int i;

  printf("%s, chart %d\n",RoutineName,chart);fflush(stdout);
  for(i=0;i<A->nBnd;i++)
   {
    printf("      bnd[%d]=%d\n",i,A->bnd[i]);fflush(stdout);
    if(A->bnd[i]==chart)
     {
      MFAtlasRemoveChartFromBoundaryList(A,i);
      return;
     }
   }
  printf("  chart %d not found on boundary list!\n",chart);fflush(stdout);
  return;
 }

double MFAtlasChartSuggestedRadius(MFAtlas A,int chart)
 {
  static char RoutineName[]={"MFAtlasChartSuggestedRadius"};

#ifdef MFDEBUG
  _heap_check();
#endif

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFAtlasCenterOfChart+=clock()-starttime;
#endif
    return -1.;
   }

  if(chart<0||chart>=A->nCharts)
   {
    sprintf(MFAtlasErrorMsg,"Chart %d (argument 2) is invalid. Must be positive and less than %d (total number of charts).",chart,A->nCharts);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFAtlasCenterOfChart+=clock()-starttime;
#endif
    return -1.;
   }

  return MFChartSuggestedRadius(A->chart[chart]);
 }

void MFAtlasAddClipF(MFAtlas this,double (*clipf)(MFNVector))
 {
  static char RoutineName[]={"MFAtlasAddClipF"};
  int verbose=0;

  if(this->nclipf>=this->mclipf)
   {
    this->mclipf+=10;
    this->clipf=(double (**)(MFNVector))realloc(this->clipf,(this->mclipf)*sizeof(double (*)(MFNVector)));
    this->clipIndx=(int*)realloc(this->clipIndx,(this->mclipf)*sizeof(int));
   }
  this->clipf[this->nclipf]=clipf;
  this->clipIndx[this->nclipf]=this->nclipIndx;
  if(verbose){printf("%s, Just added Clipping function %d index %d\n",RoutineName,this->nclipf,this->nclipIndx);fflush(stdout);}
  this->nclipf++;
  this->nclipIndx--;
 }

MFBinaryTree MFAtlasGetBB(MFAtlas A)
 {
  return A->BTree;
 }

void MFAtlasSetExpFactor(MFAtlas A, double s)
 {
  A->expFactor=s;
  return;
 }

double MFAtlasGetExpFactor(MFAtlas A)
 {
  return A->expFactor;
  return;
 }

void MFAtlasClearClipF(MFAtlas this)
 {
  static char RoutineName[]={"MFAtlasClearClipF"};
  int i;
  int verbose=0;

  for(i=0;i<this->nclipf;i++)this->clipf[i]=NULL;
  this->nclipf=0;
  return;
 }

void MFAtlasClosePlotfile(MFAtlas this)
 {
  static char RoutineName[]={"MFAtlasClosePlotfile"};

  if(this->plotFileFid!=(FILE*)NULL)
   {
    fclose(this->plotFileFid);
    this->plotFileFid=(FILE*)NULL;
   }
  return;
 }

void MFAtlasCloseCenterfile(MFAtlas this)
 {
  static char RoutineName[]={"MFAtlasCloseCenterfile"};

  if(this->centerFileFid!=(FILE*)NULL)
   {
    fclose(this->centerFileFid);
    this->centerFileFid=(FILE*)NULL;
   }
  return;
 }

void MFAtlasSetNearRtn(MFAtlas A,int (*isNear)(MFAtlas,MFChart,MFChart))
 {
  A->isNear=isNear;
 }

#ifdef OLDHS

int MFAtlasHalfSpaceIndex(MFAtlas A,int i)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceIndex"};

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
    return -1;
   }

  if(i<0||i>=A->nHS)
   {
    sprintf(MFAtlasErrorMsg,"Halfspace %d (argument 2) is invalid. Must be positive and less than %d (total number of Halfspaces).",i,A->nHS);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
    return -1;
   }

  return A->globalIndex[i];
 }

int MFAtlasHalfSpaceLeftChart(MFAtlas A,int i)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceLeftChart"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasHalfSpaceLeftChart++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceLeftChart+=clock()-starttime;
#endif
    return -1;
   }

  if(i<0||i>=A->nHS)
   {
    sprintf(MFAtlasErrorMsg,"Halfspace %d (argument 2) is invalid. Must be positive and less than %d (total number of Halfspaces).",i,A->nHS);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceLeftChart+=clock()-starttime;
#endif
    return -1;
   }

  if(A->leftChart[i]<0||A->leftChart[i]>=A->nCharts)
   {
    sprintf(MFAtlasErrorMsg,"Warning, Halfspace %d has an invalid left chart (%d). Must be positive and less than %d (total number of charts).",i,A->leftChart[i],A->nCharts);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceLeftChart+=clock()-starttime;
#endif
    return -1;
   }

#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceLeftChart+=clock()-starttime;
#endif
  return A->leftChart[i];
 }

int MFAtlasHalfSpaceRightChart(MFAtlas A,int i)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceRightChart"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasHalfSpaceRightChart++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceRightChart+=clock()-starttime;
#endif
    return -1;
   }

  if(i<0||i>=A->nHS)
   {
    sprintf(MFAtlasErrorMsg,"Halfspace %d (argument 2) is invalid. Must be positive and less than %d (total number of Halfspaces).",i,A->nHS);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceRightChart+=clock()-starttime;
#endif
    return -1;
   }

  if(A->rightChart[i]<0||A->rightChart[i]>=A->nCharts)
   {
    sprintf(MFAtlasErrorMsg,"Warning, Halfspace %d has an invalid right chart (%d). Must be positive and less than %d (total number of charts).",i,A->rightChart[i],A->nCharts);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceRightChart+=clock()-starttime;
#endif
    return -1;
   }

#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceRightChart+=clock()-starttime;
#endif
  return A->rightChart[i];
 }

MFKVector MFAtlasHalfSpaceNormal(MFAtlas A,int i)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceNormal"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasHalfSpaceNormal++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceNormal+=clock()-starttime;
#endif
    return (MFKVector)NULL;
   }

  if(i<0||i>=A->nHS)
   {
    sprintf(MFAtlasErrorMsg,"Halfspace %d (argument 2) is invalid. Must be positive and less than %d (total number of Halfspaces).",i,A->nHS);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceNormal+=clock()-starttime;
#endif
    return (MFKVector)NULL;
   }

#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceNormal+=clock()-starttime;
#endif
  return A->nrm[i];
 }

double MFAtlasHalfSpaceOrigin(MFAtlas A,int i)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceOrigin"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasHalfSpaceOrigin++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceOrigin+=clock()-starttime;
#endif
    return DBL_QNAN;
   }

  if(i<0||i>=A->nHS)
   {
    sprintf(MFAtlasErrorMsg,"Halfspace %d (argument 2) is invalid. Must be positive and less than %d (total number of Halfspaces).",i,A->nHS);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceOrigin+=clock()-starttime;
#endif
    return DBL_QNAN;
   }

#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceOrigin+=clock()-starttime;
#endif
  return A->onrm[i];
 }

void MFAtlasHalfSpaceSetIndex(MFAtlas A,int i,int gi)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceSetIndex"};
  int result;

  A->globalIndex[i]=gi;

  return;
 }

void MFAtlasHalfSpaceSetOrigin(MFAtlas A,int i, double o)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceSetOrigin"};

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(i<0||i>=A->nHS)
   {
    sprintf(MFAtlasErrorMsg,"Halfspace %d (argument 2) is invalid. Must be positive and less than %d (total number of Halfspaces).",i,A->nHS);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
    return;
   }

  A->onrm[i]=o;

  return;
 }

void MFAtlasHalfSpaceSetNormal(MFAtlas A,int i, MFKVector n)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceSetNormal"};

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(i<0||i>=A->nHS)
   {
    sprintf(MFAtlasErrorMsg,"Halfspace %d (argument 2) is invalid. Must be positive and less than %d (total number of Halfspaces).",i,A->nHS);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
    return;
   }

  A->nrm[i]=n;
  MFRefKVector(n);

  return;
 }

void MFAtlasHalfSpaceSetLeftChart(MFAtlas A,int i, int left)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceSetLeftChart"};

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(i<0||i>=A->nHS)
   {
    sprintf(MFAtlasErrorMsg,"Halfspace %d (argument 2) is invalid. Must be positive and less than %d (total number of Halfspaces).",i,A->nHS);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
    return;
   }

  A->leftChart[i]=left;

  return;

 }

void MFAtlasHalfSpaceSetRightChart(MFAtlas A,int i, int right)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceSetRightChart"};

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(i<0||i>=A->nHS)
   {
    sprintf(MFAtlasErrorMsg,"Halfspace %d (argument 2) is invalid. Must be positive and less than %d (total number of Halfspaces).",i,A->nHS);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
    return;
   }

  A->rightChart[i]=right;

  return;
 }

void MFAtlasHalfSpaceIncreaseSize(MFAtlas A)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceIncreaseSize"};
  int verbose=0;
  int i;

  printf("    Changing mHS from %d to %d\n",A->mHS,A->mHS+MFAtlasIncrementToAllocate);
  MFReportMemory();
  fflush(stdout);
  A->mHS+=MFAtlasIncrementToAllocate;

  A->globalIndex=(int*)realloc((void*)A->globalIndex,A->mHS*sizeof(int)); /*done*/
  if(A->globalIndex==(int*)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Out of memory (1), trying to allocate %d bytes\n",A->mHS*sizeof(int));
    printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
    printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
    MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
MFTimeMFAddHalfSpaceToAtlas+=clock()-starttime;
#endif
    return;
   }
  for(i=A->nHS;i<A->mHS;i++)A->globalIndex[i]=-1;

  A->onrm=(double*)realloc((void*)A->onrm,A->mHS*sizeof(double)); /*done*/
  if(A->onrm==(double*)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Out of memory (2), trying to allocate %d bytes\n",A->mHS*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
    printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
    MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAddHalfSpaceToAtlas+=clock()-starttime;
#endif
    return;
   }
  for(i=A->nHS;i<A->mHS;i++)A->onrm[i]=DBL_INFINITY;

  A->nrm=(MFKVector*)realloc((void*)A->nrm,A->mHS*sizeof(MFKVector)); /*done*/
  if(A->nrm==(MFKVector*)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Out of memory (3), trying to allocate %d bytes",A->mHS*sizeof(MFKVector));
    printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
    printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
    MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAddHalfSpaceToAtlas+=clock()-starttime;
#endif
    return;
   }
  for(i=A->nHS;i<A->mHS;i++)A->nrm[i]=(MFKVector)NULL;

  A->leftChart=(int*)realloc((void*)A->leftChart,A->mHS*sizeof(int)); /*done*/
  if(A->leftChart==(int*)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Out of memory (4), trying to allocate %d bytes",A->mHS*sizeof(int));
    printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
    printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
    MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAddHalfSpaceToAtlas+=clock()-starttime;
#endif
    return;
   }
  for(i=A->nHS;i<A->mHS;i++)A->leftChart[i]=-1;

  A->rightChart=(int*)realloc((void*)A->rightChart,A->mHS*sizeof(int)); /*done*/
  if(A->rightChart==(int*)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Out of memory (5), trying to allocate %d bytes",A->mHS*sizeof(int));
    printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
    printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
    MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAddHalfSpaceToAtlas+=clock()-starttime;
#endif
    return;
   }
  for(i=A->nHS;i<A->mHS;i++)A->rightChart[i]=-1;
  printf("    done changing mHS\n");
  MFReportMemory();
  fflush(stdout);
  if(verbose){printf("%s -- done allocs\n",RoutineName);fflush(stdout);}
  return;
 }

void MFAtlasHalfSpaceFree(MFAtlas A)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceFree"};
  int verbose=0;
  int i;

  if(verbose){printf("  %s\n",RoutineName);fflush(stdout);}
  if(verbose){printf("  Delete List of onrm\n");fflush(stdout);}
  if(A->onrm!=(double*)NULL)free(A->onrm);
  if(verbose){printf("  Delete List of globalIndex\n");fflush(stdout);}
  if(A->globalIndex!=(int*)NULL)free(A->globalIndex);
  if(verbose){printf("  Delete List of leftChart\n");fflush(stdout);}
  if(A->leftChart!=(int*)NULL)free(A->leftChart);
  if(verbose){printf("  Delete List of rightChart\n");fflush(stdout);}
  if(A->rightChart!=(int*)NULL)free(A->rightChart);
  if(A->nrm!=(MFKVector*)NULL)
   {
    if(verbose){printf("  Delete List of nrm\n");fflush(stdout);}
    for(i=0;i<A->nHS;i++)
      if(A->nrm[i]!=(MFKVector)NULL)
        MFFreeKVector(A->nrm[i]);
    free(A->nrm);
   }
  return;
 }

void MFAtlasHalfSpaceInit(MFAtlas A)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceInit"};

  A->nrm=(MFKVector*)NULL;
  A->onrm=(double*)NULL;
  A->leftChart=(int*)NULL;
  A->rightChart=(int*)NULL;
  return;
 }

#else

int MFAtlasHalfSpaceIndex(MFAtlas A,int i)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceIndex"};
  int I,J;

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
    return -1;
   }

  if(i<0||i>=A->nHS)
   {
    sprintf(MFAtlasErrorMsg,"Halfspace %d (argument 2) is invalid. Must be positive and less than %d (total number of Halfspaces).",i,A->nHS);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
    return -1;
   }

  I=i/MFAtlasIncrementToAllocate;
  J=i-I*MFAtlasIncrementToAllocate;

  return A->globalIndex[I][J];
 }

int MFAtlasHalfSpaceLeftChart(MFAtlas A,int i)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceLeftChart"};
  int I,J;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasHalfSpaceLeftChart++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceLeftChart+=clock()-starttime;
#endif
    return -1;
   }

  if(i<0||i>=A->nHS)
   {
    sprintf(MFAtlasErrorMsg,"Halfspace %d (argument 2) is invalid. Must be positive and less than %d (total number of Halfspaces).",i,A->nHS);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceLeftChart+=clock()-starttime;
#endif
    return -1;
   }

  I=i/MFAtlasIncrementToAllocate;
  J=i-I*MFAtlasIncrementToAllocate;

  if(A->leftChart[I][J]<0||A->leftChart[I][J]>=A->nCharts)
   {
    sprintf(MFAtlasErrorMsg,"Warning, Halfspace %d has an invalid left chart (%d). Must be positive and less than %d (total number of charts).",i,A->leftChart[i],A->nCharts);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceLeftChart+=clock()-starttime;
#endif
    return -1;
   }

#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceLeftChart+=clock()-starttime;
#endif
  return A->leftChart[I][J];
 }

int MFAtlasHalfSpaceRightChart(MFAtlas A,int i)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceRightChart"};
  int I,J;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasHalfSpaceRightChart++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceRightChart+=clock()-starttime;
#endif
    return -1;
   }

  if(i<0||i>=A->nHS)
   {
    sprintf(MFAtlasErrorMsg,"Halfspace %d (argument 2) is invalid. Must be positive and less than %d (total number of Halfspaces).",i,A->nHS);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceRightChart+=clock()-starttime;
#endif
    return -1;
   }

  I=i/MFAtlasIncrementToAllocate;
  J=i-I*MFAtlasIncrementToAllocate;

  if(A->rightChart[I][J]<0||A->rightChart[I][J]>=A->nCharts)
   {
    sprintf(MFAtlasErrorMsg,"Warning, Halfspace %d has an invalid right chart (%d). Must be positive and less than %d (total number of charts).",i,A->rightChart[i],A->nCharts);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceRightChart+=clock()-starttime;
#endif
    return -1;
   }

#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceRightChart+=clock()-starttime;
#endif
  return A->rightChart[I][J];
 }

MFKVector MFAtlasHalfSpaceNormal(MFAtlas A,int i)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceNormal"};
  int I,J;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasHalfSpaceNormal++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceNormal+=clock()-starttime;
#endif
    return (MFKVector)NULL;
   }

  if(i<0||i>=A->nHS)
   {
    sprintf(MFAtlasErrorMsg,"Halfspace %d (argument 2) is invalid. Must be positive and less than %d (total number of Halfspaces).",i,A->nHS);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceNormal+=clock()-starttime;
#endif
    return (MFKVector)NULL;
   }

  I=i/MFAtlasIncrementToAllocate;
  J=i-I*MFAtlasIncrementToAllocate;

#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceNormal+=clock()-starttime;
#endif
  return A->nrm[I][J];
 }

double MFAtlasHalfSpaceOrigin(MFAtlas A,int i)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceOrigin"};
  int I,J;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFAtlasHalfSpaceOrigin++;
  starttime=clock();
#endif

#ifdef MFDEBUG
  _heap_check();
#endif

  if(A==(MFAtlas)NULL)
   {
    sprintf(MFAtlasErrorMsg,"Pointer to Atlas (argument 1) is NULL");
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceOrigin+=clock()-starttime;
#endif
    return DBL_QNAN;
   }

  if(i<0||i>=A->nHS)
   {
    sprintf(MFAtlasErrorMsg,"Halfspace %d (argument 2) is invalid. Must be positive and less than %d (total number of Halfspaces).",i,A->nHS);
    MFSetError(4,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceOrigin+=clock()-starttime;
#endif
    return DBL_QNAN;
   }

#ifdef MFTIMINGS
  MFTimeMFAtlasHalfSpaceOrigin+=clock()-starttime;
#endif

  I=i/MFAtlasIncrementToAllocate;
  J=i-I*MFAtlasIncrementToAllocate;
  return A->onrm[I][J];
 }

void MFAtlasHalfSpaceSetIndex(MFAtlas A,int i,int gi)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceSetIndex"};
  int result;
  int I,J;

  I=i/MFAtlasIncrementToAllocate;
  J=i-I*MFAtlasIncrementToAllocate;
  if(I>=A->nnHS)MFAtlasHalfSpaceIncreaseSize(A);
  A->globalIndex[I][J]=gi;

  return;
 }

void MFAtlasHalfSpaceSetOrigin(MFAtlas A,int i, double o)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceSetOrigin"};
  int I,J;

  I=i/MFAtlasIncrementToAllocate;
  J=i-I*MFAtlasIncrementToAllocate;
  if(I>=A->nnHS)MFAtlasHalfSpaceIncreaseSize(A);

  A->onrm[I][J]=o;

  return;
 }

void MFAtlasHalfSpaceSetNormal(MFAtlas A,int i, MFKVector n)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceSetNormal"};
  int I,J;

  I=i/MFAtlasIncrementToAllocate;
  J=i-I*MFAtlasIncrementToAllocate;
  if(I>=A->nnHS)MFAtlasHalfSpaceIncreaseSize(A);

  A->nrm[I][J]=n;
  MFRefKVector(n);

  return;
 }

void MFAtlasHalfSpaceSetLeftChart(MFAtlas A,int i, int left)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceSetLeftChart"};
  int I,J;

  I=i/MFAtlasIncrementToAllocate;
  J=i-I*MFAtlasIncrementToAllocate;
  if(I>=A->nnHS)MFAtlasHalfSpaceIncreaseSize(A);

  A->leftChart[I][J]=left;

  return;

 }

void MFAtlasHalfSpaceSetRightChart(MFAtlas A,int i, int right)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceSetRightChart"};
  int I,J;

  I=i/MFAtlasIncrementToAllocate;
  J=i-I*MFAtlasIncrementToAllocate;
  if(I>=A->mmHS)MFAtlasHalfSpaceIncreaseSize(A);

  A->rightChart[I][J]=right;

  return;
 }

void MFAtlasHalfSpaceIncreaseSize(MFAtlas A)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceIncreaseSize"};
  int verbose=0;
  int i;
  int I;

  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

  I=A->nHS/MFAtlasIncrementToAllocate;
  if(I>=A->mmHS)
   {
    A->mmHS+=MFAtlasIncrementToAllocate;
    A->mHS+=MFAtlasIncrementToAllocate;
    A->globalIndex=(int**)realloc((void*)A->globalIndex,A->mmHS*sizeof(int*)); /*done*/
    if(A->globalIndex==(int**)NULL)
     {
      sprintf(MFAtlasErrorMsg,"Out of memory (1), trying to allocate %d bytes\n",A->mmHS*sizeof(int*));
      printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
      MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
      return;
     }
    for(i=A->nnHS;i<A->mmHS;i++)A->globalIndex[i]=(int*)NULL;

    A->onrm=(double**)realloc((void*)A->onrm,A->mmHS*sizeof(double*)); /*done*/
    if(A->onrm==(double**)NULL)
     {
      sprintf(MFAtlasErrorMsg,"Out of memory (1), trying to allocate %d bytes\n",A->mmHS*sizeof(double*));
      printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
      MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
      return;
     }
    for(i=A->nnHS;i<A->mmHS;i++)A->onrm[i]=(double*)NULL;

    A->nrm=(MFKVector**)realloc((void*)A->nrm,A->mmHS*sizeof(MFKVector*)); /*done*/
    if(A->nrm==(MFKVector**)NULL)
     {
      sprintf(MFAtlasErrorMsg,"Out of memory (1), trying to allocate %d bytes\n",A->mmHS*sizeof(MFKVector*));
      printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
      MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
      return;
     }
    for(i=A->nnHS;i<A->mmHS;i++)A->nrm[i]=(MFKVector*)NULL;

    A->leftChart=(int**)realloc((void*)A->leftChart,A->mmHS*sizeof(int*)); /*done*/
    if(A->leftChart==(int**)NULL)
     {
      sprintf(MFAtlasErrorMsg,"Out of memory (1), trying to allocate %d bytes\n",A->mmHS*sizeof(int*));
      printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
      MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
      return;
     }
    for(i=A->nnHS;i<A->mmHS;i++)A->leftChart[i]=(int*)NULL;

    A->rightChart=(int**)realloc((void*)A->rightChart,A->mmHS*sizeof(int*)); /*done*/
    if(A->rightChart==(int**)NULL)
     {
      sprintf(MFAtlasErrorMsg,"Out of memory (1), trying to allocate %d bytes\n",A->mmHS*sizeof(int*));
      printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
      MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
      return;
     }
    for(i=A->nnHS;i<A->mmHS;i++)A->rightChart[i]=(int*)NULL;
   }

  if(I>=A->nnHS)
   {
    A->nnHS++;

    A->globalIndex[I]=(int*)malloc(MFAtlasIncrementToAllocate*sizeof(int)); /*done*/
    if(A->globalIndex[I]==(int*)NULL)
     {
      sprintf(MFAtlasErrorMsg,"Out of memory (1), trying to allocate %d bytes\n",MFAtlasIncrementToAllocate*sizeof(int));
      printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
      MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
      return;
     }
    for(i=0;i<MFAtlasIncrementToAllocate;i++)A->globalIndex[I][i]=-1;

    A->onrm[I]=(double*)malloc(MFAtlasIncrementToAllocate*sizeof(double)); /*done*/
    if(A->onrm[I]==(double*)NULL)
     {
      sprintf(MFAtlasErrorMsg,"Out of memory (1), trying to allocate %d bytes\n",MFAtlasIncrementToAllocate*sizeof(double));
      printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
      MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
      return;
     }
    for(i=0;i<MFAtlasIncrementToAllocate;i++)A->onrm[I][i]=-DBL_MAX;

    A->nrm[I]=(MFKVector*)malloc(MFAtlasIncrementToAllocate*sizeof(MFKVector)); /*done*/
    if(A->nrm[I]==(MFKVector*)NULL)
     {
      sprintf(MFAtlasErrorMsg,"Out of memory (1), trying to allocate %d bytes\n",MFAtlasIncrementToAllocate*sizeof(MFKVector));
      printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
      MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
      return;
     }
    for(i=0;i<MFAtlasIncrementToAllocate;i++)A->nrm[I][i]=(MFKVector)NULL;

    A->leftChart[I]=(int*)malloc(MFAtlasIncrementToAllocate*sizeof(int)); /*done*/
    if(A->leftChart[I]==(int*)NULL)
     {
      sprintf(MFAtlasErrorMsg,"Out of memory (1), trying to allocate %d bytes\n",MFAtlasIncrementToAllocate*sizeof(int));
      printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
      MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
      return;
     }
    for(i=0;i<MFAtlasIncrementToAllocate;i++)A->leftChart[I][i]=-1;

    A->rightChart[I]=(int*)malloc(MFAtlasIncrementToAllocate*sizeof(int)); /*done*/
    if(A->rightChart[I]==(int*)NULL)
     {
      sprintf(MFAtlasErrorMsg,"Out of memory (1), trying to allocate %d bytes\n",MFAtlasIncrementToAllocate*sizeof(int));
      printf("%s -- %s\n",RoutineName,MFAtlasErrorMsg);fflush(stdout);
      printf("  line %d of file %s\n",__LINE__,__FILE__);fflush(stdout);
      MFSetError(12,RoutineName,MFAtlasErrorMsg,__LINE__,__FILE__);
      return;
     }
    for(i=0;i<MFAtlasIncrementToAllocate;i++)A->rightChart[I][i]=-1;
   }

  if(verbose){printf("%s -- done allocs\n",RoutineName);fflush(stdout);}
  return;
 }

void MFAtlasHalfSpaceFree(MFAtlas A)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceFree"};
  int verbose=0;
  int i,j;

  if(verbose){printf("  %s\n",RoutineName);fflush(stdout);}
  if(verbose){printf("  Delete List of onrm\n");fflush(stdout);}
  for(i=0;i<A->nnHS;i++)
   {
    if(A->onrm[i]!=(double*)NULL)free(A->onrm[i]);
    if(A->globalIndex[i]!=(int*)NULL)free(A->globalIndex[i]);
    if(A->leftChart[i]!=(int*)NULL)free(A->leftChart[i]);
    if(A->rightChart[i]!=(int*)NULL)free(A->rightChart[i]);
    if(A->nrm[i]!=(MFKVector*)NULL)
     {
      for(j=0;j<MFAtlasIncrementToAllocate;j++)
        if(A->nrm[i][j]!=(MFKVector)NULL)MFFreeKVector(A->nrm[i][j]);
      free(A->nrm[i]);
     }
   }
  if(A->onrm!=(double**)NULL)free(A->onrm);
  if(A->globalIndex!=(int**)NULL)free(A->globalIndex);
  if(A->leftChart!=(int**)NULL)free(A->leftChart);
  if(A->rightChart!=(int**)NULL)free(A->rightChart);
  if(A->nrm!=(MFKVector**)NULL)free(A->nrm);

  return;
 }

void MFAtlasHalfSpaceInit(MFAtlas A)
 {
  static char RoutineName[]={"MFAtlasHalfSpaceInit"};

  A->nnHS=0;
  A->mmHS=0;
  A->globalIndex=(int**)NULL;
  A->nrm=(MFKVector**)NULL;
  A->onrm=(double**)NULL;
  A->leftChart=(int**)NULL;
  A->rightChart=(int**)NULL;
  return;
 }
#endif
