/* 
    @(#)MFBinaryTree.c	1.13
    03/02/26 21:10:44
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/

/*  Author: Michael E. Henderson mhender@watson.ibm.com               */

#include <MFBinaryTree.h>
#include <MFListOfCharts.h>
#include <MFAtlas.h>
#include <MFPrint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <MFTime.h>
#ifdef MFDEBUG
#include <umalloc.h>
#endif

void MFSetError(int,char*,char*,int,char*);
static char MFBinaryTreeErrorMsg[256]="";

int MFBINARYTREESEQUENCENUMBER=0;
#define MFBINARYTREEMAXNUMBEROFCHARTS 5

struct MFBinaryTreeLeafSt;
typedef struct MFBinaryTreeLeafSt *MFBinaryTreeLeaf;

struct MFBinaryTreeSt {
                       int nCharts;
                       int k;
                       MFBinaryTreeLeaf root;
                      };

struct MFBinaryTreeLeafSt {
                     int sequenceNumber;
                     int nCharts;
                     int *chart;
                     double *chartR;
                     double **chartCenter;
                     double *value;
                     int d;
                     double split;
                     double *leftB;
                     double *rightB;
                     MFBinaryTreeLeaf parent;
                     MFBinaryTreeLeaf leftT;
                     MFBinaryTreeLeaf rightT;
                    };

MFBinaryTreeLeaf MFCreateBinaryTreeLeaf(int,int);
void MFFreeBinaryTreeLeaf(MFBinaryTreeLeaf);
void MFBinaryTreeLeafAddChart(MFBinaryTreeLeaf,int,int,double,double*);
void MFBinaryTreeGetIntersectingCharts(MFBinaryTreeLeaf,int,int,double*,double,int*,int**);
void MFBinaryTreeGetNeighboringCharts(MFBinaryTreeLeaf,int,double*,double,int*,int**);

static int nInTree=0;

MFBinaryTree MFCreateBinaryTree(int k)
 {
  static char RoutineName[]={"MFCreateBinaryTree"};
  int i;
  MFBinaryTree T;
  MFBinaryTreeLeaf TLeaf;
  int verbose;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFCreateBinaryTree++;
  starttime=clock();
#endif

#ifdef MFDEBUG
 _heap_check();
#endif

  verbose=0;
  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

  if(k<1)
   {
    sprintf(MFBinaryTreeErrorMsg,"Dimension %d, (Argument 1) must be positive.",k);
    printf("%s -- %s\n",RoutineName,MFBinaryTreeErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFBinaryTreeErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateBinaryTree+=clock()-starttime;
#endif
    return (MFBinaryTree)NULL;
   }

  T=(MFBinaryTree)malloc(sizeof(struct MFBinaryTreeSt)); /*done*/
  if(T==(MFBinaryTree)NULL)
   {
    sprintf(MFBinaryTreeErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFBinaryTreeSt));
    printf("%s -- %s\n",RoutineName,MFBinaryTreeErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFBinaryTreeErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateBinaryTree+=clock()-starttime;
#endif
    return (MFBinaryTree)NULL;
   }
  if(verbose){printf("%s, T=0x%8.8x\n",RoutineName,T);fflush(stdout);}

  T->nCharts=0;
  T->k=k;
  TLeaf=MFCreateBinaryTreeLeaf(T->k,0);
  TLeaf->parent=(MFBinaryTreeLeaf)NULL;
  T->root=TLeaf;
  nInTree=0;

#ifdef MFTIMINGS
    MFTimeMFCreateBinaryTree+=clock()-starttime;
#endif
  return T;
 }

MFBinaryTreeLeaf MFCreateBinaryTreeLeaf(int k,int d)
 {
  static char RoutineName[]={"MFCreateBinaryTreeLeaf"};
  int i;
  MFBinaryTreeLeaf TLeaf;
  double t;
  int verbose;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFCreateBinaryTreeLeaf++;
  starttime=clock();
#endif

#ifdef MFDEBUG
 _heap_check();
#endif

  verbose=0;

  if(d<0 || !(d<k))
   {
    sprintf(MFBinaryTreeErrorMsg,"direction %d (argument 3) is invalid. Must be positive and less than %d",d,k);
    printf("%s -- %s\n",RoutineName,MFBinaryTreeErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFBinaryTreeErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateBinaryTreeLeaf+=clock()-starttime;
#endif
    return (MFBinaryTreeLeaf)NULL;
   }

  TLeaf=(MFBinaryTreeLeaf)malloc(sizeof(struct MFBinaryTreeLeafSt)); /*done*/
  if(TLeaf==(MFBinaryTreeLeaf)NULL)
   {
    sprintf(MFBinaryTreeErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFBinaryTreeLeafSt));
    printf("%s -- %s\n",RoutineName,MFBinaryTreeErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFBinaryTreeErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateBinaryTreeLeaf+=clock()-starttime;
#endif
    return (MFBinaryTreeLeaf)NULL;
   }
  if(verbose){printf("%s, T=0x%8.8x\n",RoutineName,TLeaf);fflush(stdout);}

  TLeaf->sequenceNumber=MFBINARYTREESEQUENCENUMBER;
  MFBINARYTREESEQUENCENUMBER++;
  TLeaf->d=d;
  TLeaf->split=0.;
  TLeaf->nCharts=0;

  TLeaf->chart=(int*)malloc(MFBINARYTREEMAXNUMBEROFCHARTS*sizeof(int)); /*done*/
  if(TLeaf->chart==(int*)NULL)
   {
    sprintf(MFBinaryTreeErrorMsg,"Out of memory, trying to allocate %d bytes",MFBINARYTREEMAXNUMBEROFCHARTS*sizeof(int));
    printf("%s -- %s\n",RoutineName,MFBinaryTreeErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFBinaryTreeErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateBinaryTreeLeaf+=clock()-starttime;
#endif
    return (MFBinaryTreeLeaf)NULL;
   }
  if(verbose){printf("     T->chart=0x%8.8x\n",TLeaf->chart);fflush(stdout);}

  TLeaf->chartR=(double*)malloc(MFBINARYTREEMAXNUMBEROFCHARTS*sizeof(double)); /*done*/
  if(TLeaf->chartR==(double*)NULL)
   {
    sprintf(MFBinaryTreeErrorMsg,"Out of memory, trying to allocate %d bytes",MFBINARYTREEMAXNUMBEROFCHARTS*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFBinaryTreeErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFBinaryTreeErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateBinaryTreeLeaf+=clock()-starttime;
#endif
    return (MFBinaryTreeLeaf)NULL;
   }
  if(verbose){printf("     T->chartR=0x%8.8x\n",TLeaf->chartR);fflush(stdout);}

  TLeaf->chartCenter=(double**)malloc(MFBINARYTREEMAXNUMBEROFCHARTS*k*sizeof(double)); /*done*/
  if(TLeaf->chartCenter==(double**)NULL)
   {
    sprintf(MFBinaryTreeErrorMsg,"Out of memory, trying to allocate %d bytes",MFBINARYTREEMAXNUMBEROFCHARTS*k*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFBinaryTreeErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFBinaryTreeErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateBinaryTreeLeaf+=clock()-starttime;
#endif
    return (MFBinaryTreeLeaf)NULL;
   }
  if(verbose){printf("     T->chartCenter=0x%8.8x\n",TLeaf->chartCenter);fflush(stdout);}

  TLeaf->value=(double*)malloc(MFBINARYTREEMAXNUMBEROFCHARTS*sizeof(double)); /*done*/
  if(TLeaf->value==(double*)NULL)
   {
    sprintf(MFBinaryTreeErrorMsg,"Out of memory, trying to allocate %d bytes",MFBINARYTREEMAXNUMBEROFCHARTS*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFBinaryTreeErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFBinaryTreeErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateBinaryTreeLeaf+=clock()-starttime;
#endif
    return (MFBinaryTreeLeaf)NULL;
   }
  if(verbose){printf("     T->value=0x%8.8x\n",TLeaf->value);fflush(stdout);}

  TLeaf->leftB=(double*)malloc(k*sizeof(double)); /*done*/
  if(TLeaf->leftB==(double*)NULL)
   {
    sprintf(MFBinaryTreeErrorMsg,"Out of memory, trying to allocate %d bytes",MFBINARYTREEMAXNUMBEROFCHARTS*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFBinaryTreeErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFBinaryTreeErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateBinaryTreeLeaf+=clock()-starttime;
#endif
    return (MFBinaryTreeLeaf)NULL;
   }
  if(verbose){printf("     T->leftB=0x%8.8x\n",TLeaf->leftB);fflush(stdout);}

  TLeaf->rightB=(double*)malloc(k*sizeof(double)); /*done*/
  if(TLeaf->rightB==(double*)NULL)
   {
    sprintf(MFBinaryTreeErrorMsg,"Out of memory, trying to allocate %d bytes",MFBINARYTREEMAXNUMBEROFCHARTS*sizeof(double));
    printf("%s -- %s\n",RoutineName,MFBinaryTreeErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFBinaryTreeErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateBinaryTreeLeaf+=clock()-starttime;
#endif
    return (MFBinaryTreeLeaf)NULL;
   }
  if(verbose){printf("     T->rightB=0x%8.8x\n",TLeaf->rightB);fflush(stdout);}

  for(i=0;i<k;i++)
   {
    TLeaf->leftB[i]=1.e20;
    TLeaf->rightB[i]=-1.e20;
   }

  TLeaf->parent=(MFBinaryTreeLeaf)NULL;
  TLeaf->leftT=(MFBinaryTreeLeaf)NULL;
  TLeaf->rightT=(MFBinaryTreeLeaf)NULL;

#ifdef MFTIMINGS
    MFTimeMFCreateBinaryTreeLeaf+=clock()-starttime;
#endif
  return(TLeaf);
 }

void MFFreeBinaryTree(MFBinaryTree T)
 {
  static char RoutineName[]={"MFFreeBinaryTree"};
  int verbose;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFFreeBinaryTree++;
  starttime=clock();
#endif

#ifdef MFDEBUG
 _heap_check();
#endif

  verbose=0;

  if(verbose){printf("%s, T=0x%8.8x\n",RoutineName,T);fflush(stdout);}
#ifdef MFTIMINGS
    MFTimeMFFreeBinaryTree+=clock()-starttime;
#endif
  if(T==(MFBinaryTree)NULL)return;

  if(T->root!=(MFBinaryTreeLeaf)NULL)MFFreeBinaryTreeLeaf(T->root);
  free(T);
  if(verbose){printf("done %s, T=0x%8.8x\n",RoutineName,T);fflush(stdout);}
#ifdef MFTIMINGS
    MFTimeMFFreeBinaryTree+=clock()-starttime;
#endif
  return;
 }

void MFFreeBinaryTreeLeaf(MFBinaryTreeLeaf T)
 {
  static char RoutineName[]={"MFFreeBinaryTreeLeaf"};
  int i;
  int verbose;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFFreeBinaryTreeLeaf++;
  starttime=clock();
#endif

#ifdef MFDEBUG
 _heap_check();
#endif

  verbose=0;
  if(verbose){printf("%s, T=0x%8.8x\n",RoutineName,T);fflush(stdout);}
  if(verbose){printf("     T->chart=0x%8.8x\n",T->chart);fflush(stdout);}
  if(verbose){printf("     T->chartR=0x%8.8x\n",T->chartR);fflush(stdout);}
  if(verbose){printf("     T->chartCenter=0x%8.8x\n",T->chartCenter);fflush(stdout);}
  if(verbose){printf("     T->value=0x%8.8x\n",T->value);fflush(stdout);}
  if(verbose){printf("     T->leftB=0x%8.8x\n",T->leftB);fflush(stdout);}
  if(verbose){printf("     T->rightB=0x%8.8x\n",T->rightB);fflush(stdout);}

  if(T==(MFBinaryTreeLeaf)NULL)
   {
#ifdef MFTIMINGS
    MFTimeMFFreeBinaryTreeLeaf+=clock()-starttime;
#endif
    return;
   }
  if(T->chart!=(int*)NULL)free(T->chart);
  T->chart=(int*)NULL;
  if(T->chartR!=(double*)NULL)free(T->chartR);
  T->chartR=(double*)NULL;
  if(T->chartCenter!=(double**)NULL)
   {
    for(i=0;i<T->nCharts;i++)
      if(T->chartCenter[i]!=(double*)NULL)free(T->chartCenter[i]);
    free(T->chartCenter);
    T->chartCenter=(double**)NULL;
   }
  if(T->value!=(double*)NULL)free(T->value);
  T->value=(double*)NULL;
  if(T->leftB!=(double*)NULL)free(T->leftB);
  T->leftB=(double*)NULL;
  if(T->rightB!=(double*)NULL)free(T->rightB);
  T->rightB=(double*)NULL;
  if(T->leftT!=(MFBinaryTreeLeaf)NULL)MFFreeBinaryTreeLeaf(T->leftT);
  T->leftT=(MFBinaryTreeLeaf)NULL;
  if(T->rightT!=(MFBinaryTreeLeaf)NULL)MFFreeBinaryTreeLeaf(T->rightT);
  T->rightT=(MFBinaryTreeLeaf)NULL;
  free(T);
  if(verbose){printf("done %s, T=0x%8.8x\n",RoutineName,T);fflush(stdout);}
#ifdef MFTIMINGS
    MFTimeMFFreeBinaryTreeLeaf+=clock()-starttime;
#endif
  return;
 }

void MFBinaryTreeAddChart(MFBinaryTree TRoot,int chartNo,double *center, double R)
 {
  static char RoutineName[]={"MFBinaryTreeAddChart"};
  MFBinaryTreeLeaf T;
  int k;
  int verbose;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFBinaryTreeAddChart++;
  starttime=clock();
#endif

#ifdef MFDEBUG
 _heap_check();
#endif
  nInTree++;

  verbose=0;
  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

  T=TRoot->root;
  k=TRoot->k;
  MFBinaryTreeLeafAddChart(T,k,chartNo,R,center);

#ifdef MFTIMINGS
    MFTimeMFBinaryTreeAddChart+=clock()-starttime;
#endif
  return;
 }

void MFBinaryTreeLeafAddChart(MFBinaryTreeLeaf T,int k, int chart, double R, double *center)
 {
  static char RoutineName[]={"MFBinaryTreeLeafAddChart"};
  int i,n;
  double t;
  int verbose;
  int j,it;
  double *pt;
  double maxwidth,nextmaxwidth;
  double x0,x1;
  int nextdir;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFBinaryTreeLeafAddChart++;
  starttime=clock();
#endif

#ifdef MFDEBUG
 _heap_check();
#endif

  verbose=0;

  t=center[T->d];

  if(verbose)
   {
    printf("%s, %d, center (",RoutineName,T->sequenceNumber);
    for(i=0;i<k;i++){if(i>0)printf(",");printf("%lf",center[i]);}
    printf(") radius %lf, direction %d, t=%lf\n",R,T->d,t);fflush(stdout);
   }

  if(T->leftT==(MFBinaryTreeLeaf)NULL && T->rightT==(MFBinaryTreeLeaf)NULL &&
     T->nCharts<MFBINARYTREEMAXNUMBEROFCHARTS)
   {
    if(verbose){printf("  this leaf (%d) is not a branch pt, and has space (%d charts/%d)\n",T->sequenceNumber,T->nCharts,MFBINARYTREEMAXNUMBEROFCHARTS);fflush(stdout);}
    if(T->nCharts>0)
     {
      for(i=0;i<T->nCharts;i++)if(T->chart[i]==chart){printf("Error, list already contains chart %d\n",i);fflush(stdout);}
      n=0;
      while(n<T->nCharts && T->value[n]<t)n++;
      for(i=T->nCharts;i>n;i--)
       {
        T->chart[i]=T->chart[i-1];
        T->chartR[i]=T->chartR[i-1];
        T->chartCenter[i]=T->chartCenter[i-1];
        T->value[i]=T->value[i-1];
       }
      T->chart[n]=chart;
      T->chartR[n]=R;
      T->chartCenter[n]=(double*)malloc(k*sizeof(double));
      if(T->chartCenter[n]==(double*)NULL)abort();
      for(i=0;i<k;i++)(T->chartCenter[n])[i]=center[i];
      T->value[n]=t;
     }else{
      T->chart[0]=chart;
      T->chartR[0]=R;
      T->chartCenter[0]=(double*)malloc(k*sizeof(double));
      if(T->chartCenter[0]==(double*)NULL)abort();
      for(i=0;i<k;i++)(T->chartCenter[0])[i]=center[i];
      T->value[0]=t;
     }
    T->nCharts++;
    for(i=0;i<k;i++)
     {
      if(T->nCharts==1)
       {
        T->leftB[i]=center[i]-R;
        T->rightB[i]=center[i]+R;
       }else{
        t=center[i]-R;
        if(t<T->leftB[i])T->leftB[i]=t;
        t=center[i]+R;
        if(t>T->rightB[i])T->rightB[i]=t;
       }
     }
   }else if(T->leftT==(MFBinaryTreeLeaf)NULL && T->rightT==(MFBinaryTreeLeaf)NULL)
   {
    if(verbose){printf("  this leaf (%d) is not a branch pt, and has no space left (%d charts/%d)\n",T->sequenceNumber,T->nCharts,MFBINARYTREEMAXNUMBEROFCHARTS);fflush(stdout);}

/* recompute d and values */

     if(verbose){printf("  recomputing split and direction. k=%d\n",k);fflush(stdout);}
     maxwidth=-1.;
     T->split=0.;
     nextmaxwidth=0.;
     nextdir=0;
     for(i=0;i<k;i++)
      {
       x0=center[i];
       x1=center[i];
       for(j=0;j<T->nCharts;j++)
        {
         if((T->chartCenter[j])[i]>x1)x1=(T->chartCenter[j])[i];
         if((T->chartCenter[j])[i]<x0)x0=(T->chartCenter[j])[i];
        }
       if(verbose){printf("      direction %d, range is [%lf,%lf]\n",i,x0,x1);fflush(stdout);}
       if(x1-x0>maxwidth)
        {
         maxwidth=x1-x0;
         T->split=(x1+x0)/2;
         T->d=i;
        }else if(x1-x0>nextmaxwidth)
        {
         nextmaxwidth=x1-x0;
         nextdir=i;
        }
      }
    if(verbose){printf("      direction %d, range is [%lf,%lf]\n",i,x0,x1);fflush(stdout);}

/* Recompute Values */
    for(i=0;i<T->nCharts;i++)
      T->value[i]=(T->chartCenter[i])[T->d];

    T->leftT=MFCreateBinaryTreeLeaf(k,nextdir);
    (T->leftT)->parent=T;
    T->rightT=MFCreateBinaryTreeLeaf(k,nextdir);
    (T->rightT)->parent=T;
    if(verbose)
     {
      printf("  this leaf (%d) is not a branch pt, and has no space\n",T->sequenceNumber);
      printf("     values [");
      for(i=0;i<T->nCharts;i++)
       {
        if(i>0)printf(",");
        printf("%lf",T->value[i]);
       }
      printf("], new value=%lf\n",center[T->d]);
      printf("     split is at %lf\n",T->split);
      fflush(stdout);
     }

    for(i=0;i<T->nCharts;i++)
     {
      if(T->value[i]<T->split)
       {
        if(verbose){printf("  move chart %d into left (%d), split=%lf, value=%lf\n",i,(T->leftT)->sequenceNumber,T->split,t);fflush(stdout);}
        MFBinaryTreeLeafAddChart(T->leftT,k,T->chart[i],T->chartR[i],T->chartCenter[i]);
       }else if(T->value[i]>T->split){
        if(verbose){printf("  move chart %d into right (%d), split=%lf, value=%lf\n",i,(T->rightT)->sequenceNumber,T->split,t);fflush(stdout);}
        MFBinaryTreeLeafAddChart(T->rightT,k,T->chart[i],T->chartR[i],T->chartCenter[i]);
       }else{
        if((T->leftT)->nCharts<(T->rightT)->nCharts)
         {
          if(verbose){printf("  put chart %d in left (%d), (same split value, left has fewer points) split=%lf, value=%lf\n",i,(T->leftT)->sequenceNumber,T->split,t);fflush(stdout);}
          MFBinaryTreeLeafAddChart(T->leftT,k,T->chart[i],T->chartR[i],T->chartCenter[i]);
         }else{
          if(verbose){printf("  put chart %d in right (%d), (same split value, right has fewer points) split=%lf, value=%lf\n",i,(T->rightT)->sequenceNumber,T->split,t);fflush(stdout);}
          MFBinaryTreeLeafAddChart(T->rightT,k,T->chart[i],T->chartR[i],T->chartCenter[i]);
         }
       }
     }

    t=center[T->d];
    if(t<T->split) /* if == T->split put in least full */
     {
      if(verbose){printf("  put new in left (%d) which has %d, split=%lf, value=%lf\n",(T->leftT)->sequenceNumber,(T->leftT)->nCharts,T->split,t);fflush(stdout);}
      MFBinaryTreeLeafAddChart(T->leftT,k,chart,R,center);
     }else if(t>T->split)
     {
      if(verbose){printf("  put new in right (%d), which has %d, split=%lf, value=%lf\n",(T->rightT)->sequenceNumber,(T->rightT)->nCharts,T->split,t);fflush(stdout);}
      MFBinaryTreeLeafAddChart(T->rightT,k,chart,R,center);
     }else{
      if(verbose){printf("  left has %d, right has %d\n",(T->leftT)->nCharts,(T->rightT)->nCharts);fflush(stdout);}
      if((T->leftT)->nCharts<(T->rightT)->nCharts)
       {
        if(verbose){printf("  put new in left (%d), (same split value, left has fewer points) split=%lf, value=%lf\n",(T->leftT)->sequenceNumber,T->split,t);fflush(stdout);}
        MFBinaryTreeLeafAddChart(T->leftT,k,chart,R,center);
       }else{
        if(verbose){printf("  put new in right (%d), (same split value, right has fewer points) split=%lf, value=%lf\n",(T->rightT)->sequenceNumber,T->split,t);fflush(stdout);}
        MFBinaryTreeLeafAddChart(T->rightT,k,chart,R,center);
       }
     }

    free(T->chart);
    T->chart=(int*)NULL;
    for(i=0;i<T->nCharts;i++)
     {
      if(T->chartCenter[i]!=(double*)NULL)free(T->chartCenter[i]);
     }
    free(T->chartCenter);
    T->chartCenter=(double**)NULL;
    free(T->chartR);
    T->chartR=(double*)NULL;
    T->nCharts=0;
   }else{
    if(verbose){printf("  this leaf (%d) is a branch pt\n",T->sequenceNumber);fflush(stdout);}
    t=center[T->d];
    if(t<T->split) /* if == T->split put in least full */
     {
      if(verbose){printf("  put new in left (%d), split=%lf, value=%lf\n",(T->leftT)->sequenceNumber,T->split,t);fflush(stdout);}
      MFBinaryTreeLeafAddChart(T->leftT,k,chart,R,center);
     }else if(t>T->split)
     {
      if(verbose){printf("  put new in right (%d), split=%lf, value=%lf\n",(T->rightT)->sequenceNumber,T->split,t);fflush(stdout);}
      MFBinaryTreeLeafAddChart(T->rightT,k,chart,R,center);
     }else{
      if((T->rightT)->nCharts<(T->rightT)->nCharts)
       {
        if(verbose){printf("  put new in left (%d), (fewer points) split=%lf, value=%lf\n",(T->leftT)->sequenceNumber,T->split,t);fflush(stdout);}
        MFBinaryTreeLeafAddChart(T->leftT,k,chart,R,center);
       }else{
        if(verbose){printf("  put new in right (%d), (fewer points) split=%lf, value=%lf\n",(T->rightT)->sequenceNumber,T->split,t);fflush(stdout);}
        MFBinaryTreeLeafAddChart(T->rightT,k,chart,R,center);
       }
     }
   }

  for(i=0;i<k;i++)
   {
    if(T->leftT!=(MFBinaryTreeLeaf)NULL)
     {
      if(T->leftB[i]>(T->leftT)->leftB[i])T->leftB[i]=(T->leftT)->leftB[i];
      if(T->rightB[i]<(T->leftT)->rightB[i])T->rightB[i]=(T->leftT)->rightB[i];
     }
    if(T->rightT!=(MFBinaryTreeLeaf)NULL)
     {
      if((T->rightT)->leftB[i]<T->leftB[i])T->leftB[i]=(T->rightT)->leftB[i];
      if((T->rightT)->rightB[i]>T->rightB[i])T->rightB[i]=(T->rightT)->rightB[i];
     }
   }

#ifdef MFTIMINGS
    MFTimeMFBinaryTreeLeafAddChart+=clock()-starttime;
#endif
  return;
 }

void MFBinaryTreeGetIntersectingCharts(MFBinaryTreeLeaf T,int k,int chart, double *center, double R,int *n,int **list)
 {
  static char RoutineName[]={"MFBinaryTreeGetIntersectingCharts"};
  int i,j;
  double t;
  int notin;

  int verbose;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFBinaryTreeGetIntersectingCharts++;
  starttime=clock();
#endif

#ifdef MFDEBUG
 _heap_check();
#endif

  verbose=0;

  if(T==(MFBinaryTreeLeaf)NULL)return;
  if(R<0)abort();
  if(k<0)abort();

  if(T->leftT==(MFBinaryTreeLeaf)NULL && T->nCharts==0)
   {
#ifdef MFTIMINGS
    MFTimeMFBinaryTreeGetIntersectingCharts+=clock()-starttime;
#endif
    return;
   }

/* Check Bounding box */

  notin=0;
  if(verbose)printf("   leaf %d bb=\n",T->sequenceNumber);
  if(T->leftB==(double*)NULL)abort();
  if(T->rightB==(double*)NULL)abort();
  for(i=0;i<k;i++)
   {
    t=center[i];
    if(verbose)printf("   %d BB %lf,%lf Disk %lf,%lf\n",i,T->leftB[i],T->rightB[i],t-R,t+R);
    if(t+R<T->leftB[i] || t-R>T->rightB[i])notin=1;
   }
  if(verbose&&notin)printf("     not in\n");
  if(verbose&&!notin)printf("     in\n");
  if(notin)
   {
#ifdef MFTIMINGS
    MFTimeMFBinaryTreeGetIntersectingCharts+=clock()-starttime;
#endif
    return;
   }

/* Chart Intersects Bounding box, add charts to list, or check sub-trees */

  if(T->leftT==(MFBinaryTreeLeaf)NULL)
   {
    *list=(int*)realloc((void*)(*list),(*n+T->nCharts)*sizeof(int)); /*done*/
    if(*list==(int*)NULL)
     {
      sprintf(MFBinaryTreeErrorMsg,"Out of memory, trying to reallocate %d bytes",(*n+T->nCharts)*sizeof(int));
      printf("%s -- %s\n",RoutineName,MFBinaryTreeErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFBinaryTreeErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFBinaryTreeGetIntersectingCharts+=clock()-starttime;
#endif
      return;
     }

    for(i=0;i<T->nCharts;i++)
     {
      if(chart>-1 && T->chart[i]!=chart)
       {
/* !!! */
        notin=0;
        for(j=0;j<k;j++)
         {
          t=center[j];
          if(t+R<(T->chartCenter[i])[j]-T->chartR[i] || t-R>(T->chartCenter[i])[j]+T->chartR[i])notin=1;
         }
        if(!notin)
/* !!! */
         {
          (*list)[*n]=T->chart[i];
          (*n)++;
         }
       }
     }
   }else{
    MFBinaryTreeGetIntersectingCharts(T->leftT,k,chart,center,R,n,list);
    MFBinaryTreeGetIntersectingCharts(T->rightT,k,chart,center,R,n,list);
   }

#ifdef MFTIMINGS
    MFTimeMFBinaryTreeGetIntersectingCharts+=clock()-starttime;
#endif
  return;
 }

MFListOfCharts MFCreateListOfIntersectingCharts(MFBinaryTree TRoot,int chart,double *center, double R)
 {
  static char RoutineName[]={"MFCreateListOfIntersectingCharts"};
  int i,n,m;
  int *charts;
  MFListOfCharts result;
  int report;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFCreateListOfIntersectingCharts++;
  starttime=clock();
#endif

#ifdef MFDEBUG
 _heap_check();
#endif

  report=0;

#ifdef BRUTEFORCE
  n=TRoot->nCharts-1;
  if(n>0)
   {
    charts=(int*)malloc(n*sizeof(int)); /*done*/
    if(charts==(int*)NULL)
     {
      sprintf(MFBinaryTreeErrorMsg,"Out of memory, trying to allocate %d bytes",n*sizeof(int));
      printf("%s -- %s\n",RoutineName,MFBinaryTreeErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFBinaryTreeErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateListOfIntersectingCharts+=clock()-starttime;
#endif
      return (MFListOfCharts)NULL;
     }
   }else{
    charts=(int*)NULL;
   }

  m=0;
  for(i=0;i<n;i++)
   if(i!=chart){charts[m]=i;m++;}

  result=MFCreateListOfCharts(m,charts);
#else

/* Start at top of tree. */
/*   if n!=0, test each member */
/*     else test bounding box of each subtree */

  n=0;
  charts=(int*)malloc(10*sizeof(int)); /*done*/
  if(charts==(int*)NULL)
   {
    sprintf(MFBinaryTreeErrorMsg,"Out of memory, trying to allocate %d bytes",10*sizeof(int));
    printf("%s -- %s\n",RoutineName,MFBinaryTreeErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFBinaryTreeErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateListOfIntersectingCharts+=clock()-starttime;
#endif
    return (MFListOfCharts)NULL;
   }

  MFBinaryTreeGetIntersectingCharts(TRoot->root,TRoot->k,chart,center,R,&n,&charts);
  result=MFCreateListOfCharts(n,charts);

#endif

  if(report)
   {
    printf("There are %d intersecting charts\n [",n);
    for(i=0;i<n;i++)
     {
      if(i>0)printf(",");
      if(i%30==29)printf("\n  ");
      printf("%d",charts[i]);
     }
    printf("]\n");
    fflush(stdout);
   }

#ifdef MFTIMINGS
    MFTimeMFCreateListOfIntersectingCharts+=clock()-starttime;
#endif
  return(result);
 }

void MFPrintBinaryTreeLeaf(FILE *fid, int k, MFBinaryTreeLeaf T)
 {
  int i,j;

#ifdef MFDEBUG
 _heap_check();
#endif

  if(T==(MFBinaryTreeLeaf)NULL)
   {
    fprintf(fid,"     NULL\n");
    return;
   }
  fprintf(fid,"     sequenceNumber %d\n",T->sequenceNumber);
  fprintf(fid,"     Split is direction %d, threshold %lf\n",T->d,T->split);
  fprintf(fid,"     number of charts %d\n",T->nCharts);fflush(stdout);
  for(i=0;i<T->nCharts;i++)
   {
    fprintf(fid,"chart %d",T->chart[i]);
    fprintf(fid," Radius %lf, center  (",T->chartR[i]);
    for(j=0;j<k;j++){if(j>0)fprintf(fid,",");fprintf(fid,"%lf",(T->chartCenter[i])[j]);}
    fprintf(fid,") value %lf\n",T->value[i]);
   }
  fprintf(fid,"     Bounding box (");
  for(i=0;i<k;i++)
   {
    if(i>0)fprintf(fid,",");
    fprintf(fid,"%lf",T->leftB[i]);
   }
  fprintf(fid,")->(");
  for(i=0;i<k;i++)
   {
    if(i>0)fprintf(fid,",");
    fprintf(fid,"%lf",T->rightB[i]);
   }
  fprintf(fid,")\n");
  fprintf(fid,"LeftLeaf\n");
  MFPrintBinaryTreeLeaf(fid,k,T->leftT);
  fprintf(fid,"RightLeaf\n");
  MFPrintBinaryTreeLeaf(fid,k,T->rightT);

  return;
 }

void MFPrintBinaryTree(FILE *fid, MFBinaryTree T)
 {
#ifdef MFDEBUG
 _heap_check();
#endif

  fprintf(fid,"Binary Tree, dimension %d\n",T->k);
  fprintf(fid,"Root\n");
  MFPrintBinaryTreeLeaf(fid,T->k,T->root);

  return;
 }

void MFWriteBinaryTree(FILE *fid,MFBinaryTree u)
 {
  static char RoutineName[]={"MFWriteBinaryTree"};
  int i;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFWriteBinaryTree++;
  starttime=clock();
#endif

#ifdef MFDEBUG
 _heap_check();
#endif

  fprintf(fid,"%s\n","BinaryTree");
#ifdef MFTIMINGS
    MFTimeMFWriteBinaryTree+=clock()-starttime;
#endif
  return;
 }

MFBinaryTree MFReadBinaryTree(FILE *fid)
 {
  static char RoutineName[]={"MFReadBinaryTree"};
  int i;
  MFBinaryTree B;
  char tag[100]="";

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFReadBinaryTree++;
  starttime=clock();
#endif

#ifdef MFDEBUG
 _heap_check();
#endif

  fscanf(fid,"%s\n",tag);
  if(strcmp(tag,"BinaryTree"))
   {
    sprintf(MFBinaryTreeErrorMsg,"Next Object is not a BinaryTree! (%s)\n",RoutineName,tag);
    MFSetError(12,RoutineName,MFBinaryTreeErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFReadBinaryTree+=clock()-starttime;
#endif
    return;
   }
  B=(MFBinaryTree)NULL;
#ifdef MFTIMINGS
    MFTimeMFReadBinaryTree+=clock()-starttime;
#endif
  return B;
 }

void MFRecomputeBoundingBoxes(MFBinaryTree T, int chart, double *center, double R)
 {
  MFBinaryTreeLeaf leaf;
  int ic;
  int i,j;

#ifdef MFDEBUG
 _heap_check();
#endif

  leaf=T->root;
  while(leaf->nCharts==0)
   {
    if ( center[leaf->d]<leaf->split)
      leaf=leaf->leftT;
     else
      leaf=leaf->rightT;
   }

  ic=-1;
  for(i=0;i<leaf->nCharts;i++)
   {
    if(leaf->chart[i]==chart)ic=i;
   }
  leaf->chartR[ic]=R;

  for(i=0;i<T->k;i++)
   {
    leaf->leftB[i]=center[i]-leaf->chartR[0];
    leaf->rightB[i]=center[i]+leaf->chartR[0];;
    for(j=1;j<leaf->nCharts;j++)
     {
      if(center[i]-leaf->chartR[j]<leaf->leftB[i])leaf->leftB[i]=center[i]-leaf->chartR[j];
      if(center[i]+leaf->chartR[j]>leaf->rightB[i])leaf->rightB[i]=center[i]+leaf->chartR[j];
     }
   }
  leaf=leaf->parent;

  while(leaf!=(MFBinaryTreeLeaf)NULL)
   {
    for(i=0;i<T->k;i++)
     {
      if(leaf->leftB!=(double*)NULL)
       {
        leaf->leftB[i]=(leaf->leftT)->leftB[i];
        if(leaf->rightB!=(double*)NULL&&leaf->leftB[i]>(leaf->leftT)->leftB[i])leaf->leftB[i]=(leaf->rightT)->leftB[i];
       }else if(leaf->rightB!=(double*)NULL)
        leaf->leftB[i]=(leaf->rightT)->leftB[i];

      if(leaf->rightB!=(double*)NULL)
       {
        leaf->rightB[i]=(leaf->rightT)->rightB[i];
        if(leaf->rightB!=(double*)NULL&&leaf->rightB[i]>(leaf->rightT)->rightB[i])leaf->rightB[i]=(leaf->rightT)->rightB[i];
       }else if(leaf->rightB!=(double*)NULL)
        leaf->rightB[i]=(leaf->rightT)->rightB[i];
     }
    leaf=leaf->parent;
   }

 }

MFListOfCharts MFCreateListOfNearbyCharts(MFBinaryTree TRoot,double *center,double R)
 {
  static char RoutineName[]={"MFCreateListOfNearbyCharts"};
  int i,n,m;
  int *charts;
  MFListOfCharts result;
  int report;

#ifdef MFDEBUG
 _heap_check();
#endif

  report=0;

#ifdef BRUTEFORCE
  n=TRoot->nCharts-1;
  if(n>0)
   {
    charts=(int*)malloc(n*sizeof(int)); /*done*/
    if(charts==(int*)NULL)
     {
      sprintf(MFBinaryTreeErrorMsg,"Out of memory, trying to allocate %d bytes",n*sizeof(int));
      printf("%s -- %s\n",RoutineName,MFBinaryTreeErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFBinaryTreeErrorMsg,__LINE__,__FILE__);
      return (MFListOfCharts)NULL;
     }
   }else{
    charts=(int*)NULL;
   }

  m=0;
  for(i=0;i<n;i++)
   if(i!=chart){charts[m]=i;m++;}

  result=MFCreateListOfCharts(m,charts);
#else

/* Start at top of tree. */
/*   if n!=0, test each member */
/*     else test bounding box of each subtree */

  n=0;
  charts=(int*)malloc(10*sizeof(int)); /*done*/
  if(charts==(int*)NULL)
   {
    sprintf(MFBinaryTreeErrorMsg,"Out of memory, trying to allocate %d bytes",10*sizeof(int));
    printf("%s -- %s\n",RoutineName,MFBinaryTreeErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFBinaryTreeErrorMsg,__LINE__,__FILE__);
    return (MFListOfCharts)NULL;
   }

  MFBinaryTreeGetNeighboringCharts(TRoot->root,TRoot->k,center,R,&n,&charts);
  result=MFCreateListOfCharts(n,charts);

#endif

  if(report)
   {
    printf("There are %d intersecting charts\n [",n);
    for(i=0;i<n;i++)
     {
      if(i>0)printf(",");
      if(i%30==29)printf("\n  ");
      printf("%d",charts[i]);
     }
    printf("]\n");
    fflush(stdout);
   }

  return(result);
 }

void MFBinaryTreeGetNeighboringCharts(MFBinaryTreeLeaf T,int k, double *center, double R,int *n,int **list)
 {
  static char RoutineName[]={"MFBinaryTreeGetIntersectingCharts"};
  int i,j;
  double t;
  int notin;

  int verbose;

#ifdef MFDEBUG
 _heap_check();
#endif

  verbose=0;

  if(T==(MFBinaryTreeLeaf)NULL)return;
  if(R<0)abort();
  if(k<0)abort();

  if(T->leftT==(MFBinaryTreeLeaf)NULL && T->nCharts==0)return;

/* Check Bounding box */

  notin=0;
  if(verbose)printf("   leaf %d bb=\n",T->sequenceNumber);
  if(T->leftB==(double*)NULL)abort();
  if(T->rightB==(double*)NULL)abort();
  for(i=0;i<k;i++)
   {
    t=center[i];
    if(verbose)printf("   %d BB %lf,%lf Disk %lf,%lf\n",i,T->leftB[i],T->rightB[i],t-R,t+R);
    if(t+R<T->leftB[i] || t-R>T->rightB[i])notin=1;
   }
  if(verbose&&notin)printf("     not in\n");
  if(verbose&&!notin)printf("     in\n");
  if(notin)
   {
#ifdef MFTIMINGS
    MFTimeMFBinaryTreeGetIntersectingCharts+=clock()-starttime;
#endif
    return;
   }

/* Chart Intersects Bounding box, add charts to list, or check sub-trees */

  if(T->leftT==(MFBinaryTreeLeaf)NULL)
   {
    *list=(int*)realloc((void*)(*list),(*n+T->nCharts)*sizeof(int)); /*done*/
    if(*list==(int*)NULL)
     {
      sprintf(MFBinaryTreeErrorMsg,"Out of memory, trying to reallocate %d bytes",(*n+T->nCharts)*sizeof(int));
      printf("%s -- %s\n",RoutineName,MFBinaryTreeErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFBinaryTreeErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFBinaryTreeGetIntersectingCharts+=clock()-starttime;
#endif
      return;
     }

    for(i=0;i<T->nCharts;i++)
     {
      notin=0;
      for(j=0;j<k;j++)
       {
        t=center[j];
        if(t+R<(T->chartCenter[i])[j]-T->chartR[i] || t-R>(T->chartCenter[i])[j]+T->chartR[i])notin=1;
       }
      if(!notin)
       {
        (*list)[*n]=T->chart[i];
        (*n)++;
       }
     }
   }else{
    MFBinaryTreeGetNeighboringCharts(T->leftT,k,center,R,n,list);
    MFBinaryTreeGetNeighboringCharts(T->rightT,k,center,R,n,list);
   }

  return;
 }
