/* 
    %W%
    %D% %T%
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */

#include <MFAtlas.h>
#include <MFChart.h>
#include <MFKVector.h>
#include <MFNVector.h>
#include <MFNKMatrix.h>
#include <MFImplicitMF.h>
#include <MFPrint.h>
#include <stdlib.h>
#include <string.h>
#include <MFTime.h>

#define REALLYTHROWITAWAY

void MFSetError(int,char*,char*,int,char*);
static char MFChartErrorMsg[256]="";
double *MFNV_CStar(MFNVector);
int MFPolytopeTotallyInterior(MFPolytope,MFKVector,double);
int *MFPolytopeVertexIndexSet(MFPolytope,int);
int MFPolytopeVertexNumberOfIndices(MFPolytope,int);

struct MFChartSt {
                  int k;
                  int n;
                  MFImplicitMF M;

                  MFPolytope P;
                  int        allAlphasPositive;
                  int        singular;
                  double     R;
                  double     SuggestR;
                  MFNVector  u;
                  MFNKMatrix Phi;
                  int        posInBList;

                  FILE       *fid;
                  int        paged;
                  int        nearBoundary;
                  int        changed;
                  long       indexInPageFile;

                  int        refNumber;
                  int        nRefs;
                 };

#define DENSE 0
#define AUTO 1
#define LOCA 2


MFChart MFCreateChart(MFImplicitMF M,MFNVector u,MFNKMatrix TS, double R)
 {
  static char RoutineName[]={"MFCreateChart"};
  MFChart this;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFCreateChart++;
  starttime=clock();
#endif

  if(M==(MFImplicitMF)NULL)
   {
    sprintf(MFChartErrorMsg,"M (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(8,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateChart+=clock()-starttime;
#endif
    return;
   }

  if(MFIMF_K(M)<1)
   {
    sprintf(MFChartErrorMsg,"Manifold has illegal dimension %d. Must be positive",MFIMF_K(M));
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateChart+=clock()-starttime;
#endif
    return;
   }

  if(strcmp(MFNVGetId(u),"LOCA") && MFIMF_N(M)<MFIMF_K(M))
   {
    sprintf(MFChartErrorMsg,"Manifold has illegal embedding dimension (%d). Must be greater or equal to it\'s dimension %d.",MFIMF_N(M),MFIMF_K(M));
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateChart+=clock()-starttime;
#endif
    return;
   }

  if(u==(MFNVector)NULL)
   {
    sprintf(MFChartErrorMsg,"u (argument 2) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(8,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateChart+=clock()-starttime;
#endif
    return;
   }

  this=(MFChart)malloc(sizeof(struct MFChartSt)); /*done*/
  if(this==(MFChart)NULL)
   {
    sprintf(MFChartErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFChartSt));
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateChart+=clock()-starttime;
#endif
    return((MFChart)NULL);
   }

  this->M=M;
  MFRefImplicitMF(M);
  this->u=u;
  MFRefNVector(u);
  this->n=MFIMF_N(M);
  this->k=MFIMF_K(M);

  this->Phi=TS;
  MFRefNKMatrix(this->Phi);
  this->R=R;
  this->SuggestR=R;
  this->refNumber=0;

#ifndef MFSIMPLEX
  this->P=MFCreateHyperCubeAtOrigin(this->k,1.05*this->R);
#else
  this->P=MFCreateSimplexAtOrigin(this->k,1.05*this->R);
#endif
  if(TS==(MFNKMatrix)NULL)
   {
    sprintf(MFChartErrorMsg,"TangentSpace (argument 3) is NULL ");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateChart+=clock()-starttime;
#endif
    return((MFChart)NULL);
   }

  this->allAlphasPositive=1;
  this->singular=0;
  this->fid=(FILE*)NULL;
  this->paged=0;
  this->nearBoundary=0;
  this->changed=1;
  this->indexInPageFile=-1;
  this->posInBList=-1;
  this->nRefs=1;

#ifdef MFTIMINGS
    MFTimeMFCreateChart+=clock()-starttime;
#endif
  return(this);
 }

void MFSubtractHalfSpaceFromChart(MFChart chart,int i,MFKVector n,double o)
 {
  static char RoutineName[]={"MFSubtractHalfSpaceFromChart"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFSubtractHalfSpaceFromChart++;
  starttime=clock();
#endif

  if(chart==(MFChart)NULL)
   {
    sprintf(MFChartErrorMsg,"Pointer to Chart (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFSubtractHalfSpaceFromChart+=clock()-starttime;
#endif
    return;
   }

  MFSubtractHalfSpaceFromPolytope(chart->P,i,n,o);
  if(o<0.)chart->allAlphasPositive=0;
#ifdef MFTIMINGS
    MFTimeMFSubtractHalfSpaceFromChart+=clock()-starttime;
#endif
  chart->changed=1;
  return;
 }

void MFRefChart(MFChart chart)
 {
  static char RoutineName[]={"MFRefChart"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFRefChart++;
  starttime=clock();
#endif

  if(chart==(MFChart)NULL)
   {
    sprintf(MFChartErrorMsg,"Pointer to Chart (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFRefChart+=clock()-starttime;
#endif
    return;
   }

  chart->nRefs++;
#ifdef MFTIMINGS
    MFTimeMFRefChart+=clock()-starttime;
#endif
  return;
 }

void MFFreeChart(MFChart chart)
 {
  static char RoutineName[]={"MFFreeChart"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFFreeChart++;
  starttime=clock();
#endif

  if(chart==(MFChart)NULL)
   {
    sprintf(MFChartErrorMsg,"Pointer to Chart (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFFreeChart+=clock()-starttime;
#endif
    return;
   }

  chart->nRefs--;

  if(chart->nRefs>0)
   {
#ifdef MFTIMINGS
    MFTimeMFFreeChart+=clock()-starttime;
#endif
    return;
   }
  if(chart->M!=(MFImplicitMF)NULL){MFFreeImplicitMF(chart->M);chart->M=(MFImplicitMF)NULL;};
  if(chart->u!=(MFNVector)NULL){MFFreeNVector(chart->u);chart->u=(MFNVector)NULL;};
  if(chart->P!=(MFPolytope)NULL)
   {
    MFFreePolytope(chart->P);
    chart->P=(MFPolytope)NULL;
   }
  if(chart->Phi!=(MFNKMatrix)NULL){MFFreeNKMatrix(chart->Phi);chart->Phi=(MFNKMatrix)NULL;};
  free(chart);

#ifdef MFTIMINGS
    MFTimeMFFreeChart+=clock()-starttime;
#endif
  return;
 }

MFPolytope MFChartPolytope(MFChart chart)
 {
  static char RoutineName[]={"MFChartPolytope"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFChartPolytope++;
  starttime=clock();
#endif

  if(chart==(MFChart)NULL)
   {
    sprintf(MFChartErrorMsg,"Pointer to Chart (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFChartPolytope+=clock()-starttime;
#endif
    return;
   }

#ifdef MFTIMINGS
    MFTimeMFChartPolytope+=clock()-starttime;
#endif
  return(chart->P);
 }

void MFSetChartPolytope(MFChart chart, MFPolytope P)
 {
  static char RoutineName[]={"MFSetChartPolytope"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFSetChartPolytope++;
  starttime=clock();
#endif

  if(chart==(MFChart)NULL)
   {
    sprintf(MFChartErrorMsg,"Pointer to Chart (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFSetChartPolytope+=clock()-starttime;
#endif
    return;
   }

  chart->P=P;

#ifdef MFTIMINGS
    MFTimeMFSetChartPolytope+=clock()-starttime;
#endif
  return;
 }

void MFSetChartRadius(MFChart chart, double R)
 {
  static char RoutineName[]={"MFSetChartRadius"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFSetChartRadius++;
  starttime=clock();
#endif

  if(chart==(MFChart)NULL)
   {
    sprintf(MFChartErrorMsg,"Pointer to Chart (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFSetChartRadius+=clock()-starttime;
#endif
    return;
   }

  chart->R=R;

#ifdef MFTIMINGS
    MFTimeMFSetChartRadius+=clock()-starttime;
#endif
  return;
 }

MFNVector MFChartCenter(MFChart chart)
 {
  static char RoutineName[]={"MFChartCenter"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFChartCenter++;
  starttime=clock();
#endif

  if(chart==(MFChart)NULL)
   {
    sprintf(MFChartErrorMsg,"Pointer to Chart (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFChartCenter+=clock()-starttime;
#endif
    return;
   }

  if(chart->paged)MFChartPageIn(chart,chart->fid);

#ifdef MFTIMINGS
    MFTimeMFChartCenter+=clock()-starttime;
#endif
  return(chart->u);
 }

MFNKMatrix MFChartTangentSpace(MFChart chart)
 {
  static char RoutineName[]={"MFChartTangentSpace"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFChartTangentSpace++;
  starttime=clock();
#endif

  if(chart==(MFChart)NULL)
   {
    sprintf(MFChartErrorMsg,"Pointer to Chart (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFChartTangentSpace+=clock()-starttime;
#endif
    return;
   }

  if(chart->paged)MFChartPageIn(chart,chart->fid);

#ifdef MFTIMINGS
    MFTimeMFChartTangentSpace+=clock()-starttime;
#endif
  return(chart->Phi);
 }

double MFChartRadius(MFChart chart)
 {
  static char RoutineName[]={"MFChartRadius"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFChartRadius++;
  starttime=clock();
#endif

  if(chart==(MFChart)NULL)
   {
    sprintf(MFChartErrorMsg,"Pointer to Chart (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFChartRadius+=clock()-starttime;
#endif
    return;
   }

#ifdef MFTIMINGS
    MFTimeMFChartRadius+=clock()-starttime;
#endif
  return(chart->R);
 }

int MFChartEvaluate(MFChart chart,MFKVector s,MFNVector u)
 {
  static char RoutineName[]={"MFChartEvaluate"};
  MFNVector u0;
  int rc;
  int verbose=0;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFChartEvaluate++;
  starttime=clock();
#endif

  if(chart==(MFChart)NULL)
   {
    sprintf(MFChartErrorMsg,"Pointer to Chart (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFChartEvaluate+=clock()-starttime;
#endif
    return 0;
   }

  if(chart->paged)MFChartPageIn(chart,chart->fid);

/*u0=MFCloneNVector(u);
  if(verbose){fprintf(stderr,"0x%8.8x=u0\n",u0);fprintf(stderr,"0x%8.8x=u0(0x%8.8x).data\n",MFNV_CStar(u0),u0);fflush(stderr);}
  if(chart->paged)MFChartPageIn(chart,chart->fid);
  MFChartPointInTangentSpace(chart,s,u0);
  rc=MFIMFProject(chart->M,u0,chart->Phi,u);
  MFFreeNVector(u0);*/

  rc=MFIMFProjectFromCenter(chart->M,chart->u,chart->Phi,s,u);
#ifdef MFTIMINGS
    MFTimeMFChartEvaluate+=clock()-starttime;
#endif
  return (rc);
 }

int MFChartInterior(MFChart chart,MFKVector s)
 {
  static char RoutineName[]={"MFChartInterior"};
  int result;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFChartInterior++;
  starttime=clock();
#endif

  if(chart==(MFChart)NULL)
   {
    sprintf(MFChartErrorMsg,"Pointer to Chart (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFChartInterior+=clock()-starttime;
#endif
    return;
   }

  if(MFKVNorm(s)>chart->R)
   { 
#ifdef MFTIMINGS
    MFTimeMFChartInterior+=clock()-starttime;
#endif
    return 0;
   }else{
    result=MFPolytopeInterior(chart->P,s);
#ifdef MFTIMINGS
    MFTimeMFChartInterior+=clock()-starttime;
#endif
    return result;
   }
 }

void MFChartProjectIntoTangentSpace(MFChart chart,MFNVector u,MFKVector s)
 {
  static char RoutineName[]={"MFChartProjectIntoTangentSpace"};
  MFNVector v;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFChartProjectIntoTangentSpace++;
  starttime=clock();
#endif

  if(chart==(MFChart)NULL)
   {
    sprintf(MFChartErrorMsg,"Chart (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFChartProjectIntoTangentSpace+=clock()-starttime;
#endif
    return;
   }

  if(u==(MFNVector)NULL)
   {
    sprintf(MFChartErrorMsg,"u (argument 2) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(8,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFChartProjectIntoTangentSpace+=clock()-starttime;
#endif
    return;
   }

  if(s==(MFKVector)NULL)
   {
    sprintf(MFChartErrorMsg,"s (argument 3) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(8,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFChartProjectIntoTangentSpace+=clock()-starttime;
#endif
    return;
   }

  if(chart->n==0)
   {
    sprintf(MFChartErrorMsg,"Chart has zero dimension");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(8,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFChartProjectIntoTangentSpace+=clock()-starttime;
#endif
    return;
   }

  if(chart->paged)MFChartPageIn(chart,chart->fid);

  if(chart->u==(MFNVector)NULL)
   {
    sprintf(MFChartErrorMsg,"Chart center is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(8,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFChartProjectIntoTangentSpace+=clock()-starttime;
#endif
    return;
   }

  v=MFCloneNVector(u);

  MFNSpaceDirection(MFIMFNSpace(chart->M),chart->u,u,v);
  MFMVMulT(MFIMFNSpace(chart->M),chart->Phi,v,s);
  MFFreeNVector(v);

#ifdef MFTIMINGS
    MFTimeMFChartProjectIntoTangentSpace+=clock()-starttime;
#endif
  return;
 }

void MFChartProjectVectorIntoTangentSpace(MFChart chart,MFNVector u,MFKVector s)
 {
  static char RoutineName[]={"MFChartProjectVectorIntoTangentSpace"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFChartProjectVectorIntoTangentSpace++;
  starttime=clock();
#endif

  if(chart==(MFChart)NULL)
   {
    sprintf(MFChartErrorMsg,"Chart (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFChartProjectVectorIntoTangentSpace+=clock()-starttime;
#endif
    return;
   }

  if(u==(MFNVector)NULL)
   {
    sprintf(MFChartErrorMsg,"u (argument 2) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(8,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFChartProjectVectorIntoTangentSpace+=clock()-starttime;
#endif
    return;
   }

  if(s==(MFKVector)NULL)
   {
    sprintf(MFChartErrorMsg,"s (argument 3) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(8,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFChartProjectVectorIntoTangentSpace+=clock()-starttime;
#endif
    return;
   }

  if(chart->n==0)
   {
    sprintf(MFChartErrorMsg,"Chart has zero dimension");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(8,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFChartProjectVectorIntoTangentSpace+=clock()-starttime;
#endif
    return;
   }

  if(chart->paged)MFChartPageIn(chart,chart->fid);

  if(chart->u==(MFNVector)NULL)
   {
    sprintf(MFChartErrorMsg,"Chart center is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(8,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFChartProjectVectorIntoTangentSpace+=clock()-starttime;
#endif
    return;
   }

  MFMVMulT(MFIMFNSpace(chart->M),chart->Phi,u,s);

#ifdef MFTIMINGS
    MFTimeMFChartProjectVectorIntoTangentSpace+=clock()-starttime;
#endif
  return;
 }

void MFChartPointInTangentSpace(MFChart chart,MFKVector s,MFNVector u)
 {
  static char RoutineName[]={"MFChartPointInTangentSpace"};
  MFNVector v;
  int verbose;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFChartPointInTangentSpace++;
  starttime=clock();
#endif

  verbose=0;
  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

  if(chart==(MFChart)NULL)
   {
    sprintf(MFChartErrorMsg,"Pointer to Chart (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);
    MFSetError(4,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFChartPointInTangentSpace+=clock()-starttime;
#endif
    return;
   }

  if(chart->paged)MFChartPageIn(chart,chart->fid);

  v=MFCloneNVector(u);

  MFMVMul(MFIMFNSpace(chart->M),chart->Phi,s,v);
  MFNSpaceAdd(MFIMFNSpace(chart->M),v,chart->u,u);
  if(verbose){MFPrintNVector(stdout,chart->u);printf("+");MFPrintNKMatrix(stdout,chart->Phi);MFPrintKVector(stdout,s);printf("=");MFPrintNVector(stdout,u);printf("\n");fflush(stdout);}

  MFFreeNVector(v);

#ifdef MFTIMINGS
    MFTimeMFChartPointInTangentSpace+=clock()-starttime;
#endif
  return;
 }

int MFChartHasBoundary(MFChart chart)
 {
  static char RoutineName[]={"MFChartHasBoundary"};
  int vertex,nVertices;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFChartHasBoundary++;
  starttime=clock();
#endif

  if(chart==(MFChart)NULL)
   {
    sprintf(MFChartErrorMsg,"Pointer to Chart (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);
    MFSetError(4,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFChartHasBoundary+=clock()-starttime;
#endif
    return;
   }

  nVertices=MFPolytopeNumberOfVertices(chart->P);
  for(vertex=0;vertex<nVertices;vertex++)
    if(MFPolytopeRadiusOfVertex(chart->P,vertex)>chart->R)
     {
#ifdef MFTIMINGS
    MFTimeMFChartHasBoundary+=clock()-starttime;
#endif
      return 1;
     }
#ifdef MFTIMINGS
    MFTimeMFChartHasBoundary+=clock()-starttime;
#endif
  return 0;
 }

int MFChartK(MFChart chart)
 {
  static char RoutineName[]={"MFChartK"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFChartK++;
  starttime=clock();
#endif

  if(chart==(MFChart)NULL)
   {
    sprintf(MFChartErrorMsg,"Pointer to Chart (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFChartK+=clock()-starttime;
#endif
    return;
   }

#ifdef MFTIMINGS
    MFTimeMFChartK+=clock()-starttime;
#endif
  return(chart->k);
 }

int MFChartN(MFChart chart)
 {
  static char RoutineName[]={"MFChartN"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFChartN++;
  starttime=clock();
#endif

  if(chart==(MFChart)NULL)
   {
    sprintf(MFChartErrorMsg,"Pointer to Chart (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFChartN+=clock()-starttime;
#endif
    return;
   }

#ifdef MFTIMINGS
    MFTimeMFChartN+=clock()-starttime;
#endif
  return(chart->n);
 }

void MFWriteChart(FILE *fid, MFChart chart)
 {
  static char RoutineName[]={"MFWriteChart"};
  int i;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFWriteChart++;
  starttime=clock();
#endif

  if(chart->paged)
   {
/*  MFChartPageIn(chart,chart->fid);*/
    fprintf(fid,"%s\n","Chart");
    fprintf(fid,"%d %d %d %d %d %lf %d %d\n",chart->n,chart->k,chart->allAlphasPositive,chart->paged,chart->nearBoundary,chart->R,chart->singular,chart->nRefs,chart->refNumber);
    MFWritePolytope(fid,chart->P);
   }else{
    fprintf(fid,"%s\n","Chart");
    fprintf(fid,"%d %d %d %d %d %lf %d %d\n",chart->n,chart->k,chart->allAlphasPositive,chart->paged,chart->nearBoundary,chart->R,chart->singular,chart->nRefs,chart->refNumber);
    MFWritePolytope(fid,chart->P);
    MFWriteNVector(fid,chart->u);
    MFWriteNKMatrix(fid,chart->Phi);
   }

#ifdef MFTIMINGS
    MFTimeMFWriteChart+=clock()-starttime;
#endif
  return;
 }

MFChart MFReadChart(FILE *fid,MFAtlas A)
 {
  static char RoutineName[]={"MFReadChart"};
  MFChart chart;
  char tag[100]="";

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFReadChart++;
  starttime=clock();
#endif

  fscanf(fid,"%s\n",tag);
  if(strcmp(tag,"Chart"))
   {
    sprintf(MFChartErrorMsg,"Next Object is not a Chart! (%s)\n",RoutineName,tag);
    MFSetError(12,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFReadChart+=clock()-starttime;
#endif
    return;
   }

  chart=(MFChart)malloc(sizeof(struct MFChartSt)); /*done*/
  if(chart==(MFChart)NULL)
   {
    sprintf(MFChartErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFChartSt));
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFReadChart+=clock()-starttime;
#endif
    return (MFChart)NULL;
   }
  fscanf(fid,"%d %d %d %d %d %lf %d %d\n",&(chart->n),&(chart->k),&(chart->allAlphasPositive),&(chart->paged),&(chart->nearBoundary),&(chart->R),&(chart->singular),&(chart->nRefs),&(chart->refNumber));

  chart->M=MFAtlasMF(A);
  chart->P=MFReadPolytope(fid);
  if(chart->paged)
   {
    chart->u=(MFNVector)NULL;
    chart->Phi=(MFNKMatrix)NULL;
   }else{
    chart->u=MFReadNVector(fid);
    chart->Phi=MFReadNKMatrix(fid);
   }
  chart->posInBList=-1;

#ifdef MFTIMINGS
    MFTimeMFReadChart+=clock()-starttime;
#endif
  return chart;
 }

int MFChartTotallyInterior(MFChart chart,MFKVector s,double eps)
 {
  static char RoutineName[]={"MFChartTotallyInterior"};
  int result;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFChartTotallyInterior++;
  starttime=clock();
#endif

  if(chart==(MFChart)NULL)
   {
    sprintf(MFChartErrorMsg,"Pointer to Chart (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFChartTotallyInterior+=clock()-starttime;
#endif
    return;
   }

  if(MFKVNorm(s)>chart->R)
   { 
#ifdef MFTIMINGS
    MFTimeMFChartTotallyInterior+=clock()-starttime;
#endif
    return 0;
   }else{
    result=MFPolytopeTotallyInterior(chart->P,s,eps);
#ifdef MFTIMINGS
    MFTimeMFChartTotallyInterior+=clock()-starttime;
#endif
    return result;
   }
 }

int MFChartGetPositionInBoundaryList(MFChart this)
 {
  return this->posInBList;
 }

void MFChartSetPositionInBoundaryList(MFChart this, int pos)
 {
  this->posInBList=pos;

  return;
 }

MFImplicitMF MFChartGetManifold(MFChart this)
 {
  return this->M;
 }

int MFChartIsSingular(MFChart this)
 {
  return this->singular>0;
 }

void MFChartSetSingular(MFChart this)
 {
  this->singular=1;
  return;
 }

void MFChartSetNonSingular(MFChart this)
 {
  this->singular=0;
  return;
 }

int MFChartReferenceNumber(MFChart this)
 {
  return this->refNumber;
 }

void MFChartSetReferenceNumber(MFChart this, int n)
 {
  this->refNumber=n;
  return;
 }

double *MFNV_CStar(MFNVector);
double *MFNKM_CStar(MFNKMatrix);

int MFChartPageOut(MFChart this, FILE *fid, int index)
 {
  static char RoutineName[]={"MFChartPageOut"};
  static int i,j;
  double *c;
  int verbose=0;
  int rc;

  if(this==(MFChart)NULL)
   {
    sprintf(MFChartErrorMsg,"Chart is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(0,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateChart+=clock()-starttime;
#endif
    return 0;
   }

/* Returns 1 if written, 0 if already paged out */

  if(this->paged)
   {
    if(verbose){printf("Already paged out\n");fflush(stdout);}
    return 0;
   }

#ifdef REALLYPAGE
  if(verbose){fprintf(stderr,"PAGING out chart, fseek to end of file 0x%8.8x\n",fid);fflush(stderr);}
  rc=fseek(fid,(long)0,SEEK_END);
  if(verbose){fprintf(stderr,"  return code from fseek is %d\n",rc);fprintf(stderr,"ftell= %d\n",ftell(fid));fflush(stderr);}
  if(verbose){fprintf(stderr,"  write to position %d\n",index);fprintf(stderr,"ftell= %d\n",ftell(fid));fflush(stderr);}
  c=MFNV_CStar(this->u);
  if(verbose){fprintf(stderr,"PAGING   u starts (%lf,%lf,%lf...\n",c[0],c[1],c[2]);fflush(stderr);}
  this->fid=fid;
  this->paged=1;
  this->indexInPageFile=ftell(fid);
  for(i=0;i<this->n;i++)fprintf(fid,"%lf ",c[i]);
  fprintf(fid,"\n");

  c=MFNKM_CStar(this->Phi);
  if(verbose){fprintf(stderr,"PAGING   Phi starts (%lf,%lf,%lf...\n",c[0],c[1],c[2]);fflush(stderr);}
  for(j=0;j<this->k;j++)
  for(i=0;i<this->n;i++)fprintf(fid,"%lf ",c[i+this->n*j]);
  fprintf(fid,"\n");
#endif

#ifdef REALLYTHROWITAWAY
  if(verbose){printf("Discarding contents of chart\n");fflush(stdout);}
  MFFreeNVector(this->u);this->u=(MFNVector)NULL;
  MFFreeNKMatrix(this->Phi);this->Phi=(MFNKMatrix)NULL;
  MFFreePolytope(this->P);this->P=(MFPolytope)NULL;
#endif

  this->paged=1;

  return 1;
 }

int MFChartPageIn(MFChart this, FILE *fid)
 {
  static int i,j;
  double *c;
  long pos;
  int verbose=0;
  static char RoutineName[]={"MFChartPageIn"};

  printf("%s!!!!!\n",RoutineName);fflush(stdout);abort();

/* Returns 1 if paged in, 0 if not paged out */

  if(!(this->paged))return 0;

  if(verbose){fprintf(stderr,"PAGING IN CHART! %d in file\n",this->indexInPageFile);fflush(stderr);}
  if(this->u!=(MFNVector)NULL && this->Phi!=(MFNKMatrix)NULL)
   {
    if(verbose){fprintf(stderr,"PAGING Already in\n");fflush(stderr);}
    return 1;
   }

  if(this->u!=(MFNVector)NULL)MFFreeNVector(this->u);
  this->u=MFCreateNVector(this->n);

  if(this->Phi!=(MFNKMatrix)NULL)MFFreeNKMatrix(this->Phi);
  this->Phi=MFCreateNKMatrixWithData(this->n,this->k,(double*)NULL);

  if(verbose){fprintf(stderr,"PAGING IN CHART! read from file\n");}
  fseek(fid,this->indexInPageFile,SEEK_SET);
  if(verbose){fprintf(stderr,"ftell= %d\n",ftell(fid));}

  c=MFNV_CStar(this->u);
  for(i=0;i<this->n;i++)fscanf(fid,"%lf ",&(c[i]));
  fscanf(fid,"\n");
  if(verbose){fprintf(stderr,"PAGING   u starts (%lf,%lf,%lf...\n",c[0],c[1],c[2]);fflush(stderr);}

  c=MFNKM_CStar(this->Phi);
  for(j=0;j<this->k;j++)
  for(i=0;i<this->n;i++)fscanf(fid,"%lf ",&(c[i+this->n*j]));
  fscanf(fid,"\n");
  if(verbose){fprintf(stderr,"PAGING   Phi starts (%lf,%lf,%lf...\n",c[0],c[1],c[2]);fflush(stderr);}

  fseek(fid,0,SEEK_END);

  return 1;
 }

void MFChartClean(MFChart this)
 {
  if(!(this->paged))return;

  if(this->u!=(MFNVector)NULL){MFFreeNVector(this->u);this->u=(MFNVector)NULL;}
  if(this->Phi!=(MFNKMatrix)NULL){MFFreeNKMatrix(this->Phi);this->Phi=(MFNKMatrix)NULL;}

  return;
 }

int MFChartPaged(MFChart this)
 {
  static char RoutineName[]={"MFChartPaged"};

  if(this==(MFChart)NULL)
   {
    sprintf(MFChartErrorMsg,"Chart is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(0,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateChart+=clock()-starttime;
#endif
    return 0;
   }

  return this->paged;
 }

int MFChartNearBoundary(MFChart this)
 {
  return this->nearBoundary;
 }

void MFChartSetNearBoundary(MFChart this,int nearBoundary)
 {
  this->nearBoundary=nearBoundary;
  return;
 }

int MFChartNumberOfFaces(MFChart this)
 {
  return MFPolytopeNumberOfFaces(this->P);
 }

void MFSetSuggestedChartRadius(MFChart chart, double R)
 {
  static char RoutineName[]={"MFSetSuggestedChartRadius"};

  if(chart==(MFChart)NULL)
   {
    sprintf(MFChartErrorMsg,"Pointer to Chart (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
    return;
   }

  chart->SuggestR=R;

  return;
 }

double MFChartSuggestedRadius(MFChart chart)
 {
  static char RoutineName[]={"MFChartSuggestedRadius"};

  if(chart==(MFChart)NULL)
   {
    sprintf(MFChartErrorMsg,"Pointer to Chart (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(4,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
    return;
   }

  return(chart->SuggestR);
 }

void MFChartResetChangedFlag(MFChart chart)
 {
  chart->changed=0;
  return;
 }

int MFChartChangedFlag(MFChart chart)
 {
  return chart->changed;
 }

void MFChartSetPaged(MFChart this)
 {
  static char RoutineName[]={"MFChartSetPaged"};

  if(this==(MFChart)NULL)
   {
    sprintf(MFChartErrorMsg,"Chart is NULL");
    printf("%s -- %s\n",RoutineName,MFChartErrorMsg);fflush(stdout);
    MFSetError(0,RoutineName,MFChartErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateChart+=clock()-starttime;
#endif
    return;
   }

  this->paged=1;
  return;
 }
