/* 
    %W%
    %D% %T%
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
#include <MFEnumPolytope.h>
#include <MFEnumDualPolytope.h>
#include <MFPolytope.h>
#include <MFAtlas.h>
#include <MFAtlasFriends.h>
#include <MFChart.h>
#include <MFKVector.h>
#include <MFPrint.h>
#include <MFPrintEnum.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <MFError.h>

/* Routines special for access to other data structs that aren't public */

MFChart MFAtlasChart(MFAtlas,int);
int MFPolytopeIntersectIndexSets(MFPolytope,int,int,int*);

struct MFCellSt {
             int nIndices;
             int *indices;
             int mFaceCells;
             int nFaceCells;
             int *faceCells;
             int nCellFaces;
             int mCellFaces;
             int *cellFaces;
            };

typedef struct MFCellSt *MFCell;

struct MFEnumPolytopeSt {
                  MFKVector *v;
                  int k;
                  int n;
                  int *nCells;
                  int *mCells;
                  MFCell **cells;
                 };

#define EPNumberOfCellsToAllocate 1
/* #define EPTraceFrees*/

int MFIntersectIndices(int,int*,int,int*,int*);
void MFCreateZeroCellV(MFCell*,int,...);
void MFCreateZeroCell(MFCell*,int,int*);
void MFCreateZeroCellWithEdgeList(MFCell*,int,int*);
void MFFreeCell(MFCell);
void MFCreateEmptyKComplex(int,int,MFEnumPolytope*,int);
int MFIndicesEqual(int,int*,int,int*);
void EPSortPolytopeIndices(int,int*);
void EPCollapseIndices(int*,int*);
int MFEnumPolytopeVerticesOnCell(MFEnumPolytope,int,int,int,int**);
int MFEnumPolytopeIsCellExterior(MFAtlas,int,MFEnumPolytope,int,int);

int MFTestEnumPolytope(MFEnumPolytope,int);

void MFSetError(int,char*,char*,int,char*);
static char MFEnumPolytopeErrorMsg[256]="";

MFEnumPolytope MFEnumeratePolytope(MFPolytope P)
 {
  static char RoutineName[]={"MFEnumeratePolytope"};
  int    i,j,n;
  MFEnumPolytope EP=(MFEnumPolytope)NULL;
  int verbose;
  int *indices;
  MFKVector s;
  int III;

  verbose=0;
  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

/* In 2d, boundary complex is a polygon, put edges head to tail, so interior is
                   to left */
/* In 3d, boundary complex is a polyhedron, put edges of each 2-cell head to
                   tail with outward normal */

  MFCreateEmptyKComplex(MFPolytopeDimension(P),MFPolytopeDimension(P),&EP,MFPolytopeNumberOfVertices(P));
  if(MFPolytopeNumberOfFaces(P)>0)
    indices=(int*)malloc(MFPolytopeNumberOfFaces(P)*sizeof(int)); /*done*/
   else
    indices=(int*)NULL;
  if(MFPolytopeNumberOfFaces(P)> 0 && indices==(int*)NULL)
   {
    sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",MFPolytopeNumberOfFaces(P)*sizeof(int));
    printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
    return (MFEnumPolytope)NULL;
   }

  for(III=0;III<MFPolytopeNumberOfFaces(P);III++)indices[III]=-1;

  for(i=0;i<MFPolytopeNumberOfVertices(P);i++)
   {
    n=MFPolytopeNumberOfVertexIndices(P,i);
    for(j=0;j<n;j++)indices[j]=MFPolytopeVertexIndex(P,i,j);
    MFCreateZeroCell(&((EP->cells[0])[i]),n,indices);

    s=MFCreateKVector(MFPolytopeDimension(P));
    MFPolytopeVertex(P,i,s);
    EP->v[i]=s;
   }

  if(verbose){printf("  call MFEnumerate\n");fflush(stdout);}
  MFEnumerate(EP);
  if(verbose){printf("  back from MFEnumerate\n");fflush(stdout);}
  if(indices!=(int*)NULL)free(indices);

  if(verbose){printf("  done %s\n",RoutineName);fflush(stdout);}
  return(EP);
 }

void MFEnumerate(MFEnumPolytope EP)
 {
  static char RoutineName[]={"MFEnumerate"};
  int    intersection[1000]={0.}; /* ??? */

  int    nv,ne,ni;
  int    iv,jv;
  int     n1, n2;
  int    *i1,*i2;
  int    d,k;
  int    mf;
  int    i,n;
  int    alreadyThere;
  int    j,f;
  int    verbose;
  MFCell tmp;
  int III;

  verbose=0;
  if(verbose){printf("\n\n\n\nMFEnumerate\n");fflush(stdout);}

/* Pairwise intersection of vertices */

  k=EP->k;

  for(d=1;d<k+1;d++)
   {
    if(verbose){printf("  intersecting %d cells to build %d cells\n",d-1,d);fflush(stdout);}
    nv=EP->nCells[d-1];
    if(verbose){printf("  There are %d %d-cells\n",nv,d-1);fflush(stdout);}
    if(EP->mCells[d]==0)
     {
      EP->mCells[d]=EPNumberOfCellsToAllocate;
      EP->cells[d]=(MFCell*)realloc(EP->cells[d],EP->mCells[d]*sizeof(MFCell)); /*done*/
      if(EP->cells[d]==(MFCell*)NULL)
       {
        sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",EP->mCells[d]*sizeof(MFCell));
        printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
        MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
        return;
       }
/*    for(III=EP->nCells[d];III<EP->mCells[d];III++)(EP->cells[d])[III]=(MFCell)NULL;*/
     }
    if(verbose){printf("  There is space for %d %d-cells\n",EP->mCells[d],d);fflush(stdout);}
    if(verbose){printf("  There are %d %d-cells\n",nv,d);fflush(stdout);}
    ne=0;
    for(iv=0;iv<nv-1;iv++)
     {
      for(jv=iv+1;jv<nv;jv++)
       {
        n1=((EP->cells[d-1])[iv])->nIndices;
        i1=((EP->cells[d-1])[iv])->indices;
        n2=((EP->cells[d-1])[jv])->nIndices;
        i2=((EP->cells[d-1])[jv])->indices;
        ni=MFIntersectIndices(n1,i1,n2,i2,intersection);
        if(verbose)
         {
          fprintf(stdout,"pair %2d,%2d ",iv,jv);
          fprintf(stdout,"      indices [");
          for(i=0;i<n1;i++)
           {
            if(i>0)fprintf(stdout," ");
            fprintf(stdout,"%2d",i1[i]);
           }
          fprintf(stdout,"], and [");
          for(i=0;i<n2;i++)
           {
            if(i>0)fprintf(stdout," ");
            fprintf(stdout,"%2d",i2[i]);
           }
          fprintf(stdout,"]      intersection [");
          for(i=0;i<ni;i++)
           {
            if(i>0)fprintf(stdout," ");
            fprintf(stdout,"%2d",intersection[i]);
           }
          fprintf(stdout,"]\n");fflush(stdout);
         }
        if(ni>=k-d)
         {
          if(verbose){printf("  Intersection large enough\n");fflush(stdout);}
          alreadyThere=0;
          n=-1;
          for(i=0;i<ne;i++)
           {
            n2=((EP->cells[d])[i])->nIndices;
            i2=((EP->cells[d])[i])->indices;
            if(MFIndicesEqual(ni,intersection,n2,i2)){alreadyThere=1;n=i;}
           }
          if(!alreadyThere)
           {
            if(verbose){printf("  Found a new %d-cell\n",d);fflush(stdout);}
            if(EP->mCells[d]==0)
             {
              EP->mCells[d]=EPNumberOfCellsToAllocate;
              if(verbose){printf("  increase size of EP->cells[%d] to %d\n",d,EP->mCells[d]);fflush(stdout);}
              EP->cells[d]=(MFCell*)realloc(EP->cells[d],EP->mCells[d]*sizeof(MFCell)); /*done*/
              if(EP->cells[d]==(MFCell*)NULL)
               {
                sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",EP->mCells[d]*sizeof(MFCell));
                printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
                MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
                return;
               }
/*            for(III=EP->nCells[d];III<EP->mCells[d];III++)(EP->cells[d])[III]=(MFCell)NULL;*/
             }else if(ne>=EP->mCells[d])
             {
              EP->mCells[d]+=EPNumberOfCellsToAllocate;
              if(verbose){printf("  allocate %d EP->cells[%d]\n",EP->mCells[d],d);fflush(stdout);}
              EP->cells[d]=(MFCell*)realloc(EP->cells[d],EP->mCells[d]*sizeof(MFCell)); /*done*/
              if(EP->cells[d]==(MFCell*)NULL)
               {
                sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",EP->mCells[d]*sizeof(MFCell));
                printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
                MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
                return;
               }
/*            for(III=EP->nCells[d];III<EP->mCells[d];III++)(EP->cells[d])[III]=(MFCell)NULL;*/
             }
            if(verbose){printf("  This will be %d-cell number %d, address=0x%8.8x, location=0x0%8.8x\n",d,ne,(EP->cells[d])+ne,(EP->cells[d])[ne]);fflush(stdout);}
            if(verbose){printf("  sizeof(struct MFCellSt)=%d\n",sizeof(struct MFCellSt));fflush(stdout);}
            tmp=(MFCell)malloc(sizeof(struct MFCellSt)); /*done*/
            if(verbose){printf("  done malloc\n");fflush(stdout);}
            (EP->cells[d])[ne]=tmp;
            if(verbose){printf("  done assignment\n");fflush(stdout);}
#ifdef EPTraceFrees
            fprintf(stderr,"Allocated MFCell at 0x%8.8x, line %d of file %s\n",(EP->cells[d])[ne],__LINE__-2,__FILE__);
#endif
            if(EP->cells[d][ne]==(MFCell)NULL)
             {
              sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFCellSt));
              printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
              MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
              return;
             }
            if(verbose){printf("  done allocating cell\n");fflush(stdout);}
            ((EP->cells[d])[ne])->nIndices=ni;
            if(verbose){printf("  done setting number of indices\n");fflush(stdout);}
            if(ni>0)
             {
              if(verbose){printf("  allocating space for %d indices\n",ni);fflush(stdout);}
              ((EP->cells[d])[ne])->indices=(int*)malloc(ni*sizeof(int)); /*done*/
              if((EP->cells[d][ne])->indices==(int*)NULL)
               {
                sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",ni*sizeof(int));
                printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
                MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
                return;
               }
              if(verbose){printf("  copying indices\n");fflush(stdout);}
              for(i=0;i<ni;i++)
                ((EP->cells[d])[ne])->indices[i]=intersection[i];
             }else
              ((EP->cells[d])[ne])->indices=(int*)NULL;

            if(verbose){printf("  done indices\n");fflush(stdout);}
            ((EP->cells[d])[ne])->mCellFaces=2;
            ((EP->cells[d])[ne])->nCellFaces=2;
            if(verbose){printf("  allocating space for cell faces\n");fflush(stdout);}
            ((EP->cells[d])[ne])->cellFaces=(int*)malloc(((EP->cells[d])[ne])->mCellFaces*sizeof(int)); /*done*/
            if((EP->cells[d][ne])->cellFaces==(int*)NULL)
             {
              sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",((EP->cells[d])[ne])->mCellFaces*sizeof(int));
              printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
              MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
              return;
             }
            if(verbose){printf("  set cell faces\n");fflush(stdout);}
            ((EP->cells[d])[ne])->cellFaces[0]=iv;
            ((EP->cells[d])[ne])->cellFaces[1]=jv;

            ((EP->cells[d])[ne])->mFaceCells=0;
            ((EP->cells[d])[ne])->nFaceCells=0;
            ((EP->cells[d])[ne])->faceCells=(int*)NULL;
            if(verbose){printf("  done creating %d-cell number %d\n",d,ne);fflush(stdout);}

            n=ne;
            ne++;
           }else{
            if(verbose){printf("  Found an old %d-cell\n",d);fflush(stdout);}
            alreadyThere=0;
            mf=((EP->cells[d])[n])->nCellFaces;
            for(i=0;i<mf;i++)
              if(((EP->cells[d])[n])->cellFaces[i]==iv)alreadyThere=1;
            if(!alreadyThere)
             {
              if(((EP->cells[d])[n])->nCellFaces>=((EP->cells[d])[n])->mCellFaces)
               {
                ((EP->cells[d])[n])->mCellFaces++;
                ((EP->cells[d])[n])->cellFaces=(int*)realloc(((EP->cells[d])[n])->cellFaces,((EP->cells[d])[n])->mCellFaces*sizeof(int)); /*done*/
                if((EP->cells[d][n])->cellFaces==(int*)NULL)
                 {
                  sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",((EP->cells[d])[n])->mCellFaces*sizeof(int));
                  printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
                  MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
                  return;
                 }
               }
              if(verbose){printf("  Add a cell face\n",d);fflush(stdout);}
              ((EP->cells[d])[n])->cellFaces[((EP->cells[d])[n])->nCellFaces]=iv;
              (((EP->cells[d])[n])->nCellFaces)++;
             }
            alreadyThere=0;
            mf=((EP->cells[d])[n])->nCellFaces;
            for(i=0;i<mf;i++)
              if(((EP->cells[d])[n])->cellFaces[i]==jv)alreadyThere=1;
            if(!alreadyThere)
             {
              if(((EP->cells[d])[n])->nCellFaces>=((EP->cells[d])[n])->mCellFaces)
               {
                ((EP->cells[d])[n])->mCellFaces++;
                ((EP->cells[d])[n])->cellFaces=(int*)realloc(((EP->cells[d])[n])->cellFaces,((EP->cells[d])[n])->mCellFaces*sizeof(int)); /*done*/
                if((EP->cells[d][n])->cellFaces==(int*)NULL)
                 {
                  sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",((EP->cells[d])[n])->mCellFaces*sizeof(int));
                  printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
                  MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
                  return;
                 }
               }
              ((EP->cells[d])[n])->cellFaces[((EP->cells[d])[n])->nCellFaces]=jv;
              (((EP->cells[d])[n])->nCellFaces)++;
             }
           }

/* Face Cells of one lower dimension */

          mf=((EP->cells[d])[n])->nCellFaces;
          for(i=0;i<mf;i++)
           {
            f=((EP->cells[d])[n])->cellFaces[i];
            if(((EP->cells[d-1])[f])->nFaceCells>0)
             {
              alreadyThere=0;
              for(j=0;j<((EP->cells[d-1])[f])->nFaceCells;j++)
                if(((EP->cells[d-1])[f])->faceCells[j]==n)alreadyThere=1;
              if(!alreadyThere)
               {
                j=((EP->cells[d-1])[f])->nFaceCells+1;
                if(j>=((EP->cells[d-1])[f])->mFaceCells)
                 {
                  ((EP->cells[d-1])[f])->mFaceCells=j;
                  ((EP->cells[d-1])[f])->faceCells=(int*)realloc(((EP->cells[d-1])[f])->faceCells,((EP->cells[d-1])[f])->mFaceCells*sizeof(int)); /*done*/
                  if((EP->cells[d-1][f])->faceCells==(int*)NULL)
                   {
                    sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",((EP->cells[d-1])[f])->mFaceCells*sizeof(int));
                    printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
                    MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
                    return;
                   }
                 }
                ((EP->cells[d-1])[f])->nFaceCells=j;
                ((EP->cells[d-1])[f])->faceCells[j-1]=n;
               }
             }else{
              if(0>=((EP->cells[d-1])[f])->mFaceCells)
               {
                ((EP->cells[d-1])[f])->mFaceCells=1;
                ((EP->cells[d-1])[f])->faceCells=(int*)malloc(((EP->cells[d-1])[f])->mFaceCells*sizeof(int)); /*done*/
                if((EP->cells[d-1][f])->faceCells==(int*)NULL)
                 {
                  sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",((EP->cells[d-1])[f])->mFaceCells*sizeof(int));
                  printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
                  MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
                  return;
                 }
               }
              ((EP->cells[d-1])[f])->nFaceCells=1;
              ((EP->cells[d-1])[f])->faceCells[0]=n;
             }
           }
         }
       }
      EP->nCells[d]=ne;
     }
    if(verbose){printf("  done intersecting %d cells to build %d cells\n",d-1,d);fflush(stdout);}
   }

  if(verbose)MFPrintEnumPolytope(stdout,EP);
  if(verbose){printf("done %s\n",RoutineName);fflush(stdout);}
  return;
 }

int MFIntersectIndices(int n1,int *inter1,int n2,int *inter2,int *inter)
 {
  static char RoutineName[]={"MFIntersectIndices"};
  int n;
  int i1,i2;

/*if(!MFTestVertexIndexOrder(n1,inter1))
   {
    printf("Error: MFIntersectIndices at entry, indices not sorted\n");
    fflush(stdout);
   }
  if(!MFTestVertexIndexOrder(n2,inter2))
   {
    printf("Error: MFIntersectIndices at entry, indices not sorted\n");
    fflush(stdout);
   }*/

  n=0;
  i1=0;i2=0;
  while(i1<n1&&i2<n2)
   {
    if(inter1[i1]==inter2[i2])
     {
      if(inter!=(int*)NULL)inter[n]=inter1[i1];
      n++;
      i1++;i2++;
     }else if(inter1[i1]<inter2[i2])
     {
      i1++;
     }else if(inter1[i1]>inter2[i2])
     {
      i2++;
     }
   }

/*if(inter!=(int*)NULL && !MFTestVertexIndexOrder(n,inter))
   {
    printf("Error: MFIntersectIndices at exit, indices not sorted\n");
    fflush(stdout);
   }*/

  return(n);
 }

void MFCreateZeroCellV(MFCell *cell,int n,...)
 {
  static char RoutineName[]={"MFCreateZeroCellV"};
  va_list indices;
  int i,j;

  *cell=(MFCell)malloc(sizeof(struct MFCellSt)); /*done*/
#ifdef EPTraceFrees
        fprintf(stderr,"Allocated MFCell at 0x%8.8x, line %d of file %s\n",*cell,__LINE__-2,__FILE__);
#endif
  if(*cell==(MFCell)NULL)
   {
    sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFCellSt));
    printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
    return;
   }
  (*cell)->nIndices=n;
  (*cell)->indices=(int*)malloc(n*sizeof(int)); /*done*/
  if((*cell)->indices==(int*)NULL)
   {
    sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",n*sizeof(int));
    printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
    return;
   }
  (*cell)->mCellFaces=0;
  (*cell)->nCellFaces=0;
  (*cell)->cellFaces=(int*)NULL;
  (*cell)->mFaceCells=0;
  (*cell)->nFaceCells=0;
  (*cell)->faceCells=(int*)NULL;

  va_start(indices,n);
  for(i=0;i<n;i++)
    (*cell)->indices[i]=va_arg(indices,int);
  va_end(indices);

  return;
 }

void MFCreateZeroCell(MFCell *cell,int n,int* indices)
 {
  static char RoutineName[]={"MFCreateZeroCell"};
  int i,j;

  *cell=(MFCell)malloc(sizeof(struct MFCellSt)); /*done*/
#ifdef EPTraceFrees
  fprintf(stderr,"Allocated MFCell at 0x%8.8x, line %d of file %s\n",*cell,__LINE__-2,__FILE__);
#endif
  if(*cell==(MFCell)NULL)
   {
    sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFCellSt));
    printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
    return;
   }
  (*cell)->nIndices=n;
  if(n>0)
   {
    (*cell)->indices=(int*)malloc(n*sizeof(int)); /*done*/
    if((*cell)->indices==(int*)NULL)
     {
      sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",n*sizeof(int));
      printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
      return;
     }
   }else{
    (*cell)->indices=(int*)NULL;
   }
  (*cell)->mCellFaces=0;
  (*cell)->nCellFaces=0;
  (*cell)->cellFaces=(int*)NULL;
  (*cell)->mFaceCells=0;
  (*cell)->nFaceCells=0;
  (*cell)->faceCells=(int*)NULL;

  for(i=0;i<n;i++)(*cell)->indices[i]=indices[i];

  return;
 }

void  MFCreateEmptyKComplex(int k, int n, MFEnumPolytope *EP,int m)
 {
  static char RoutineName[]={"MFCreateEmptyKComplex"};
  int i,j;

  *EP=(MFEnumPolytope)malloc(sizeof(struct MFEnumPolytopeSt)); /*done*/
  if(*EP==(MFEnumPolytope)NULL)
   {
    sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFEnumPolytopeSt));
    printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
    return;
   }
  (*EP)->k=k;
  (*EP)->n=n;
  (*EP)->nCells=(int*)malloc((k+1)*sizeof(int)); /*done*/
  if((*EP)->nCells==(int*)NULL)
   {
    sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",(k+1)*sizeof(int));
    printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
    return;
   }
  (*EP)->mCells=(int*)malloc((k+1)*sizeof(int)); /*done*/
  if((*EP)->mCells==(int*)NULL)
   {
    sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",(k+1)*sizeof(int));
    printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
    return;
   }
  (*EP)->cells=(MFCell**)malloc((k+1)*sizeof(MFCell*)); /*done*/
  if((*EP)->cells==(MFCell**)NULL)
   {
    sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",(k+1)*sizeof(MFCell*));
    printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
    return;
   }
  (*EP)->mCells[0]=m;
  (*EP)->nCells[0]=m;
  if(m>0)
   {
    (*EP)->cells[0]=(MFCell*)malloc((*EP)->mCells[0]*sizeof(MFCell)); /*done*/
    if((*EP)->cells[0]==(MFCell*)NULL)
     {
      sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",((*EP)->mCells[0])*sizeof(MFCell));
      printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
      return;
     }
    (*EP)->v=(MFKVector*)malloc((*EP)->mCells[0]*sizeof(MFKVector)); /*done*/
    if((*EP)->v==(MFKVector*)NULL)
     {
      sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",((*EP)->mCells[0])*sizeof(MFKVector));
      printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
      return;
     }
    for(j=0;j<(*EP)->mCells[0];j++)
     {
      ((*EP)->cells[0])[j]=(MFCell)NULL;
      (*EP)->v[j]=(MFKVector)NULL;
     }
   }else{
    (*EP)->mCells[0]=EPNumberOfCellsToAllocate;
    (*EP)->nCells[0]=0;
    (*EP)->cells[0]=(MFCell*)malloc((*EP)->mCells[0]*sizeof(MFCell)); /*done*/
    if((*EP)->cells[0]==(MFCell*)NULL)
     {
      sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",((*EP)->mCells[0])*sizeof(MFCell));
      printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
      return;
     }
    (*EP)->v=(MFKVector*)malloc((*EP)->mCells[0]*sizeof(MFKVector)); /*done*/
    if((*EP)->v==(MFKVector*)NULL)
     {
      sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",((*EP)->mCells[0])*sizeof(MFKVector));
      printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
      return;
     }
    for(j=0;j<(*EP)->mCells[0];j++)
     {
      ((*EP)->cells[0])[j]=(MFCell)NULL;
      (*EP)->v[j]=(MFKVector)NULL;
     }
   }

  for(i=1;i<=k;i++)
   {
    (*EP)->mCells[i]=EPNumberOfCellsToAllocate;
    (*EP)->nCells[i]=0;
    (*EP)->cells[i]=(MFCell*)malloc((*EP)->mCells[i]*sizeof(MFCell)); /*done*/
    if((*EP)->cells[i]==(MFCell*)NULL)
     {
      sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",((*EP)->mCells[i])*sizeof(MFCell));
      printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
      return;
     }
    for(j=0;j<(*EP)->mCells[i];j++)((*EP)->cells[i])[j]=(MFCell)NULL;
   }

  return;
 }

int MFIndicesEqual(int n1,int *inter1,int n2,int *inter2)
 {
  static char RoutineName[]={"MFIndicesEqual"};
  int result;
  int i;

  result=0;
  if(n1!=n2)
   {
    result=0;
   }else{
    result=1;
    for(i=0;i<n1;i++)
      if(inter1[i]!=inter2[i])result=0;
   } 
  return(result);
 }

void MFFreeEnumPolytope(MFEnumPolytope EP)
 {
  static char RoutineName[]={"MFFreeEnumPolytope"};
  int i,j;
  int verbose;

  verbose=0;
  if(verbose){printf("MFFreeEnumPolytope\n");fflush(stdout);}

  if(EP==(MFEnumPolytope)NULL)return;

  if(verbose){printf("  delete vertices\n");fflush(stdout);}
  for(i=0;i<EP->nCells[0];i++)
   {
    if(verbose){printf("  %d \n",i);fflush(stdout);}
    if(EP->v[i]!=(MFKVector)NULL)MFFreeKVector(EP->v[i]);
   }
  if(verbose){printf("  delete vertex list\n");fflush(stdout);}
  if(EP->v!=(MFKVector*)NULL)
   {
#ifdef EPTraceFrees
  fprintf(stderr,"Free called at line %d in file %s\n",__LINE__+2,__FILE__);
#endif
    free(EP->v);
   }

  if(verbose){printf("  delete cells\n");fflush(stdout);}
  if(EP->nCells!=(int*)NULL)
   {
    for(j=0;j<=EP->k;j++)
     {
      if(verbose){printf("    dimension %d\n",j);fflush(stdout);}
      for(i=0;i<EP->nCells[j];i++)
       {
        if(verbose){printf("      %d\n",i);fflush(stdout);}
        if((EP->cells[j])[i]!=(MFCell)NULL)MFFreeCell((EP->cells[j])[i]);
       }
      if(verbose){printf("   delete list of cells of dimension %d\n",j);fflush(stdout);}
      if(EP->cells[j]!=(MFCell*)NULL)
       {
#ifdef EPTraceFrees
  fprintf(stderr,"Free called at line %d in file %s\n",__LINE__+2,__FILE__);
#endif
        free(EP->cells[j]);
       }
     }
   }

  if(verbose){printf("   delete cell counts\n");fflush(stdout);}
  if(EP->nCells!=(int*)NULL)
   {
#ifdef EPTraceFrees
  fprintf(stderr,"Free called at line %d in file %s\n",__LINE__+2,__FILE__);
#endif
    free(EP->nCells);
   }
  if(EP->mCells!=(int*)NULL)
   {
#ifdef EPTraceFrees
  fprintf(stderr,"Free called at line %d in file %s\n",__LINE__+2,__FILE__);
#endif
    free(EP->mCells);
   }
  if(verbose){printf("   delete list of list of cells\n");fflush(stdout);}
  if(EP->cells!=(MFCell**)NULL)
   {
#ifdef EPTraceFrees
  fprintf(stderr,"Free called at line %d in file %s\n",__LINE__+2,__FILE__);
#endif
    free(EP->cells);
   }
     
  if(verbose){printf("   free data structure\n");fflush(stdout);}
#ifdef EPTraceFrees
  fprintf(stderr,"Free called at line %d in file %s\n",__LINE__+2,__FILE__);
#endif
  free(EP);
  if(verbose){printf("done MFFreeEnumPolytope\n");fflush(stdout);}
  return;
 }

void MFFreeCell(MFCell C)
 {
  static char RoutineName[]={"MFFreeCell"};
  int verbose;

  verbose=0;
  if(verbose){printf("MFFreeCell\n");fflush(stdout);}

  if(C==(MFCell)NULL)return;
  if(verbose){printf("   delete indices\n");fflush(stdout);}
  if(C->indices!=(int*)NULL)
   {
#ifdef EPTraceFrees
    fprintf(stderr,"Free called at line %d in file %s\n",__LINE__+2,__FILE__);
    fprintf(stderr,"C=0x%8.8x\n",C);
#endif
    free(C->indices);
   }
  if(verbose){printf("   delete faceCells\n");fflush(stdout);}
  if(C->faceCells!=(int*)NULL)
   {
#ifdef EPTraceFrees
    fprintf(stderr,"Free called at line %d in file %s\n",__LINE__+2,__FILE__);
    fprintf(stderr,"C=0x%8.8x\n",C);
#endif
    free(C->faceCells);
   }
  if(verbose){printf("   delete cellFaces\n");fflush(stdout);}
  if(C->cellFaces!=(int*)NULL)
   {
#ifdef EPTraceFrees
    fprintf(stderr,"Free called at line %d in file %s\n",__LINE__+2,__FILE__);
    fprintf(stderr,"C=0x%8.8x\n",C);
#endif
    free(C->cellFaces);
   }
  if(verbose){printf("   free data structure\n");fflush(stdout);}
#ifdef EPTraceFrees
  fprintf(stderr,"Free called at line %d in file %s\n",__LINE__+2,__FILE__);
#endif
  free(C);

  if(verbose){printf("done MFFreeCell\n");fflush(stdout);}
  return;
 }

long MFEnumPolytopeNumberOfVertices(MFEnumPolytope EP)
 {
  static char RoutineName[]={"MFEnumPolytopeNumberOfVertices"};
  return(EP->nCells[0]);
 }

int MFEnumPolytopeNumberOfCells(MFEnumPolytope EP,int dim)
 {
  static char RoutineName[]={"MFEnumPolytopeNumberOfCells"};
  if(dim>EP->n)return 0;
  return(EP->nCells[dim]);
 }

int MFEnumPolytopeNumberOfCellFaces(MFEnumPolytope EP,int dim,long cell)
 {
  static char RoutineName[]={"MFEnumPolytopeNumberOfCellFaces"};
  return(((EP->cells[dim])[cell])->nCellFaces);
 }

long MFEnumPolytopeCellFace(MFEnumPolytope EP,int dim,long cell,int face)
 {
  static char RoutineName[]={"MFEnumPolytopeCellFace"};
  int result;

  /*@@*/
/*printf("%s\n",RoutineName);fflush(stdout);
*/
/*printf("(0x303dee70)[0]=%d\n",((int*)(0x303dee70))[0]);fflush(stdout);
*/

  if(dim<0||dim>EP->k)fprintf(stderr,"Error at line %d in file %s, you asked for a cell face of a %d-cell, but k=%d\n",__LINE__,__FILE__,dim,EP->k);
  if(cell<0||cell>=EP->nCells[dim])fprintf(stderr,"Error at line %d in file %s, you asked for a cell face of %d-cell %d, but there are %d cells of this dimension \n",__LINE__,__FILE__,dim,cell,EP->nCells[dim]);
  if(face<0||face>=((EP->cells[dim])[cell])->nCellFaces)fprintf(stderr,"Error at line %d in file %s, you asked for cell face %d of %d-cell %d, but there are %d cell faces\n",__LINE__,__FILE__,face,dim,cell,((EP->cells[dim])[cell])->nCellFaces);

/* Signs: if cell<0 return -( (...[-cell])->cellFaces[face]) ? */

  result=((EP->cells[dim])[cell])->cellFaces[face];

  return result;
 }

int MFEnumPolytopeNumberOfFaceCells(MFEnumPolytope EP,int dim,long face)
 {
  static char RoutineName[]={"MFEnumPolytopeNumberOfFaceCells"};
  return(((EP->cells[dim])[face])->nFaceCells);
 }

int MFEnumPolytopeNumberOfCellIndices(MFEnumPolytope EP,int dim,long cell)
 {
  static char RoutineName[]={"MFEnumPolytopeNumberOfCellIndices"};
  if(EP==(MFEnumPolytope)NULL)
   {
    printf("EP NULL in MFEnumPolytopeNumberOfCellIndices\n");
    fflush(stdout);abort();
   }
  if(dim<0 || dim>EP->k)
   {
    printf("In MFEnumPolytopeNumberOfCellIndices, dim (%d) is invalid. Must be positice and less than or equal to %d.\n",dim,EP->k);
    fflush(stdout);abort();
   }
  if(cell<0 || !(cell<EP->nCells[dim]))
   {
    printf("In MFEnumPolytopeNumberOfCellIndices, cell (%d) is invalid. Must be positice and less than %d.\n",cell,EP->nCells[dim]);
    fflush(stdout);abort();
   }
  return(((EP->cells[dim])[cell])->nIndices);
 }

int MFEnumPolytopeCellIndex(MFEnumPolytope EP,int dim,long cell,int i)
 {
  static char RoutineName[]={"MFEnumPolytopeCellIndex"};
  return(((EP->cells[dim])[cell])->indices[i]);
 }

long MFEnumPolytopeFaceCell(MFEnumPolytope EP,int dim,long face,int cell)
 {
  static char RoutineName[]={"MFEnumPolytopeFaceCell"};
  return(((EP->cells[dim])[face])->faceCells[cell]);
 }

void EPSortPolytopeIndices(int n,int *list)
 {
  static char RoutineName[]={"EPSortPolytopeIndices"};
  int i,j,t;

  for(i=0;i<n-1;i++)
   for(j=i;j<n;j++)
    {
     if(list[j]<list[i])
      {
       t=list[j];
       list[j]=list[i];
       list[i]=t;
      }
    }
 }

void EPCollapseIndices(int *n,int *list)
 {
  static char RoutineName[]={"EPCollapseIndices"};
  int i,j;
  EPSortPolytopeIndices(*n,list);
  for(i=0;i<*n-1;i++)
   {
    if(list[i]==list[i+1])
     {
      (*n)--;
      for(j=i+1;j<*n;j++)list[j]=list[j+1];
      i--;
     }
   }
  return;
 }

void MFCreateZeroCellWithEdgeList(MFCell *cell,int n,int *edges)
 {
  static char RoutineName[]={"MFCreateZeroCellWithEdgeList"};
  int i,j;

  *cell=(MFCell)malloc(sizeof(struct MFCellSt)); /*done*/
#ifdef EPTraceFrees
  fprintf(stderr,"Allocated MFCell at 0x%8.8x, line %d of file %s\n",*cell,__LINE__-2,__FILE__);
#endif
  if(*cell==(MFCell)NULL)
   {
    sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFCellSt));
    printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
    return;
   }
  (*cell)->nIndices=0;
  (*cell)->indices=(int*)NULL;
  (*cell)->mCellFaces=0;
  (*cell)->nCellFaces=0;
  (*cell)->cellFaces=(int*)NULL;
  (*cell)->mFaceCells=n;
  (*cell)->nFaceCells=n;
  (*cell)->faceCells=(int*)malloc(n*sizeof(int)); /*done*/
  if((*cell)->faceCells==(int*)NULL)
   {
    sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",n*sizeof(int));
    printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
    return;
   }

  for(i=0;i<n;i++)(*cell)->faceCells[i]=edges[i];

  return;
 }

int MFEnumPolytopeDimension(MFEnumPolytope EP)
 {
  static char RoutineName[]={"MFEnumPolytopeDimension"};
  return(EP->k);
 }

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

int MFTestEnumPolytope(MFEnumPolytope EP,int closed)
 {
  static char RoutineName[]={"MFTestEnumPolytope"};
  int k;
  int d;
  int nCells,nFaces,nAdj;
  long cell,face,adj;
  long cellface,adjcell;
  int nFaceAdjs,nAdjFaces;
  long faceadj,adjface;
  int i;
  int present,presentTwice,passed;

  k=MFEnumPolytopeDimension(EP);

  passed=1;
  for(d=0;d<k+1;d++)
   {
    nCells=MFEnumPolytopeNumberOfCells(EP,d);
    for(cell=0;cell<nCells;cell++)
     {
      if(d-1>-1)
       {
        nFaces=MFEnumPolytopeNumberOfCellFaces(EP,d,cell);
        if(closed && nFaces<d+1)
         {
          passed=0;
          printf("MFTestEnumPolytope:\n %d-cell %ld has %d faces, fewer than %d\n",d,cell,nFaces,d+1);
         }
        for(face=0;face<nFaces;face++)
         {
          cellface=MFEnumPolytopeCellFace(EP,d,cell,face);
          presentTwice=0;
          for(i=0;i<nFaces;i++)
           {
            if(i!=face && cellface==MFEnumPolytopeCellFace(EP,d,cell,i))presentTwice=1;
           }
          if(presentTwice)
           {
            passed=0;
            printf("MFTestEnumPolytope:\n A face of %d-cell %ld (%d-cell %ld) appears twice in the face list.\n",d,cell,d-1,cellface);
            printf("The face list of %d-cell %ld is [",d,cell);
            for(i=0;i<nFaces;i++)
             {
              if(i>0)printf(",");
              printf("%d",MFEnumPolytopeCellFace(EP,d,cell,i));
             }
            printf("]\n");
           }
          nFaceAdjs=MFEnumPolytopeNumberOfFaceCells(EP,d-1,cellface);
          present=0;
          for(faceadj=0;faceadj<nFaceAdjs;faceadj++)
           {
            if(MFEnumPolytopeFaceCell(EP,d-1,cellface,faceadj)==cell)present=1;
           }
          if(!present)
           {
            passed=0;
            printf("MFTestEnumPolytope:\n Face of %d-cell %ld (%d-cell %ld) does not list %d-cell %ld as a cell it is contained in\n",d,cell,d-1,cellface,d,cell);
            printf("%d-cell %ld is contained in cells (%d-cells) [",d-1,cellface,d);
            for(faceadj=0;faceadj<nFaceAdjs;faceadj++)
             {
              if(faceadj>0)printf(",");
              printf("%d",MFEnumPolytopeFaceCell(EP,d-1,cellface,faceadj));
             }
            printf("]\n");
           }
         }
       }
      if(d+1<k)
       {
        nAdj=MFEnumPolytopeNumberOfFaceCells(EP,d,cell);
        for(adj=0;adj<nAdj;adj++)
         {
          adjcell=MFEnumPolytopeFaceCell(EP,d,cell,adj);
          presentTwice=0;
          for(i=0;i<nAdj;i++)
           {
            if(i!=adj && adjcell==MFEnumPolytopeFaceCell(EP,d,cell,i))presentTwice=1;
           }
          if(presentTwice)
           {
            passed=0;
            printf("MFTestEnumPolytope:\n A containing cell of %d-cell %ld (%d-cell %ld) appears twice in the list of containing cells.\n",d,cell,d+1,adjcell);
            printf("The containing cell list of %d-cell %ld is [",d,cell);
            for(i=0;i<nAdj;i++)
             {
              if(i>0)printf(",");
              printf("%d",MFEnumPolytopeFaceCell(EP,d,cell,i));
             }
            printf("]\n");
           }
          nAdjFaces=MFEnumPolytopeNumberOfCellFaces(EP,d+1,adjcell);
          present=0;
          for(adjface=0;adjface<nAdjFaces;adjface++)
           {
            if(MFEnumPolytopeCellFace(EP,d+1,adjcell,adjface)==cell)present=1;
           }
          if(!present)
           {
            passed=0;
            printf("MFTestEnumPolytope:\n Containing cell of %d-cell %ld (%d-cell %ld) does not list %d-cell %ld as a face\n",d,cell,d+1,adjcell,d,cell);
            printf("%d-cell %ld has faces (%d-cells) [",d+1,adjcell,d);
            for(adjface=0;adjface<nAdjFaces;adjface++)
             {
              if(adjface>0)printf(",");
              printf("%d",MFEnumPolytopeCellFace(EP,d+1,adjcell,adjface));
             }
            printf("]\n");
           }
         }
       }
     }
   }
  return(passed);
 }

int MFEnumPolytopeIsCellExterior(MFAtlas A,int chart,MFEnumPolytope P,int dim,int cell)
 {
  static char RoutineName[]={"MFEnumPolytopeIsCellExterior"};
  int hasHyperCube,hasNotInBoth;
  int indicesOneIn;
  int result;
  int nV;
  int *Vs;
  MFCell c;
  int i;
  int *indices;
  int imin;
  double rmin,r,vij;
  MFPolytope PP;
  MFKVector vi;
  MFKVector vj;
  MFKVector vk;
  MFKVector v,diff;
  MFKVector diffi,diffj;
  int verbose;
  double s,t,R;
  double a11,a12,a22,b1,b2;

  verbose=0;
  if(verbose){printf("MFEnumPolytopeIsCellExterior P=0x%8.8x--\n",P);fflush(stdout);}

  if(dim==MFEnumPolytopeDimension(P))return(0);

  PP=MFChartPolytope(MFAtlasChart(A,chart));

  if(verbose){printf("  %d-cell %d\n",dim,cell);fflush(stdout);}

  if(verbose&&dim==1)
   {
    indices=(int*)NULL;
    nV=0;
    nV=MFEnumPolytopeVerticesOnCell(P,dim,cell,nV,&indices);
    printf("There are %d vertices lying on edge %d [",nV,cell);
    for(i=0;i<nV;i++)
     {
      if(i>0)printf(",");
      printf("%d",indices[i]);
     }
    printf("]\n");fflush(stdout);
    free(indices);
   }

  c=(P->cells[dim])[cell];

  result=0;
  for(i=0;i<c->nIndices;i++)
   {
    if(MFAtlasIsHalfSpaceHyperCube(A,c->indices[i]))
     {
      result=1;
      if(verbose){printf(" Cell lies on a hypercube face\n");fflush(stdout);}
      return(result);
     }
    if(!MFAtlasIsHalfSpaceInBothPolytopes(A,c->indices[i]))
     {
      result=1;
      if(verbose){printf(" Cell does not lie on a face shared by an adjacent polytope.\n");fflush(stdout);}
      return(result);
     }
   }

  result=0;

  nV=0;
  indices=(int*)NULL;
  nV=MFEnumPolytopeVerticesOnCell(P,dim,cell,nV,&indices);

  indicesOneIn=0;

  if(verbose)
   {
    printf("There are %d vertices lying on the cell [",nV);
    for(i=0;i<nV;i++)
     {
      if(i>0)printf(",");
      printf("%d",indices[i]);
     }
    printf("]\n");
   }

  for(i=0;i<nV;i++)
   {
    if(verbose)
     {
      printf("  Vertex %d is ",indices[i]);
      vi=MFCreateKVector(MFAtlasK(A));
      MFPolytopeVertex(PP,indices[i],vi);
      MFPrintKVector(stdout,vi);
      MFFreeKVector(vi);
      printf(" radius %lf",MFPolytopeRadiusOfVertex(PP,indices[i]));
     }
    if(MFPolytopeRadiusOfVertex(PP,indices[i])<MFAtlasChartRadius(A,chart))
     {
      if(verbose)printf(" and is in chart %d (radius %lf)\n",chart,MFAtlasChartRadius(A,chart));
      indicesOneIn=1;
     }
   }

  if(indicesOneIn)
   {
    if(verbose)printf(" One vertex is interior to the chart\n");
    result=0;
    if(indices!=(int*)NULL)free(indices);
    return(result);
   }else{
    if(verbose)printf(" All vertices are exterior to the chart\n");
   }

  if(dim==0)
   {
    result=0;
    if(MFPolytopeRadiusOfVertex(MFChartPolytope(MFAtlasChart(A,chart)),cell)>MFAtlasChartRadius(A,chart))result=1;
/* }else if(dim==1)
   {
    vi=MFCreateKVector(MFAtlasK(A));
    vj=MFCreateKVector(MFAtlasK(A));
    diff=MFCreateKVector(MFAtlasK(A));
    v=MFCreateKVector(MFAtlasK(A));
    MFPolytopeVertex(PP,indices[0],vi);
    MFPolytopeVertex(PP,indices[1],vj);
    MFKVDiff(vi,vj,diff);
    t=MFKVDot(vi,diff)/MFKVNorm(diff);
    if(t<0.)t=0.;
    if(t>1.)t=1.;
    MFKVScaleMul(t,diff,vj);
    MFKVAdd(vi,vj,v);
    R=MFKVNorm(v);
    
    MFFreeKVector(vi);
    MFFreeKVector(vj);
    MFFreeKVector(diff);
    MFFreeKVector(v);
    result=0;
    if(R>MFAtlasChartRadius(A,chart))result=1;
   }else if(dim==2)
   {
    vi=MFCreateKVector(MFAtlasK(A));
    vj=MFCreateKVector(MFAtlasK(A));
    vk=MFCreateKVector(MFAtlasK(A));
    diffi=MFCreateKVector(MFAtlasK(A));
    diffj=MFCreateKVector(MFAtlasK(A));
    v=MFCreateKVector(MFAtlasK(A));
    MFPolytopeVertex(PP,indices[0],vi);
    MFPolytopeVertex(PP,indices[1],vj);
    MFPolytopeVertex(PP,indices[2],vk);
    MFKVDiff(vi,vj,diffi);
    MFKVDiff(vi,vk,diffj);
    a11=MFKVNorm(diffi);
    a12=MFKVDot(diffi,diffj);
    a22=MFKVNorm(diffj);
    b1=-MFKVDot(vi,diffi);
    b2=-MFKVDot(vi,diffj);
    s=( b1*a22-b2*a12)/(a11*a22-a12*a12);
    t=(-b1*a12+b2*a22)/(a11*a22-a12*a12);
    if(s<0.)s=0.;
    if(t<0.)t=0.;
    MFKVScaleMul(t,diffi,vj);
    MFKVScaleMul(s,diffj,vk);
    MFKVAdd(vi,vj,diffi);
    MFKVAdd(diffi,diffj,v);
    R=MFKVNorm(v);

    MFFreeKVector(vi);
    MFFreeKVector(vj);
    MFFreeKVector(vk);
    MFFreeKVector(diffi);
    MFFreeKVector(diffj);
    MFFreeKVector(v);
    result=0;
    if(R>MFAtlasChartRadius(A,chart))result=1;
*/
    if(indices!=(int*)NULL)free(indices);
    return result;
   }

  result=0;
  if(0 && nV>1)
   {
    if(verbose)printf("Punt -- Trying to eliminate a cell of dimension %d\n",dim);
    imin=0;
    rmin=MFPolytopeRadiusOfVertex(PP,indices[0]);
    if(verbose)printf("   Vertex %d is radius %lf\n",indices[0],rmin);
    for(i=1;i<nV;i++)
     {
      r=MFPolytopeRadiusOfVertex(PP,indices[i]);
      if(r<rmin)
       {
        rmin=r;
        imin=i;
       }
     }
    if(verbose)printf(" The closest vertex to the center is Vertex %d, radius %lf, ball is radius %lf\n",indices[imin],rmin,MFAtlasChartRadius(A,chart));

    vi=MFCreateKVector(MFAtlasK(A));
    vj=MFCreateKVector(MFAtlasK(A));
    MFPolytopeVertex(PP,indices[imin],vi);
    r=rmin*rmin;
    result=0;
    for(i=0;i<nV;i++)
     {
      if(i!=imin)
       {
        MFPolytopeVertex(PP,indices[i],vj);
        vij=MFKVDot(vi,vj);
        if(vij>=r)result=1;
        if(verbose)
         {
          printf("  v[%d]=",imin);
          MFPrintKVector(stdout,vi);
          printf("\n");
          printf("  v[%d]=",i);
          MFPrintKVector(stdout,vj);
          printf("\n");
          printf("    v[%d].v[%d]=%lf, v[%d].v[%d]=%lf\n",indices[i],indices[imin],vij,indices[imin],indices[imin],r);
          printf("    v[%d].(v[%d]-v[%d])=%lf, v[%d].(v[%d]-v[%d])=%lf\n",indices[i],indices[i],indices[imin],r-vij,indices[imin],indices[imin],indices[i],MFKVDot(vj,vj)-vij);
         }
       }
     }
    if(result&&rmin<MFAtlasChartRadius(A,chart))result=0;
    if(verbose)printf("   result is %d\n",result);

    MFFreeKVector(vi);
    MFFreeKVector(vj);
   }
   
  if(indices!=(int*)NULL)free(indices);

  return(result);
 }

int MFEnumPolytopeVerticesOnCell(MFEnumPolytope P,int dim,int cell,int nV,int **indices)
 {
  static char RoutineName[]={"MFEnumPolytopeVerticesOnCell"};
  int nF,f;
  int n;
  MFCell c;
  int verbose;

  verbose=0;
  if(verbose){printf("Counting the vertices contained on %d-cell %d, nV=%d\n",dim,cell,nV);fflush(stdout);}
  if(dim==0)
   {
    if(verbose){printf("   Single vertex %d, checking to see if it is in the list nV=%d\n",cell,nV);fflush(stdout);}

    if(*indices!=(int*)NULL)
     {
      n=0;
      for(f=0;f<nV;f++)if(((*indices)[f])==cell)n=1;
      if(n==1)
       {
        if(verbose){printf("     Already there, returning\n");fflush(stdout);}
        return(nV);
       }
     }

    if(verbose){printf("     Not there, adding it\n");fflush(stdout);}
    if(*indices==(int*)NULL)
     {
      if(verbose){printf("     *indices is NULL, mallocing it\n");fflush(stdout);}
      *indices=(int*)malloc(10*sizeof(int)); /*done*/
      if(*indices==(int*)NULL)
       {
        sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",10*sizeof(int));
        printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
        MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
        return;
       }
     }else{
      if(verbose){printf("     *indices is not NULL, reallocing it\n");fflush(stdout);}
      *indices=(int*)realloc((void*)(*indices),(nV+10)*sizeof(int)); /*done*/
      if(*indices==(int*)NULL)
       {
        sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",(nV+10)*sizeof(int));
        printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
        MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
        return -1;
       }
     }

    (*indices)[nV]=cell;
    return(nV+1);
   }


  c=(P->cells[dim])[cell];

  nF=c->nCellFaces;
  if(verbose)
   {
    printf("   Pushing down to check the faces of this %d-cell [",dim);
    for(f=0;f<nF;f++)
     {
      if(f>0)printf(",");
      printf("%d",c->cellFaces[f]);
     }
    printf("]\n");
    fflush(stdout);
   }
  if(verbose){printf("     There are %d faces\n",nF);fflush(stdout);}
  n=nV;
  for(f=0;f<nF;f++)
   {
    if(verbose){printf("     Check face %d, %d-cell %d\n",f,dim-1,c->faceCells[f]);fflush(stdout);}
    n=MFEnumPolytopeVerticesOnCell(P,dim-1,c->cellFaces[f],n,indices);
   }

  return(n);
 }

MFEnumPolytope MFEnumerateAtlas(MFAtlas A)
 {
  static char RoutineName[]={"MFEnumerateAtlas"};
  int chart,nCharts;
  int dim;
  int k;
  int i,l;
  MFCell mfcell;
  MFCell cellFace;
  MFCell faceCell;
  int chart2;
  long nCells2,cell2;
  MFCell mfcell2;
  long cell,nCells;
  MFPolytope P;
  MFEnumPolytope *EPS;
  MFEnumPolytope Complex=(MFEnumPolytope)NULL;
  int m;
  int indices[10000]={0}; /*!!!*/
  int verbose;
  int verboseResult;
  int verboseTerse;
  int hasHyperCube;
  int hasNotInBoth;
  int found;
  int failed;

  int nSame;
  int mSame;
  int maxSame;
  MFCell mfcellClosest;
  int cellClosest;

/* In 2d, edge dual takes you to neighboring polygon, and then the next edge CCW
                   which gives the next neighbor (sign?)*/
/* In 3d, face dual takes you to neighboring polyhedron. Edge on face in primal
                   walks to next face whcih gives next neighbor */

  verbose=0;
  verboseTerse=0;
  verboseResult=0;
  if(verbose){printf("In MFEnumeratePolytopes\n");fflush(stdout);}

  nCharts=MFAtlasNumberOfCharts(A);
  k=MFAtlasK(A);
  EPS=(MFEnumPolytope*)malloc(nCharts*sizeof(MFEnumPolytope)); /*done*/
  if(EPS==(MFEnumPolytope*)NULL)
   {
    sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",nCharts*sizeof(MFEnumPolytope));
    printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
    return (MFEnumPolytope)NULL;
   }

  if(verbose){printf("Atlas is dimension %d, with %d charts.\n",k,nCharts);fflush(stdout);}

  MFCreateEmptyKComplex(k,MFAtlasN(A),&Complex,0);

  if(verbose)
   {
    if(verbose){printf("In MFEnumeratePolytopes\n");fflush(stdout);}
    for(chart=0;chart<nCharts;chart++)
     {
      P=MFChartPolytope(MFAtlasChart(A,chart));
      printf("\nPolytope %d\n\n",chart);
      MFPrintPolytopeTerse(stdout,P);
     }
    printf("\nFaces\n\n");
    MFPrintAtlasFaceList(stdout,A);
    fflush(stdout);
   }

  for(chart=0;chart<nCharts;chart++)
   {
    if(verbose){printf("chart %d/%d\n",chart,nCharts);fflush(stdout);}
    if(MFAtlasChart(A,chart)==(MFChart)NULL)
     {
      {printf("chart %d/%d is NULL! ,atlas 0x%8.8x\n",chart,nCharts,A);fflush(stdout);}
       continue;
     }
    P=MFChartPolytope(MFAtlasChart(A,chart));

/*  for(face=0;face<MFPolytopeNumberOfFaces(P);face++)
     {
      face1=MFPolytopeFaceIndex(P,face);
     Merge all vertices on a face if the face is not also a face of
     the polytope across the face 
     } */

/*  MFPolytopeMergeCloseVertices(P,.05*MFAtlasChartRadius(A,chart));
    MFPolytopeUpdateFaceList(P);*/
    if(verbose){printf("call MFEnumeratePolytope, chart %d\n",chart);fflush(stdout);}
    if(MFPolytopeNumberOfVertices(P)>0)
      EPS[chart]=MFEnumeratePolytope(P);
     else
      EPS[chart]=(MFEnumPolytope)NULL;

    if(verbose){printf("call MFTestEnumPolytope, chart %d\n",chart);fflush(stdout);}
    if(0 && EPS[chart]!=(MFEnumPolytope)NULL && (failed=!MFTestEnumPolytope(EPS[chart],1)))
     {
      printf("----------------------------------------------------\n");
      printf("----------------------------------------------------\n");
      printf("\n%s EPS[%d] failed tests:\n\n",RoutineName,chart);
      printf("P[%d]=0x%8.8x\n",chart,P);
      MFPrintPolytope(stdout,P);
      printf("\n");
      printf("EPS[%d]=0x%8.8x\n",chart,EPS[chart]);
      MFPrintEnumPolytope(stdout,EPS[chart]);
      printf("\n");
      printf("----------------------------------------------------\n");
      printf("----------------------------------------------------\n");
      fflush(stdout);
      abort();
     }
    if(verbose)
     {
      printf("  Chart %d:\n",chart);
      printf("EPS[%d]=0x%8.8x\n",chart,EPS[chart]);
      printf("\nCenter:");fflush(stdout);
      if(MFChartCenter(MFAtlasChart(A,chart))==(MFNVector)NULL)
       {
        printf("NULL\n");fflush(stdout);abort();
       }else
        MFPrintNVector(stdout,MFChartCenter(MFAtlasChart(A,chart)));
      fflush(stdout);
      printf("\n");fflush(stdout);
      printf("\nPolytope:\n\n");fflush(stdout);
      MFPrintPolytope(stdout,MFChartPolytope(MFAtlasChart(A,chart)));fflush(stdout);
      printf("\n");fflush(stdout);
      printf("\nEnumPolytope:\n\n");fflush(stdout);
      if(EPS[chart]!=(MFEnumPolytope)NULL){MFPrintEnumPolytope(stdout,EPS[chart]);fflush(stdout);}
      printf("\n");
      fflush(stdout);
     }
/*  if(failed)abort();*/

    if(verbose){printf("In middle of MFEnumeratePolytopes\n");fflush(stdout);}

    if(EPS[chart]!=(MFEnumPolytope)NULL)
     {
    for(dim=0;dim<k+1;dim++)
     {
      nCells=MFEnumPolytopeNumberOfCells(EPS[chart],dim);
      if(verbose){printf("    There are %d cells of dimension %d\n",nCells,dim);fflush(stdout);}
      for(cell=0;cell<nCells;cell++)
       {
        mfcell=((EPS[chart])->cells[dim])[cell];
        if(verbose)
         {
          printf("      %d [",cell);
          for(i=0;i<mfcell->nIndices;i++)
           {
            if(i>0)printf(",");
            printf("%d",mfcell->indices[i]);
           }
          printf("]");
          fflush(stdout);
         }

        if(!MFEnumPolytopeIsCellExterior(A,chart,EPS[chart],dim,cell))
         {
          indices[0]=chart;
          m=1;
          for(i=0;i<mfcell->nIndices;i++)
           {
            if((indices[m]=MFAtlasLeftPolytope(A,mfcell->indices[i]))==chart)
                indices[m]=MFAtlasRightPolytope(A,mfcell->indices[i]);
            m++;
           }
          EPSortPolytopeIndices(m,indices);
          mfcell->indices=(int*)realloc((void*)mfcell->indices,(m+1)*sizeof(int)); /*done*/
          if(mfcell->indices==(int*)NULL)
           {
            sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",(m+1)*sizeof(int));
            printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
            MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
            return (MFEnumPolytope)NULL;
           }
          mfcell->nIndices=m;
          for(i=0;i<m;i++)mfcell->indices[i]=indices[i];
          if(verbose)
           {
            printf("        Polytopes [");
            for(i=0;i<m;i++)
             {
              if(i>0)printf(",");
              printf("%d",mfcell->indices[i]);
             }
            printf("]\n");
            fflush(stdout);
           }
          mfcell->indices[m]=-1;
         }else{
          mfcell->nIndices=0;
          if(mfcell->indices!=(int*)NULL)
           {
#ifdef EPTraceFrees
            fprintf(stderr,"Free called at line %d in file %s\n",__LINE__+2,__FILE__);
#endif
            free(mfcell->indices);
           }
          mfcell->indices=(int*)NULL;
          if(verbose){printf("        Not a real %d-cell\n",dim);fflush(stdout);}
         }
       }
      if(verbose){printf("done print dim %d in middle of MFEnumeratePolytopes\n",dim);fflush(stdout);}
     }
    if(verbose){printf("done print in middle of MFEnumeratePolytopes\n");fflush(stdout);}
     }
   }

/* Now have for each cell of each polytope a list of the k-cells to which it belongs */

  if(verbose){printf("\nSecond step: assign global numbers to the cells in the polytope of each chart.\n\n");fflush(stdout);}

  for(i=0;i<k+1;i++)Complex->nCells[i]=0;

  for(chart=0;chart<nCharts;chart++)
   {
    if(verbose){printf("  Chart %d - \n",chart);fflush(stdout);}
    if(EPS[chart]!=(MFEnumPolytope)NULL)
     {
    for(dim=0;dim<k+1;dim++)
     {
      nCells=MFEnumPolytopeNumberOfCells(EPS[chart],dim);
      if(verbose){printf("    There are %d cells of dimension %d \n",nCells,dim);fflush(stdout);}
      found=0;
      for(cell=0;cell<nCells;cell++)
       {
        mfcell=((EPS[chart])->cells[dim])[cell];
        if(mfcell->nIndices>0 && mfcell->indices[mfcell->nIndices]==-1)
         {
          found=1;
          if(verbose)
           {
            printf("     %d-cell %d is unassigned, %d, signature [",dim,cell,mfcell->indices[mfcell->nIndices]);
            for(i=0;i<mfcell->nIndices;i++)
             {
              if(i>0)printf(",");
              printf("%d",mfcell->indices[i]);
             }
            printf("]\n");
            fflush(stdout);
           }
          for(i=0;i<mfcell->nIndices;i++)
           {
            chart2=mfcell->indices[i];
            if(verbose){printf("       Polytope %d shares this cell\n",chart2);fflush(stdout);}
            if(chart2!=chart && EPS[chart2]!=(MFEnumPolytope)NULL)
             {
              nCells2=MFEnumPolytopeNumberOfCells(EPS[chart2],dim);
              maxSame=0;
              mSame=0;
              mfcellClosest=(MFCell)NULL;
              cellClosest=-1;
              for(cell2=0;cell2<nCells2;cell2++)
               {
                mfcell2=((EPS[chart2])->cells[dim])[cell2];
                if(mfcell2->nIndices==0)
                 {
                  if(verbose){printf("         cell %d is not a real cell\n",cell2);fflush(stdout);}
                 }else{
                  nSame=MFIntersectIndices(mfcell->nIndices,mfcell->indices,mfcell2->nIndices,mfcell2->indices,(int*)NULL);
                  if(verbose)
                   {
                    printf("         cell %d, %d, signature [",cell2,mfcell2->indices[mfcell2->nIndices]);
                    for(l=0;l<mfcell->nIndices;l++)
                     {
                      if(l>0)printf(",");
                      printf("%d",mfcell2->indices[l]);
                     }
                    printf("] matches %d, max so far is %d\n",nSame,maxSame);
                    fflush(stdout);
                   }
/* Here don't want to require match. Know that one of these vertices is the one */
                  if(nSame>maxSame)
                   {
                    maxSame=nSame;
                    mSame=1;
                    mfcellClosest=mfcell2;
                    cellClosest=cell2;
                   }else if(nSame==maxSame) mSame++;
                 }
               }
              if(verbose){printf("       blah %d \n",chart2);fflush(stdout);}
/*            if(mfcellClosest!=(MFCell)NULL && mSame==1 )*/
              if(mfcellClosest!=(MFCell)NULL && maxSame==mfcell->nIndices )
               {
                mfcell2=mfcellClosest;
                if(verbose)
                 {
                  printf("         match! with cell %d, chart %d signature [",cellClosest,chart2);
                  for(l=0;l<mfcell2->nIndices;l++)
                   {
                    if(l>0)printf(",");
                    printf("%d",mfcell2->indices[l]);
                   }
                  printf("]");
                  printf(" %d matches with chart %d cell %d [",maxSame,chart,cell);
                  for(l=0;l<mfcell->nIndices;l++)
                   {
                    if(l>0)printf(",");
                    printf("%d",mfcell->indices[l]);
                   }
                  printf("]\n");
                  fflush(stdout);
                 }
                if(mfcell->indices[mfcell->nIndices]==-1
                && mfcell2->indices[mfcell2->nIndices]==-1)
                 {
                  mfcell->indices[mfcell->nIndices]=Complex->nCells[dim];
                  if(verbose)printf("     unassigned %d-cell %d on chart %d will be assigned index %d\n",dim,cell,chart,mfcell->indices[mfcell->nIndices]);
                  Complex->nCells[dim]++;
                 }
                if(mfcell2->indices[mfcell2->nIndices]==-1)
                 {
                  mfcell2->indices[mfcell2->nIndices]=mfcell->indices[mfcell->nIndices];
                  if(verbose)printf("     matching %d-cell %d on chart %d will also be assigned index %d\n",dim,cellClosest,chart2,mfcell2->indices[mfcell2->nIndices]);
                 }else{
                  mfcell->indices[mfcell->nIndices]=mfcell2->indices[mfcell2->nIndices];
                  if(verbose)printf("     unassigned %d-cell %d on chart %d will be assigned index %d from matching cell\n",dim,cell,chart,mfcell2->indices[mfcell2->nIndices]);
                 }
               }
             } /* if chart!=chart2 */
            if(verbose){printf("       blah2 %d \n",chart2);fflush(stdout);}
           } /* for Indices */
          if(mfcell->indices[mfcell->nIndices]==-1 && dim==k)
           {
            mfcell->indices[mfcell->nIndices]=Complex->nCells[dim];
            if(verbose)printf("     unassigned %d-cell %d on chart %d matches nothing. Will be assigned index %d\n",dim,cell,chart,mfcell->indices[mfcell->nIndices]);
            Complex->nCells[dim]++;
           }
         }   /* if unassigned */
       }     /* for cell */
      if(verbose&&!found){printf("        None become new ones\n");fflush(stdout);}
     } /* for dim */
     }
   }   /* for chart */

/* At this point we have found all the vertices which are simple. */

/* Now have for each cell of each polytope a global index, and a count of the cells of each dimension */
/* May have unmatched cells (index -1) */

/* If k==2 can walk around 0-cells. 
                           1-cells are already identified, 
                           2-cells are unique */

/* If k>2 can walk around (k-2)-cells. 
                          (k-1)-cells are already identified, 
                           k-cells are unique */

  if(verbose){printf("\nThird step: Allocate space for cells in the Complex.\n\n");fflush(stdout);}

  for(dim=0;dim<k+1;dim++)
   {
    if(dim==0)
     {
      Complex->mCells[dim]=Complex->nCells[dim];
      Complex->cells[dim]=(MFCell*)realloc((void*)(Complex->cells[dim]),Complex->mCells[dim]*sizeof(MFCell)); /*done*/
      if(Complex->cells[dim]==(MFCell*)NULL && Complex->mCells[dim]!=0)
       {
        sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",Complex->mCells[dim]*sizeof(MFCell));
        printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
        MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
        return (MFEnumPolytope)NULL;
       }
      Complex->v=(MFKVector*)realloc((void*)(Complex->v),Complex->mCells[dim]*sizeof(MFKVector)); /*done*/
      if(Complex->v==(MFKVector*)NULL && Complex->mCells[dim]!=0)
       {
        sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",Complex->mCells[dim]*sizeof(MFKVector));
        printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
        MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
        return (MFEnumPolytope)NULL;
       }
      for(i=0;i<Complex->mCells[dim];i++)Complex->v[i]=(MFKVector)NULL;
     }else{
      Complex->mCells[dim]=Complex->nCells[dim];
      Complex->cells[dim]=(MFCell*)realloc((void*)Complex->cells[dim],Complex->mCells[dim]*sizeof(MFCell)); /*done*/
      if(Complex->cells[dim]==(MFCell*)NULL && Complex->mCells[dim]!=0)
       {
        sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",Complex->mCells[dim]*sizeof(MFCell));
        printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
        MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
        return (MFEnumPolytope)NULL;
       }
     }
    if(verbose){printf("    There are %d cells of dimension %d\n",Complex->nCells[dim],dim);fflush(stdout);}
    for(i=0;i<Complex->nCells[dim];i++)
     {
      mfcell=(MFCell)malloc(sizeof(struct MFCellSt)); /*done*/
#ifdef EPTraceFrees
        fprintf(stderr,"Allocated MFCell at 0x%8.8x, line %d of file %s\n",mfcell,__LINE__-2,__FILE__);
#endif
      if(mfcell==(MFCell)NULL)
       {
        sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFCellSt));
        printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
        MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
        return (MFEnumPolytope)NULL;
       }
      (Complex->cells[dim])[i]=mfcell;
      mfcell->nIndices=0;
      mfcell->indices=(int*)NULL;
      mfcell->nFaceCells=0;
      mfcell->mFaceCells=0;
      mfcell->faceCells=(int*)NULL;
      mfcell->nCellFaces=0;
      mfcell->mCellFaces=0;
      mfcell->cellFaces=(int*)NULL;
     }
   }

  if(verbose){printf("\nFourth step: Place the k-cells from the polytopes into the Complex.\n\n");fflush(stdout);}

  for(chart=0;chart<nCharts;chart++)
   {
    if(EPS[chart]!=(MFEnumPolytope)NULL)
     {
    if(verboseTerse)
     {
      printf("\n  Chart %d:\n\n",chart);
      MFPrintEnumPolytope(stdout,EPS[chart]);
     }
    if(verbose){printf("    %d-cell %d\n",k,chart);fflush(stdout);}
    mfcell2=((EPS[chart])->cells[k])[0];
    if(Complex->cells[k]!=(MFCell*)NULL)
      mfcell=(Complex->cells[k])[chart];
     else
      mfcell=(MFCell)NULL;

    if(chart<Complex->nCells[k] && mfcell!=(MFCell)NULL)
     {
      if(verbose){printf("        mfcell is (Complex->cells[%d])[%d] = 0x%8.8x\n",k,chart,mfcell);fflush(stdout);}
      mfcell->nFaceCells=0;
      mfcell->mFaceCells=0;
      mfcell->faceCells=(int*)NULL;
  
      m=0;
      for(i=0;i<mfcell2->nCellFaces;i++)
       {
        cellFace=(EPS[chart]->cells[k-1])[mfcell2->cellFaces[i]];
        if(cellFace->nIndices>0 && cellFace->indices[cellFace->nIndices]>-1 )m++;
       }
      if(m==0)
       {
        mfcell->nCellFaces=0;
        mfcell->mCellFaces=0;
        mfcell->cellFaces=(int*)NULL;
        if(verbose){printf("           has no cell faces\n");fflush(stdout);}
       }else{
        mfcell->nCellFaces=0;
        mfcell->mCellFaces=m;
        mfcell->cellFaces=(int*)malloc(mfcell->mCellFaces*sizeof(int)); /*done*/
        if(mfcell->cellFaces==(int*)NULL)
         {
          sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",mfcell->mCellFaces*sizeof(int));
          printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
          MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
          return (MFEnumPolytope)NULL;
         }
  
        for(i=0;i<mfcell2->nCellFaces;i++)
         {
          cellFace=(EPS[chart]->cells[k-1])[mfcell2->cellFaces[i]];
          if(cellFace->nIndices>0 && cellFace->indices[cellFace->nIndices]>-1)
           {
            mfcell->cellFaces[mfcell->nCellFaces]=cellFace->indices[cellFace->nIndices];
            (mfcell->nCellFaces)++;
           }
         }
        if(verbose)
         {
          printf("        will have %d cell faces [",mfcell->nCellFaces);
          for(i=0;i<mfcell->nCellFaces;i++)
           {
            if(i>0)printf(",");
            printf("%d",mfcell->cellFaces[i]);
           }
          printf("]\n");
         }
       }
     }
     }
   }

  if(verbose){printf("\nFifth step: Merge the cell lists from the polytopes into the Complex.\n\n");fflush(stdout);}

  for(chart=0;chart<nCharts;chart++)
   {
    if(verbose){printf("    chart %d\n",chart);fflush(stdout);}
    if(EPS[chart]!=(MFEnumPolytope)NULL)
     {
    for(dim=0;dim<k;dim++)
     {
      nCells=MFEnumPolytopeNumberOfCells(EPS[chart],dim);
      if(verbose){printf("      There are %d cells of dimension %d\n",nCells,dim);fflush(stdout);}
      for(cell=0;cell<nCells;cell++)
       {
        mfcell=(EPS[chart]->cells[dim])[cell];
        if(mfcell->nIndices>0 && mfcell->indices[mfcell->nIndices]!=-1)
         {
          i=mfcell->indices[mfcell->nIndices];
          mfcell2=Complex->cells[dim][i];

          m=0;
          for(i=0;i<mfcell->nCellFaces;i++)
           {
            cellFace=(EPS[chart]->cells[dim-1])[mfcell->cellFaces[i]];
            if(cellFace->nIndices>0 && cellFace->indices[cellFace->nIndices]>-1 )m++;
           }

          if(verbose){printf("        cell %d is global %d-cell %d\n",cell,dim,mfcell->indices[mfcell->nIndices]);fflush(stdout);}
          if(m==0)
           {
            if(verbose){printf("          has no new cellFaces to contribute\n");fflush(stdout);}
           }else{
            if(verbose)
             {
              printf("          has %d new cellFaces to contribute [",m);
              for(i=0;i<mfcell->nCellFaces;i++)
               {
                cellFace=((EPS[chart])->cells[dim-1])[mfcell->cellFaces[i]];
                if(i>0)printf(",");
                if(cellFace->nIndices>0 && cellFace->indices[cellFace->nIndices]>-1)
                  printf("%d",cellFace->indices[cellFace->nIndices]);
                 else
                  printf("--");
               }
              printf("]\n");
              fflush(stdout);
             }
  
            if(mfcell2->mCellFaces==0)
             {
              mfcell2->mCellFaces=m;
              if(mfcell2->mCellFaces>-1)
               {
                mfcell2->cellFaces=(int*)malloc(mfcell2->mCellFaces*sizeof(int)); /*done*/
                if(mfcell2->cellFaces==(int*)NULL)
                 {
                  sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",mfcell->mCellFaces*sizeof(int));
                  printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
                  MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
                  return (MFEnumPolytope)NULL;
                 }
               }
             }else{
              mfcell2->mCellFaces+=m;
              if(mfcell2->mCellFaces>-1)
               {
                mfcell2->cellFaces=(int*)realloc((void*)(mfcell2->cellFaces),mfcell2->mCellFaces*sizeof(int)); /*done*/
                if(mfcell2->cellFaces==(int*)NULL)
                 {
                  sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",mfcell2->mCellFaces*sizeof(int));
                  printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
                  MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
                  return (MFEnumPolytope)NULL;
                 }
               }
             }
    
            for(i=0;i<mfcell->nCellFaces;i++)
             {
              cellFace=((EPS[chart])->cells[dim-1])[mfcell->cellFaces[i]];
              if(cellFace->nIndices>0 && cellFace->indices[cellFace->nIndices]>-1)
               {
                mfcell2->cellFaces[mfcell2->nCellFaces]=cellFace->indices[cellFace->nIndices];
                if(verbose){printf("             add cellFace %d\n",mfcell2->cellFaces[mfcell2->nCellFaces]);fflush(stdout);}
                mfcell2->nCellFaces++;
               }
             }
            if(verbose)
             {
              printf("        before collapse: cell faces [");
              for(i=0;i<mfcell2->nCellFaces;i++)
               {
                if(i>0)printf(",");
                printf("%d",mfcell2->cellFaces[i]);
               }
              printf("]\n");
              fflush(stdout);
             }
            EPCollapseIndices(&(mfcell2->nCellFaces),mfcell2->cellFaces);
            if(verbose)
             {
              printf("        after collapse: cell faces [");
              for(i=0;i<mfcell2->nCellFaces;i++)
               {
                if(i>0)printf(",");
                printf("%d",mfcell2->cellFaces[i]);
               }
              printf("]\n");
              fflush(stdout);
             }
           }

          m=0;
          for(i=0;i<mfcell->nFaceCells;i++)
           {
            faceCell=(EPS[chart]->cells[dim+1])[mfcell->faceCells[i]];
            if(faceCell->nIndices>0 && faceCell->indices[faceCell->nIndices]>-1)m++;
           }

          if(m==0)
           {
            if(verbose){printf("        has no new faceCells to contribute\n");fflush(stdout);}
           }else{
            if(verbose)
             {
              printf("          has %d new faceCells to contribute [",m);
              for(i=0;i<mfcell->nFaceCells;i++)
               {
                faceCell=(EPS[chart]->cells[dim+1])[mfcell->faceCells[i]];
                if(i>0)printf(",");
                if(faceCell->nIndices>0 && faceCell->indices[faceCell->nIndices]>-1)
                  printf("%d",faceCell->indices[faceCell->nIndices]);
                 else
                  printf("--");
               }
              printf("]\n");
              fflush(stdout);
             }
            if(mfcell2->mFaceCells==0)
             {
              mfcell2->mFaceCells=m+1;
              if(mfcell2->mFaceCells>-1)
               {
                mfcell2->faceCells=(int*)malloc(mfcell2->mFaceCells*sizeof(int)); /*done*/
                if(mfcell2->faceCells==(int*)NULL)
                 {
                  sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",mfcell2->mFaceCells*sizeof(int));
                  printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
                  MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
                  return (MFEnumPolytope)NULL;
                 }
               }
             }else{
              mfcell2->mFaceCells+=m+1;
              if(mfcell2->mFaceCells>-1)
               {
                mfcell2->faceCells=(int*)realloc((void*)(mfcell2->faceCells),mfcell2->mFaceCells*sizeof(int)); /*done*/
                if(mfcell2->faceCells==(int*)NULL)
                 {
                  sprintf(MFEnumPolytopeErrorMsg,"Out of memory, trying to allocate %d bytes",mfcell2->mFaceCells*sizeof(int));
                  printf("%s -- %s\n",RoutineName,MFEnumPolytopeErrorMsg);fflush(stdout);
                  MFSetError(12,RoutineName,MFEnumPolytopeErrorMsg,__LINE__,__FILE__);
                  return (MFEnumPolytope)NULL;
                 }
               }
             }
  
            for(i=0;i<mfcell->nFaceCells;i++)
             {
              faceCell=((EPS[chart])->cells[dim+1])[mfcell->faceCells[i]];
              if(faceCell->nIndices>0 && faceCell->indices[faceCell->nIndices]>-1)
               {
                mfcell2->faceCells[mfcell2->nFaceCells]=faceCell->indices[faceCell->nIndices];
                if(verbose){printf("             add faceCell %d\n",mfcell2->faceCells[mfcell2->nFaceCells]);fflush(stdout);}
                mfcell2->nFaceCells++;
               }
             }
            if(verbose)
             {
              printf("        before collapse: face cells [");
              for(i=0;i<mfcell2->nFaceCells;i++)
               {
                if(i>0)printf(",");
                printf("%d",mfcell2->faceCells[i]);
               }
              printf("]\n");
              fflush(stdout);
             }
            EPCollapseIndices(&(mfcell2->nFaceCells),mfcell2->faceCells);
            if(verbose)
             {
              printf("        after collapse: face cells [");
              for(i=0;i<mfcell2->nFaceCells;i++)
               {
                if(i>0)printf(",");
                printf("%d",mfcell2->faceCells[i]);
               }
              printf("]\n");
              fflush(stdout);
             }
           }
         }
       }
     }
     }  /*if(EPS[chart]!=(MFEnumPolytope)NULL)*/
   }

  for(chart=0;chart<nCharts;chart++)
   if(EPS[chart]!=(MFEnumPolytope)NULL)MFFreeEnumPolytope(EPS[chart]);
  free(EPS);

  if((failed=!MFTestEnumPolytope(Complex,0)) || verboseResult)
   {
    printf("----------------------------------------------------\n");
    printf("----------------------------------------------------\n");
    printf("\nComplex:\n\n");
    MFPrintEnumPolytope(stdout,Complex);
    printf("\n");
    printf("----------------------------------------------------\n");
    printf("----------------------------------------------------\n");
    fflush(stdout);
   }

  if(verboseTerse)
   {
    int halfspace;

    printf("\nHalf Spaces:\n\n");
    for(halfspace=0;halfspace<MFAtlasNumberOfHalfSpaces(A);halfspace++)
     {
      printf("      %d       [%d,%d]\n",halfspace,MFAtlasLeftPolytope(A,halfspace),MFAtlasRightPolytope(A,halfspace));
     }
    fflush(stdout);
   }
  return(Complex);
 }

MFKVector MFEnumPolytopeVertex(MFEnumPolytope P,long i)
 {
  static char RoutineName[]={"MFEnumPolytopeVertex"};
  return P->v[i];
 }
