/*
    @(#)MFNKMatrix.c	1.23
    03/02/25 16:32:04
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   November 11, 1997                     */
/*              February 2, 1999   Ported to C        */

#include <MFError.h>
#include <MFNKMatrix.h>
#include <MFNVector.h>
#include <MFKVector.h>
#include <float.h>
#include <stdlib.h>
#include <string.h>
#include <MFTime.h>
#include <MFPrint.h>
#include <math.h>

void MFSetError(int,char*,char*,int,char*);
static char MFNKMatrixErrorMsg[256]="";

double *MFNV_CStar(MFNVector);

#define DENSE 0
#define VECTOR 1

struct MFNKMatrixSt
 {
  int type;
  int n;
  int k;
  double *data;
  MFNVector *cols;
  int nRefs;
 };

void MFFreeNKMatrix(MFNKMatrix this)
 {
  static char RoutineName[]={"MFFreeNKMatrix"};
  int i;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFFreeNKMatrix++;
  starttime=clock();
#endif

  if(this==(MFNKMatrix)NULL)
   {
    sprintf(MFNKMatrixErrorMsg,"Pointer to Matrix (argument 1) is NULL");
    MFSetError(4,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFFreeNKMatrix+=clock()-starttime;
#endif
    return;
   }

  this->nRefs--;

  if(this->nRefs<1)
   {
    if(this->data!=(double*)NULL)free(this->data);
    if(this->cols!=(MFNVector*)NULL)
     {
      for(i=0;i<this->k;i++)MFFreeNVector(this->cols[i]);
      free(this->cols);
     }
    free(this);
   }
#ifdef MFTIMINGS
    MFTimeMFFreeNKMatrix+=clock()-starttime;
#endif
  return;
 }

void MFRefMatrix(MFNKMatrix this)
 {
  static char RoutineName[]={"MFRefMatrix"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFRefMatrix++;
  starttime=clock();
#endif

  if(this==(MFNKMatrix)NULL)
   {
    sprintf(MFNKMatrixErrorMsg,"Pointer to Matrix (argument 1) is NULL");
    MFSetError(4,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFRefMatrix+=clock()-starttime;
#endif
    return;
   }

  this->nRefs++;
 }

MFNKMatrix MFCreateNKMatrix(int k,MFNVector *col)
 {
  static char RoutineName[]={"MFCreateNKMatrix"};
  MFNKMatrix this;
  int i,j;
  double *u;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFCreateNKMatrix++;
  starttime=clock();
#endif

  if(k<1)
   {
    sprintf(MFNKMatrixErrorMsg,"Number of Columns %d (argument 1) is Illegal. Must be positive.",k);
    MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateNKMatrix+=clock()-starttime;
#endif
    return((MFNKMatrix)NULL);
   }

  if(col==(MFNVector*)NULL)
   {
    sprintf(MFNKMatrixErrorMsg,"List of Columns (argument 2) is NULL.");
    MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateNKMatrix+=clock()-starttime;
#endif
    return((MFNKMatrix)NULL);
   }

  this=(MFNKMatrix)malloc(sizeof(struct MFNKMatrixSt)); /*done*/
  if(this==(MFNKMatrix)NULL)
   {
    sprintf(MFNKMatrixErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFNKMatrixSt));
    MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateNKMatrix+=clock()-starttime;
#endif
    return((MFNKMatrix)NULL);
   }

  this->n=-1;
  this->k=k;

  if(!strcmp(MFNVGetId(col[0]),"DENSE"))
   {
    this->n=MFNV_NC(col[0]);
    this->data=(double*)malloc(this->n*this->k*sizeof(double)); /*done*/
    if(this->data==(double*)NULL)
     {
      sprintf(MFNKMatrixErrorMsg,"Out of memory, trying to allocate %d bytes",this->n*this->k*sizeof(double));
      MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
      free(this);
#ifdef MFTIMINGS
      MFTimeMFCreateNKMatrix+=clock()-starttime;
#endif
      return((MFNKMatrix)NULL);
     }
    this->cols=(MFNVector*)NULL;

    for(j=0;j<k;j++)
     {
      u=MFNV_CStar(col[j]);
      for(i=0;i<this->n;i++)
       {
        this->data[i+this->n*j]=u[i];
       }
     }
    this->type=DENSE;
   }else{
    this->data=(double*)NULL;

    this->cols=(MFNVector*)malloc(this->k*sizeof(MFNVector)); /*done*/
    if(this->cols==(MFNVector*)NULL)
     {
      sprintf(MFNKMatrixErrorMsg,"Out of memory, trying to allocate %d bytes",this->k*sizeof(MFNVector));
      MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
      free(this);
#ifdef MFTIMINGS
      MFTimeMFCreateNKMatrix+=clock()-starttime;
#endif
      return((MFNKMatrix)NULL);
     }

    for(j=0;j<k;j++)
     {
      this->cols[j]=col[j];
      MFRefNVector(col[j]);
     }

    if(!strcmp(MFNVGetId(col[0]),"DENSE"))this->type=DENSE;
     else this->type=VECTOR;
   }

  this->nRefs=1;

#ifdef MFTIMINGS
    MFTimeMFCreateNKMatrix+=clock()-starttime;
#endif
  return(this);
 }

MFNKMatrix MFCreateNKMatrixWithData(int n,int k, double *data)
 {
  static char RoutineName[]={"MFCreateNKMatrixWithData"};
  MFNKMatrix this;
  int i,j;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFCreateNKMatrixWithData++;
  starttime=clock();
#endif

  if(n<1)
   {
    sprintf(MFNKMatrixErrorMsg,"Number of Rows %d (argument 1) is Illegal. Must be positive.",n);
    MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateNKMatrixWithData+=clock()-starttime;
#endif
    return((MFNKMatrix)NULL);
   }

  if(k<1)
   {
    sprintf(MFNKMatrixErrorMsg,"Number of Columns %d (argument 2) is Illegal. Must be positive.",k);
    MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateNKMatrixWithData+=clock()-starttime;
#endif
    return((MFNKMatrix)NULL);
   }

  this=(MFNKMatrix)malloc(sizeof(struct MFNKMatrixSt)); /*done*/
  if(this==(MFNKMatrix)NULL)
   {
    sprintf(MFNKMatrixErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFNKMatrixSt));
    MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFCreateNKMatrixWithData+=clock()-starttime;
#endif
    return((MFNKMatrix)NULL);
   }
  this->n=n;
  this->k=k;

  this->data=(double*)malloc(this->n*this->k*sizeof(double)); /*done*/
  if(this->data==(double*)NULL)
   {
    sprintf(MFNKMatrixErrorMsg,"Out of memory, trying to allocate %d bytes",this->n*this->k*sizeof(double));
    MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
    free(this);
#ifdef MFTIMINGS
    MFTimeMFCreateNKMatrixWithData+=clock()-starttime;
#endif
    return((MFNKMatrix)NULL);
   }

  this->nRefs=1;

  if(data!=(double*)NULL)
   {
    for(j=0;j<k;j++)
     {
      for(i=0;i<n;i++)
       {
        this->data[i+n*j]=data[i+n*j];
       }
     }
   }else{
    for(i=0;i<n*k;i++)this->data[i]=0.;
   }

  this->cols=(MFNVector*)NULL;
  this->type=DENSE;

#ifdef MFTIMINGS
    MFTimeMFCreateNKMatrixWithData+=clock()-starttime;
#endif
  return(this);
 }

void MFMVMul(MFNSpace R, MFNKMatrix this,MFKVector s,MFNVector u)
 {
  static char RoutineName[]={"MFMVMul"};
  int i,j;
  MFNVector col;
  MFNVector v;
  double t;
  static int verbose;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFMVMul++;
  starttime=clock();
#endif

  verbose=0;
  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

/* Need to use Add and Scale from NSpace */
  v=MFCloneNVector(u);

  for(i=0;i<this->k;i++)
   {
    if(this->type==DENSE)
      col=MFCreateNVectorWithData(this->n,this->data+i*this->n);
     else
      col=MFMColumn(this,i);

    if(col==(MFNVector)NULL)
     {
      sprintf(MFNKMatrixErrorMsg,"col %d of this is NULL");
      MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
      MFTimeMFMVMul+=clock()-starttime;
#endif
      return;
     }

    t=MFKV_C(s,i);
    if(i==0)
     {
      MFNSpaceScale(R,t,col,u);
     }else{
      MFNSpaceScale(R,t,col,v);
      MFNSpaceAdd(R,u,v,u);
     }
    MFFreeNVector(col);
   }
  MFFreeNVector(v);
#ifdef MFTIMINGS
    MFTimeMFMVMul+=clock()-starttime;
#endif
  return;
 }

void MFMVMulT(MFNSpace R, MFNKMatrix this,MFNVector u,MFKVector s)
 {
  static char RoutineName[]={"MFMVMulT"};
  int i,j;
  double t;
  MFNVector col;
  int verbose;

  verbose=0;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFMVMulT++;
  starttime=clock();
#endif

  if(verbose){printf("%s, n=%d\n",RoutineName,this->n);fflush(stdout);}
  for(j=0;j<this->k;j++)
   {
    col=MFMColumn(this,j);
    t=MFNSpaceInner(R,col,u);
    if(verbose)
     {
      printf("   s[%d]=",j);MFPrintNVector(stdout,col);
      printf(".");MFPrintNVector(stdout,u);
      printf("=%lf\n",t);
     }
    MFKVSetC(s,j,t);
    MFFreeNVector(col);
   }

#ifdef MFTIMINGS
    MFTimeMFMVMulT+=clock()-starttime;
#endif
  return;
 }

double *MFNKM_CStar(MFNKMatrix this)
 {
  static char RoutineName[]={"MFNKM_CStar"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNKM_CStar++;
  starttime=clock();
#endif

  if(this==(MFNKMatrix)NULL)
   {
    sprintf(MFNKMatrixErrorMsg,"Pointer to Matrix (argument 1) is NULL");
    MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFNKM_CStar+=clock()-starttime;
#endif
    return (double*)NULL;
   }

  if(this->type!=DENSE)
   {
    sprintf(MFNKMatrixErrorMsg,"Trying to get dense arrays from non-dense Vector type %d",this->type);
    printf("%s\n",MFNKMatrixErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFNKM_CStar+=clock()-starttime;
#endif
    return (double*)NULL;
   }

#ifdef MFTIMINGS
    MFTimeMFNKM_CStar+=clock()-starttime;
#endif
  return(this->data);
 }

MFNVector MFMColumn(MFNKMatrix this,int j)
 {
  static char RoutineName[]={"MFMColumn"};
  int i;
  MFNVector c;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFMColumn++;
  starttime=clock();
#endif

  if(this==(MFNKMatrix)NULL)
   {
    sprintf(MFNKMatrixErrorMsg,"Pointer to Matrix (argument 1) is NULL");
    MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFNKM_CStar+=clock()-starttime;
#endif
    return (MFNVector)NULL;
   }

  if(j>this->k)
   {
    sprintf(MFNKMatrixErrorMsg,"Request for invalid column, %d must be in [0,%d)",j,this->k);
    MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFNKM_CStar+=clock()-starttime;
#endif
    return (MFNVector)NULL;
   }

  switch(this->type)
   {
    case DENSE:
     c=MFCreateWrappedNVector(this->n,this->data+j*this->n);
     break;
    default:
     c=this->cols[j];
     MFRefNVector(c);
   }

#ifdef MFTIMINGS
    MFTimeMFMColumn+=clock()-starttime;
#endif
  return c;
 }

void MFMRow(MFNKMatrix this,int i,MFKVector r)
 {
  static char RoutineName[]={"MFMRow"};
  int j;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFMRow++;
  starttime=clock();
#endif

  switch(this->type)
   {
    case DENSE:
     for(j=0;j<this->k;j++)
       MFKVSetC(r,j,this->data[i+this->n*j]);
     break;
    default:
     for(j=0;j<this->k;j++)
       MFKVSetC(r,j,MFNV_C(this->cols[j],i));
     break;
   }

#ifdef MFTIMINGS
    MFTimeMFMRow+=clock()-starttime;
#endif
  return;
 }

int MFNKMatrixK(MFNKMatrix this)
 {
  return this->k;
 }

int MFNKMatrixN(MFNKMatrix this)
 {
  return this->n;
 }

void MFWriteNKMatrix(FILE *fid,MFNKMatrix L)
 {
  static char RoutineName[]={"MFWriteNKMatrix"};
  int i;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFWriteNKMatrix++;
  starttime=clock();
#endif

  fprintf(fid,"%s\n","NKMatrix");
  fprintf(fid,"%d %d %d %d\n",L->type,L->n,L->k,L->nRefs);
  if(L->type==DENSE)
   {
    for(i=0;i<L->n*L->k;i++)
     {
      if(i>0)fprintf(fid," ");
      fprintf(fid,"%lf",L->data[i]);
     }
    fprintf(fid,"\n");
   }else{
    for(i=0;i<L->k;i++)
      MFWriteNVector(fid,L->cols[i]);
   }

#ifdef MFTIMINGS
    MFTimeMFWriteNKMatrix+=clock()-starttime;
#endif
  return;
 }

MFNKMatrix MFReadNKMatrix(FILE *fid)
 {
  static char RoutineName[]={"MFReadNKMatrix"};
  int i;
  MFNKMatrix L;
  char tag[100]="";

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFReadNKMatrix++;
  starttime=clock();
#endif

  fscanf(fid,"%s\n",tag);
  if(strcmp(tag,"NKMatrix"))
   {
    sprintf(MFNKMatrixErrorMsg,"Next Object is not a NKMatrix! (%s)\n",RoutineName,tag);
    MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFReadNKMatrix+=clock()-starttime;
#endif
    return;
   }

  L=(MFNKMatrix)malloc(sizeof(struct MFNKMatrixSt)); /*done*/
  if(L==(MFNKMatrix)NULL)
   {
    sprintf(MFNKMatrixErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFNKMatrixSt));
    printf("%s -- %s\n",RoutineName,MFNKMatrixErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFReadNKMatrix+=clock()-starttime;
#endif
    return (MFNKMatrix)NULL;
   }

  fscanf(fid,"%d %d %d %d\n",&(L->type),&(L->n),&(L->k),&(L->nRefs));

  if(L->type==DENSE)
   {
    L->data=(double*)malloc(L->n*L->k*sizeof(double)); /*done*/
    if(L->data==(double*)NULL)
     {
      sprintf(MFNKMatrixErrorMsg,"Out of memory, trying to allocate %d bytes",L->n*L->k*sizeof(double));
      printf("%s -- %s\n",RoutineName,MFNKMatrixErrorMsg);fflush(stdout);
      MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
      MFTimeMFReadNKMatrix+=clock()-starttime;
#endif
      return (MFNKMatrix)NULL;
     }

    for(i=0;i<L->n*L->k;i++)
     {
      if(i>0)fscanf(fid," ");
      fscanf(fid,"%lf",&(L->data[i]));
     }
    fscanf(fid,"\n");
    L->cols=(MFNVector*)NULL;
   }else{
    L->cols=(MFNVector*)malloc(L->k*sizeof(MFNVector)); /*done*/
    if(L->cols==(MFNVector*)NULL)
     {
      sprintf(MFNKMatrixErrorMsg,"Out of memory, trying to allocate %d bytes",L->k*sizeof(MFNVector));
      MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
      free(L);
#ifdef MFTIMINGS
      MFTimeMFCreateNKMatrix+=clock()-starttime;
#endif
      return((MFNKMatrix)NULL);
     }
    for(i=0;i<L->k;i++)
      L->cols[i]=MFReadNVector(fid);
    L->data=(double*)NULL;
   }

#ifdef MFTIMINGS
    MFTimeMFReadNKMatrix+=clock()-starttime;
#endif
  return L;
 }

void MFNKMSetC(MFNKMatrix A,int i,int j,double Aij)
 {
  static char RoutineName[]={"MFNKMSetC"};

  if(A==(MFNKMatrix)NULL)
   {
    sprintf(MFNKMatrixErrorMsg,"Pointer to Matrix (argument 1) is NULL");
    MFSetError(4,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFFreeNKMatrix+=clock()-starttime;
#endif
    return;
   }

  if(i<0)
   {
    sprintf(MFNKMatrixErrorMsg,"column index (argument 2) is less than zero");
    MFSetError(4,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
    return;
   }
  if(i>=A->n)
   {
    sprintf(MFNKMatrixErrorMsg,"column index (argument 2) is too large must be < %d)",A->n);
    MFSetError(4,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(j<0)
   {
    sprintf(MFNKMatrixErrorMsg,"row index (argument 3) is less than zero");
    MFSetError(4,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
    return;
   }
  if(j>=A->k)
   {
    sprintf(MFNKMatrixErrorMsg,"row index (argument 3) is too large must be < %d)",A->k);
    MFSetError(4,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
    return;
   }

  switch(A->type)
   {
    case DENSE:
     A->data[i+A->n*j]=Aij;
     break;
    default:
     MFNVSetC(A->cols[j],i,Aij);
     break;
   }

  return;
 }

void MFRefNKMatrix(MFNKMatrix this)
 {
  static char RoutineName[]={"MFRefNKMatrix"};

  if(this==(MFNKMatrix)NULL)
   {
    sprintf(MFNKMatrixErrorMsg,"Pointer to Matrix (argument 1) is NULL");
    MFSetError(4,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFNKM_CStar+=clock()-starttime;
#endif
    return;
   }
  this->nRefs++;

  return;
 }

void MFGramSchmidt(MFNSpace space,MFNKMatrix A)
 {
  static char RoutineName[]={"MFGramSchmidt"};
  double *a;
  int i,j,n;
  int k;
  MFNVector u;
  MFNVector v;
  MFNVector w;
  double inner;
  int verbose=0;

  n=MFNKMatrixN(A);
  k=MFNKMatrixK(A);

  switch(A->type)
   {
    case DENSE:
     a=MFNKM_CStar(A);
     MFGramSchmidtNoMat(n,k,a);
    default:
     u=MFMColumn(A,0);
     w=MFCloneNVector(u);
     MFFreeNVector(u);
     if(verbose){printf("Gram Schmidt:\n");fflush(stdout);}
     for(i=0;i<k;i++)
      {
       u=MFMColumn(A,i);

       inner=MFNSpaceInner(space,u,u);
       if(verbose){printf("u_%d\n",i);MFPrintNVector(stdout,u);printf("\n");fflush(stdout);
                   printf("<u_%d,u_%d>=%lf\n",i,i,inner);fflush(stdout);}
       inner=1./sqrt(inner);
       MFNSpaceScale(space,inner,u,u);
       if(verbose){printf("normalized u_%d\n",i);MFPrintNVector(stdout,u);printf("\n");fflush(stdout);}
   
       for(j=i+1;j<k;j++)
        {
         v=MFMColumn(A,j);
         if(verbose){printf("u_%d\n",j);MFPrintNVector(stdout,v);printf("\n");fflush(stdout);}
         inner=MFNSpaceInner(space,u,v);
         if(verbose){printf("<u_%d,u_%d>=%lf\n",i,j,inner);fflush(stdout);}
         MFNSpaceScale(space,inner,u,w);
         if(verbose){printf("<u_%d,u_%d>u_%d\n",j,i,i);MFPrintNVector(stdout,w);printf("\n");fflush(stdout);}
         MFNVDiff(v,w,v);
         if(verbose){printf("u_%d-<u_%d,u_%d>u_%d\n",j,j,i,i);MFPrintNVector(stdout,v);printf("\n");fflush(stdout);}
         MFFreeNVector(v);
        }
       MFFreeNVector(u);
      }
     MFFreeNVector(w);
   }

/* Test */

#ifdef JUNK
  printf("Verify Gram Schmidt:\n");fflush(stdout);
  for(i=0;i<k;i++)
   {
    u=MFMColumn(A,i);
    for(j=0;j<k;j++)
     {
      v=MFMColumn(A,j);
      printf("  <phi[%d],phi[%d]>=%lf\n",i,j,MFNSpaceInner(space,u,v));fflush(stdout);
      MFFreeNVector(v);
     }
    MFFreeNVector(u);
   }
#endif

  return;
 }

void MFGramSchmidtNoMat(int n, int k, double *a)
 {
  static char RoutineName[]={"MFGramSchmidtNoMat"};
  double inner;
  int i,j,l;
  double *u;
  double *v;
  int verbose;

  verbose=0;
  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

  for(i=0;i<k;i++)
   {
    u=a+n*i;

    inner=0.;for(l=0;l<n;l++)inner+=u[l]*u[l];
    inner=1./sqrt(inner);
    for(l=0;l<n;l++)u[l]=u[l]*inner;

    for(j=i+1;j<k;j++)
     {
      v=a+n*j;
      inner=0.;for(l=0;l<n;l++)inner+=u[l]*v[l];
      for(l=0;l<n;l++)v[l]=v[l]-inner*u[l];
     }
   }

  return;
 }

void MFNKMProjectTangentForBranchSwitch(MFNSpace space, MFNKMatrix A,MFNVector phi,MFNKMatrix Phi)
 {
  static char RoutineName[]={"MFNKMProjectTangentForBranchSwitch"};

  int i,j;
  int k,m;
  MFNVector u,w;
  double *p;
  double pmax;
  MFNVector pi,pj;
  int verbose;

  verbose=0;
  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}
  if(verbose){printf("  A=");MFPrintNKMatrix(stdout,A);fflush(stdout);}
  if(verbose){printf("  phi=");MFPrintNVector(stdout,phi);printf("\n");fflush(stdout);}

  k=MFNKMatrixK(A);

  p=(double*)malloc(k*sizeof(double));

  j=0;
  pmax=0.;
  for(i=0;i<k;i++)
   {
    u=MFMColumn(A,i);
    p[i]=MFNSpaceInner(space,phi,u);
    if(fabs(p[i])>pmax){pmax=fabs(p[i]);j=i;}
    MFFreeNVector(u);
    if(verbose){printf("phi^* A_[%d]=%lf\n",i,p[i]);fflush(stdout);}
   }
  if(verbose){printf("max of phi^* A_j is for j=%d, =%lf\n",j,pmax);fflush(stdout);}

  m=0;
  u=MFMColumn(Phi,m);
  w=MFCloneNVector(u);
  MFNSpaceScale(space,1.,phi,u);
  if(verbose){printf("column 0 of Phi is phi ");MFPrintNVector(stdout,u);printf("\n");(stdout);}
  if(Phi->type==DENSE)MFMSetColumn(Phi,m,u);
  MFFreeNVector(u);
  m++;

  pj=MFMColumn(A,j);
  for(i=0;i<k;i++)
   {
    if(i!=j)
     {
      pi=MFMColumn(A,i);
      u=MFMColumn(Phi,m);

      MFNSpaceScale(space, p[i],pj,u);
      MFNSpaceScale(space,-p[j],pi,w);
      MFNSpaceAdd(space,u,w,u);
      if(verbose){printf("column %d of Phi is %lf A_%d - %lf A_%d ",m,p[i],j,p[j],i);MFPrintNVector(stdout,u);printf("\n");(stdout);}
      if(Phi->type==DENSE)MFMSetColumn(Phi,m,u);
      m++;
      MFFreeNVector(u);
      MFFreeNVector(pi);
     }
   }
  MFFreeNVector(pj);
  free(p);
  if(verbose){printf("Before G-S:\n");MFPrintNKMatrix(stdout,Phi);fflush(stdout);}

  MFGramSchmidt(space,Phi);

  if(verbose){printf("After G-S:\n");MFPrintNKMatrix(stdout,Phi);fflush(stdout);}

  if(verbose){printf("done %s\n",RoutineName);fflush(stdout);}
  return;
 }

MFNKMatrix MFCloneNKMatrix(MFNKMatrix A)
 {
  static char RoutineName[]={"MFCloneNKMatrix"};
  MFNKMatrix this;
  int i,j;
  double *u;

  if(A==(MFNKMatrix)NULL)
   {
    sprintf(MFNKMatrixErrorMsg,"Cloning a NULL Matrix!");
    MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
    return((MFNKMatrix)NULL);
   }

  this=(MFNKMatrix)malloc(sizeof(struct MFNKMatrixSt)); /*done*/
  if(this==(MFNKMatrix)NULL)
   {
    sprintf(MFNKMatrixErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFNKMatrixSt));
    MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
    return((MFNKMatrix)NULL);
   }

  this->n=A->n;
  this->k=A->k;
  this->type=A->type;
  if(this->type==DENSE)
   {
    this->data=(double*)malloc(this->n*this->k*sizeof(double)); /*done*/
    if(this->data==(double*)NULL)
     {
      sprintf(MFNKMatrixErrorMsg,"Out of memory, trying to allocate %d bytes",this->n*this->k*sizeof(double));
      MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
      free(this);
      return((MFNKMatrix)NULL);
     }
    this->cols=(MFNVector*)NULL;

    for(j=0;j<this->k;j++)
     {
      for(i=0;i<this->n;i++)this->data[i+this->n*j]=A->data[i+this->n*j];
     }
   }else{
    this->data=(double*)NULL;

    this->cols=(MFNVector*)malloc(this->k*sizeof(MFNVector)); /*done*/
    if(this->cols==(MFNVector*)NULL)
     {
      sprintf(MFNKMatrixErrorMsg,"Out of memory, trying to allocate %d bytes",this->k*sizeof(MFNVector));
      MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
      free(this);
      return((MFNKMatrix)NULL);
     }

    for(j=0;j<this->k;j++)
      this->cols[j]=MFCloneNVector(A->cols[j]);
   }

  this->nRefs=1;

  return(this);
 }

void MFMSetColumn(MFNKMatrix this,int j, MFNVector u)
 {
  static char RoutineName[]={"MFMSetColumn"};
  int i;
  MFNVector c;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFMColumn++;
  starttime=clock();
#endif

  switch(this->type)
   {
    case DENSE:
     for(i=0;i<this->n;i++)
      this->data[i+this->n*j]=MFNV_C(u,i);
     break;
    default:
     for(i=0;i<this->n;i++)
      MFNVSetC(this->cols[j],i,MFNV_C(u,i));
   }

#ifdef MFTIMINGS
    MFTimeMFMColumn+=clock()-starttime;
#endif
  return;
 }

void MFGramSchmidtReplace(MFNSpace space,MFNKMatrix A,MFNVector phi0, MFNVector phi1)
 {
  static char RoutineName[]={"MFGramSchmidtReplace"};
  double *a;
  int i,j,n;
  int k;
  MFNVector u;
  MFNVector v;
  MFNVector w;
  double inner;
  int verbose=0;
  int zero;
  double *B=(double*)NULL;
  int sign0,sign1;

/* Remove vector u from the basis, and replaces it with the vector v */
  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

  n=MFNKMatrixN(A);
  k=MFNKMatrixK(A);

  B=(double*)malloc(k*k*sizeof(double));if(B==(double*)NULL)abort();
  for(i=0;i<k;i++)
   {
    u=MFMColumn(A,i);
    for(j=0;j<k;j++)
     {
      v=MFMColumn(A,j);
      B[i+k*j]=MFNSpaceInner(space,u,v);
      MFFreeNVector(v);
     }
    MFFreeNVector(u);
   }
  sign0=MFAtlasDet(k,B)>0;

  u=MFMColumn(A,0);
  w=MFCloneNVector(u);
  MFFreeNVector(u);
  if(verbose){printf("Gram Schmidt:\n");fflush(stdout);}
  zero=-1;
  for(i=-1;i<k;i++)
   {
    if(i>-1)u=MFMColumn(A,i);
     else{u=phi0;MFRefNVector(phi0);}

    inner=MFNSpaceInner(space,u,u);
    if(verbose){printf("u_%d\n",i);MFPrintNVector(stdout,u);printf("\n");fflush(stdout);
                printf("<u_%d,u_%d>=%lf\n",i,i,inner);fflush(stdout);}
    if(inner>1.e-7)
     {
      inner=1./sqrt(inner);
      MFNSpaceScale(space,inner,u,u);
      if(verbose){printf("normalized u_%d\n",i);MFPrintNVector(stdout,u);printf("\n");fflush(stdout);}
     }else{
      zero=i;
      if(verbose){printf("zero vector u_%d\n",i);MFPrintNVector(stdout,u);printf("\n");fflush(stdout);}
     }
 
    for(j=i+1;j<k;j++)
     {
      v=MFMColumn(A,j);
      if(verbose){printf("u_%d\n",j);MFPrintNVector(stdout,v);printf("\n");fflush(stdout);}
      inner=MFNSpaceInner(space,u,v);
      if(verbose){printf("<u_%d,u_%d>=%lf\n",i,j,inner);fflush(stdout);}
      MFNSpaceScale(space,inner,u,w);
      if(verbose){printf("<u_%d,u_%d>u_%d\n",j,i,i);MFPrintNVector(stdout,w);printf("\n");fflush(stdout);}
      MFNVDiff(v,w,v);
      if(verbose){printf("u_%d-<u_%d,u_%d>u_%d\n",j,j,i,i);MFPrintNVector(stdout,v);printf("\n");fflush(stdout);}
      MFFreeNVector(v);
     }
    MFFreeNVector(u);
    if(verbose){printf("\n");fflush(stdout);}
   }
  MFFreeNVector(w);

/* Delete the zero column (move it to the first position) */

  for(i=zero;i>-1;i--)
   {
    u=MFMColumn(A,i-1);
    v=MFMColumn(A,i);
    MFMSetColumn(A,i,u);
    MFFreeNVector(u);
    MFFreeNVector(v);
   }

/* Replace the zero column with phi1 */
  u=MFMColumn(A,0);
  MFFreeNVector(u);
  MFMSetColumn(A,0,phi1);

 MFGramSchmidt(space,A);

  for(i=0;i<k;i++)
   {
    u=MFMColumn(A,i);
    for(j=0;j<k;j++)
     {
      v=MFMColumn(A,j);
      B[i+k*j]=MFNSpaceInner(space,u,v);
      MFFreeNVector(v);
     }
    MFFreeNVector(u);
   }
  sign1=MFAtlasDet(k,B)>0;

  if(sign0&&!sign1||sign1&&!sign0)
   {
    u=MFMColumn(A,0);
    MFNSpaceScale(space,-1.,u,u);
    MFFreeNVector(u);
   }

  free(B);
  if(verbose){printf("done %s\n",RoutineName);fflush(stdout);}

  return;
 }

void MFMMMul(MFNSpace R, MFNKMatrix Phi0,MFNKMatrix Phi1,double *prod)
 {
  static char RoutineName[]={"MFMMMul"};
  int i,j;
  MFNVector col0,col1;
  static int verbose;

/* Phi_0^T Phi_1 */

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFMVMul++;
  starttime=clock();
#endif

  verbose=0;
  if(verbose){printf("%s\n",RoutineName);fflush(stdout);}

  if(Phi0->n!=Phi1->n)
   {
    sprintf(MFNKMatrixErrorMsg,"Can't do product of incompatible matrices. A.n=%d, B.n=%d",Phi0->n,Phi1->n);
    MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
    return;
   }

  if(Phi0->type!=Phi1->type)
   {
    sprintf(MFNKMatrixErrorMsg,"Can't do product of incompatible matrices. A.type=%d, B.type=%d",Phi0->type,Phi1->type);
    MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
    return;
   }

  for(i=0;i<Phi0->k;i++)
   {
    if(Phi0->type==DENSE)
      col0=MFCreateNVectorWithData(Phi0->n,Phi0->data+i*Phi0->n);
     else
      col0=MFMColumn(Phi0,i);
    if(col0==(MFNVector)NULL)
     {
      sprintf(MFNKMatrixErrorMsg,"col %d of A is NULL",i);
      MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
      return;
     }
    for(j=0;j<Phi1->k;j++)
     {
      if(Phi1->type==DENSE)
        col1=MFCreateNVectorWithData(Phi1->n,Phi1->data+i*Phi1->n);
       else
        col1=MFMColumn(Phi1,j);
  
      if(col1==(MFNVector)NULL)
       {
        sprintf(MFNKMatrixErrorMsg,"col %d of B is NULL",j);
        MFSetError(12,RoutineName,MFNKMatrixErrorMsg,__LINE__,__FILE__);
        return;
       }
      prod[i+Phi0->k*j]=MFNSpaceInner(R,col0,col1);
      MFFreeNVector(col1);
     }
    MFFreeNVector(col0);
   }
  return;
 }
