/* 
    %W%
    %D% %T%
   
    PROGRAM NAME:  Manifold

    (c) COPYRIGHT INTERNATIONAL BUSINESS MACHINES
    CORPORATION 12/1/2001.  ALL RIGHTS RESERVED.

    Please refer to the LICENSE file in the top directory

*/
/*      author: Mike Henderson mhender@watson.ibm.com */
/*      date:   March 1, 1999                         */

#include <MFNSpace.h>
#include <MFNVector.h>
#include <MFNKMatrix.h>
#include <MFError.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <MFTime.h>

void MFSetError(int,char*,char*,int,char*);
static char MFNSpaceErrorMsg[256]="";

struct MFNSpaceSt
 {
  int nRefs;
  char *id;
  void *d;

  double (*distance)(MFNSpace,MFNVector,MFNVector,void*);
  double (*inner)(MFNSpace,MFNVector,MFNVector,void*);
  void (*direction)(MFNSpace,MFNVector,MFNVector,MFNVector,void*);
  void (*add)(MFNSpace,MFNVector,MFNVector,MFNVector,void*);
  void (*scale)(MFNSpace,double,MFNVector,MFNVector,void*);
  void (*writedata)(FILE*,MFNSpace,void*);
  void (*freedata)(void *);
 };


void MFRefNSpace(MFNSpace this)
 {
  static char RoutineName[]={"MFRefNSpace"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFRefNSpace++;
  starttime=clock();
#endif
  this->nRefs++;

#ifdef MFTIMINGS
    MFTimeMFRefNSpace+=clock()-starttime;
#endif
  return;
 }

void MFFreeNSpace(MFNSpace this)
 {
  static char RoutineName[]={"MFFreeNSpace"};

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFFreeNSpace++;
  starttime=clock();
#endif

  this->nRefs--;
  if(this->nRefs<1)
   {
    if(this->id!=(char*)NULL)free(this->id);
    if(this->freedata!=NULL && this->d!=(void*)NULL)this->freedata(this->d);
    free(this);
   }
#ifdef MFTIMINGS
    MFTimeMFFreeNSpace+=clock()-starttime;
#endif
  return;
 }

double MFNSpaceDistance(MFNSpace this,MFNVector v0,MFNVector v1)
 {
  static char RoutineName[]={"MFNSpaceDistance"};
  double result;
  int i;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNSpaceDistance++;
  starttime=clock();
#endif

  result=this->distance(this,v0,v1,this->d);

#ifdef MFTIMINGS
    MFTimeMFNSpaceDistance+=clock()-starttime;
#endif
  return result;
 }

void MFNSpaceDirection(MFNSpace this,MFNVector v0,MFNVector v1,MFNVector diff)
 {
  static char RoutineName[]={"MFNSpaceDirection"};
  double d;
  int i;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNSpaceDirection++;
  starttime=clock();
#endif

  this->direction(this,v0,v1,diff,this->d);

#ifdef MFTIMINGS
    MFTimeMFNSpaceDirection+=clock()-starttime;
#endif
  return;
 }

double MFNSpaceInner(MFNSpace this,MFNVector v0,MFNVector v1)
 {
  static char RoutineName[]={"MFNSpaceInner"};
  double result;
  int i;
  int verbose;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNSpaceInner++;
  starttime=clock();
#endif

  result=this->inner(this,v0,v1,this->d);

#ifdef MFTIMINGS
    MFTimeMFNSpaceInner+=clock()-starttime;
#endif
  return result;
 }

void MFWriteNSpace(FILE *fid,MFNSpace u)
 {
  static char RoutineName[]={"MFWriteNSpace"};
  int i;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFWriteNSpace++;
  starttime=clock();
#endif

  fprintf(fid,"%s\n","NSpace");
  fprintf(fid,"%d\n",strlen(u->id));
  fprintf(fid,"%s\n",u->id);
  fprintf(fid,"%d\n",u->nRefs);
  u->writedata(fid,u,u->d);

#ifdef MFTIMINGS
    MFTimeMFWriteNSpace+=clock()-starttime;
#endif
  return;
 }

void MFReadWeightedNSpaceData(FILE *fid,MFNSpace);
#ifdef TPBVP
void MFReadTPBVPNSpaceData(FILE *fid,MFNSpace);
#endif
#ifdef AUTO
void MFReadAUTONSpaceData(FILE *fid,MFNSpace);
#endif

MFNSpace MFReadNSpace(FILE *fid)
 {
  static char RoutineName[]={"MFReadNSpace"};
  int i;
  MFNSpace Omega;
  char tag[100]="";
  int n=0;
  char *id;
  int verbose;

  verbose=0;
#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFReadNSpace++;
  starttime=clock();
#endif

  fscanf(fid,"%s\n",tag);
  if(strcmp(tag,"NSpace"))
   {
    sprintf(MFNSpaceErrorMsg,"Next Object is not a NSpace! (%s) -- %s\n",RoutineName,tag);
    MFSetError(12,RoutineName,MFNSpaceErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFReadNSpace+=clock()-starttime;
#endif
    return;
   }
  if(verbose)
   {
    printf("%s\n",RoutineName);
    printf("  tag %s\n",tag);fflush(stdout);
   }

  fscanf(fid,"%d\n",&n);
  if(verbose){printf("  len(id) %d\n",n);fflush(stdout);}

  id=(char*)malloc((n+1)*sizeof(char));
  if(id==(char*)NULL)
   {
    sprintf(MFNSpaceErrorMsg,"Out of memory, trying to allocate %d bytes",(n+1)*sizeof(char));
    printf("%s -- %s\n",RoutineName,MFNSpaceErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNSpaceErrorMsg,__LINE__,__FILE__);
#ifdef MFTIMINGS
    MFTimeMFReadNSpace+=clock()-starttime;
#endif
    return (MFNSpace)NULL;
   }
  fscanf(fid,"%s\n",id);
  if(verbose){printf("  id %s\n",id);fflush(stdout);}

  Omega= MFCreateNSpaceBaseClass(id);
  fscanf(fid,"%d\n",&(Omega->nRefs));

  if(!strcmp(id,"WeightedNSpace"))
   {
    MFReadWeightedNSpaceData(fid,Omega);
#ifdef TPBVP
  }else if(!strcmp(id,"TPBVPNSpace"))
   {
    MFReadTPBVPNSpaceData(fid,Omega);
#endif
#ifdef AUTO
  }else if(!strcmp(id,"AUTONSpace"))
   {
    MFReadAUTONSpaceData(fid,Omega);
#endif
   }else{
    sprintf(MFNSpaceErrorMsg,"Object is not a known NSpace! (%s)\n",RoutineName,tag);
    MFSetError(12,RoutineName,MFNSpaceErrorMsg,__LINE__,__FILE__);
    free(id);
#ifdef MFTIMINGS
    MFTimeMFReadNSpace+=clock()-starttime;
#endif
    return;
   }
  free(id);

#ifdef MFTIMINGS
    MFTimeMFReadNSpace+=clock()-starttime;
#endif
  return Omega;
 }

void MFNSpaceAdd(MFNSpace this,MFNVector v0,MFNVector v1,MFNVector sum)
 {
  static char RoutineName[]={"MFNSpaceAdd"};
  double d;
  int i;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNSpaceAdd++;
  starttime=clock();
#endif

  this->add(this,v0,v1,sum,this->d);

#ifdef MFTIMINGS
    MFTimeMFNSpaceAdd+=clock()-starttime;
#endif
  return;
 }

double MFNSpaceTangentDistance(MFNSpace this, MFNKMatrix Phi0, MFNKMatrix Phi1)
 {
  int i,j,n,k;
  double id[9];
  MFNVector phi0,phi1;

  k=MFNKMatrixK(Phi0);
  n=MFNKMatrixN(Phi0);
  if(k>2)return 0.;


  for(i=0;i<k;i++)
   {
    phi0=MFMColumn(Phi0,i);
    for(j=0;j<k;j++)
     {
      phi1=MFMColumn(Phi1,j);
      id[i+k*j]=MFNSpaceInner(this,phi0,phi1);
      MFFreeNVector(phi1);
     }
    MFFreeNVector(phi0);
   }

  if(k==1)return fabs(id[0]-1.);
   else if(k==2)return sqrt(fabs(id[0]*id[3]-id[1]*id[2]-1.));

  return 0.;
 }

#ifdef AUTO
void MFReadAUTONSpaceData(FILE *fid,MFNSpace Omega)
 {
  static char RoutineName[]={"MFReadAUTONSpaceData"};

  return;
 }
#endif

void MFNSpaceScale(MFNSpace this,double s,MFNVector v,MFNVector prod)
 {
  static char RoutineName[]={"MFNSpaceScale"};
  double d;
  int i;

#ifdef MFTIMINGS
  clock_t starttime;

  MFCalledMFNSpaceScale++;
  starttime=clock();
#endif

  this->scale(this,s,v,prod,this->d);

#ifdef MFTIMINGS
    MFTimeMFNSpaceScale+=clock()-starttime;
#endif
  return;
 }

MFNSpace MFCreateNSpaceBaseClass(char *id)
 {
  static char RoutineName[]={"MFCreateNSpaceBaseClass"};
  MFNSpace this;

  this=(MFNSpace)malloc(sizeof(struct MFNSpaceSt)); /*done*/
  if(this==(MFNSpace)NULL)
   {
    sprintf(MFNSpaceErrorMsg,"Out of memory, trying to allocate %d bytes",sizeof(struct MFNSpaceSt));
    printf("%s -- %s\n",RoutineName,MFNSpaceErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNSpaceErrorMsg,__LINE__,__FILE__);
    return (MFNSpace)NULL;
   }

  this->id=(char*)malloc((strlen(id)+1)*sizeof(char));
  if(this->id==(char*)NULL)
   {
    sprintf(MFNSpaceErrorMsg,"Out of memory, trying to allocate %d bytes",(strlen(id)+1)*sizeof(char));
    printf("%s -- %s\n",RoutineName,MFNSpaceErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNSpaceErrorMsg,__LINE__,__FILE__);
    return (MFNSpace)NULL;
   }

  strcpy(this->id,id);
  this->nRefs=1;
  this->d=(void*)NULL;
  this->distance=NULL;
  this->inner=NULL;
  this->direction=NULL;
  this->add=NULL;
  this->scale=NULL;
  this->writedata=NULL;
  this->freedata=NULL;

  return this;
 }

void MFNSpaceSetDistance(MFNSpace this,double (*distance)(MFNSpace,MFNVector,MFNVector,void*))
 {
  static char RoutineName[]={"MFNSpaceSetDistance"};

  if(this==(MFNSpace)NULL)
   {
    sprintf(MFNSpaceErrorMsg,"NSpace, (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNSpaceErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNSpaceErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->distance=distance;
  return;
 }

void MFNSpaceSetInnerProduct(MFNSpace this,double (*inner)(MFNSpace,MFNVector,MFNVector,void*))
 {
  static char RoutineName[]={"MFNSpaceSetDistance"};

  if(this==(MFNSpace)NULL)
   {
    sprintf(MFNSpaceErrorMsg,"NSpace, (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNSpaceErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNSpaceErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->inner=inner;
  return;
 }

void MFNSpaceSetDirection(MFNSpace this,void (*direction)(MFNSpace,MFNVector,MFNVector,MFNVector,void*))
 {
  static char RoutineName[]={"MFNSpaceSetDistance"};

  if(this==(MFNSpace)NULL)
   {
    sprintf(MFNSpaceErrorMsg,"NSpace, (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNSpaceErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNSpaceErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->direction=direction;
  return;
 }

void MFNSpaceSetAdd(MFNSpace this,void (*add)(MFNSpace,MFNVector,MFNVector,MFNVector,void*))
 {
  static char RoutineName[]={"MFNSpaceSetDistance"};

  if(this==(MFNSpace)NULL)
   {
    sprintf(MFNSpaceErrorMsg,"NSpace, (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNSpaceErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNSpaceErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->add=add;
  return;
 }

void MFNSpaceSetScale(MFNSpace this,void (*scale)(MFNSpace,double,MFNVector,MFNVector,void*))
 {
  static char RoutineName[]={"MFNSpaceSetDistance"};

  if(this==(MFNSpace)NULL)
   {
    sprintf(MFNSpaceErrorMsg,"NSpace, (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNSpaceErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNSpaceErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->scale=scale;
  return;
 }

void MFNSpaceSetWriteData(MFNSpace this,void (*writedata)(FILE*,MFNSpace,void*))
 {
  static char RoutineName[]={"MFNSpaceSetDistance"};

  if(this==(MFNSpace)NULL)
   {
    sprintf(MFNSpaceErrorMsg,"NSpace, (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNSpaceErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNSpaceErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->writedata=writedata;
  return;
 }

void MFNSpaceSetFreeData(MFNSpace this,void (*freedata)(void *))
 {
  static char RoutineName[]={"MFNSpaceSetDistance"};

  if(this==(MFNSpace)NULL)
   {
    sprintf(MFNSpaceErrorMsg,"NSpace, (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNSpaceErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNSpaceErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->freedata=freedata;
  return;
 }

void MFNSpaceSetData(MFNSpace this,void *data)
 {
  static char RoutineName[]={"MFNSpaceSetDistance"};

  if(this==(MFNSpace)NULL)
   {
    sprintf(MFNSpaceErrorMsg,"NSpace, (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNSpaceErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNSpaceErrorMsg,__LINE__,__FILE__);
    return;
   }

  this->d=data;
  return;
 }

void *MFNSpaceGetData(MFNSpace this)
 {
  static char RoutineName[]={"MFNSpaceGetData"};

  if(this==(MFNSpace)NULL)
   {
    sprintf(MFNSpaceErrorMsg,"NSpace, (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNSpaceErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNSpaceErrorMsg,__LINE__,__FILE__);
    return (void*)NULL;
   }

  return this->d;
 }

char *MFNSpaceGetId(MFNSpace this)
 {
  static char RoutineName[]={"MFNSpaceGetId"};

  if(this==(MFNSpace)NULL)
   {
    sprintf(MFNSpaceErrorMsg,"NSpace, (argument 1) is NULL");
    printf("%s -- %s\n",RoutineName,MFNSpaceErrorMsg);fflush(stdout);
    MFSetError(12,RoutineName,MFNSpaceErrorMsg,__LINE__,__FILE__);
    return (void*)NULL;
   }

  return this->id;
 }

